/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.27-src\Nachtblitz\I2P.cs
 * PURPOSE: SAM Bridge interface and related code
 * VERSION: 1.4.27
 * DATE: March 23, 2016
 * COPYRIGHT: Copyright (C) 2012-2016 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using NB.I2P.Internal;
using NB.Network;
using NB.Network.eD2k;
using Support;
using Support.Encodings;
using Support.HashFunctions;

namespace NB.I2P.Internal
{
    internal class NetworkStreamHelper
    {
        // This class helps reading text and binary data
        // from a stream (it solves the problem of incompatibility
        // between the IO.StreamReader and IO.BinaryReader classes).

        public const ushort c_MaxExpectedLineLength = 2048;
        public const ushort c_MaxReadAheadSize = (c_MaxExpectedLineLength + Sam3DatagramControlClient.c_MaxDatagramPayloadIngoing);

        private static readonly LingerOption s_LingerOption = new LingerOption(true, 0);

        private readonly TcpClient f_TcpClnt;
        private readonly Socket f_Socket;
        private byte[] f_Buf;
        private bool f_DontUseBuffer;
        private int f_BufferUsage;
        private int f_ScannedBytes;
        private SendAsyncFields f_SendAsyncFields;
        private Time f_NextLifeCheckTime;
        private Time f_ReadLineSurrenderTime;

        public IPEndPoint p_RemoteEndpoint
        {
            get { return (IPEndPoint)f_Socket.RemoteEndPoint; }
        }

        public bool p_IsAlive
        {
            get { return CheckLife(false); }
        }

        public bool p_IsBusyWriting
        {
            get { return f_SendAsyncFields.IsBusyWriting(); }
        }

        public uint p_AvailableBytes
        {
            get { return ((uint)f_BufferUsage + (uint)p_SocketAvailableCatched); }
        }

        private int p_SocketAvailableCatched
        {
            get
            {
                try
                {
                    return f_Socket.Available;
                }
                catch (SocketException)
                {
                }
                catch (ObjectDisposedException)
                {
                }

                return 0;
            }
        }

        public NetworkStreamHelper(TcpClient a_TcpClnt)
        {
            f_TcpClnt = a_TcpClnt;
            f_Socket = f_TcpClnt.Client;
            f_Socket.LingerState = s_LingerOption;
            f_Buf = new byte[c_MaxReadAheadSize];
        }

        public void AdjustOSBufferSizes(int a_RecvBufSize,
                                        int a_SendBufSize)
        {
            f_Socket.ReceiveBufferSize = a_RecvBufSize;
            f_Socket.SendBufferSize = a_SendBufSize;
        }

        public void Close()
        {
            // Call this routine even if the connection was closed
            // remotely in order to release resources. This routine
            // is not thread safe.

            bool l_StillConnected = CheckIfSocketIsConnected();

            if (true == l_StillConnected)
            {
                CloseSocket();
            }

            // NOTE: This also closes the network stream and the socket
            // (debugger proven).
            f_TcpClnt.Close();
        }

        private bool CheckIfSocketIsConnected()
        {
            // This routine is thread-safe, as the f_Socket class is.

            try
            {
                return f_Socket.Connected;
            }
            catch (SocketException)
            {
            }
            catch (ObjectDisposedException)
            {
            }

            return false;
        }

        private void CloseSocket()
        {
            // This routine is thread-safe, as the f_Socket class is.

            try
            {
                if (true == f_Socket.Connected)
                {
                    f_Socket.Shutdown(SocketShutdown.Both);
                    f_Socket.Close();
                }
            }
            catch (SocketException)
            {
            }
            catch (ObjectDisposedException)
            {
            }
        }

        public void StopUsingOwnBuffer()
        {
            // Without the buffer TryReadLine() may not be
            // used anymore, but the performance of TryReadBytes()
            // increases.

            f_DontUseBuffer = true;

            if (0 == f_BufferUsage)
            {
                f_Buf = null;
            }

            f_Socket.NoDelay = true;
        }

        public bool CheckLife(bool a_Force)
        {
            Time l_Time = Software.sp_MainForm.f_Inst.GetTime();

            if (true == a_Force ||
                l_Time >= f_NextLifeCheckTime)
            {
                const uint l_AutoUpdateIntervalMs = 400;
                const uint CONN_STAT_DEAD = 1;

                f_NextLifeCheckTime = l_Time.AddMilliseconds(l_AutoUpdateIntervalMs);

                uint l_ConnStatus;
                bool l_Succ = Util.CheckConnectionStatus(f_Socket.Handle, out l_ConnStatus);

                if (true == l_Succ &&
                    CONN_STAT_DEAD == l_ConnStatus)
                {
                    Close();

                    return false;
                }
            }

            return CheckIfSocketIsConnected();
        }

        private int ReadSync(ref ByteArray a_Dest)
        {
            int l_Read;
            SocketError l_Err;

            try
            {
                l_Read = f_Socket.Receive(a_Dest.f_Arr, a_Dest.f_Idx, a_Dest.f_Len, SocketFlags.None, out l_Err);
            }
            catch
            {
                Close();

                return 0;
            }

            if (SocketError.Success != l_Err)
            {
                Close();

                return 0;
            }

            return l_Read;
        }

        private void ReadChunkSync(int a_MaxSizeNeeded)
        {
            if (a_MaxSizeNeeded <= f_BufferUsage)
            {
                // No need to read.

                return;
            }

            // If we have to read, we read as much as possible.
            int l_Avail = Math.Min(p_SocketAvailableCatched, f_Buf.Length - f_BufferUsage);

            if (0 < l_Avail)
            {
                ByteArray l_Dest = new ByteArray(f_Buf, f_BufferUsage, l_Avail);

                f_BufferUsage += ReadSync(ref l_Dest);
            }
        }

        private void KillBufferContent(int a_ByteCount)
        {
            int l_Rest = (f_BufferUsage - a_ByteCount);

            if (0 == l_Rest)
            {
                if (true == f_DontUseBuffer)
                {
                    f_Buf = null;
                }

                f_BufferUsage = 0;
                f_ScannedBytes = 0;
            }
            else if (0 < a_ByteCount)
            {
                Buffer.BlockCopy(f_Buf, a_ByteCount, f_Buf, 0, l_Rest);

                f_BufferUsage = l_Rest;
                f_ScannedBytes = Math.Max(0, f_ScannedBytes - a_ByteCount);
            }
        }

        private int FindByte(int a_StartingIdx,
                             byte a_Byte)
        {
            for (int i = a_StartingIdx; f_BufferUsage > i; i++)
            {
                if (f_Buf[i] == a_Byte)
                {
                    return i;
                }
            }

            return (-1);
        }

        public string TryReadLine()
        {
            ReadChunkSync(c_MaxExpectedLineLength);

            int l_Idx = FindByte(f_ScannedBytes, (byte)'\n');

            if (0 <= l_Idx)
            {
                string Res = Encoding.ASCII.GetString(f_Buf, 0, l_Idx);

                KillBufferContent(l_Idx + sizeof(byte));

                return Res;
            }

            return null;
        }

        public byte[] TryReadBytes(int a_Count)
        {
            ReadChunkSync(a_Count);

            if (a_Count <= f_BufferUsage)
            {
                byte[] l_Res = Util.CopyBytes(f_Buf, 0, a_Count);

                KillBufferContent(a_Count);

                return l_Res;
            }

            return null;
        }

        public int ReadAvailableBytes(ref ByteArray a_Bytes)
        {
            // a_Bytes.f_Len may not exceed the available bytes.
            // Returns the bytes actually read.

            if (0 == f_BufferUsage)
            {
                return ReadSync(ref a_Bytes);
            }

            int l_BufLen = Math.Min(f_BufferUsage, a_Bytes.f_Len);

            Buffer.BlockCopy(f_Buf, 0, a_Bytes.f_Arr, a_Bytes.f_Idx, l_BufLen);

            KillBufferContent(l_BufLen);

            if (a_Bytes.f_Len > l_BufLen)
            {
                ByteArray l_Dest = new ByteArray(a_Bytes.f_Arr, a_Bytes.f_Idx + l_BufLen, a_Bytes.f_Len - l_BufLen);

                l_BufLen += ReadSync(ref l_Dest);
            }

            return l_BufLen;
        }

        public void Flush()
        {
            while (true == p_IsAlive &&
                   true == p_IsBusyWriting)
            {
                Util.MiniSleep();
            }
        }

        public bool WriteLineAsync(string a_Str)
        {
            // Don't use Environment.NewLine for this.
            byte[] l_Arr = Encoding.ASCII.GetBytes(a_Str + "\n");
            ByteArray l_Data = new ByteArray(l_Arr);

            return WriteBytesAsync(ref l_Data);
        }

        public bool WriteBytesAsync(ref ByteArray a_Src)
        {
            // NOTE: As any routine in this class, this routine
            // is not thread-safe.

            // If this routine returns true, it only means that
            // no error was encountered yet, but it doesn't mean the
            // data was actually sent.

            bool l_Success = f_SendAsyncFields.SendAsync(this, ref a_Src);

            return l_Success;
        }

        private void SendAsyncCallback(IAsyncResult a_AsyncRes)
        {
            // Callback of Socket.BeginSend(). We land here on
            // sync and async completion.

            f_SendAsyncFields.HandleCompletionLegacy(this, a_AsyncRes);
        }

#if !PURE_DOT_NET_2_WITHOUT_ANY_SP
        private void SendAsyncCompleted(object a_Sender,
                                        EventArgs e)
        {
            // Callback of Socket.SendAsync(). We only land here if the
            // operation completes asynchronously.

            bool l_Success = f_SendAsyncFields.HandleAsyncCompletionAdvanced(this, e);
        }
#endif // !PURE_DOT_NET_2_WITHOUT_ANY_SP

        public string WriteLineAndWaitForAnswer(string a_Str)
        {
            Flush();
            WriteLineAsync(a_Str);

            return ReadLine();
        }

        public string ReadLine()
        {
            // Return can be null.

            f_ReadLineSurrenderTime = new Time();

            for (uint i = 0; true == p_IsAlive;)
            {
                string l_Answer = TryReadLine();

                if (null != l_Answer)
                {
                    return l_Answer;
                }

                if (200 > i)
                {
                    i++;
                }
                else
                {
                    Time l_Now = Time.Now();

                    if (false == f_ReadLineSurrenderTime.IsSet())
                    {
                        f_ReadLineSurrenderTime = l_Now.AddSeconds(Time.c_SecsPerMinute);
                    }
                    else if (f_ReadLineSurrenderTime <= l_Now)
                    {
                        break;
                    }
                }

                Util.MiniSleep();
            }

            return null;
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct SendAsyncFields
        {
            // This object supports ('legacy') .NET 2.0 and ('advanced')
            // .NET 2.0 SP1 semantics, while preferring SP1 semantics if
            // available. The PURE_DOT_NET_2_WITHOUT_ANY_SP compilation
            // symbol can be used to enfore legacy semantics.

#if !PURE_DOT_NET_2_WITHOUT_ANY_SP
            private static readonly bool s_SupportsSocketAsyncEventArgs;

            private object f_AsyncRes;
#else // !PURE_DOT_NET_2_WITHOUT_ANY_SP
            private IAsyncResult f_AsyncRes;
#endif // !PURE_DOT_NET_2_WITHOUT_ANY_SP

            private IAsyncResult p_AsyncResultLegacy
            {
                get { return (IAsyncResult)f_AsyncRes; }
            }

#if !PURE_DOT_NET_2_WITHOUT_ANY_SP
            static SendAsyncFields()
            {
                try
                {
                    CheckForSocketAsyncEventArgsSupport();
                }
                catch (TypeLoadException)
                {
                    return;
                }

                s_SupportsSocketAsyncEventArgs = true;
            }

            private static void CheckForSocketAsyncEventArgsSupport()
            {
                SocketAsyncEventArgs l_Dummy = new SocketAsyncEventArgs();
                l_Dummy.Dispose();
            }
#endif // !PURE_DOT_NET_2_WITHOUT_ANY_SP

            public bool IsBusyWriting()
            {
                // This routine may (of course) be called during an
                // ongoing asynchronous write.

                if (null == f_AsyncRes)
                {
                    return false;
                }

                bool l_Busy;

#if !PURE_DOT_NET_2_WITHOUT_ANY_SP
                if (false == s_SupportsSocketAsyncEventArgs)
                {
                    l_Busy = IsBusyWritingLegacy();
                }
                else
                {
                    l_Busy = IsBusyWritingAdvanced();
                }
#else // !PURE_DOT_NET_2_WITHOUT_ANY_SP
                l_Busy = IsBusyWritingLegacy();
#endif // !PURE_DOT_NET_2_WITHOUT_ANY_SP

                if (false == l_Busy)
                {
                    f_AsyncRes = null;

                    return false;
                }

                return true;
            }

            public bool SendAsync(NetworkStreamHelper a_Hlpr,
                                  ref ByteArray a_Data)
            {
#if !PURE_DOT_NET_2_WITHOUT_ANY_SP
                if (true == s_SupportsSocketAsyncEventArgs)
                {
                    return SendAsyncAdvanced(a_Hlpr, ref a_Data);
                }
#endif // !PURE_DOT_NET_2_WITHOUT_ANY_SP

                return SendAsyncLegacy(a_Hlpr, ref a_Data);
            }

            private bool IsBusyWritingLegacy()
            {
                return !p_AsyncResultLegacy.IsCompleted;
            }

            private bool SendAsyncLegacy(NetworkStreamHelper a_Hlpr,
                                         ref ByteArray a_Data)
            {
                IAsyncResult l_AsyncRes = null;
                bool l_Succ = false;

                try
                {
                    l_AsyncRes = a_Hlpr.f_Socket.BeginSend(a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len, SocketFlags.None, a_Hlpr.SendAsyncCallback, (uint)a_Data.f_Len);
                    l_Succ = true;
                }
                catch (SocketException)
                {
                }
                catch (ObjectDisposedException)
                {
                }

                if (false == l_Succ)
                {
                    a_Hlpr.Close();

                    return false;
                }

                if (false == l_AsyncRes.CompletedSynchronously)
                {
                    f_AsyncRes = l_AsyncRes;
                }

                return true;
            }

            public void HandleCompletionLegacy(NetworkStreamHelper a_Hlpr,
                                               IAsyncResult a_AsyncRes)
            {
                bool l_Succ = false;

                try
                {
                    SocketError l_Error;
                    int l_BytesSent = a_Hlpr.f_Socket.EndSend(a_AsyncRes, out l_Error);

                    if (SocketError.Success == l_Error &&
                        (uint)l_BytesSent == (uint)a_AsyncRes.AsyncState)
                    {
                        l_Succ = true;
                    }
                }
                catch (SocketException)
                {
                }
                catch (ObjectDisposedException)
                {
                }

                if (false == l_Succ)
                {
                    a_Hlpr.CloseSocket();
                }
            }

#if !PURE_DOT_NET_2_WITHOUT_ANY_SP
            private bool IsBusyWritingAdvanced()
            {
                SocketAsyncEventArgs l_Args = (SocketAsyncEventArgs)f_AsyncRes;

                return (SocketError.Success != l_Args.SocketError ||
                        l_Args.Count != l_Args.BytesTransferred);
            }

            private bool SendAsyncAdvanced(NetworkStreamHelper a_Hlpr,
                                           ref ByteArray a_Data)
            {
                SocketAsyncEventArgs l_Args = new SocketAsyncEventArgs();
                l_Args.SetBuffer(a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len);
                l_Args.Completed += a_Hlpr.SendAsyncCompleted;
                f_AsyncRes = l_Args;

                bool l_Succ = false;
                bool l_Pending = false;

                try
                {
                    if (true == a_Hlpr.f_Socket.SendAsync(l_Args))
                    {
                        l_Pending = true;
                    }

                    l_Succ = true;
                }
                catch (SocketException)
                {
                }
                catch (NotSupportedException)
                {
                }
                catch (InvalidOperationException)
                {
                    // This includes ObjectDisposedException.
                }

                if (false == l_Succ)
                {
                    l_Args.Dispose();
                    f_AsyncRes = null;

                    a_Hlpr.Close();

                    return false;
                }

                if (false == l_Pending)
                {
                    l_Succ = HandleAsyncCompletionAdvanced(a_Hlpr, l_Args);
                    f_AsyncRes = null;
                }

                return l_Succ;
            }

            public bool HandleAsyncCompletionAdvanced(NetworkStreamHelper a_Hlpr,
                                                      EventArgs a_Args)
            {
                SocketAsyncEventArgs l_Args = (SocketAsyncEventArgs)a_Args;
                SocketError l_Error = l_Args.SocketError;

                l_Args.Dispose();

                if (SocketError.Success != l_Error)
                {
                    a_Hlpr.CloseSocket();

                    return false;
                }

                return true;
            }
#endif // !PURE_DOT_NET_2_WITHOUT_ANY_SP
        }
    }

    abstract class SamControlClient
    {
        public const ushort c_DefaultSamPort = 7656;
        public const ushort c_DefaultBobPort = 2827;

        private const int c_BufferSize = (4 * 1024);

#pragma warning disable 429
        private const byte c_DefaultSigType = (false == NetworkManager.c_RestrictToClassicAddresses ? (byte)3 : (byte)0);
#pragma warning restore 429

        private NetworkStreamHelper f_StrmHlpr;
        protected BridgeStatus f_Status;

        // f_KeyPair can be null.
        private I2PKeyPair f_KeyPair;

        public bool p_IsAlive
        {
            get
            {
                if (null != f_StrmHlpr)
                {
                    if (true == f_StrmHlpr.p_IsAlive)
                    {
                        return true;
                    }

                    Close();
                }

                return false;
            }
        }

        public bool p_IsBusyWriting
        {
            get { return f_StrmHlpr.p_IsBusyWriting; }
        }

        public uint p_AvailableBytes
        {
            get { return f_StrmHlpr.p_AvailableBytes; }
        }

        public BridgeStatus p_Status
        {
            get { return f_Status; }
        }

        public I2PKeyPair p_KeyPair
        {
            get { return f_KeyPair; }
        }

        protected SamControlClient()
        {
        }

        public virtual void Close()
        {
            if (null != f_StrmHlpr)
            {
                f_StrmHlpr.Close();
                f_StrmHlpr = null;
            }

            f_Status = BridgeStatus.Disconnected;
        }

        public bool WriteBytesAsync(ref ByteArray a_Bytes)
        {
            return f_StrmHlpr.WriteBytesAsync(ref a_Bytes);
        }

        protected int ReadAvailableBytes(ref ByteArray a_Bytes)
        {
            return f_StrmHlpr.ReadAvailableBytes(ref a_Bytes);
        }

        protected byte[] TryReadBytes(int a_Count)
        {
            return f_StrmHlpr.TryReadBytes(a_Count);
        }

        protected string TryReadLine()
        {
            return f_StrmHlpr.TryReadLine();
        }

        protected void SamBeginControlSession(IPEndPoint a_SamEndPoint,
                                              string a_SessionID,
                                              I2PKeyPair a_KeyPair,
                                              string a_Options,
                                              string a_Style)
        {
            // a_PrivateKey can be null (in this case the SAM
            // bridge will create a new destination).

            TcpClient l_TcpClnt = Connect(a_SamEndPoint);

            if (null == l_TcpClnt)
            {
                return;
            }

            bool l_SAMv31;
            NetworkStreamHelper l_Hlpr = SamCreateClientAndHandleHello(l_TcpClnt, out l_SAMv31);

            if (null == l_Hlpr)
            {
                return;
            }

            l_Hlpr.AdjustOSBufferSizes(c_BufferSize, c_BufferSize);

            string l_PrivKeyStr;
            bool l_UsePassedKey = (null != a_KeyPair &&
                                   ((false == NetworkManager.c_RestrictToClassicAddresses &&
                                     true == l_SAMv31) ||
                                    true == a_KeyPair.f_PublicKey.p_IsClassicAddress));

            if (false == l_UsePassedKey)
            {
                l_PrivKeyStr = ("TRANSIENT" + (false == l_SAMv31 ? string.Empty : " SIGNATURE_TYPE=" + Util.NumberToInvString(c_DefaultSigType)));
            }
            else
            {
                l_PrivKeyStr = a_KeyPair.PrivateKeyAsBase64Str();
            }

            l_Hlpr.Flush();
            if (false == l_Hlpr.WriteLineAsync("SESSION CREATE STYLE=" + a_Style + " DESTINATION=" +
                                               l_PrivKeyStr + " ID=" + a_SessionID + a_Options))
            {
                l_Hlpr.Close();

                return;
            }

            f_StrmHlpr = l_Hlpr;
            f_Status = BridgeStatus.TryingToConnect;
            f_KeyPair = (false == l_UsePassedKey ? null : a_KeyPair);
        }

        protected void SamContinueBeginControlSession()
        {
            string l_Answer = TryReadLine();

            if (null == l_Answer)
            {
                // Wait.

                return;
            }

            if ((37 + I2PKeyPair.c_MinimumPubAndPrivKeyBase64Len) > l_Answer.Length ||
                false == Util.StartsWith(l_Answer, "SESSION STATUS RESULT=OK DESTINATION=") ||
                ((l_Answer.Length - 37) % 4) != 0)
            {
                Close();

                return;
            }

            byte[] l_PrivKey = Base64.FromBase64(l_Answer, 37);

            if (null == l_PrivKey)
            {
                Close();

                return;
            }

            I2PKeyPair l_KeyPair = I2PKeyPair.FromPrivateKey(l_PrivKey);

            if (null == l_KeyPair ||
                (true == NetworkManager.c_RestrictToClassicAddresses &&
                 false == l_KeyPair.f_PublicKey.p_IsClassicAddress))
            {
                if (null != l_KeyPair)
                {
                    l_KeyPair.Release();
                }

                Close();

                return;
            }

            l_KeyPair.f_PublicKey.SetValid();

            if (null == f_KeyPair ||
                false == f_KeyPair.ValueEquals(l_KeyPair, false))
            {
                f_KeyPair = l_KeyPair;
            }
            else
            {
                l_KeyPair.Release();
            }

            f_Status = BridgeStatus.Connected;
        }

        public static NetworkStreamHelper SamCreateClientAndHandleHello(TcpClient a_TcpClnt)
        {
            bool l_Junk;

            return SamCreateClientAndHandleHello(a_TcpClnt, out l_Junk);
        }

        public static NetworkStreamHelper SamCreateClientAndHandleHello(TcpClient a_TcpClnt,
                                                                        out bool a_SAMv31)
        {
            // Return can be null.

            a_SAMv31 = false;

            NetworkStreamHelper l_Hlpr = new NetworkStreamHelper(a_TcpClnt);

            string l_Answer = l_Hlpr.WriteLineAndWaitForAnswer("HELLO VERSION MIN=3.0 MAX=3.1");

            if (null != l_Answer)
            {
                if ("HELLO REPLY RESULT=OK VERSION=3.0" == l_Answer)
                {
                    return l_Hlpr;
                }

                if ("HELLO REPLY RESULT=OK VERSION=3.1" == l_Answer)
                {
                    a_SAMv31 = true;

                    return l_Hlpr;
                }
            }

            l_Hlpr.Close();

            return null;
        }

        public static TcpClient Connect(IPEndPoint a_EndPoint)
        {
            // Return can be null.

            TcpClient l_TcpClnt = new TcpClient();

            try
            {
                // This won't block for long.
                l_TcpClnt.Connect(a_EndPoint);
            }
            catch (SocketException)
            {
                l_TcpClnt.Close();

                return null;
            }

            return l_TcpClnt;
        }
    }

    class Sam3DatagramControlClient : SamControlClient
    {
        private const int c_ConnBufSize = (1024 * 1024);
        public const ushort c_MaxDatagramPayloadOutgoing = (10 * 1024);
        public const ushort c_MaxDatagramPayloadIngoing = (32 * 1024);
        // This is the maximum size in ascii chars we prepend
        // when sending a datagram.
        public const ushort c_MaxDatagramOverhead = (38 + I2PAddress.c_MaximumBase64CharLen);

        private string f_SessionID;
        private bool f_HeaderReceived;
        private I2PAddress f_Origin;
        private ushort f_Size;

        public Sam3DatagramControlClient()
        {
        }

        public void Start(IPEndPoint a_SamEndPoint,
                          string a_SessionID,
                          I2PKeyPair a_KeyPair,
                          string a_Options)
        {
            base.SamBeginControlSession(a_SamEndPoint, a_SessionID, a_KeyPair, a_Options, "DATAGRAM");

            if (BridgeStatus.Disconnected != f_Status)
            {
                f_SessionID = a_SessionID;
            }
        }

        public void ContinueStart()
        {
            base.SamContinueBeginControlSession();
        }

        public byte[] ReceiveDatagram(out I2PAddress a_Origin)
        {
            START:

            if (false == f_HeaderReceived)
            {
                string l_Str = base.TryReadLine();

                if (null != l_Str)
                {
                    int l_StrLen = l_Str.Length;

                    if (553 > l_StrLen ||
                        false == Util.StartsWith(l_Str, "DATAGRAM RECEIVED DESTINATION="))
                    {
                        goto FAIL;
                    }

                    int l_Idx = l_Str.LastIndexOf(" SIZE=", StringComparison.InvariantCulture);

                    if (546 > l_Idx)
                    {
                        goto FAIL;
                    }

                    f_Origin = I2PAddress.TryCreateFromString(l_Str, 30, l_Idx - 30);

                    if (null != f_Origin)
                    {
                        f_Origin.SetValid();
                    }

                    l_Idx += 6;
                    int l_SizeDigits = (l_StrLen - l_Idx);

                    if (0 == l_SizeDigits ||
                        5 < l_SizeDigits ||
                        false == Util.TryParseUInt16Inv(l_Str, l_Idx, l_SizeDigits, out f_Size))
                    {
                        goto FAIL;
                    }

                    f_HeaderReceived = true;
                }
            }

            if (true == f_HeaderReceived)
            {
                byte[] l_Body = base.TryReadBytes(f_Size);

                if (null != l_Body)
                {
                    bool l_SkipCurrentDatagram = (null == f_Origin);
                    a_Origin = f_Origin;

                    f_HeaderReceived = false;
                    f_Origin = null;
                    f_Size = 0;

                    if (true == l_SkipCurrentDatagram)
                    {
                        goto START;
                    }

                    return l_Body;
                }
            }

            EXIT:

            a_Origin = null;

            return null;

            FAIL:

            Close();

            goto EXIT;
        }

        public void Shutdown()
        {
            Close();
        }

        public static byte[] FormatDatagramHeader(I2PAddress a_Addr,
                                                  int a_BodySize)
        {
            string l_Line = ("DATAGRAM SEND DESTINATION=" + a_Addr.ToFullString() +
                             " SIZE=" + Util.NumberToInvString(a_BodySize) + "\n");

            return Encoding.ASCII.GetBytes(l_Line);
        }
    }

    class Sam3StreamControlClient : SamControlClient
    {
        private IPEndPoint f_SamEndPoint;
        private string f_SessionID;
        private NetworkStreamHelper f_AcceptHelper;

        public Sam3StreamControlClient()
        {
        }

        public void Start(IPEndPoint a_SamEndPoint,
                          string a_SessionID,
                          I2PKeyPair a_KeyPair,
                          string a_Options)
        {
            base.SamBeginControlSession(a_SamEndPoint, a_SessionID, a_KeyPair, a_Options, "STREAM");

            if (BridgeStatus.Disconnected != f_Status)
            {
                f_SamEndPoint = a_SamEndPoint;
                f_SessionID = a_SessionID;

                if (BridgeStatus.Connected == f_Status)
                {
                    FinishInitialization();
                }
            }
        }

        public void ContinueStart()
        {
            base.SamContinueBeginControlSession();

            if (BridgeStatus.Connected == f_Status)
            {
                FinishInitialization();
            }
        }

        private void FinishInitialization()
        {
            if (false == CreateNewAcceptor())
            {
                base.Close();
            }
        }

        public override void Close()
        {
            if (null != f_AcceptHelper)
            {
                f_AcceptHelper.Close();
                f_AcceptHelper = null;
            }

            base.Close();
        }

        private bool CreateNewAcceptor()
        {
            // This routine does not close the connection
            // or update f_Status.

            TcpClient l_TcpClnt = Connect(f_SamEndPoint);

            if (null == l_TcpClnt)
            {
                return false;
            }

            f_AcceptHelper = SamControlClient.SamCreateClientAndHandleHello(l_TcpClnt);

            if (null == f_AcceptHelper)
            {
                return false;
            }

            string l_Answer = f_AcceptHelper.WriteLineAndWaitForAnswer("STREAM ACCEPT ID=" + f_SessionID +
                                                                       " SILENCE=false");

            if (null == l_Answer ||
                false == l_Answer.Contains("OK"))
            {
                f_AcceptHelper.Close();

                return false;
            }

            return true;
        }

        public ConnectResult BeginConnect(I2PAddress a_Destination,
                                          out I2PConnectCookie a_Cookie)
        {
            // This routine will return ConnectResult.Pending
            // on success.

            TcpClient l_TcpClnt = Connect(f_SamEndPoint);

            if (null == l_TcpClnt)
            {
                Close();
                a_Cookie = null;

                return ConnectResult.CantReachBridge;
            }

            NetworkStreamHelper l_Hlpr = SamControlClient.SamCreateClientAndHandleHello(l_TcpClnt);

            if (null == l_Hlpr)
            {
                Close();
                a_Cookie = null;

                return ConnectResult.BridgeDialogError;
            }

            l_Hlpr.Flush();
            l_Hlpr.WriteLineAsync("STREAM CONNECT ID=" + f_SessionID +
                                  " DESTINATION=" + a_Destination.ToFullString() +
                                  " SILENCE=false");

            // Don't wait for an answer.

            a_Cookie = new I2PConnectCookie(l_Hlpr, a_Destination);

            return ConnectResult.Pending;
        }

        public ConnectResult TryFinishConnect(I2PConnectCookie a_Cookie,
                                              out I2PStream a_Stream)
        {
            string l_Answer = a_Cookie.f_Hlpr.TryReadLine();

            if (null == l_Answer)
            {
                a_Stream = null;

                return ConnectResult.Pending;
            }

            if (false == Util.StartsWith(l_Answer, "STREAM STATUS RESULT="))
            {
                Close();
                a_Cookie.Close();
                a_Stream = null;

                return ConnectResult.BridgeDialogError;
            }

            ConnectResult l_ConnRes;

            switch (l_Answer.Substring(21))
            {
                case "OK":
                    a_Cookie.f_RemoteAddr.SetValid();
                    a_Stream = I2PStream.CreateOutgoing(a_Cookie.f_Hlpr, a_Cookie.f_RemoteAddr);
                    return ConnectResult.Success;

                case "CANT_REACH_PEER":
                    l_ConnRes = ConnectResult.NoResponse;
                    break;

                case "INVALID_KEY":
                    l_ConnRes = ConnectResult.MalformedAddress;
                    break;

                case "INVALID_ID":
                    l_ConnRes = ConnectResult.BridgeDialogError;
                    break;

                case "I2P_ERROR":
                    l_ConnRes = ConnectResult.InternalI2PError;
                    break;

                default:
                    l_ConnRes = ConnectResult.UnknownBridgeError;
                    break;
            }

            a_Cookie.Close();
            a_Stream = null;

            return l_ConnRes;
        }

        public I2PStream TryAccept()
        {
            string l_AddrStr = f_AcceptHelper.TryReadLine();

            if (null == l_AddrStr)
            {
                return null;
            }

            I2PAddress l_RemoteAddr = I2PAddress.TryCreateFromString(l_AddrStr, 0, l_AddrStr.Length);

            if (null == l_RemoteAddr)
            {
                goto ADDRESS_FAIL;
            }

            l_RemoteAddr.SetValid();
            I2PStream l_Stream = I2PStream.CreateIncoming(f_AcceptHelper, l_RemoteAddr);
            f_AcceptHelper = null;

            if (false == CreateNewAcceptor())
            {
                Close();

                l_Stream.Close();

                return null;
            }

            return l_Stream;

            ADDRESS_FAIL:

            f_AcceptHelper.Close();
            f_AcceptHelper = null;

            if (false == CreateNewAcceptor())
            {
                Close();
            }

            return null;
        }
    }
}

namespace NB.I2P
{
    enum ConnectResult : byte
    {
        None,

        // Used by I2P.cs
        Pending,
        Success,
        MalformedAddress,
        NoResponse,
        DestinationBanned,
        InternalI2PError,
        CantReachBridge,
        BridgeDialogError,
        ConnectionToBridgeNotReady,
        UnknownBridgeError,

        // Used by Network.cs
        HandshakeError,
        Shutdown,
        UserBanned,
    }

    class I2PConnectCookie
    {
        public readonly NetworkStreamHelper f_Hlpr;
        public readonly I2PAddress f_RemoteAddr;

        public I2PConnectCookie(NetworkStreamHelper a_Hlpr,
                                I2PAddress a_RemoteAddr)
        {
            f_Hlpr = a_Hlpr;
            f_RemoteAddr = a_RemoteAddr;
        }

        public void Close()
        {
            f_Hlpr.Close();
        }
    }

    class I2PKeyPair
    {
        public const ushort c_Magic = 0x4A7B;

        public const ushort c_MinimumPubAndPrivKeyByteSize = (I2PAddress.c_MinimumByteSize + 1);
        public const ushort c_MinimumPubAndPrivKeyBase64Len = (I2PAddress.c_MinimumBase64CharLen + 4);

        private readonly Time f_CreationTime;
        private readonly SensitiveBuffer f_PrivateKey;
        public readonly I2PAddress f_PublicKey;

        private I2PKeyPair(Time a_CreationDate,
                           SensitiveBuffer a_PrivateKey,
                           I2PAddress a_PublicKey)
        {
            f_CreationTime = a_CreationDate;
            f_PrivateKey = a_PrivateKey;
            f_PublicKey = a_PublicKey;
        }

        public static I2PKeyPair FromPrivateKey(byte[] a_PrivateKey)
        {
            // Return can be null on error. On success this
            // instance takes ownership of a_PrivateKey.

            if (c_MinimumPubAndPrivKeyByteSize > a_PrivateKey.Length)
            {
                return null;
            }

            int l_AddrSize;

            if (false == I2PAddress.TryDetermineAddressSize(a_PrivateKey, 0, a_PrivateKey.Length, out l_AddrSize))
            {
                return null;
            }

            I2PAddress l_PubKey = I2PAddress.TryCreateFromCopyOfBytes(a_PrivateKey, 0, l_AddrSize);

            if (null == l_PubKey)
            {
                return null;
            }

            return new I2PKeyPair(Time.SysTimeLastMidnight(), SensitiveBuffer.FromByteArray(a_PrivateKey), l_PubKey);
        }

        public static I2PKeyPair FromBuffer(BufferIO a_Buffer)
        {
            // Return can be null on error.

            if (c_Magic != a_Buffer.ReadUInt16LE())
            {
                return null;
            }

            Time l_CreationDate = a_Buffer.ReadFormerDateTime();
            SensitiveBuffer l_PrivateKey = SensitiveBuffer.FromBuffer(a_Buffer);

            if (null == l_PrivateKey ||
                c_MinimumPubAndPrivKeyByteSize > l_PrivateKey.p_Length)
            {
                return null;
            }

            I2PAddress l_PubKey;

            byte[] l_RawKey = l_PrivateKey.BeginExpose();
            {
                int l_AddrSize;

                if (false == I2PAddress.TryDetermineAddressSize(l_RawKey, 0, l_RawKey.Length, out l_AddrSize))
                {
                    l_PubKey = null;
                }
                else
                {
                    l_PubKey = I2PAddress.TryCreateFromCopyOfBytes(l_RawKey, 0, l_AddrSize);
                }
            }
            l_PrivateKey.EndExpose();

            if (null == l_PubKey)
            {
                return null;
            }

            return new I2PKeyPair(l_CreationDate, l_PrivateKey, l_PubKey);
        }

        public void ToBuffer(BufferIO a_Buffer)
        {
            a_Buffer.WriteUInt16LE(c_Magic);
            a_Buffer.WriteTime(f_CreationTime);
            f_PrivateKey.ToBuffer(a_Buffer);
        }

        public bool ValueEquals(I2PKeyPair a_Other,
                                bool a_IncludeTime)
        {
            if (true == a_IncludeTime &&
                this.f_CreationTime != a_Other.f_CreationTime)
            {
                return false;
            }

            return this.f_PublicKey.ValueEquals(a_Other.f_PublicKey);
        }

        public void Release()
        {
            f_PrivateKey.Release();
        }

        public string PrivateKeyAsBase64Str()
        {
            string l_ResStr;

            byte[] l_RawKey = f_PrivateKey.BeginExpose();
            {
                l_ResStr = Base64.ToBase64(l_RawKey, 0, (int)f_PrivateKey.p_Length);
            }
            f_PrivateKey.EndExpose();

            return l_ResStr;
        }

        public long CalcAgeInDays(Time a_RefTime)
        {
            // The return value will never be negative.

            if (f_CreationTime >= a_RefTime)
            {
                return 0;
            }

            return a_RefTime.FullDaysSince(f_CreationTime);
        }
    }

    class I2PAddress : IEquatable<I2PAddress>, IComparable, IComparable<I2PAddress>
    {
        public const ushort c_MinimumByteSize = 387;
        private const ushort c_MinimumByteSizeCertType5 = (c_MinimumByteSize + sizeof(uint));
        // The maximum size is not fix, it can be lifted future I2P
        // versions.
        public const ushort c_MaximumByteSize = 775;

        public const ushort c_MinimumBase64CharLen = 516;
        public const ushort c_MaximumBase64CharLen = 1036;

        private const ushort c_IsValidFlag = 0x8000;

        private static SipHash.Key s_SipHashKey;
        // We can't use I2PAddress as value directly, since
        // the destructor wouldn't work in this case (we
        // would always hold at least one reference), so we
        // use InternalAddress instead.
        private static readonly MultiMap<int, InternalAddress> s_Interns;
        private static uint s_InternsRevision;
        // The number of interened addresses.
        private static uint s_InternsCount;

        // All equal I2PAddress instances will point to the same
        // InternalAddress instance.
        private readonly InternalAddress f_InternalAddress;

        public bool p_IsClassicAddress
        {
            get { return (SigningKeyTypes.DSA_SHA1_160 == p_SigningKeyType); }
        }

        public SigningKeyTypes p_SigningKeyType
        {
            get { return f_InternalAddress.f_SigningKeyType; }
        }

        public bool p_IsValid
        {
            get { return f_InternalAddress.p_IsValid; }
        }

        static I2PAddress()
        {
            SipHash.Key.CreateRandom(out s_SipHashKey);
            MultiMap<int, InternalAddress>.Create(out s_Interns);
        }

        private I2PAddress(InternalAddress a_InternalAddress)
        {
            f_InternalAddress = a_InternalAddress;
        }

        ~I2PAddress()
        {
            Deref(f_InternalAddress);
        }

        private static I2PAddress TryCreateFromBytesValidated(byte[] a_Bytes)
        {
            // a_Bytes must be valid. The new instance takes ownership
            // of a_Bytes.

            return new I2PAddress(Intern(a_Bytes));
        }

        public static I2PAddress TryCreateFromBytes(byte[] a_Bytes)
        {
            // Returns null on error. Otherwise the new instance takes
            // ownership of a_Bytes.

            if (false == InternalAddress.VerifyAddressBytes(a_Bytes))
            {
                return null;
            }

            return TryCreateFromBytesValidated(a_Bytes);
        }

        public static I2PAddress TryCreateFromCopyOfBytes(byte[] a_Bytes,
                                                          int a_Idx,
                                                          int a_Len)
        {
            // Returns null on error. Unlike TryCreateFromBytes(), this
            // routine does not take ownership of a_Bytes.

            if (false == InternalAddress.VerifyAddressBytes(a_Bytes, a_Idx, a_Len))
            {
                return null;
            }

            byte[] l_AddrBytesCopy = new byte[a_Len];
            Buffer.BlockCopy(a_Bytes, a_Idx, l_AddrBytesCopy, 0, a_Len);

            return TryCreateFromBytesValidated(l_AddrBytesCopy);
        }

        public static bool TryDetermineAddressSize(byte[] a_Bytes,
                                                   int a_Idx,
                                                   int a_MaxLen,
                                                   out int a_AddrSize)
        {
            return InternalAddress.TryDetermineAddressSize(a_Bytes, a_Idx, a_MaxLen, out a_AddrSize);
        }

        public static I2PAddress TryCreateFromString(string a_Base64Str,
                                                     int a_Idx,
                                                     int a_Len)
        {
            // Returns null on error.

            if (c_MinimumBase64CharLen > a_Len ||
                c_MaximumBase64CharLen < a_Len ||
                0 != (a_Len % 4))
            {
                return null;
            }

            byte[] l_AddrBytes = Base64.FromBase64(a_Base64Str, a_Idx, a_Len);

            if (null == l_AddrBytes)
            {
                return null;
            }

            return TryCreateFromBytes(l_AddrBytes);
        }

        public static BufferIO.LoadResult TryCreateFromBufferRaw(BufferIO a_Buffer,
                                                                 out I2PAddress a_Addr)
        {
            // On failure a_Addr will be null on return in any case.
            // The buffer may be at the end.

            uint l_Remaining = a_Buffer.GetRemaining();

            if (c_MinimumByteSize > l_Remaining)
            {
                return HandleStreamError(false, out a_Addr);
            }

            uint l_StartPos = a_Buffer.p_Position;
            a_Buffer.SetPosition(l_StartPos + c_MinimumByteSize - sizeof(ushort));
            uint l_AddrSize = ((uint)c_MinimumByteSize + Util.ByteSwapUInt16(a_Buffer.ReadUInt16LE()));

            if (l_AddrSize > l_Remaining)
            {
                return HandleStreamError(false, out a_Addr);
            }

            if (c_MaximumByteSize < l_AddrSize)
            {
                a_Buffer.JumpAhead(l_StartPos + l_AddrSize);

                return HandleStreamError(true, out a_Addr);
            }

            a_Buffer.SetPosition(l_StartPos);
            byte[] l_Bytes = a_Buffer.ReadBytes(l_AddrSize);
            I2PAddress l_Addr = TryCreateFromBytes(l_Bytes);

            if (null == l_Addr)
            {
                return HandleStreamError(true, out a_Addr);
            }

            a_Addr = l_Addr;

            return BufferIO.LoadResult.Success;
        }

        public static BufferIO.LoadResult TryCreateFromBufferOwnFormat(BufferIO a_Buffer,
                                                                       out I2PAddress a_Addr)
        {
            // On failure a_Addr will be null on return in any case.
            // The buffer may be at the end.

            uint l_Remaining = a_Buffer.GetRemaining();

            if (sizeof(ushort) >= l_Remaining)
            {
                return HandleStreamError(false, out a_Addr);
            }

            uint l_AddrSize = a_Buffer.ReadUInt16LE();
            bool l_IsValid = (0 != (l_AddrSize & c_IsValidFlag));
            l_AddrSize = (l_AddrSize & ~(uint)c_IsValidFlag);
            l_Remaining -= sizeof(ushort);

            if (l_AddrSize > l_Remaining)
            {
                return HandleStreamError(false, out a_Addr);
            }

            if (c_MinimumByteSize > l_AddrSize ||
                c_MaximumByteSize < l_AddrSize)
            {
                // From the size we know that we won't accept this
                // address.

                a_Buffer.JumpAhead(l_AddrSize);

                return HandleStreamError(true, out a_Addr);
            }

            byte[] l_Bytes = a_Buffer.ReadBytes(l_AddrSize);
            I2PAddress l_Addr = TryCreateFromBytes(l_Bytes);

            if (null == l_Addr)
            {
                // Address not accepted.

                return HandleStreamError(true, out a_Addr);
            }

            if (true == l_IsValid)
            {
                l_Addr.SetValid();
            }

            a_Addr = l_Addr;

            return BufferIO.LoadResult.Success;
        }

        public static BufferIO.LoadResult TryCreateFromBufferLegayFormat(BufferIO a_Buffer,
                                                                         bool a_OldAddrFormat,
                                                                         out I2PAddress a_Addr)
        {
            // On failure a_Addr will be null on return in any case.
            // The buffer may be at the end.

            if (false == a_OldAddrFormat)
            {
                return TryCreateFromBufferOwnFormat(a_Buffer, out a_Addr);
            }

            const ushort lc_MinSize = (sizeof(byte) + c_MinimumByteSize);
            uint l_Remaining = a_Buffer.GetRemaining();

            if (lc_MinSize > l_Remaining ||
                0 != a_Buffer.ReadByte())
            {
                return HandleStreamError(false, out a_Addr);
            }

            ulong l_Pos = (a_Buffer.p_Position + c_MinimumByteSize);
            BufferIO.LoadResult l_Res = TryCreateFromBufferRaw(a_Buffer, out a_Addr);

            if (BufferIO.LoadResult.Success == l_Res &&
                a_Buffer.p_Position != l_Pos)
            {
                // The read went longer than expected.

                return HandleStreamError(false, out a_Addr);
            }

            return l_Res;
        }

        private static BufferIO.LoadResult HandleStreamError(bool a_Recoverable,
                                                             out I2PAddress a_Addr)
        {
            a_Addr = null;

            return (false == a_Recoverable ? BufferIO.LoadResult.Corruption : BufferIO.LoadResult.InvalidDataButBufferPosIsStillGood);
        }

        public void ToBufferRaw(BufferIO a_Buffer)
        {
            a_Buffer.WriteBytes(f_InternalAddress.f_Bytes);
        }

        public void ToBufferOwnFormat(BufferIO a_Buffer)
        {
            uint l_AddrSize = Util.GetArrayLength(f_InternalAddress.f_Bytes);

            if (true == p_IsValid)
            {
                l_AddrSize |= c_IsValidFlag;
            }

            a_Buffer.WriteUInt16LE((ushort)l_AddrSize);
            a_Buffer.WriteBytes(f_InternalAddress.f_Bytes);
        }

        public override bool Equals(object a_Other)
        {
            // Warn us if someone operates non-generic.
            throw (new ApplicationException());
        }

        public bool Equals(I2PAddress a_Other)
        {
            return object.ReferenceEquals(this.f_InternalAddress, a_Other.f_InternalAddress);
        }

        public override int GetHashCode()
        {
            return f_InternalAddress.GetHashCode();
        }

        public bool ValueEquals(I2PAddress a_Other)
        {
            return this.Equals(a_Other);
        }

        public int CompareTo(object a_Other)
        {
            return CompareTo((I2PAddress)a_Other);
        }

        public int CompareTo(I2PAddress a_Other)
        {
            if (this.f_InternalAddress == a_Other.f_InternalAddress)
            {
                return 0;
            }

            return Buffer2.BlockCompare(this.f_InternalAddress.f_Bytes, a_Other.f_InternalAddress.f_Bytes);
        }

        public static I2PAddress BuildRandom()
        {
            byte[] l_Bytes = Util.GetRandomBytes(c_MinimumByteSize);
            l_Bytes[c_MinimumByteSize - 3] = 0;
            l_Bytes[c_MinimumByteSize - 2] = 0;
            l_Bytes[c_MinimumByteSize - 1] = 0;

            return TryCreateFromBytes(l_Bytes);
        }

        public uint GetCompatibleHashCode(bool a_ForSecondiMulePort)
        {
            return (false == a_ForSecondiMulePort ? f_InternalAddress.f_iMuleCompatibleHashCode : Util.ReadUInt32LE(f_InternalAddress.f_Bytes, 0));
        }

        public string ToShortString()
        {
            return f_InternalAddress.f_AbbrevBase64;
        }

        public string ToFullString()
        {
            return Base64.ToBase64(f_InternalAddress.f_Bytes);
        }

        public static uint GetInternedAddressCount()
        {
            // We do this lockless (since 32 bit).

            return Util.VolatileRead(ref s_InternsCount);
        }

        private static InternalAddress Intern(byte[] a_Bytes)
        {
            // First marshal any address that can be represented in
            // the classic form to the classic form, so that we only
            // have one legal representation of each address.

            if (c_MinimumByteSizeCertType5 == Util.GetArrayLength(a_Bytes) &&
                (byte)SigningKeyTypes.DSA_SHA1_160 == a_Bytes[c_MinimumByteSize + 1])
            {
                Array.Resize(ref a_Bytes, c_MinimumByteSize);
                a_Bytes[c_MinimumByteSize - 3] = (byte)CertificateType.CERTIFICATE_TYPE_NULL;
                a_Bytes[c_MinimumByteSize - 1] = 0;
            }

            int l_HashCode = InternalAddress.CalcInternalHashCode(a_Bytes);
            InternalAddress l_EffAddr;

            Enter();
            {
                // Check if we've got this address already.

                l_EffAddr = FindInternalAddress(a_Bytes, l_HashCode);

                if (null == l_EffAddr)
                {
                    uint l_StoredRevision = s_InternsRevision;

                    Leave();

                    InternalAddress l_NewAddr = new InternalAddress(a_Bytes, l_HashCode);

                    Enter();

                    if (l_StoredRevision != s_InternsRevision)
                    {
                        // Recheck.

                        l_EffAddr = FindInternalAddress(a_Bytes, l_HashCode);
                    }

                    if (null == l_EffAddr)
                    {
                        // Still not there.

                        bool l_Added = s_Interns.Add(l_HashCode, l_NewAddr);
                        s_InternsRevision++;
                        s_InternsCount++;

                        l_EffAddr = l_NewAddr;
                    }
                }

                l_EffAddr.IncRefCount();
            }
            Leave();

            return l_EffAddr;
        }

        private static void Deref(InternalAddress a_InternalAddress)
        {
            Enter();
            {
                a_InternalAddress.DefRefCount();

                if (false == a_InternalAddress.IsReferenced())
                {
                    bool l_Removed = s_Interns.Remove(a_InternalAddress.f_HashCode, a_InternalAddress);
                    s_InternsRevision++;
                    s_InternsCount--;
                }
            }
            Leave();
        }

        private static InternalAddress FindInternalAddress(byte[] a_Bytes,
                                                           int a_HashCode)
        {
            // The caller must hold the lock. Return can be null, if no
            // such address exists.

            MultiMap<int, InternalAddress>.Enumerator l_Enum;
            if (true == s_Interns.CreateEnumerator(a_HashCode, out l_Enum))
            {
                InternalAddress l_Addr;
                while (true == l_Enum.GetNext(out l_Addr))
                {
                    if (true == l_Addr.HasSameBytes(a_Bytes))
                    {
                        return l_Addr;
                    }
                }
            }

            return null;
        }

        public void SetValid()
        {
            f_InternalAddress.SetValid();
        }

        private static void Enter()
        {
            Util.MonitorEnter(typeof(InternalAddress));
        }

        private static void Leave()
        {
            Util.MonitorLeave(typeof(InternalAddress));
        }

        private class InternalAddress : IEquatable<InternalAddress>
        {
            private const byte c_ShortAddressStrPartLen = 12;

            public readonly byte[] f_Bytes;
            public readonly SigningKeyTypes f_SigningKeyType;
            public readonly string f_AbbrevBase64;

            // We use f_HashCode only internally for our dictionaries,
            // so any algo could be used.
            public readonly int f_HashCode;
            public readonly uint f_iMuleCompatibleHashCode;

            private uint f_RefCount;

            private bool f_IsValid;

            public bool p_IsValid
            {
                get { return f_IsValid; }
            }

            public InternalAddress(byte[] a_Bytes,
                                   int a_HashCode)
            {
                // Should only be called from I2PAddress.Intern().
                // a_Bytes must match a_HashCode.

                f_Bytes = a_Bytes;
                f_SigningKeyType = (SigningKeyTypes)(c_MinimumByteSizeCertType5 > Util.GetArrayLength(a_Bytes) ? (byte)0 : a_Bytes[c_MinimumByteSizeCertType5 - 3]);
                f_AbbrevBase64 = Util.ToShortString(Base64.ToBase64(a_Bytes), c_ShortAddressStrPartLen);

                f_HashCode = a_HashCode;
                f_iMuleCompatibleHashCode = CalcCompatibleHashCode(f_Bytes);
            }

            public override bool Equals(object obj)
            {
                // Warn us if someone operates non-generic.
                throw (new ApplicationException());
            }

            public bool Equals(InternalAddress a_Other)
            {
                return (this.f_HashCode == a_Other.f_HashCode &&
                        true == this.HasSameBytes(a_Other.f_Bytes));
            }

            public override int GetHashCode()
            {
                return f_HashCode;
            }

            public bool HasSameBytes(byte[] a_Bytes)
            {
                return Buffer2.ValueEquals(this.f_Bytes, a_Bytes);
            }

            public void IncRefCount()
            {
                f_RefCount++;
            }

            public void DefRefCount()
            {
                f_RefCount--;
            }

            public bool IsReferenced()
            {
                return (0 != f_RefCount);
            }

            public static int CalcInternalHashCode(byte[] a_Bytes)
            {
                return SipHash.CalculateHashCode(ref I2PAddress.s_SipHashKey, a_Bytes, 0, a_Bytes.Length);
            }

            private static uint CalcCompatibleHashCode(byte[] a_Bytes)
            {
                // This is the checksum iMule uses. It never fills
                // the uint's high bits (max possible value is
                // 0x1817D). It can be forged too easily.

                uint l_Num = 0;

                foreach (byte l_Byte in a_Bytes)
                {
                    l_Num += l_Byte;
                }

                return l_Num;
            }

            public void SetValid()
            {
                f_IsValid = true;
            }

            public static bool VerifyAddressBytes(byte[] a_Bytes)
            {
                return VerifyAddressBytes(a_Bytes, 0, a_Bytes.Length);
            }

            public static bool VerifyAddressBytes(byte[] a_Bytes,
                                                  int a_Idx,
                                                  int a_Len)
            {
                // Verification will fail if the address is not of a
                // known type.

                int l_AddrSize;

                if (c_MaximumByteSize < a_Len ||
                    false == TryDetermineAddressSize(a_Bytes, a_Idx, a_Len, out l_AddrSize) ||
                    l_AddrSize != a_Len)
                {
                    return false;
                }

                CertificateType l_CertType = (CertificateType)a_Bytes[a_Idx + c_MinimumByteSize - 3];

                if (CertificateType.CERTIFICATE_TYPE_NULL == l_CertType)
                {
                    if (c_MinimumByteSize != a_Len)
                    {
                        return false;
                    }

                    return true;
                }

                if (CertificateType.CERTIFICATE_TYPE_KEY != l_CertType)
                {
                    return false;
                }

                if (c_MinimumByteSizeCertType5 > a_Len)
                {
                    return false;
                }

                uint l_CryptoAndSigPubKeyType = Util.ByteSwapUInt32(Util.ReadUInt32LE(a_Bytes, a_Idx + c_MinimumByteSizeCertType5 - sizeof(uint)));
                uint l_ExpectedSize;

                return (true == TryDetermineCertType5AddressSize(l_CryptoAndSigPubKeyType, out l_ExpectedSize) &&
                        l_ExpectedSize == a_Len);
            }

            public static bool TryDetermineAddressSize(byte[] a_Bytes,
                                                       int a_Idx,
                                                       int a_MaxLen,
                                                       out int a_AddrSize)
            {
                // Determines the addresses size from the certificate. On
                // successful return, a_AddrSize can exceed c_MaximumByteSize.
                // This routine is able to process any kind of future
                // address type.

                if (c_MinimumByteSize > a_MaxLen)
                {
                    a_AddrSize = 0;

                    return false;
                }

                ushort l_CertPayloadSize = Util.ByteSwapUInt16(Util.ReadUInt16LE(a_Bytes, a_Idx + c_MinimumByteSize - 2));
                a_AddrSize = (c_MinimumByteSize + l_CertPayloadSize);

                return true;
            }

            private static bool TryDetermineCertType5AddressSize(uint a_CryptoAndSigPubKeyType,
                                                                 out uint a_ByteSize)
            {
                // a_CryptoAndSigPubKeyType is expected to contain the
                // signing public key type in the low ushort and the
                // crypto public key type in the high ushort. Returns
                // zero if the input is not recognized.

                ushort l_CryptoPubKeyType = (ushort)a_CryptoAndSigPubKeyType;
                ushort l_SigPubKeyType = (ushort)(a_CryptoAndSigPubKeyType >> 16);

                if (0 != l_CryptoPubKeyType ||
                    byte.MaxValue < l_SigPubKeyType)
                {
                    goto FAIL;
                }

                uint l_CertPayload;

                switch ((SigningKeyTypes)l_SigPubKeyType)
                {
                    case SigningKeyTypes.DSA_SHA1_160:
                    case SigningKeyTypes.ECDSA_SHA256_P256:
                    case SigningKeyTypes.ECDSA_SHA384_P384:
                    case SigningKeyTypes.EdDSA_SHA512_Ed25519:
                        l_CertPayload = 0;
                        break;

                    case SigningKeyTypes.ECDSA_SHA512_P521:
                        l_CertPayload = 4;
                        break;

                    case SigningKeyTypes.RSA_SHA256_2048:
                        l_CertPayload = 128;
                        break;

                    case SigningKeyTypes.RSA_SHA384_3072:
                        l_CertPayload = 256;
                        break;

                    case SigningKeyTypes.RSA_SHA512_4096:
                        l_CertPayload = 384;
                        break;

                    default:
                        goto FAIL;
                }

                a_ByteSize = (c_MinimumByteSizeCertType5 + l_CertPayload);

                return true;

                FAIL:

                a_ByteSize = 0;

                return false;
            }
        }

        public enum SigningKeyTypes : byte
        {
            DSA_SHA1_160 = 0,
            ECDSA_SHA256_P256 = 1,
            ECDSA_SHA384_P384 = 2,
            ECDSA_SHA512_P521 = 3,
            RSA_SHA256_2048 = 4,
            RSA_SHA384_3072 = 5,
            RSA_SHA512_4096 = 6,
            EdDSA_SHA512_Ed25519 = 7,
        }

        private enum CertificateType : byte
        {
            CERTIFICATE_TYPE_NULL = 0,
            CERTIFICATE_TYPE_KEY = 5,
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct AdvancedTunnelSettings
    {
        public const byte c_MaxTunnels = 6;
        public const byte c_MaxHops = 6;

        public readonly bool f_Outgoing;
        public readonly byte f_Tunnels;
        public readonly byte f_BackupTunnels;
        public readonly byte f_MinHops;
        public readonly sbyte f_RandomHops;
        public readonly byte f_IPRestriction;

        public AdvancedTunnelSettings(bool a_Outgoing,
                                      uint a_Tunnels,
                                      uint a_MinHops,
                                      int a_RandomHops) :
                                 this(a_Outgoing, a_Tunnels, 0, a_MinHops, a_RandomHops, 2)
        {
        }

        public AdvancedTunnelSettings(bool a_Outgoing,
                                      uint a_Tunnels,
                                      uint a_BackupTunnels,
                                      uint a_MinHops,
                                      int a_RandomHops,
                                      uint a_IPRestriction)
        {
            f_Outgoing = a_Outgoing;
            f_Tunnels = (byte)a_Tunnels;
            f_BackupTunnels = (byte)a_BackupTunnels;
            f_MinHops = (byte)a_MinHops;
            f_RandomHops = (sbyte)a_RandomHops;
            f_IPRestriction = (byte)a_IPRestriction;
        }

        public void ToString(StringBuilder a_Bld)
        {
            string l_Prexfix = (false == f_Outgoing ? " inbound" : " outbound");

            a_Bld.Append(l_Prexfix + ".quantity=" + Util.NumberToInvString(f_Tunnels));

            if (0 != f_BackupTunnels)
            {
                a_Bld.Append(l_Prexfix + ".backupQuantity=" + Util.NumberToInvString(f_BackupTunnels));
            }

            a_Bld.Append(l_Prexfix + ".length=" + Util.NumberToInvString(f_MinHops));

            if (0 != f_RandomHops)
            {
                a_Bld.Append(l_Prexfix + ".lengthVariance=" + Util.NumberToInvString(f_RandomHops));
            }

            a_Bld.Append(l_Prexfix + ".allowZeroHop=false");

            // The manual says about IPRestriction: Number of IP
            // bytes to match to determine if two routers should
            // not be in the same tunnel. 0 to disable.
            a_Bld.Append(l_Prexfix + ".IPRestriction=" + Util.NumberToInvString(f_IPRestriction));
        }
    }

    class I2PStream
    {
        private const int c_ConnBufSize = (80 * 1024);

        private readonly NetworkStreamHelper f_StrmHlpr;
        public readonly bool f_Outgoing;
        public I2PAddress f_RemoteAddr;

        public bool p_Outgoing
        {
            get { return f_Outgoing; }
        }

        public I2PAddress p_RemoteAddr
        {
            get { return f_RemoteAddr; }
        }

        public bool p_IsAlive
        {
            get { return f_StrmHlpr.p_IsAlive; }
        }

        public bool p_IsBusyWriting
        {
            get { return f_StrmHlpr.p_IsBusyWriting; }
        }

        public uint p_AvailableBytes
        {
            get { return f_StrmHlpr.p_AvailableBytes; }
        }

        private I2PStream(NetworkStreamHelper a_StrmHlpr,
                                bool a_Outgoing,
                                I2PAddress a_RemoteAddr)
        {
            f_StrmHlpr = a_StrmHlpr;
            f_Outgoing = a_Outgoing;
            f_RemoteAddr = a_RemoteAddr;

            f_StrmHlpr.AdjustOSBufferSizes(c_ConnBufSize, c_ConnBufSize);
            f_StrmHlpr.StopUsingOwnBuffer();
        }

        public static I2PStream CreateOutgoing(NetworkStreamHelper a_StrmHlpr,
                                               I2PAddress a_RemoteAddr)
        {
            return new I2PStream(a_StrmHlpr, true, a_RemoteAddr);
        }

        public static I2PStream CreateIncoming(NetworkStreamHelper a_StrmHlpr,
                                               I2PAddress a_RemoteAddr)
        {
            return new I2PStream(a_StrmHlpr, false, a_RemoteAddr);
        }

        public void Close()
        {
            f_StrmHlpr.Close();
        }

        public bool WriteBytesAsync(ref ByteArray a_Bytes)
        {
            return f_StrmHlpr.WriteBytesAsync(ref a_Bytes);
        }

        public int ReadAvailableBytes(ref ByteArray a_Bytes)
        {
            return f_StrmHlpr.ReadAvailableBytes(ref a_Bytes);
        }
    }

    enum BridgeStatus : byte
    {
        Disconnected = 0,
        TryingToConnect = 1,
        Connected = 2,
    }

    class SamBridgeManager
    {
        // The user should check p_Status after the following
        // routines were invoked: Start(), ContinueStart(),
        // IsSane().

        public const ushort c_MaxDatagramPayloadOutgoing = Sam3DatagramControlClient.c_MaxDatagramPayloadOutgoing;
        private const byte c_MaxDatagramReceiveRetry = 50;
        public const ushort c_MaxRepairDatagramConnectionSecs = 60;

        public delegate void NotifyBannedDestHandler(I2PAddress a_BannedAddr);

        private readonly Dictionary<I2PAddress, I2PAddress> f_BannedAddresses;

        private uint f_MalformedDatagrams;
        private ulong f_WastedDatagramBandwidth;
        private uint f_BlockedIncomingConnections;

        private BridgeStatus f_Status;
        private uint f_ConnGeneration;
        private Sam3DatagramControlClient f_DatagramClient;
        private Sam3StreamControlClient f_StreamClient;

        // In tests the datagram connection sometimes broke, while
        // the stream connection was still good. The cause seems to be
        // outside the scope of this application. Our workaround is
        // to try for one minute to get a new datagram connection up
        // or, if it fails, to close the stream connection as well.
        DatagramConnectionRepairContext f_DgConnRepairCntxt;

        public event NotifyBannedDestHandler e_AddressBanned;

        public int p_BannedDestinations
        {
            get { return f_BannedAddresses.Count; }
        }

        public BridgeStatus p_Status
        {
            get { return f_Status; }
        }

        public uint p_ConnGeneration
        {
            // Returns zero if there is no connection.
            get { return (BridgeStatus.Disconnected == f_Status ? 0 : f_ConnGeneration); }
        }

        public I2PKeyPair p_DatagramKeyPair
        {
            get { return f_DgConnRepairCntxt.p_DatagramKeyPair; }
        }

        public I2PKeyPair p_StreamKeyPair
        {
            get { return f_StreamClient.p_KeyPair; }
        }

        public uint p_MalformedDatagrams
        {
            get { return f_MalformedDatagrams; }
        }

        public SamBridgeManager() :
                           base()
        {
            f_BannedAddresses = new Dictionary<I2PAddress, I2PAddress>();
        }

        public void Start(SamConnectionCookie a_Cookie)
        {
            SamBridgeParams l_SamParams = new SamBridgeParams();
            StringBuilder l_Bld = new StringBuilder(1024);

            l_SamParams.ToString(l_Bld);
            AdvancedTunnelSettings l_InDatagramSettings = new AdvancedTunnelSettings(false, a_Cookie.f_InTunnelSettings.f_DatagramQuantity, a_Cookie.f_InTunnelSettings.f_MinHops, a_Cookie.f_InTunnelSettings.f_RandomExtraHops);
            AdvancedTunnelSettings l_OutDatagramSettings = new AdvancedTunnelSettings(true, a_Cookie.f_OutTunnelSettings.f_DatagramQuantity, a_Cookie.f_OutTunnelSettings.f_MinHops, a_Cookie.f_OutTunnelSettings.f_RandomExtraHops);
            l_InDatagramSettings.ToString(l_Bld);
            l_OutDatagramSettings.ToString(l_Bld);
            string l_DatagramOptions = l_Bld.ToString();

            l_Bld.Length = 0;
            l_SamParams.ToString(l_Bld);
            AdvancedTunnelSettings l_InStreamSettings = new AdvancedTunnelSettings(false, a_Cookie.f_InTunnelSettings.f_StreamQuantity, a_Cookie.f_InTunnelSettings.f_MinHops, a_Cookie.f_InTunnelSettings.f_RandomExtraHops);
            AdvancedTunnelSettings l_OutStreamSettings = new AdvancedTunnelSettings(true, a_Cookie.f_OutTunnelSettings.f_StreamQuantity, a_Cookie.f_OutTunnelSettings.f_MinHops, a_Cookie.f_OutTunnelSettings.f_RandomExtraHops);
            l_InStreamSettings.ToString(l_Bld);
            l_OutStreamSettings.ToString(l_Bld);
            string l_StreamOptions = l_Bld.ToString();

            Start(a_Cookie.f_SamEndPoint, a_Cookie.f_ProposedDatagramKeyPair, a_Cookie.f_ProposedStreamKeyPair, l_DatagramOptions, l_StreamOptions);
        }

        private void Start(IPEndPoint a_SamEndPoint,
                           I2PKeyPair a_DatagramKeyPair,
                           I2PKeyPair a_StreamKeyPair,
                           string a_DatagramOptions,
                           string a_StreamOptions)
        {
            // The private key byte arrays can be null.

            if (BridgeStatus.Disconnected != f_Status)
            {
                Util.DebugBreak();

                Shutdown();
            }

            f_DgConnRepairCntxt = new DatagramConnectionRepairContext(a_SamEndPoint, a_DatagramOptions, a_DatagramKeyPair);
            f_DatagramClient = f_DgConnRepairCntxt.CreateAndStartClient();

            if (BridgeStatus.Disconnected == f_DatagramClient.p_Status)
            {
                Shutdown();

                return;
            }

            f_StreamClient = new Sam3StreamControlClient();
            f_StreamClient.Start(a_SamEndPoint,
                                 GetRandomSessionID(Software.c_AppName),
                                 a_StreamKeyPair,
                                 a_StreamOptions + " inbound.nickname=" + Software.c_AppName + "-Streams");

            if (BridgeStatus.Disconnected == f_StreamClient.p_Status)
            {
                Shutdown();

                return;
            }

            f_Status = BridgeStatus.TryingToConnect;
            ++f_ConnGeneration;
        }

        public void ContinueStart()
        {
            if (BridgeStatus.TryingToConnect != f_Status)
            {
                Shutdown();

                return;
            }

            if (BridgeStatus.TryingToConnect == f_DatagramClient.p_Status)
            {
                f_DatagramClient.ContinueStart();
            }

            if (BridgeStatus.TryingToConnect == f_StreamClient.p_Status)
            {
                f_StreamClient.ContinueStart();
            }

            if (BridgeStatus.Connected == f_DatagramClient.p_Status &&
                BridgeStatus.Connected == f_StreamClient.p_Status)
            {
                f_Status = BridgeStatus.Connected;
                // We possibly had a null key pair so far, so now take
                // over the final key pair.
                f_DgConnRepairCntxt.SetKeyPair(f_DatagramClient.p_KeyPair);

                return;
            }

            if (BridgeStatus.Disconnected == f_DatagramClient.p_Status ||
                BridgeStatus.Disconnected == f_StreamClient.p_Status)
            {
                Shutdown();
            }
        }

        public void CheckSanity(Time a_Time)
        {
            if (false == f_DgConnRepairCntxt.f_ConnBroken)
            {
                if (true == f_StreamClient.p_IsAlive)
                {
                    if (true == f_DatagramClient.p_IsAlive)
                    {
                        return;
                    }

                    // We will only enter repair mode if we had
                    // reached full connectivity already.

                    if (BridgeStatus.Connected == f_Status)
                    {
                        // Try to get the datagram connection up again.

                        f_DgConnRepairCntxt.SetBroken(a_Time);
                        f_DatagramClient = f_DgConnRepairCntxt.CreateAndStartClient();

                        if (BridgeStatus.Disconnected != f_DatagramClient.p_Status)
                        {
                            return;
                        }
                    }
                }

                Shutdown();

                return;
            }

            // We are in repair mode.

            if (BridgeStatus.TryingToConnect == f_DatagramClient.p_Status)
            {
                f_DatagramClient.ContinueStart();
            }

            BridgeStatus l_Status = f_DatagramClient.p_Status;

            switch (l_Status)
            {
                case BridgeStatus.Disconnected:
                    // We don't try a second time.
                    Shutdown();
                    break;

                case BridgeStatus.TryingToConnect:
                    if (true == f_DgConnRepairCntxt.f_ConnMustBeRepairedUntil.IsEarlierThan(a_Time))
                    {
                        // We didn't make it in time.

                        Shutdown();
                    }
                    break;

                case BridgeStatus.Connected:
                    f_DgConnRepairCntxt.SetRepaired();
                    break;

                default: throw (new ApplicationException());
            }
        }

        public void Shutdown()
        {
            if (BridgeStatus.Disconnected != f_Status)
            {
                f_Status = BridgeStatus.Disconnected;

                if (null != f_DatagramClient)
                {
                    f_DatagramClient.Close();
                    f_DatagramClient = null;
                }

                if (null != f_StreamClient)
                {
                    f_StreamClient.Close();
                    f_StreamClient = null;
                }

                f_DgConnRepairCntxt = new DatagramConnectionRepairContext();
            }
        }

        public bool IsAddressBanned(I2PAddress a_Addr)
        {
            return f_BannedAddresses.ContainsKey(a_Addr);
        }

        public void ResetBans()
        {
            f_BannedAddresses.Clear();
        }

        public void UnbanDestination(I2PAddress a_Addr)
        {
            bool l_Removed = f_BannedAddresses.Remove(a_Addr);
        }

        public void BanDestination(I2PAddress a_Addr)
        {
            // NOTE: Bans work only against ingoing traffic.

            if (false == IsAddressBanned(a_Addr))
            {
                f_BannedAddresses.Add(a_Addr, a_Addr);

                if (null != e_AddressBanned)
                {
                    e_AddressBanned(a_Addr);
                }
            }
        }

        public ConnectResult BeginConnect(I2PAddress a_Addr,
                                          out I2PConnectCookie a_Cookie)
        {
            // This routine will never return ConnectResult.Success.
            // If a cookie was created the return value is
            // ConnectResult.Pending.

            if (BridgeStatus.Connected != f_StreamClient.p_Status)
            {
                a_Cookie = null;

                return ConnectResult.ConnectionToBridgeNotReady;
            }

            return f_StreamClient.BeginConnect(a_Addr, out a_Cookie);
        }

        public ConnectResult TryFinishConnect(I2PConnectCookie a_Cookie,
                                              out I2PStream a_Strm)
        {
            // This routine will close the cookie unless the
            // return value is ConnectResult.Pending or
            // ConnectResult.Success.

            if (BridgeStatus.Connected != f_StreamClient.p_Status)
            {
                a_Cookie.Close();
                a_Strm = null;

                return ConnectResult.ConnectionToBridgeNotReady;
            }

            return f_StreamClient.TryFinishConnect(a_Cookie, out a_Strm);
        }

        public I2PStream TryAcceptConnection()
        {
            if (BridgeStatus.Connected != f_StreamClient.p_Status)
            {
                return null;
            }

            I2PStream l_Strm = f_StreamClient.TryAccept();

            if (null != l_Strm)
            {
                if (true == IsAddressBanned(l_Strm.f_RemoteAddr))
                {
                    f_BlockedIncomingConnections++;
                }
                else
                {
                        return l_Strm;
                }

                l_Strm.Close();
            }

            return null;
        }

        public bool IsBusySendingDatagrams()
        {
            if (true == f_DgConnRepairCntxt.f_ConnBroken)
            {
                // We are kind of busy.

                return true;
            }

            if (BridgeStatus.Connected != f_DatagramClient.p_Status)
            {
                return false;
            }

            return f_DatagramClient.p_IsBusyWriting;
        }

        public bool SendDatagrams(Packet[] a_Packets,
                                  out uint a_ConsumedBandwidth)
        {
            // We do not include the SAM overhead here.
            a_ConsumedBandwidth = 0;

            if (true == f_DgConnRepairCntxt.f_ConnBroken ||
                BridgeStatus.Connected != f_DatagramClient.p_Status)
            {
                return false;
            }

            foreach (Packet l_Packet in a_Packets)
            {
                uint l_PacketSize = l_Packet.f_TransferSize;

                a_ConsumedBandwidth += l_PacketSize;
            }

            // This includes the SAM overhead (as we see it).
            int l_MaxSamSize = checked((int)a_ConsumedBandwidth + (Sam3DatagramControlClient.c_MaxDatagramOverhead * a_Packets.Length));

            byte[] l_Dest = new byte[l_MaxSamSize];
            int l_Pos = 0;

            foreach (Packet l_Packet in a_Packets)
            {
                ByteArray l_Body;
                l_Packet.BuildByteArrayForSending(0, out l_Body);

                byte[] l_EncodedLine = Sam3DatagramControlClient.FormatDatagramHeader(l_Packet.f_RemoteAddr, l_Body.f_Len);

                Buffer.BlockCopy(l_EncodedLine, 0, l_Dest, l_Pos, l_EncodedLine.Length);
                l_Pos += l_EncodedLine.Length;
                Buffer.BlockCopy(l_Body.f_Arr, l_Body.f_Idx, l_Dest, l_Pos, l_Body.f_Len);
                l_Pos += l_Body.f_Len;
            }

            ByteArray l_Data = new ByteArray(l_Dest, 0, l_Pos);
            return f_DatagramClient.WriteBytesAsync(ref l_Data);
        }

        public Packet TryReceiveDatagram(out uint a_ConsumedBandwidth)
        {
            a_ConsumedBandwidth = 0;

            if (true == f_DgConnRepairCntxt.f_ConnBroken)
            {
                return null;
            }

            for (int i = 0; i < c_MaxDatagramReceiveRetry; i++)
            {
                if (BridgeStatus.Connected != f_DatagramClient.p_Status)
                {
                    break;
                }

                bool l_ShouldTryAgain;
                Packet l_Packet = TryReceiveDatagramCore(ref a_ConsumedBandwidth, out l_ShouldTryAgain);

                if (null != l_Packet)
                {
                    return l_Packet;
                }

                if (false == l_ShouldTryAgain)
                {
                    break;
                }
            }

            return null;
        }

        private Packet TryReceiveDatagramCore(ref uint a_ConsumedBandwidth,
                                              out bool a_ShouldTryAgain)
        {
            // a_ConsumedBandwidth gets incremented if suitable.

            I2PAddress l_Origin;
            byte[] l_Data = f_DatagramClient.ReceiveDatagram(out l_Origin);

            if (null == l_Data)
            {
                a_ShouldTryAgain = false;

                return null;
            }

            // We don't include the SAM overhead here, since
            // it's not outgoing bandwidth.
            uint l_PacketBandwidth = (uint)l_Data.Length;
            a_ConsumedBandwidth += l_PacketBandwidth;

            bool l_IsBanned = IsAddressBanned(l_Origin);

            if (true == l_IsBanned)
            {
                goto FAIL;
            }

            if (2 <= l_Data.Length)
            {
                Protocols l_Protocol = (Protocols)l_Data[0];
                KadOpcodes l_Opcode = (KadOpcodes)l_Data[1];

                if ((Protocols.OP_KADEMLIAHEADER == l_Protocol ||
                     Protocols.OP_KADEMLIAPACKEDPROT == l_Protocol) &&
                    (KadOpcodes.KADEMLIA_BOOTSTRAP_REQ <= l_Opcode &&
                     KadOpcodes.KADEMLIA_PUB_NOTES_REQ >= l_Opcode))
                {
                    ByteArray l_Bytes = new ByteArray(l_Data, 2, l_Data.Length - 2);
                    Packet l_Packet = Packet.InterpretIncoming(l_Origin,
                                                               l_Protocol,
                                                               (byte)l_Opcode,
                                                               ref l_Bytes,
                                                               Sam3DatagramControlClient.c_MaxDatagramPayloadIngoing,
                                                               l_PacketBandwidth);

                    if (null != l_Packet)
                    {
                        a_ShouldTryAgain = false;

                        return l_Packet;
                    }
                }
                else
                {
                    goto FAIL;
                }
            }

            BanDestination(l_Origin);

            FAIL:

            f_MalformedDatagrams++;
            f_WastedDatagramBandwidth += l_PacketBandwidth;
            a_ShouldTryAgain = true;

            return null;
        }

        public static string GetRandomSessionID(string a_Prefix)
        {
            return (a_Prefix + "_" + Base64.ToBase64(Util.GetRandomBytes(6)));
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct DatagramConnectionRepairContext
        {
            private readonly IPEndPoint f_SamEndPoint;
            private readonly string f_DatagramOptions;
            private I2PKeyPair f_DatagramKeyPair;

            public bool f_ConnBroken;
            public Time f_ConnMustBeRepairedUntil;
            public uint f_ConnFailureCount;

            public I2PKeyPair p_DatagramKeyPair
            {
                get { return f_DatagramKeyPair; }
            }

            public DatagramConnectionRepairContext(IPEndPoint a_SamEndPoint,
                                                   string a_DatagramOptions,
                                                   I2PKeyPair a_DatagramKeyPair)
            {
                f_SamEndPoint = a_SamEndPoint;
                f_DatagramOptions = a_DatagramOptions;
                f_DatagramKeyPair = a_DatagramKeyPair;

                f_ConnBroken = false;
                f_ConnMustBeRepairedUntil = new Time();
                f_ConnFailureCount = 0;
            }

            public void SetKeyPair(I2PKeyPair a_DatagramKeyPair)
            {
                f_DatagramKeyPair = a_DatagramKeyPair;
            }

            public bool HasNonzeroKeyPair()
            {
                return (null != f_DatagramKeyPair);
            }

            public Sam3DatagramControlClient CreateAndStartClient()
            {
                Sam3DatagramControlClient l_Client = new Sam3DatagramControlClient();
                l_Client.Start(f_SamEndPoint,
                               GetRandomSessionID(Software.c_AppName),
                               f_DatagramKeyPair,
                               f_DatagramOptions + " inbound.nickname=" + Software.c_AppName + "-Datagrams");

                return l_Client;
            }

            public void SetBroken(Time a_Time)
            {
                f_ConnBroken = true;
                f_ConnMustBeRepairedUntil = a_Time.AddSeconds(c_MaxRepairDatagramConnectionSecs);
                f_ConnFailureCount++;
            }

            public void SetRepaired()
            {
                f_ConnBroken = false;
            }
        }

        public static void SamTest(SamTestContext a_Cntxt)
        {
            SamTestContext.States l_State;
            NetworkStreamHelper l_Hlpr = null;
            Time l_StartTime = Software.sp_MainForm.f_Inst.GetTime();
            Time l_EndTime = l_StartTime.AddMilliseconds(a_Cntxt.f_TimeoutMs);
            TcpClient l_TcpClient = SamControlClient.Connect(a_Cntxt.f_SamEndPoint);

            if (null == l_TcpClient)
            {
                l_State = SamTestContext.States.FailedToConnect;

                goto EXIT;
            }

            l_Hlpr = new NetworkStreamHelper(l_TcpClient);

            if (false == l_Hlpr.WriteLineAsync("HELLO VERSION MIN=3.0 MAX=3.0"))
            {
                l_State = SamTestContext.States.ConnectionError;

                goto EXIT;
            }

            while (true)
            {
                if (false == l_Hlpr.p_IsAlive)
                {
                    l_State = SamTestContext.States.ConnectionError;

                    break;
                }

                string l_Answer = l_Hlpr.TryReadLine();

                if (null != l_Answer)
                {
                    if ("HELLO REPLY RESULT=OK VERSION=3.0" == l_Answer)
                    {
                        l_State = SamTestContext.States.GoodAnswer;
                    }
                    else
                    {
                        l_State = SamTestContext.States.UnexpectedAnswer;
                    }

                    break;
                }

                Time l_Time = Software.sp_MainForm.f_Inst.GetTime();

                if (l_Time >= l_EndTime)
                {
                    l_State = SamTestContext.States.Timeout;

                    break;
                }

                a_Cntxt.SetProgress((ulong)l_Time.MillisecondsSince(l_StartTime));

                Util.MiniSleep();

                if (true == a_Cntxt.p_AbortRequested)
                {
                    l_State = SamTestContext.States.AbortedViaController;

                    break;
                }
            }

            EXIT:

            if (null != l_Hlpr)
            {
                l_Hlpr.Close();
            }

            a_Cntxt.Finish(l_State);
        }

        public class SamTestContext : NB.Gui.ControllerContext
        {
            public readonly IPEndPoint f_SamEndPoint;
            public readonly uint f_TimeoutMs;

            private bool f_AbortRequested;
            private States f_State;

            public bool p_AbortRequested
            {
                get { return f_AbortRequested; }
            }

            public States p_State
            {
                get
                {
                    return f_State;
                }
            }

            public SamTestContext(IPEndPoint a_SamEndPoint,
                                  uint a_TimeoutMs) :
                             base(true)
            {
                f_SamEndPoint = a_SamEndPoint;
                f_TimeoutMs = a_TimeoutMs;
            }

            public override void Abort()
            {
                f_AbortRequested = true;
            }

            public void SetProgress(ulong l_MsElapsed)
            {
                base.SetPercentage(l_MsElapsed, f_TimeoutMs);
            }

            public void Finish(States a_State)
            {
                f_State = a_State;
                base.SetFinishedBase();
            }

            public enum States : byte
            {
                Undefined = 0,
                AbortedViaController = 1,
                FailedToConnect = 2,
                ConnectionError = 3,
                Timeout = 4,
                UnexpectedAnswer = 5,
                GoodAnswer = 6,
            }
        }
    }

    class SamBridgeParams
    {
        public const ushort c_DefaultMTU = 1730;

        public readonly ushort f_ConnectTimeoutSecs;
        public readonly bool f_FlushOnEveryWrite;
        public readonly ushort f_StreamingMTU;

        public SamBridgeParams()
        {
            // Default is five minutes.
            f_ConnectTimeoutSecs = (5 * 60);
            f_FlushOnEveryWrite = true;
            f_StreamingMTU = 0;
        }

        public void ToString(StringBuilder a_Bld)
        {
            a_Bld.Append(" i2cp.messageReliability=None");

            a_Bld.Append(" i2p.streaming.connectTimeout=" +
                         Util.NumberToInvString(f_ConnectTimeoutSecs * 1000));

            if (0 != f_StreamingMTU)
            {
                a_Bld.Append(" i2p.streaming.maxMessageSize=" + Util.NumberToInvString(f_StreamingMTU));
            }

            a_Bld.Append(" sam.forceFlush=" + (false == f_FlushOnEveryWrite ? "false" : "true"));
        }
    }
}

// End of file.
