/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.27-src\Nachtblitz\Kad2.cs
 * PURPOSE: Kademlia related code (2)
 * VERSION: 1.4.27
 * DATE: March 23, 2016
 * COPYRIGHT: Copyright (C) 2012-2016 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *            Some portions of this file are loosely based on the
 *            GPLed code of iMule v1.4.5.
 *            Copyright (C) 2004-2006 iMule Project.
 *            Copyright (C) 2004-2006 Angel Vidal (Kry).
 *            Copyright (C) 2003 Barry Dunne.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Text;
using NB.Gui;
using NB.I2P;
using Support;
using Support.HashFunctions;

namespace NB.Network.eD2k
{
    enum KadOpcodes : byte
    {
        NONE,

        KADEMLIA_BOOTSTRAP_REQ = 1,
        KADEMLIA_BOOTSTRAP_RES = 2,

        KADEMLIA_HELLO_REQ = 3,
        KADEMLIA_HELLO_RES = 4,

        KADEMLIA_REQ = 5,
        KADEMLIA_RES = 6,

        KADEMLIA_SEARCH_REQ = 7,
        KADEMLIA_SEARCH_RES = 8,

        KADEMLIA_SRC_NOTES_REQ = 9,

        KADEMLIA_PUBLISH_REQ = 10,
        KADEMLIA_PUBLISH_RES = 11,

        KADEMLIA_PUB_NOTES_REQ = 12,
    }

    enum ContactVitality : byte
    {
        Alive = 0,
        NeedAnswer1 = 1,
        NeedAnswer2 = 2,
        NeedAnswer3 = 3,
        Dead = 4,
    }

    class Contact
    {
        public const ushort c_MinimumSize = (ushort)(UInt128.c_ByteSize + (2 * I2PAddress.c_MinimumByteSize) + sizeof(byte));
        public const ushort c_MaximumSize = (ushort)(UInt128.c_ByteSize + (2 * I2PAddress.c_MaximumByteSize) + sizeof(byte));

        public readonly ClientID f_ClientID;
        public readonly Distance f_Distance;

        private ContactVitality f_Vitality;
        private I2PAddress f_DatagramAddr;
        private I2PAddress f_StreamAddr;

        // Contacts that are not in the zone are called
        // temporary contacts. They are mostly required
        // during kad searched.
        private bool f_InZone;
        private Time f_ExpectPongTime;
        private Time f_LastActivityTime;
        private Time f_PinTime;
        private bool f_Verified;

        public I2PAddress p_DatagramAddr
        {
            get { return f_DatagramAddr; }
        }

        public I2PAddress p_StreamAddr
        {
            get { return f_StreamAddr; }
        }

        public bool p_HasClassicAddresses
        {
            get
            {
                return (true == f_DatagramAddr.p_IsClassicAddress &&
                        true == f_StreamAddr.p_IsClassicAddress);
            }
        }

        public bool p_IsInRoutingZone
        {
            get { return f_InZone; }
        }

        public bool p_IsVerified
        {
            get { return f_Verified; }
        }

        public Contact(ClientID a_ClientID,
                       Distance a_Distance)
        {
            f_ClientID = a_ClientID;
            f_Distance = a_Distance;
        }

        public Contact(ClientID a_ClientID) :
                  this(a_ClientID, new Distance(UInt128.s_Zero))
        {
            // This is for creating our own contact. Our
            // f_Vitality will be always be ContactVitality.Alive.
        }

        public bool HasClientID(ClientID a_ClientID)
        {
            return f_ClientID.f_Val.Equals(a_ClientID.f_Val);
        }

        public static BufferIO.LoadResult FromBuffer(Kademlia a_Kad,
                                                     BufferIO a_Buffer,
                                                     out Contact a_Contact)
        {
            // Return can be null. The buffer may be at the end.

            if (Contact.c_MinimumSize > a_Buffer.GetRemaining())
            {
                goto FAIL;
            }

            ClientID l_ClientID = new ClientID(a_Buffer.ReadUInt128ME());

            I2PAddress l_DatagramAddr;
            BufferIO.LoadResult l_Res = I2PAddress.TryCreateFromBufferRaw(a_Buffer, out l_DatagramAddr);

            if (BufferIO.LoadResult.Corruption == l_Res)
            {
                goto FAIL;
            }

            I2PAddress l_StreamAddr;
            l_Res = I2PAddress.TryCreateFromBufferRaw(a_Buffer, out l_StreamAddr);

            if (BufferIO.LoadResult.Corruption == l_Res ||
                true == a_Buffer.EndReached())
            {
                goto FAIL;
            }

            ContactVitality l_Vitality = (ContactVitality)a_Buffer.ReadByte();

            if (ContactVitality.Dead < l_Vitality ||
                null == l_DatagramAddr ||
                null == l_StreamAddr)
            {
                a_Contact = null;

                return BufferIO.LoadResult.InvalidDataButBufferPosIsStillGood;
            }

            a_Contact = new Contact(l_ClientID, a_Kad.CalcDistance(l_ClientID));
            a_Contact.f_DatagramAddr = l_DatagramAddr;
            a_Contact.f_StreamAddr = l_StreamAddr;
            a_Contact.f_Vitality = l_Vitality;

            return BufferIO.LoadResult.Success;

            FAIL:

            a_Contact = null;

            return BufferIO.LoadResult.Corruption;
        }

        public void ToBuffer(BufferIO a_Buffer,
                             bool a_RestrictToClassicAddresses)
        {
            ToBufferWithSpecificVitality(a_Buffer, f_Vitality, a_RestrictToClassicAddresses);
        }

        public void ToBufferWithSpecificVitality(BufferIO a_Buffer,
                                                 ContactVitality a_Vitality,
                                                 bool a_RestrictToClassicAddresses)
        {
            a_Buffer.WriteUInt128ME(f_ClientID.f_Val);
            f_DatagramAddr.ToBufferRaw(a_Buffer);
            f_StreamAddr.ToBufferRaw(a_Buffer);
            a_Buffer.WriteByte((byte)a_Vitality);
        }

        public void TurnRoutingZoneStateOn()
        {
            f_InZone = true;
        }

        public void TurnRoutingZoneStateOff()
        {
            f_InZone = false;
        }

        public bool HasAnyConflict(Contact a_Contact)
        {
            return (true == this.HasClientID(a_Contact.f_ClientID) ||
                    true == this.HasSameAddressesWithCrossCheck(a_Contact));
        }

        public bool HasExactSameAddresses(Contact a_Other)
        {
            return (true == this.p_DatagramAddr.ValueEquals(a_Other.p_DatagramAddr) &&
                    true == this.p_StreamAddr.ValueEquals(a_Other.p_StreamAddr));
        }

        public bool HasSameAddressesWithCrossCheck(Contact a_Other)
        {
            // Checks if there are any identical addresses.

            return (true == this.p_DatagramAddr.ValueEquals(a_Other.p_DatagramAddr) ||
                    true == this.p_DatagramAddr.ValueEquals(a_Other.p_StreamAddr) ||
                    true == this.p_StreamAddr.ValueEquals(a_Other.p_DatagramAddr) ||
                    true == this.p_StreamAddr.ValueEquals(a_Other.p_StreamAddr));
        }

        public void SetDatagramAddr(I2PAddress a_Addr)
        {
            f_DatagramAddr = a_Addr;
        }

        public void SetStreamAddr(I2PAddress a_Addr)
        {
            f_StreamAddr = a_Addr;
        }

        public bool AreDestinationsBanned(NetworkManager a_NetworkMngr)
        {
            return (true == a_NetworkMngr.IsAddressBanned(p_DatagramAddr) ||
                    true == a_NetworkMngr.IsAddressBanned(p_StreamAddr));
        }

        public Time GetLastActivityTime()
        {
            return f_LastActivityTime;
        }

        public ContactVitality GetVitality()
        {
            return f_Vitality;
        }

        public void SetVitality(ContactVitality a_Vitality)
        {
            f_Vitality = a_Vitality;
        }

        public bool WaitingForPong(Time a_Time)
        {
            // f_ExpectPongTime can be zero.

            return a_Time.IsEarlierThan(f_ExpectPongTime);
        }

        public void NotifyPing(Time a_Time)
        {
            if (ContactVitality.Dead > f_Vitality)
            {
                if (false == WaitingForPong(a_Time))
                {
                    f_Vitality++;
                }

                // Even a contact that just became dead has one
                // last grace period where we see if there's
                // a pong or not.

                f_ExpectPongTime = a_Time.AddSeconds(Time.c_SecsPerMinute);
            }
        }

        public void PinUntil(Time a_Time)
        {
            // Keeps this contact from deletion until a_Time.

            // We grow only.
            if (true == f_PinTime.IsEarlierThan(a_Time))
            {
                f_PinTime = a_Time;
            }
        }

        public bool IsPinned(Time a_Time)
        {
            // f_PinTime can be zero.

            return a_Time.IsEarlierThan(f_PinTime);
        }

        public bool HadActivitySince(Time a_Time)
        {
            // f_LastActivityTime can be zero.

            return a_Time.IsEarlierThan(f_LastActivityTime);
        }

        public void NotifyActivity(Time a_Time)
        {
            // May only be used by the ContactManager.

            f_Vitality = ContactVitality.NeedAnswer1;

            f_LastActivityTime = a_Time;
            f_Verified = true;
        }
    }

    enum SearchTypes : byte
    {
        NODE = 0,
        NODECOMPLETE = 1,
        FILE = 2,
        KEYWORD = 3,
        NOTES = 4,
    }

    class ClientSearch : Search<ClientID>
    {
        public ClientSearch(Kademlia a_Kad,
                            ClientID a_Key,
                            bool a_Complete) :
                       base(a_Kad,
                            a_Key,
                            a_Key,
                            (false == a_Complete ? SearchTypes.NODE : SearchTypes.NODECOMPLETE))
        {
        }

        public override void AddToTypeSpecificCollection()
        {
            f_Kad.f_SearchMngr.p_ClientSearches.Add(f_Key, this);
        }

        public override bool RemoveFromTypeSpecificCollection()
        {
            return f_Kad.f_SearchMngr.p_ClientSearches.Remove(f_Key);
        }
    }

    class SourceSearch : Search<FileID>
    {
        public delegate void NotifySourceHandler(FileEntry a_FileEntry);

        private readonly ulong f_FileSize;
        private readonly NotifySourceHandler f_NotifySource;

        public SourceSearch(Kademlia a_Kad,
                            FileID a_Key,
                            ulong a_FileSize,
                            NotifySourceHandler a_NotifySource) :
                       base(a_Kad,
                            new ClientID(a_Key.f_Val),
                            a_Key,
                            SearchTypes.FILE)
        {
            f_FileSize = a_FileSize;
            f_NotifySource = a_NotifySource;

            MemoryIO l_Buffer = MemoryIO.CreateWritable(UInt128.c_ByteSize + sizeof(bool));
            l_Buffer.WriteUInt128ME(f_Key.f_Val);
            l_Buffer.WriteBool(true);
            f_Data = new ByteArray(l_Buffer.ToArray(true));
        }

        public override void AddToTypeSpecificCollection()
        {
            f_Kad.f_SearchMngr.p_SourceSearches.Add(f_Key, this);
        }

        public override bool RemoveFromTypeSpecificCollection()
        {
            return f_Kad.f_SearchMngr.p_SourceSearches.Remove(f_Key);
        }

        protected override bool ProcessSearchResult(Contact a_Contact,
                                                    UInt128 a_UserID,
                                                    TagList a_Info)
        {
            FileEntry l_Entry = FileEntry.FromTagList(f_Kad,
                                                      f_Key,
                                                      new UserID(a_UserID),
                                                      a_Info,
                                                      // NOTE: We don't know the real datagram address so far.
                                                      null);

            if (null == l_Entry)
            {
                return false;
            }

            if (0 != l_Entry.f_FileSize &&
                l_Entry.f_FileSize != f_FileSize)
            {
                // Right file id, wrong size.

                return false;
            }

            f_NotifySource(l_Entry);

            return true;
        }
    }

    class KeywordSearch : Search<KeywordID>
    {
        private readonly SearchTerm f_SearchTerms;

        public KeywordSearch(Kademlia a_Kad,
                             KeywordID a_Key,
                             SearchTerm a_SearchTerms,
                             GuiSearchCookie a_SearchCookie,
                             byte[] a_Bytes) :
                        base(a_Kad,
                             new ClientID(a_Key.f_Val),
                             a_Key,
                             SearchTypes.KEYWORD)
        {
            f_SearchTerms = a_SearchTerms;
            f_SearchCookie = a_SearchCookie;
            f_Data = new ByteArray(a_Bytes);
        }

        public static KeywordSearch CreateSimple(Kademlia a_Kad,
                                                 KeywordID a_Key,
                                                 GuiSearchCookie a_SearchCookie,
                                                 string a_LowerKeyword)
        {
            MemoryIO l_Buffer = MemoryIO.CreateWritable(UInt128.c_ByteSize + sizeof(bool));
            l_Buffer.WriteUInt128ME(a_Key.f_Val);
            l_Buffer.WriteBool(false);

            // Create search terms that we only use for
            // internal checking of the results sent to us.
            SearchTerm l_SearchTerms = SearchTerm.CreateStringType(a_LowerKeyword);

            return new KeywordSearch(a_Kad, a_Key, l_SearchTerms, a_SearchCookie, l_Buffer.ToArray(true));
        }

        public static KeywordSearch CreateAdvanced(Kademlia a_Kad,
                                                   KeywordID a_Key,
                                                   SearchTerm a_SearchTerms,
                                                   GuiSearchCookie a_SearchCookie)
        {
            // Return can be null.

            MemoryIO l_Buffer = MemoryIO.CreateWritable();
            l_Buffer.WriteUInt128ME(a_Key.f_Val);
            l_Buffer.WriteBool(true);

            if (false == SearchTerm.ToBuffer(l_Buffer, a_SearchTerms) ||
                l_Buffer.p_Length > PacketHandler.c_MaxKadMessageSize)
            {
                return null;
            }

            return new KeywordSearch(a_Kad, a_Key, a_SearchTerms, a_SearchCookie, l_Buffer.ToArray(true));
        }

        public override void AddToTypeSpecificCollection()
        {
            f_Kad.f_SearchMngr.p_KeywordSearches.Add(f_Key, this);
        }

        public override bool RemoveFromTypeSpecificCollection()
        {
            return f_Kad.f_SearchMngr.p_KeywordSearches.Remove(f_Key);
        }

        protected override bool ProcessSearchResult(Contact a_Contact,
                                                    UInt128 a_FileID,
                                                    TagList a_Info)
        {
            KeywordEntry l_Entry = KeywordEntry.FromTagList(f_Kad, f_Key, new FileID(a_FileID), a_Info);

            if (null == l_Entry)
            {
                return false;
            }

            if (false == KadMap.SearchTermsMatch(l_Entry, f_SearchTerms))
            {
                // If the search results don't match, still accept
                // the callers data and return true even though
                // we don't actually need this.

                f_Kad.f_Stats.NotifyRejectedKeywordSearchResult();

                return true;
            }

            f_SearchCookie.AddFile(f_Kad.f_KadPrefs.f_KadFilter, l_Entry.f_RefID, l_Entry.f_FileName, l_Entry.f_FileSize, l_Entry.f_CompleteSources, l_Entry.f_Sources);

            return true;
        }

        public void PullResultsFromKadIndex()
        {
            f_Kad.f_Map.AddValidKeywordsToSearchCookie(f_Key, f_SearchTerms, f_SearchCookie);
        }
    }

    class NotesSearch : Search<FileID>
    {
        public NotesSearch(Kademlia a_Kad,
                           FileID a_Key) :
                      base(a_Kad,
                           new ClientID(a_Key.f_Val),
                           a_Key,
                           SearchTypes.NOTES)
        {
            MemoryIO l_Buffer = MemoryIO.CreateWritable(2 * UInt128.c_ByteSize);
            l_Buffer.WriteUInt128ME(f_Key.f_Val);
            l_Buffer.WriteUInt128ME(a_Kad.f_Self.f_ClientID.f_Val);
            f_Data = new ByteArray(l_Buffer.ToArray(true));
        }

        public override void AddToTypeSpecificCollection()
        {
            f_Kad.f_SearchMngr.p_NotesSearches.Add(f_Key, this);
        }

        public override bool RemoveFromTypeSpecificCollection()
        {
            return f_Kad.f_SearchMngr.p_NotesSearches.Remove(f_Key);
        }

        protected override bool ProcessSearchResult(Contact a_Contact,
                                                    UInt128 a_ClientID,
                                                    TagList a_Info)
        {
            NoteEntry l_Entry = NoteEntry.FromTagList(f_Kad,
                                                      f_Key,
                                                      new ClientID(a_ClientID),
                                                      a_Info);

            return (null != l_Entry);
        }
    }

    abstract class SearchBase
    {
        protected const byte c_MaxBestListEntries = 3;
        private const ushort c_SearchStartMaxNodes = 50;

        protected readonly Kademlia f_Kad;
        public readonly uint f_SearchID;
        public readonly ClientID f_ClientKey;
        public readonly SearchTypes f_Type;
        protected readonly KadParams f_KadParam;
        protected readonly KadOpcodes f_RequestOpcode;
        protected readonly ushort f_MaxSearchResponses;
        protected readonly Time f_TerminationTime;
        protected readonly Time f_LastSendTime;
        protected readonly Time f_ContactPinTime;

        // f_SearchCookie and f_Data can be null.
        protected GuiSearchCookie f_SearchCookie;
        protected ByteArray f_Data;

        protected readonly Dictionary<ClientID, SearchContact> f_KadPossible;
        protected readonly Dictionary<ClientID, SearchContact> f_KadTried;
        protected readonly Dictionary<ClientID, SearchContact> f_KadResponded;
        protected readonly SortedList<Distance, SearchContact> f_Best;

        protected bool f_Running;
        protected bool f_Stoping;

        protected ushort f_SentKadRequests;
        protected ushort f_SentSearchRequests;
        protected ushort f_ReceivedKadResponses;
        protected ushort f_ReceivedSearchResponses;
        protected uint f_ReceivedResults;

        public SearchBase(Kademlia a_Kad,
                          ClientID a_ClientKey,
                          SearchTypes a_Type)
        {
            f_Kad = a_Kad;
            f_SearchID = a_Kad.f_SearchMngr.ProposeSearchID();
            f_ClientKey = a_ClientKey;
            f_Type = a_Type;

            switch (f_Type)
            {
                case SearchTypes.NODE:
                case SearchTypes.NODECOMPLETE:
                    f_KadParam = KadParams.KADEMLIA_FIND_NODE;
                    f_RequestOpcode = KadOpcodes.NONE;
                    break;

                case SearchTypes.FILE:
                case SearchTypes.KEYWORD:
                    f_KadParam = KadParams.KADEMLIA_FIND_VALUE;
                    f_RequestOpcode = KadOpcodes.KADEMLIA_SEARCH_REQ;
                    break;

                case SearchTypes.NOTES:
                    f_KadParam = KadParams.KADEMLIA_FIND_VALUE;
                    f_RequestOpcode = KadOpcodes.KADEMLIA_SRC_NOTES_REQ;
                    break;

                default:
                    throw(new ApplicationException());
            }

            f_MaxSearchResponses = a_Kad.f_SearchMngr.GetMaxSearchResponses(a_Type);
            uint l_LifetimeSecs = a_Kad.f_SearchMngr.GetLifetimeSecs(a_Type);
            f_TerminationTime = a_Kad.f_Inst.GetTime().AddSeconds(l_LifetimeSecs);
            f_LastSendTime = f_TerminationTime.SubSeconds(Time.c_SecsPerHalfMinute);
            f_ContactPinTime = f_TerminationTime.AddSeconds(Time.c_SecsPerMinute * 5);

            f_KadPossible = new Dictionary<ClientID, SearchContact>();
            f_KadTried = new Dictionary<ClientID, SearchContact>();
            f_KadResponded = new Dictionary<ClientID, SearchContact>();
            f_Best = new SortedList<Distance, SearchContact>();
        }

        public bool WentWellEnough()
        {
            if (5 <= f_ReceivedKadResponses &&
                0 != f_SentSearchRequests)
            {
                return true;
            }

            return false;
        }

        public bool IsRunning()
        {
            return f_Running;
        }

        public void Destroy()
        {
            if (true == f_Running)
            {
                f_Running = false;
            }

            if (null != f_SearchCookie)
            {
                f_SearchCookie.MakeResultsAvailableToGui();
                f_SearchCookie.Finished();
            }
        }

        public abstract void AddToTypeSpecificCollection();

        public abstract bool RemoveFromTypeSpecificCollection();

        public I2PAddress[] GetBestForPublishing(int a_MaxCount)
        {
            // Return value is null if the effective count is zero.

            SortedList<Distance, I2PAddress> l_Ls = new SortedList<Distance, I2PAddress>(f_KadResponded.Count);

            foreach (SearchContact l_SContact in f_KadResponded.Values)
            {
                if (true == l_SContact.f_TargDistance.HasMinFitness(2))
                {
                    l_Ls.Add(l_SContact.f_TargDistance, l_SContact.f_Contact.p_DatagramAddr);
                }
            }

            int l_Copy = Math.Min(l_Ls.Count, a_MaxCount);

            if (0 == l_Copy)
            {
                return null;
            }

            I2PAddress[] l_Addresses = new I2PAddress[l_Copy];
            uint l_Idx = 0;

            foreach (I2PAddress l_Addr in l_Ls.Values)
            {
                l_Addresses[l_Idx] = l_Addr;

                if (l_Copy == ++l_Idx)
                {
                    break;
                }
            }

            return l_Addresses;
        }

        public bool Go()
        {
            // If this routine returns false, the search can't
            // be done (lack of suitable known peers).

            SortedList<Distance, Contact> l_Results = new SortedList<Distance, Contact>(c_SearchStartMaxNodes);
            RoutingZone.GetClosestToArgs l_Args = new RoutingZone.GetClosestToArgs(f_Kad, f_ClientKey, c_SearchStartMaxNodes, ContactVitality.Dead, 0, false, l_Results);

            l_Args.f_VerifiedOnly = true;
            f_Kad.f_RootZone.GetClosestTo(ref l_Args);

            if (c_SearchStartMaxNodes > l_Results.Count)
            {
                l_Args.f_VerifiedOnly = false;
                f_Kad.f_RootZone.GetClosestTo(ref l_Args);
            }

            if (0 == l_Results.Count)
            {
                return false;
            }

            int l_InitialTries = (SearchTypes.NODE == f_Type ? 1 : (int)c_MaxBestListEntries);

            foreach (KeyValuePair<Distance, Contact> CurPair in l_Results)
            {
                CurPair.Value.PinUntil(f_ContactPinTime);

                SearchContact l_SContact = new SearchContact(CurPair.Value, CurPair.Key);

                if (l_InitialTries > f_Best.Count)
                {
                    SendKademliaRequest(l_SContact);

                    f_Best.Add(CurPair.Key, l_SContact);
                    f_KadTried.Add(l_SContact.f_Contact.f_ClientID, l_SContact);
                }
                else
                {
                    f_KadPossible.Add(l_SContact.f_Contact.f_ClientID, l_SContact);
                }
            }

            f_Running = true;

            if (null != f_SearchCookie)
            {
                f_SearchCookie.SearchStarted();
            }

            return true;
        }

        public bool Process(Time a_Time)
        {
            // If this routine returns false, the search should
            // be terminated without delay.

            if (null != f_SearchCookie)
            {
                if (CookieStates.Pending != f_SearchCookie.p_GuiStateVolatile)
                {
                    return false;
                }

                f_SearchCookie.MakeResultsAvailableToGui();
            }

            // See if this search reached the end of its lifetime.
            if (f_MaxSearchResponses <= f_ReceivedSearchResponses ||
                true == f_TerminationTime.IsEarlierThan(a_Time))
            {
                return false;
            }

            if (true == f_Stoping)
            {
                return true;
            }

            if (true == f_LastSendTime.IsEarlierThan(a_Time))
            {
                // It's now too late to send requests.

                StopLookingForPeers();

                return true;
            }

            if (0 < f_KadPossible.Count)
            {
                foreach (SearchContact l_SContact in f_KadPossible.Values)
                {
                    SendKademliaRequest(l_SContact);

                    f_KadTried.Add(l_SContact.f_Contact.f_ClientID, l_SContact);
                }

                f_KadPossible.Clear();
            }

            return true;
        }

        private bool TryAddToBestList(SearchContact a_SContact)
        {
            bool l_Top = false;

            if (c_MaxBestListEntries > f_Best.Count)
            {
                l_Top = true;
            }
            else
            {
                SearchContact l_CurLast = f_Best.Values[c_MaxBestListEntries - 1];

                if (0 < a_SContact.f_TargDistance.CompareTo(l_CurLast.f_TargDistance))
                {
                    // The one we've got now is better than that before.

                    f_Best.RemoveAt(c_MaxBestListEntries - 1);

                    l_Top = true;
                }
            }

            if (true == l_Top)
            {
                f_Best.Add(a_SContact.f_TargDistance, a_SContact);
            }

            return l_Top;
        }

        public void StopLookingForPeers()
        {
            f_Stoping = true;
        }

        private void SendKademliaRequest(SearchContact a_SContact)
        {
            MemoryIO l_Buffer = MemoryIO.CreateWritable(sizeof(byte) + (3 * UInt128.c_ByteSize));

            l_Buffer.WriteByte((byte)f_KadParam);
            // The id that we're looking for.
            l_Buffer.WriteUInt128ME(f_ClientKey.f_Val);
            // The id of the peer asked.
            l_Buffer.WriteUInt128ME(a_SContact.f_Contact.f_ClientID.f_Val);
            // The id of the requesting peer.
            l_Buffer.WriteUInt128ME (f_Kad.f_Self.f_ClientID.f_Val);

            f_Kad.f_PacketHndlr.SendKademliaRequest(a_SContact.f_Contact, l_Buffer);

            f_SentKadRequests++;
        }

        private void SendOtherRequest(SearchContact a_SContact)
        {
            f_Kad.f_PacketHndlr.SendPacket(f_RequestOpcode, a_SContact.f_Contact.p_DatagramAddr, ref f_Data);

            f_SentSearchRequests++;
        }

        public bool RegisterKadResponse(Contact a_Origin,
                                        List<Contact> a_Results)
        {
            // Returns false if the input doesn't seem to belong
            // to this search. Due to protocol limitations we
            // don't have a tag or something to identify the
            // correct recipient right away.

            SearchContact l_SOrigin;
            if (false == f_KadTried.TryGetValue(a_Origin.f_ClientID, out l_SOrigin))
            {
                // We didn't request this.

                return false;
            }

            if (true == f_KadResponded.ContainsKey(a_Origin.f_ClientID))
            {
                // We've already handled this.

                return false;
            }

            f_ReceivedKadResponses++;
            f_KadResponded.Add(a_Origin.f_ClientID, l_SOrigin);

            if (SearchTypes.NODE == f_Type)
            {
                StopLookingForPeers();

                return true;
            }

            if (false == f_Stoping)
            {
                if (KadOpcodes.NONE != f_RequestOpcode &&
                    false == l_SOrigin.f_TargDistance.BelongsToUpperHalf())
                {
                    SendOtherRequest(l_SOrigin);
                }

                foreach (Contact l_Contact in a_Results)
                {
                    if (true == f_KadPossible.ContainsKey(l_Contact.f_ClientID) ||
                        true == f_KadTried.ContainsKey(l_Contact.f_ClientID))
                    {
                        // A known contact.

                        continue;
                    }

                    // NOTE: This check is quite useless, since AddContact()
                    // ensures that it cannot happen.

                    l_Contact.PinUntil(f_ContactPinTime);

                    SearchContact l_SContact = new SearchContact(this, l_Contact);
                    bool l_Top = TryAddToBestList(l_SContact);

                    if (false == l_Top)
                    {
                        f_KadPossible.Add(l_Contact.f_ClientID, l_SContact);
                    }
                    else
                    {
                        SendKademliaRequest(l_SContact);

                        f_KadTried.Add(l_Contact.f_ClientID, l_SContact);
                    }
                }
            }

            return true;
        }

        public DispatchResult ProcessSearchResultList(Contact a_Contact,
                                                      BufferIO a_Buffer,
                                                      uint a_Count)
        {
            // If the input can't be processed, the original file
            // position will be restored before control leaves the
            // routine.

            if (false == f_KadResponded.ContainsKey(a_Contact.f_ClientID))
            {
                return DispatchResult.WrongID;
            }

            if (0 == a_Count)
            {
                // We'll tell the caller that we handled this, but we
                // don't increment f_ReceivedSearchResponses.

                return DispatchResult.Success;
            }

            uint l_Pos = a_Buffer.p_Position;
            bool l_OK = ProcessSearchResultListCore(a_Contact, a_Buffer, a_Count);

            if (f_MaxSearchResponses <= f_ReceivedSearchResponses)
            {
                StopLookingForPeers();
            }

            if (false == l_OK)
            {
                a_Buffer.SetPosition(l_Pos);

                return DispatchResult.Corruption;
            }

            f_ReceivedSearchResponses++;

            return DispatchResult.Success;
        }

        private bool ProcessSearchResultListCore(Contact a_Contact,
                                                 BufferIO a_Buffer,
                                                 uint a_Count)
        {
            for (; 0 != a_Count; a_Count--)
            {
                if (false == f_Kad.VerifyRemainingMin(a_Buffer, UInt128.c_ByteSize + sizeof(byte)))
                {
                    return false;
                }

                UInt128 l_Ref = a_Buffer.ReadUInt128ME();
                TagList l_TagList = TagList.ReadListUntrusted(a_Buffer, a_Buffer.ReadByte(), PacketHandler.c_ExpectNormalStrsAsUTF8);

                if (null == l_TagList ||
                    false == ProcessSearchResult(a_Contact, l_Ref, l_TagList))
                {
                    // We can't handle this, so abort.

                    f_Kad.f_Stats.NotifyCorruptValues();

                    return false;
                }

                f_ReceivedResults++;
            }

            return f_Kad.VerifyEndReached(a_Buffer);
        }

        protected virtual bool ProcessSearchResult(Contact a_Contact,
                                                   UInt128 a_Answer,
                                                   TagList a_Info)
        {
            throw(new ApplicationException());
        }

        protected class SearchContact
        {
            public readonly Contact f_Contact;
            public readonly Distance f_TargDistance;

            public SearchContact(SearchBase a_Search,
                                 Contact a_Contact) :
                            this(a_Contact, Kademlia.CalcDistance(a_Search.f_ClientKey, a_Contact.f_ClientID))
            {
            }

            public SearchContact(Contact a_Contact,
                                 Distance a_TargDistance)
            {
                f_Contact = a_Contact;
                f_TargDistance = a_TargDistance;
            }
        }

        public enum DispatchResult : byte
        {
            Success = 0,
            WrongID = 1,
            Corruption = 2,
        }
    }

    abstract class Search<T> : SearchBase
    {
        public readonly T f_Key;

        public Search(Kademlia a_Kad,
                      ClientID a_ClientKey,
                      T a_Key,
                      SearchTypes a_Type) :
                 base(a_Kad, a_ClientKey, a_Type)
        {
            f_Key = a_Key;
        }
    }

    class SearchManager
    {
        private const byte c_MinKeywordChars = 1;
        private const byte c_SelfPublishingMinKeywordChars = 3;

        private readonly ushort[] f_MaxSearchResponses = new ushort[] { 1, // SearchTypes.NODE
                                                                        10, // SearchTypes.NODECOMPLETE
                                                                        ushort.MaxValue, // SearchTypes.FILE
                                                                        ushort.MaxValue, // SearchTypes.KEYWORD
                                                                        ushort.MaxValue }; // SearchTypes.NOTES

        private readonly ushort[] f_SearchLifetimeSecs = new ushort[] { 60, // SearchTypes.NODE
                                                                        65, // SearchTypes.NODECOMPLETE
                                                                        120, // SearchTypes.FILE
                                                                        120, // SearchTypes.KEYWORD
                                                                        120 }; // SearchTypes.NOTES

        private readonly Kademlia f_Kad;

        private readonly Dictionary<uint, SearchBase> f_Searches;
        private readonly Dictionary<ClientID, ClientSearch> f_ClientSearches;
        private readonly Dictionary<FileID, SourceSearch> f_SourceSearches;
        private readonly Dictionary<KeywordID, KeywordSearch> f_KeywordSearches;
        private readonly Dictionary<FileID, NotesSearch> f_NotesSearches;

        private readonly MD_4 f_KeywordHasher;

        private uint f_NextID;

        public SearchManager(Kademlia a_Kad)
        {
            f_Kad = a_Kad;

            f_Searches = new Dictionary<uint, SearchBase>();
            f_ClientSearches = new Dictionary<ClientID, ClientSearch>();
            f_SourceSearches = new Dictionary<FileID, SourceSearch>();
            f_KeywordSearches = new Dictionary<KeywordID, KeywordSearch>();
            f_NotesSearches = new Dictionary<FileID, NotesSearch>();

            f_KeywordHasher = new MD_4();

            f_NextID = 1;
        }

        public bool StartingNewSearchIsNotRecommended()
        {
            if (5 <= f_Searches.Count)
            {
                return true;
            }

            return false;
        }

        private SearchBase GetSearch(uint a_SearchID)
        {
            SearchBase l_Search;
            if (true == f_Searches.TryGetValue(a_SearchID, out l_Search))
            {
                return l_Search;
            }

            return null;
        }

        private bool AddAndStartSearch(SearchBase a_Search)
        {
            if (true == AddSearch(a_Search))
            {
                f_NextID++;

                if (true == a_Search.Go())
                {
                    return true;
                }

                // There seem to be too few valid contacts.

                bool OK = RemoveSearch(a_Search);
            }

            return false;
        }

        private bool AddSearch(SearchBase a_Search)
        {
            try
            {
                a_Search.AddToTypeSpecificCollection();
            }
            catch (ArgumentException)
            {
                // A search of that type with that key already exists.

                return false;
            }

            f_Searches.Add(a_Search.f_SearchID, a_Search);

            return true;
        }

        private bool RemoveSearch(SearchBase a_Search)
        {
            SearchBase Temp;
            if (true == f_Searches.TryGetValue(a_Search.f_SearchID, out Temp) &&
                Temp == a_Search)
            {
                bool OK = (true == f_Searches.Remove(a_Search.f_SearchID) &&
                           true == a_Search.RemoveFromTypeSpecificCollection());

                return true;
            }

            return false;
        }

        public bool StopSearch(uint a_SearchID,
                               bool a_DelayDelete)
        {
            SearchBase l_Search = GetSearch(a_SearchID);

            if (null != l_Search)
            {
                StopSearch(l_Search, a_DelayDelete);

                return true;
            }

            return false;
        }

        private void StopSearch(SearchBase a_Search,
                                bool a_DelayDelete)
        {
            a_Search.StopLookingForPeers();

            if (false == a_DelayDelete)
            {
                a_Search.Destroy();

                RemoveSearch(a_Search);
            }
        }

        public void StopAllSearches()
        {
            foreach (SearchBase l_Search in f_Searches.Values)
            {
                l_Search.StopLookingForPeers();
                l_Search.Destroy();
            }

            f_Searches.Clear();
            f_ClientSearches.Clear();
            f_SourceSearches.Clear();
            f_KeywordSearches.Clear();
            f_NotesSearches.Clear();
        }

        public uint ProposeSearchID()
        {
            return f_NextID;
        }

        public ClientSearch FindNode(ClientID a_Target,
                                     bool a_Complete)
        {
            // a_ReceivedKadResponse may be null.

            ClientSearch l_Search = new ClientSearch(f_Kad, a_Target, a_Complete);

            if (true == AddAndStartSearch(l_Search))
            {
                return l_Search;
            }

            return null;
        }

        public SourceSearch FindSources(FileID a_FileID,
                                        ulong a_FileSize,
                                        SourceSearch.NotifySourceHandler a_NotifySource)
        {
            SourceSearch l_Search = new SourceSearch(f_Kad, a_FileID, a_FileSize, a_NotifySource);

            if (true == AddAndStartSearch(l_Search))
            {
                return l_Search;
            }

            return null;
        }

        public bool FindByKeyword(string a_LowerKeyword,
                                  GuiSearchCookie a_SearchCookie)
        {
            KeywordID l_KeywordID = new KeywordID(CalcKeywordHash(a_LowerKeyword));
            KeywordSearch l_Search = KeywordSearch.CreateSimple(f_Kad, l_KeywordID, a_SearchCookie, a_LowerKeyword);

            return AddAndStartKeywordSearch(l_Search, a_SearchCookie);
        }

        public bool FindBySearchTerms(SearchTerm a_SearchTerm,
                                      GuiSearchCookie a_SearchCookie)
        {
            // Try to select one of the and-keywords to
            // generate a unique keyword id for this search.
            // We prefer latin-1 compatible strings for this
            // (since iMule can handle them).

            UInt128 l_KeywordID = null;
            UInt128 l_GoodID = null;

            foreach (string l_Keyword in a_SearchCookie.f_SearchTerms.f_AndKeywords)
            {
                UInt128 l_TempID = CalcKeywordHash(l_Keyword);

                if (false == f_KeywordSearches.ContainsKey(new KeywordID(l_TempID)))
                {
                    // This keyword id isn't in use.

                    l_KeywordID = l_TempID;

                    if (true == Util.IsLatin1AndUTF8Equivalent(l_Keyword))
                    {
                        // Ideal solution (most client compatible).

                        l_GoodID = l_TempID;

                        break;
                    }

                    if (true == Util.IsLatin1Compatible(l_Keyword))
                    {
                        // Second best solution.

                        l_GoodID = l_TempID;
                    }
                }
            }

            if (null == l_KeywordID)
            {
                // Other searches are using all the keywords
                // we have.

                a_SearchCookie.Rejected();

                return false;
            }

            KeywordSearch l_Search = KeywordSearch.CreateAdvanced(f_Kad, new KeywordID(null == l_GoodID ? l_KeywordID : l_GoodID), a_SearchTerm, a_SearchCookie);

            return AddAndStartKeywordSearch(l_Search, a_SearchCookie);
        }

        private bool AddAndStartKeywordSearch(KeywordSearch a_Search,
                                              GuiSearchCookie a_SearchCookie)
        {
            // a_Search may be null.

            if (null != a_Search &&
                true == AddAndStartSearch(a_Search))
            {
                // Add results from the local kad index.
                a_Search.PullResultsFromKadIndex();

                return true;
            }

            a_SearchCookie.Rejected();

            return false;
        }

        public void GuiStartKeywordSearch(GeneralAPCCookie a_GeneralCookie)
        {
            GuiSearchCookie l_SearchCookie = (GuiSearchCookie)a_GeneralCookie.GetArg();

            if (KademliaStatus.Connected != f_Kad.p_KadStatus)
            {
                l_SearchCookie.Rejected();
                a_GeneralCookie.SetState(CookieStates.Failure);

                return;
            }

            if (false == l_SearchCookie.f_SearchTerms.IsSimple())
            {
                SearchTerm l_SearchTerms = SearchTerm.CreateTree(l_SearchCookie.f_SearchTerms);

                FindBySearchTerms(l_SearchTerms, l_SearchCookie);
            }
            else
            {
                FindByKeyword(l_SearchCookie.f_SearchTerms.f_AndKeywords[0], l_SearchCookie);
            }

            a_GeneralCookie.SetState(CookieStates.Success);
        }

        public bool FindNotes(FileID a_FileID)
        {
            NotesSearch l_Search = new NotesSearch(f_Kad, a_FileID);

            return AddAndStartSearch(l_Search);
        }

        public static void CalcKeywordHash(MD_4 a_MD4,
                                           string a_Keyword)
        {
            byte[] l_Msg = Encoding.UTF8.GetBytes(a_Keyword);

            a_MD4.Reset();
            a_MD4.Update(l_Msg);
            a_MD4.Finish();
        }

        public UInt128 CalcKeywordHash(string a_Keyword)
        {
            // May only be used by the network thread.

            CalcKeywordHash(f_KeywordHasher, a_Keyword);

            return f_KeywordHasher.GetDigest();
        }

        public bool VerifyKeyword(UInt128 a_KeywordHash,
                                  string a_FileName)
        {
            string[] l_Keywords = IdentifyKeywords(a_FileName, false);

            if (null != l_Keywords)
            {
                foreach (string l_Keyword in l_Keywords)
                {
                    CalcKeywordHash(f_KeywordHasher, l_Keyword);

                    if (true == f_KeywordHasher.Verify(a_KeywordHash))
                    {
                        return true;
                    }
                }
            }

            return false;
        }

        public static string[] IdentifyKeywords(string a_FileName,
                                                bool a_Sort)
        {
            return IdentifyKeywords(a_FileName, a_Sort, false);
        }

        public static string[] IdentifyKeywords(string a_FileName,
                                                bool a_Sort,
                                                bool a_SelfPublishing)
        {
            // Return can be null.

            string[] l_Arr = a_FileName.Split(Kademlia.s_InvKadKeywordChars, StringSplitOptions.RemoveEmptyEntries);
            int l_Count = l_Arr.Length;
            string l_Extension = LongPath.GetExtension(a_FileName);

            if (null != l_Extension &&
                0 < l_Count &&
                l_Arr[l_Count - 1] == l_Extension)
            {
                // Remove extension.

                l_Count--;
            }

            List<string> l_Keywords = new List<string>(l_Count);
            int l_EffMinChars = (false == a_SelfPublishing ? c_MinKeywordChars : c_SelfPublishingMinKeywordChars);

            for (int i = 0; i < l_Count; i++)
            {
                string l_Keyword = l_Arr[i].ToLowerInvariant();

                if (l_EffMinChars <= l_Keyword.Length)
                {
                    if (false == l_Keywords.Contains(l_Keyword))
                    {
                        l_Keywords.Add(l_Keyword);
                    }
                }
            }

            if (true == a_Sort)
            {
                l_Keywords.Sort();
            }

            return (0 == l_Keywords.Count ? null : l_Keywords.ToArray());
        }

        public static bool IsWholeKeyword(string a_Str,
                                          int a_Idx,
                                          int a_Len)
        {
            if (0 == a_Idx ||
                true == Util.IsCharInArray(a_Str[a_Idx - 1], Kademlia.s_InvKadKeywordChars))
            {
                int l_LastIdx = (a_Idx + a_Len);

                return (a_Str.Length == l_LastIdx ||
                        true == Util.IsCharInArray(a_Str[l_LastIdx], Kademlia.s_InvKadKeywordChars));
            }

            return false;
        }

        public static bool IsInvalidKeywordChar(char a_Char)
        {
            return Util.IsCharInArray(a_Char, Kademlia.s_InvKadKeywordChars);
        }

        public static bool ContainsKeyword(string a_String,
                                           string a_Keyword)
        {
            for (int l_Idx = 0, l_StrLen = a_String.Length, l_KeywordLen = a_Keyword.Length; true; )
            {
                l_Idx = a_String.IndexOf(a_Keyword, l_Idx, StringComparison.InvariantCultureIgnoreCase);

                if (0 > l_Idx)
                {
                    break;
                }

                // Make sure this is about a whole word and not just
                // some part of a word.

                int l_NextIdx = (l_Idx + l_KeywordLen);

                if (l_NextIdx <= l_StrLen &&
                    true == IsWholeKeyword(a_String, l_Idx, l_KeywordLen))
                {
                    return true;
                }

                if (l_NextIdx >= l_StrLen)
                {
                    break;
                }

                l_Idx = l_NextIdx;
            }

            return false;
        }

        public static bool ContainsKeywords(string a_FileName,
                                            string[] a_Keywords,
                                            bool a_AND)
        {
            // a_AND determines if the operator AND or
            // OR is used.

            foreach (string l_Keyword in a_Keywords)
            {
                if (false == ContainsKeyword(a_FileName, l_Keyword))
                {
                    if (true == a_AND)
                    {
                        return false;
                    }
                }
                else if (false == a_AND)
                {
                    // OR test positive.

                    return true;
                }
            }

            return (true == a_AND &&
                    0 < a_Keywords.Length);
        }

        public void Process(Time a_Time)
        {
            List<SearchBase> l_Remove = new List<SearchBase>();

            foreach (SearchBase l_Search in f_Searches.Values)
            {
                if (false == l_Search.Process(a_Time))
                {
                    l_Remove.Add(l_Search);
                }
            }

            foreach (SearchBase l_Search in l_Remove)
            {
                StopSearch(l_Search, false);
            }
        }

        public bool RegisterKadResponse(UInt128 a_SearchKey,
                                        Contact a_Origin,
                                        List<Contact> a_Results)
        {
            // Call this routine even if a_Results is an
            // empty list, since we need to know who responded.

            // We don't know which search was asking for this
            // peer list, so we try all (it doesn't hurt if
            // there is more than one).

            byte l_Uses = 0;

            if (true == RegisterKadResponseInternal<ClientID, ClientSearch>(new ClientID(a_SearchKey),
                                                                            f_ClientSearches,
                                                                            a_Origin,
                                                                            a_Results))
            {
                l_Uses++;
            }

            if (true == RegisterKadResponseInternal<FileID, SourceSearch>(new FileID(a_SearchKey),
                                                                        f_SourceSearches,
                                                                        a_Origin,
                                                                        a_Results))
            {
                l_Uses++;
            }

            if (true == RegisterKadResponseInternal<KeywordID, KeywordSearch>(new KeywordID(a_SearchKey),
                                                                              f_KeywordSearches,
                                                                              a_Origin,
                                                                              a_Results))
            {
                l_Uses++;
            }

            if (true == RegisterKadResponseInternal<FileID, NotesSearch>(new FileID(a_SearchKey),
                                                                         f_NotesSearches,
                                                                         a_Origin,
                                                                         a_Results))
            {
                l_Uses++;
            }

            return (0 != l_Uses);
        }

        private bool RegisterKadResponseInternal<TKey, TSearch>(TKey a_SearchKey,
                                                                Dictionary<TKey, TSearch> a_Dic,
                                                                Contact a_Origin,
                                                                List<Contact> a_Results)
                                                                where TSearch : SearchBase
        {
            TSearch l_Search;
            if (true == a_Dic.TryGetValue(a_SearchKey, out l_Search))
            {
                return l_Search.RegisterKadResponse(a_Origin, a_Results);
            }

            return false;
        }

        public bool ProcessSearchResultList(Contact a_Contact,
                                            BufferIO a_Buffer)
        {
            // Due to limitation of the protocol, we don't know if
            // we've searched for a file or a keyword or for notes,
            // so we'll try all three.

            UInt128 l_Key = a_Buffer.ReadUInt128ME();
            uint l_Count = a_Buffer.ReadUInt16LE();
            SourceSearch.DispatchResult l_Res;

            SourceSearch l_SourceSearch;
            if (true == f_SourceSearches.TryGetValue(new FileID(l_Key), out l_SourceSearch))
            {
                l_Res = l_SourceSearch.ProcessSearchResultList(a_Contact, a_Buffer, l_Count);

                if (SearchBase.DispatchResult.WrongID != l_Res)
                {
                    return DispatchResToBool(l_Res);
                }
            }

            KeywordSearch l_KeywordSearch;
            if (true == f_KeywordSearches.TryGetValue(new KeywordID(l_Key), out l_KeywordSearch))
            {
                l_Res = l_KeywordSearch.ProcessSearchResultList(a_Contact, a_Buffer, l_Count);

                if (SearchBase.DispatchResult.WrongID != l_Res)
                {
                    return DispatchResToBool(l_Res);
                }
            }

            NotesSearch l_NotesSearch;
            if (true == f_NotesSearches.TryGetValue(new FileID(l_Key), out l_NotesSearch))
            {
                l_Res = l_NotesSearch.ProcessSearchResultList(a_Contact, a_Buffer, l_Count);

                if (SearchBase.DispatchResult.WrongID != l_Res)
                {
                    return DispatchResToBool(l_Res);
                }
            }

            return false;
        }

        public ushort GetMaxSearchResponses(SearchTypes a_Type)
        {
            return f_MaxSearchResponses[(int)a_Type];
        }

        public ushort GetLifetimeSecs(SearchTypes a_Type)
        {
            return f_SearchLifetimeSecs[(int)a_Type];
        }

        private static bool DispatchResToBool(SourceSearch.DispatchResult a_DispatchRes)
        {
            return (SearchBase.DispatchResult.Success == a_DispatchRes);
        }

        // Only the SearchBase class may use the
        // following special properties.

        public Dictionary<ClientID, ClientSearch> p_ClientSearches
        {
            get { return f_ClientSearches; }
        }

        public Dictionary<FileID, SourceSearch> p_SourceSearches
        {
            get { return f_SourceSearches; }
        }

        public Dictionary<KeywordID, KeywordSearch> p_KeywordSearches
        {
            get { return f_KeywordSearches; }
        }

        public Dictionary<FileID, NotesSearch> p_NotesSearches
        {
            get { return f_NotesSearches; }
        }
    }

    class KadFoundFile
    {
        // A search result item.

        public readonly FileID f_FileID;
        // f_FileName is guaranteed to be legal.
        public readonly string f_FileName;
        public readonly ulong f_FileSize;
        private uint f_CompletedSourcesSum;
        private uint f_SourcesSum;
        private ushort f_SumOfSourcIndications;

        public ushort p_CompletedSources
        {
            get { return CalcSourcesMean(f_CompletedSourcesSum); }
        }

        public ushort p_Sources
        {
            get { return CalcSourcesMean(f_SourcesSum); }
        }

        public KadFoundFile(FileID a_FileID,
                            string a_FileName,
                            ulong a_FileSize,
                            ushort a_CompletedSources,
                            ushort a_Sources)
        {
            f_FileID = a_FileID;
            f_FileName = a_FileName;
            f_FileSize = a_FileSize;
            f_CompletedSourcesSum = a_CompletedSources;
            f_SourcesSum = a_Sources;
            f_SumOfSourcIndications = 1;
        }

        public void AddSources(ushort a_CompletedSources,
                               ushort a_Sources)
        {
            if (false == BufferIO.AddWithOverflowCheck(f_CompletedSourcesSum, a_CompletedSources, out f_CompletedSourcesSum))
            {
                f_CompletedSourcesSum = uint.MaxValue;
            }

            if (false == BufferIO.AddWithOverflowCheck(f_SourcesSum, a_Sources, out f_SourcesSum))
            {
                f_SourcesSum = uint.MaxValue;
            }

            if (ushort.MaxValue != f_SumOfSourcIndications)
            {
                f_SumOfSourcIndications++;
            }
        }

        private ushort CalcSourcesMean(uint a_Sum)
        {
            if (0 == a_Sum)
            {
                return 0;
            }

            if (1 >= f_SumOfSourcIndications)
            {
                return (ushort)a_Sum;
            }

            return (ushort)Util.MinMax(1, ushort.MaxValue, a_Sum / f_SumOfSourcIndications);
        }

        public void ToValue(StorableTime a_StorableTime,
                            byte[] a_TempBuffer,
                            out Gui.FileSearchPanel.FoundFile a_Dest)
        {
            UInt128Value l_FileID;
            f_FileID.f_Val.ToValue(out l_FileID);

            a_Dest = new FileSearchPanel.FoundFile(ref l_FileID,
                                                   f_FileName,
                                                   f_FileSize,
                                                   p_CompletedSources,
                                                   p_Sources,
                                                   a_StorableTime,
                                                   a_StorableTime,
                                                   a_TempBuffer);
        }
    }

    class PacketHandler
    {
        public const byte c_MinDesiredKadRequestFitness = 4;

        public const ushort c_MaxKadMessageSize = (SamBridgeManager.c_MaxDatagramPayloadOutgoing - 2);
        private const ushort c_ProcessBootstrapRequestMaxContacts = ((c_MaxKadMessageSize - sizeof(ushort)) / Contact.c_MaximumSize);
        private const ushort c_ProcessKademliaRequestMaxContacts = ((c_MaxKadMessageSize - UInt128.c_ByteSize - 1) / Contact.c_MaximumSize);

        public const bool c_ExpectNormalStrsAsUTF8 = true;

        private readonly Kademlia f_Kad;
        private readonly KadMap f_Map;

        private byte[] f_SelfDetails;

        public PacketHandler(Kademlia a_Kad)
        {
            f_Kad = a_Kad;
            f_Map = a_Kad.f_Map;
        }

        public void NotifyAddressUpdate()
        {
            MemoryIO l_DetailsBuf = MemoryIO.CreateWritableAndGrowable(Contact.c_MaximumSize);
            f_Kad.f_Self.ToBuffer(l_DetailsBuf, NetworkManager.c_RestrictToClassicAddresses);
            f_SelfDetails = l_DetailsBuf.ToArray(true);
        }

        public void PingContact(Contact a_Contact)
        {
            // May only be called while we are connected to Kademlia.

            PingContact(f_Kad.f_Inst.GetTime(), a_Contact);
        }

        public void PingContact(Time a_Time,
                                Contact a_Contact)
        {
            // Don't ping if a ping is going on already.
            if (false == a_Contact.WaitingForPong(a_Time))
            {
                PingSource(a_Contact.p_DatagramAddr);

                a_Contact.NotifyPing(a_Time);
            }
        }

        private void PingSource(I2PAddress a_Addr)
        {
            SendMyDetails(KadOpcodes.KADEMLIA_HELLO_REQ, a_Addr);
        }

        private void PongContact(Contact a_Contact)
        {
            SendMyDetails(KadOpcodes.KADEMLIA_HELLO_RES, a_Contact.p_DatagramAddr);
        }

        public void SendBootstrapRequest(Time a_Time,
                                         Contact a_Contact)
        {
            SendMyDetails(KadOpcodes.KADEMLIA_BOOTSTRAP_REQ, a_Contact.p_DatagramAddr);

            a_Contact.NotifyPing(a_Time);
        }

        public void SendKademliaRequest(Contact a_Contact,
                                        MemoryIO a_Buffer)
        {
            SendPacket(KadOpcodes.KADEMLIA_REQ, a_Contact.p_DatagramAddr, a_Buffer);

            a_Contact.NotifyPing(f_Kad.f_Inst.GetTime());
        }

        private void SendMyDetails(KadOpcodes a_Opcode,
                                   I2PAddress a_Dest)
        {
            ByteArray l_Data = new ByteArray(f_SelfDetails);
            SendPacket(a_Opcode, a_Dest, ref l_Data);
        }

        public void SendPacket(KadOpcodes a_Opcode,
                               I2PAddress a_Dest,
                               MemoryIO a_Buffer)
        {
            // This routine takes ownership of a_Buffer.

            ByteArray l_Data = new ByteArray(a_Buffer.ToArray(true));
            SendPacket(a_Opcode, a_Dest, ref l_Data);
        }

        public bool SendPacket(KadOpcodes a_Opcode,
                               I2PAddress a_Dest,
                               ref ByteArray a_Data)
        {
            Packet l_Packet = Packet.BuildOwnPacket(a_Dest, a_Opcode, f_Kad.f_Inst.f_NetworkMngr.f_NetworkPrefs.f_AllowDataCompression, ref a_Data);

            return f_Kad.f_Inst.f_UpDwnMngr.EnqueueDatagram(l_Packet);
        }

        public uint BroadcastPacket(KadOpcodes a_Opcode,
                                    I2PAddress[] a_Dests,
                                    ref ByteArray a_Data)
        {
            // Returns the number of packets sent.

            int l_DestCount = a_Dests.Length;

            if (0 == l_DestCount)
            {
                return 0;
            }

            uint l_Done = 0;
            Packet l_Packet = Packet.BuildOwnPacket(a_Dests[0], a_Opcode, f_Kad.f_Inst.f_NetworkMngr.f_NetworkPrefs.f_AllowDataCompression, ref a_Data);

            for (int i = 0; i < l_DestCount; i++)
            {
                if (0 < i)
                {
                    l_Packet = Packet.RecreateOwnPacket(l_Packet, a_Dests[i]);
                }

                if (true == f_Kad.f_Inst.f_UpDwnMngr.EnqueueDatagram(l_Packet))
                {
                    l_Done++;
                }
            }

            return l_Done;
        }

        public void ProcessPacket(Time a_Time,
                                  Packet a_Packet)
        {
            I2PAddress l_Origin = a_Packet.f_RemoteAddr;
            KadOpcodes l_Opcode = a_Packet.p_KadOpcode;

            // l_Contact can be null if we don't know the src yet.
            Contact l_Contact = f_Kad.f_ContactMngr.GetContact(l_Origin);

            if (null != l_Contact)
            {
                f_Kad.f_ContactMngr.NotifyActivity(a_Time, l_Contact);
            }

            BufferIO l_Buffer = a_Packet.p_Buffer;

#region Dispatch

            // A .NET function pointer array might be
            // slower than a 12 item switch, since each
            // delegate is an object (the CPU might have
            // to wait for the delegate to be transferred
            // from RAM to cache first).

            switch (l_Opcode)
            {
                case KadOpcodes.KADEMLIA_BOOTSTRAP_REQ:
                    ProcessBootstrapRequest(l_Buffer, l_Contact, l_Origin);
                    break;

                case KadOpcodes.KADEMLIA_BOOTSTRAP_RES:
                    ProcessBootstrapResponse(l_Buffer, l_Contact, l_Origin);
                    break;

                case KadOpcodes.KADEMLIA_HELLO_REQ:
                    ProcessHelloRequest(l_Buffer, l_Contact, l_Origin);
                    break;

                case KadOpcodes.KADEMLIA_HELLO_RES:
                    ProcessHelloResponse(l_Buffer, l_Contact, l_Origin);
                    break;

                case KadOpcodes.KADEMLIA_REQ:
                    ProcessKademliaRequest(l_Buffer, l_Contact, l_Origin);
                    break;

                case KadOpcodes.KADEMLIA_RES:
                    ProcessKademliaResponse(l_Buffer, l_Contact, l_Origin);
                    break;

                case KadOpcodes.KADEMLIA_SEARCH_REQ:
                    ProcessSearchRequest(l_Buffer, l_Contact, l_Origin);
                    break;

                case KadOpcodes.KADEMLIA_SEARCH_RES:
                    ProcessSearchResponse(l_Buffer, l_Contact, l_Origin);
                    break;

                case KadOpcodes.KADEMLIA_SRC_NOTES_REQ:
                    ProcessSearchNotesRequest(l_Buffer, l_Contact, l_Origin);
                    break;

                case KadOpcodes.KADEMLIA_PUBLISH_REQ:
                    ProcessPublishRequest(l_Buffer, l_Contact, l_Origin);
                    break;

                case KadOpcodes.KADEMLIA_PUBLISH_RES:
                    ProcessPublishResponse(l_Buffer, l_Contact, l_Origin);
                    break;

                case KadOpcodes.KADEMLIA_PUB_NOTES_REQ:
                    ProcessPublishNotesRequest(l_Buffer, l_Contact, l_Origin);
                    break;

                default:
                    return;
            }

#endregion

            if (null == l_Contact)
            {
                // Check if the contact is known now.

                l_Contact = f_Kad.f_ContactMngr.GetContact(l_Origin);

                if (null != l_Contact)
                {
                    f_Kad.f_ContactMngr.NotifyActivity(a_Time, l_Contact);
                }
            }
        }

        private Contact AddOrUpdateContact(BufferIO a_Buffer,
                                           I2PAddress a_Origin)
        {
            // Return can be null.

            Contact l_Contact;

            if (BufferIO.LoadResult.Success != Contact.FromBuffer(f_Kad, a_Buffer, out l_Contact))
            {
                f_Kad.f_Stats.NotifyCorruptValues();

                return null;
            }

            if (true == a_Origin.ValueEquals(l_Contact.p_DatagramAddr))
            {
                bool l_Junk;
                return f_Kad.TrySetContact(l_Contact, true, out l_Junk);
            }

            return null;
        }

        private List<Contact> ReadContactsAtBufferEnd(BufferIO a_Buffer,
                                                      uint a_Count)
        {
            // Returns null if there is no valid contact.

            uint l_ExpectedMinSize = ((uint)a_Count * Contact.c_MinimumSize);

            if (false == f_Kad.VerifyRemainingMin(a_Buffer, l_ExpectedMinSize))
            {
                return null;
            }

            List<Contact> l_Results = new List<Contact>((int)a_Count);

            bool l_Error = false;
            for (uint i = 0; i < a_Count; i++)
            {
                Contact l_Contact;
                switch (Contact.FromBuffer(f_Kad, a_Buffer, out l_Contact))
                {
                    case BufferIO.LoadResult.Success:
                        l_Results.Add(l_Contact);
                        break;

                    case BufferIO.LoadResult.InvalidDataButBufferPosIsStillGood:
                        if (false == l_Error)
                        {
                            f_Kad.f_Stats.NotifyCorruptValues();
                        }
                        l_Error = true;
                        continue;

                    case BufferIO.LoadResult.Corruption:
                        break;

                    default: throw (new ApplicationException());
                }
            }

            return (0 == l_Results.Count ? null : l_Results);
        }

        private List<Contact> AddContacts(List<Contact> a_Results,
                                          Contact a_Contact,
                                          I2PAddress a_Origin,
                                          bool a_NeedResultList)
        {
            // Returns matching contact instances that are
            // actually in the zone.

            List<Contact> a_Output = null;

            if (true == a_NeedResultList)
            {
                a_Output = new List<Contact>(a_Results.Count);
            }

            bool l_SrcHasMyAddressWrong = false;

            foreach (Contact l_ListContact in a_Results)
            {
                bool l_IsMe;
                Contact l_AddedContact = f_Kad.TrySetContact(l_ListContact, false, out l_IsMe);

                if (null == l_AddedContact)
                {
                    if (true == l_IsMe &&
                        false == f_Kad.f_Self.HasExactSameAddresses(l_ListContact))
                    {
                        l_SrcHasMyAddressWrong = true;
                    }
                }
                else if (true == a_NeedResultList)
                {
                    a_Output.Add(l_AddedContact);
                }
            }

            if (true == l_SrcHasMyAddressWrong)
            {
                f_Kad.f_Stats.NotifyPeerHasMyAddressesWrong();

                if (null != a_Contact)
                {
                    PingContact(a_Contact);
                }
                else
                {
                    PingSource(a_Origin);
                }
            }

            return a_Output;
        }

        // KADEMLIA_BOOTSTRAP_REQ
        private void ProcessBootstrapRequest(BufferIO a_Buffer,
                                             Contact a_Contact,
                                             I2PAddress a_Origin)
        {
            if (false == f_Kad.VerifyRemainingMin(a_Buffer, Contact.c_MinimumSize))
            {
                return;
            }

            Contact l_Contact = AddOrUpdateContact(a_Buffer, a_Origin);

            if (null == l_Contact)
            {
                return;
            }

            // We will always append our own contact at the
            // end of the list.

            bool l_RestrictToClassicAddresses = NetworkManager.ShouldRestrictPeerToClassicAddresses(a_Origin);
            Contact[] l_Contacts = f_Kad.f_ContactMngr.GetBootstrapContactsForPeer(l_RestrictToClassicAddresses);
            int l_CountInclMe = Math.Min(c_ProcessBootstrapRequestMaxContacts, l_Contacts.Length + 1);
            int l_CountExclMe = Math.Min(c_ProcessBootstrapRequestMaxContacts - 1, l_Contacts.Length);

            MemoryIO l_Buffer = MemoryIO.CreateWritableAndGrowable((sizeof(ushort) + ((uint)l_CountInclMe * Contact.c_MaximumSize)));
            l_Buffer.WriteUInt16LE((ushort)l_CountInclMe);

            for (int i = 0; i < l_CountExclMe; i++)
            {
                l_Contacts[i].ToBuffer(l_Buffer, l_RestrictToClassicAddresses);
            }

            if (false == l_RestrictToClassicAddresses ||
                true == f_Kad.f_Self.p_HasClassicAddresses)
            {
                f_Kad.f_Self.ToBuffer(l_Buffer, l_RestrictToClassicAddresses);
            }

            SendPacket(KadOpcodes.KADEMLIA_BOOTSTRAP_RES, a_Origin, l_Buffer);
        }

        // KADEMLIA_BOOTSTRAP_RES
        private void ProcessBootstrapResponse(BufferIO a_Buffer,
                                              Contact a_Contact,
                                              I2PAddress a_Origin)
        {
            if (false == f_Kad.VerifyRemainingMin(a_Buffer, sizeof(ushort)))
            {
                return;
            }

            uint l_Count = a_Buffer.ReadUInt16LE();
            List<Contact> l_Results = ReadContactsAtBufferEnd(a_Buffer, l_Count);

            if (null != l_Results)
            {
                AddContacts(l_Results, a_Contact, a_Origin, false);
            }
        }

        // KADEMLIA_HELLO_REQ
        private void ProcessHelloRequest(BufferIO a_Buffer,
                                         Contact a_Contact,
                                         I2PAddress a_Origin)
        {
            if (false == f_Kad.VerifyRemainingMin(a_Buffer, Contact.c_MinimumSize))
            {
                return;
            }

            Contact l_Contact = AddOrUpdateContact(a_Buffer, a_Origin);

            if (null != l_Contact)
            {
                PongContact(l_Contact);
            }
        }

        // KADEMLIA_HELLO_RES
        private void ProcessHelloResponse(BufferIO a_Buffer,
                                          Contact a_Contact,
                                          I2PAddress a_Origin)
        {
            if (false == f_Kad.VerifyRemainingMin(a_Buffer, Contact.c_MinimumSize))
            {
                return;
            }

            AddOrUpdateContact(a_Buffer, a_Origin);
        }

        // KADEMLIA_REQ
        private void ProcessKademliaRequest(BufferIO a_Buffer,
                                            Contact a_Contact,
                                            I2PAddress a_Origin)
        {
            f_Kad.f_Stats.f_KadRequests++;

            if (false == f_Kad.VerifyRemaining(a_Buffer, sizeof(byte) + (3 * UInt128.c_ByteSize)))
            {
                return;
            }

            KadParams l_Param = (KadParams)a_Buffer.ReadByte();
            // The id that's looked for.
            ClientID l_Target = new ClientID(a_Buffer.ReadUInt128ME());
            // The id of the peer now asked (should be my id).
            ClientID l_Receiver = new ClientID(a_Buffer.ReadUInt128ME());
            // The id of the requesting peer.
            ClientID l_Requestor = new ClientID(a_Buffer.ReadUInt128ME());

            if (false == f_Kad.IsMyClientID(l_Receiver))
            {
                f_Kad.f_Stats.NotifyMalformedKadReq();

                return;
            }

            if (null != a_Contact)
            {
                if (false == a_Contact.HasClientID(l_Requestor))
                {
                    f_Kad.f_Stats.NotifyCorruptValues();

                    return;
                }
            }
            else
            {
                if (true == f_Kad.f_RootZone.WouldLikeToAdd(l_Requestor))
                {
                    // Try to add the requesting peer to our
                    // contacts.

                    PingSource(a_Origin);
                }
            }

            Distance l_MyDistance = f_Kad.CalcDistance(l_Target);
            uint l_MinFitness = l_MyDistance.GetFitness();
            // We always require at least a fitness of 2,
            // but we even demand 4 if suitable.
            l_MinFitness = Math.Max(2u, Math.Min(c_MinDesiredKadRequestFitness, l_MinFitness));

            bool l_RestrictToClassicAddresses = NetworkManager.ShouldRestrictPeerToClassicAddresses(a_Origin);
            SortedList<Distance, Contact> l_Results = new SortedList<Distance, Contact>(c_ProcessKademliaRequestMaxContacts);
            RoutingZone.GetClosestToArgs l_Args = new RoutingZone.GetClosestToArgs(f_Kad, l_Target, c_ProcessKademliaRequestMaxContacts, ContactVitality.NeedAnswer2, l_MinFitness, l_RestrictToClassicAddresses, l_Results);

            l_Args.f_VerifiedOnly = true;
            f_Kad.f_RootZone.GetClosestTo(ref l_Args);

            if (0 < l_Results.Count)
            {
                uint l_Count = (uint)l_Results.Count;
                MemoryIO l_Buffer = MemoryIO.CreateWritableAndGrowable(UInt128.c_ByteSize + (l_Count * Contact.c_MaximumSize) + sizeof(byte));

                l_Buffer.WriteUInt128ME(l_Target.f_Val);
                l_Buffer.WriteByte((byte)l_Count);

                foreach (Contact l_Contact in l_Results.Values)
                {
                    l_Contact.ToBuffer(l_Buffer, l_RestrictToClassicAddresses);
                }

                SendPacket(KadOpcodes.KADEMLIA_RES, a_Origin, l_Buffer);
            }
        }

        // KADEMLIA_RES
        private void ProcessKademliaResponse(BufferIO a_Buffer,
                                             Contact a_Contact,
                                             I2PAddress a_Src)
        {
            if (false == f_Kad.VerifyRemainingMin(a_Buffer, UInt128.c_ByteSize + sizeof(byte)))
            {
                return;
            }

            if (null == a_Contact)
            {
                return;
            }

            UInt128 l_SearchKey = a_Buffer.ReadUInt128ME();
            uint l_Count = a_Buffer.ReadByte();
            List<Contact> l_Results = ReadContactsAtBufferEnd(a_Buffer, l_Count);

            if (null == l_Results)
            {
                return;
            }

            l_Results = AddContacts(l_Results, a_Contact, a_Src, true);

            // Even if there is no data in the list, we need to
            // call RegisterKadResponse(), to notify that the
            // contact has responded.

            bool OK = f_Kad.f_SearchMngr.RegisterKadResponse(l_SearchKey, a_Contact, l_Results);

            if (false == OK)
            {
                // This isn't necessarily an error (it's probably
                // relating to a search that has already expired).
            }
        }

        // KADEMLIA_SEARCH_REQ
        private void ProcessSearchRequest(BufferIO a_Buffer,
                                          Contact a_Contact,
                                          I2PAddress a_Src)
        {
            if (false == f_Kad.VerifyRemainingMin(a_Buffer, UInt128.c_ByteSize + sizeof(bool)))
            {
                return;
            }

            UInt128 l_ID = a_Buffer.ReadUInt128ME();
            bool l_Flag = a_Buffer.ReadBool();

            if (false == a_Buffer.EndReached())
            {
                bool l_HasSearchTerms = l_Flag;

                if (true == l_HasSearchTerms)
                {
                    SearchTerm l_SearchTerms = SearchTerm.FromBuffer(a_Buffer);

                    if (null == l_SearchTerms)
                    {
                        f_Kad.f_Stats.NotifyCorruptValues();
                    }
                    else if (true == f_Kad.VerifyEndReached(a_Buffer))
                    {
                        f_Map.SendValidSearchTermsResult(new KeywordID(l_ID), l_SearchTerms, a_Src);
                    }

                    return;
                }
            }
            else
            {
                bool l_SourceRequest = l_Flag;

                if (true == l_SourceRequest)
                {
                    f_Map.SendFileResult(new FileID(l_ID), a_Src);

                    return;
                }
            }

            f_Map.SendValidKeywordResult(new KeywordID(l_ID), a_Src);
        }

        // KADEMLIA_SEARCH_RES
        private void ProcessSearchResponse(BufferIO a_Buffer,
                                           Contact a_Contact,
                                           I2PAddress a_Src)
        {
            // Expect at least the key (16) and the number of
            // results (2).
            if (false == f_Kad.VerifyRemainingMin(a_Buffer, UInt128.c_ByteSize + sizeof(ushort)))
            {
                return;
            }

            if (null == a_Contact)
            {
                f_Kad.f_Stats.NotifyCorruptValues();

                return;
            }

            bool Handled = f_Kad.f_SearchMngr.ProcessSearchResultList(a_Contact, a_Buffer);

            if (false == Handled)
            {
                // This indicates that the corresponding search
                // has expired, or that the feedback we've received
                // is corrupt (unlikely).
                f_Kad.f_Stats.NotifyUnprocessableSearchResult();
            }
        }

        // KADEMLIA_PUBLISH_REQ
        private void ProcessPublishRequest(BufferIO a_Buffer,
                                           Contact a_Contact,
                                           I2PAddress a_Src)
        {
            if (false == f_Kad.VerifyRemainingMin(a_Buffer, UInt128.c_ByteSize + sizeof(ushort)))
            {
                return;
            }

            if (true == f_Kad.f_KadPrefs.f_PublAcceptOnlyTrustedUsers &&
                false == f_Kad.f_Inst.f_UserMngr.MayDatagramDestinationPublish(a_Src))
            {
                f_Kad.f_Stats.NotifyUntrustedPublishRequest();

                return;
            }

            // l_Key is the file id or keyword id that
            // gets published.
            UInt128 l_Key = a_Buffer.ReadUInt128ME();
            Distance l_Dist = f_Kad.CalcDistance(new ClientID(l_Key));

            if (false == l_Dist.HasMinFitness(2))
            {
                f_Kad.f_Stats.NotifyOddPublishRequest();

                return;
            }

            ushort l_Count = a_Buffer.ReadUInt16LE();

            bool l_HadUserID = false;
            UserID l_OriginUserID = new UserID();
            bool l_ResponseSent = false;

            for (int i = 0; i < l_Count; i++)
            {
                if (false == f_Kad.VerifyRemainingMin(a_Buffer, UInt128.c_ByteSize + sizeof(byte)))
                {
                    return;
                }

                // For sources l_Reference is the publisher's user id,
                // for keywords l_Reference is the id of the referenced file.
                UInt128 l_Reference = a_Buffer.ReadUInt128ME();
                TagList l_TagList = TagList.ReadListUntrusted(a_Buffer, a_Buffer.ReadByte(), c_ExpectNormalStrsAsUTF8);

                if (null == l_TagList)
                {
                    f_Kad.f_Stats.NotifyCorruptValues();

                    return;
                }

                if (null != l_TagList.GetTag(TagCodes.TAG_SOURCETYPE))
                {
                    f_Kad.f_Stats.f_SourcePublishRequestsRcvd++;

                    if (false == l_HadUserID)
                    {
                        l_OriginUserID = new UserID(l_Reference);

                        if (false == l_OriginUserID.IsValid())
                        {
                            f_Kad.f_Stats.NotifyCorruptValues();

                            return;
                        }

                        if (true == f_Kad.IsMyUserID(l_OriginUserID))
                        {
                            // An impostor.

                            return;
                        }

                        l_HadUserID = true;
                    }
                    else if (false == l_OriginUserID.f_Val.Equals(l_Reference))
                    {
                        f_Kad.f_Stats.NotifyCorruptValues();

                        return;
                    }

                    FileEntry l_Entry = FileEntry.FromTagList(f_Kad, new FileID(l_Key), l_OriginUserID, l_TagList, a_Src);

                    if (null == l_Entry)
                    {
                        f_Kad.f_Stats.NotifyCorruptValues();

                        continue;
                    }

                    int l_Load;
                    if (true == f_Map.AddFileSource(l_Entry, true, out l_Load) &&
                        false == l_ResponseSent)
                    {
                        uint l_CompletedSources;
                        uint l_Sources = f_Map.GetFileSourceCount(l_Entry.f_KeyID, out l_CompletedSources);

                        MemoryIO l_Buffer = MemoryIO.CreateWritable(UInt128.c_ByteSize + sizeof(uint) +
                                                                     sizeof(uint) + sizeof(byte));
                        l_Buffer.WriteUInt128ME(l_Entry.f_KeyID.f_Val);
                        l_Buffer.WriteUInt32LE(l_Sources);
                        l_Buffer.WriteUInt32LE(l_CompletedSources);
                        l_Buffer.WriteByte((byte)l_Load);

                        SendPacket(KadOpcodes.KADEMLIA_PUBLISH_RES, a_Src, l_Buffer);
                        l_ResponseSent = true;
                    }
                }
                else
                {
                    f_Kad.f_Stats.f_KeywordPublishRequestsRcvd++;

                    KeywordEntry l_Entry = KeywordEntry.FromTagList(f_Kad, new KeywordID(l_Key), new FileID(l_Reference), l_TagList);

                    if (null == l_Entry)
                    {
                            f_Kad.f_Stats.NotifyCorruptValues();

                        continue;
                    }

                    int l_Load;
                    if (false == f_Map.AddKeyword(l_Entry, out l_Load) &&
                        false == l_ResponseSent)
                    {
                        MemoryIO l_Buffer = MemoryIO.CreateWritable(UInt128.c_ByteSize + sizeof(byte));
                        l_Buffer.WriteUInt128ME(l_Entry.f_KeyID.f_Val);
                        l_Buffer.WriteByte((byte)l_Load);

                        SendPacket(KadOpcodes.KADEMLIA_PUBLISH_RES, a_Src, l_Buffer);
                        l_ResponseSent = true;
                    }
                }
            }

            f_Kad.VerifyEndReached(a_Buffer);
        }

        // KADEMLIA_PUBLISH_RES
        private void ProcessPublishResponse(BufferIO a_Buffer,
                                            Contact a_Contact,
                                            I2PAddress a_Src)
        {
            f_Kad.f_PublishMngr.NotifyPublishResponse();
        }

        // KADEMLIA_SRC_NOTES_REQ
        private void ProcessSearchNotesRequest(BufferIO a_Buffer,
                                               Contact a_Contact,
                                               I2PAddress a_Src)
        {
            if (false == f_Kad.VerifyRemainingMin(a_Buffer, 2 * UInt128.c_ByteSize))
            {
                return;
            }

            FileID l_FileID = new FileID(a_Buffer.ReadUInt128ME());

            if (null != a_Contact)
            {
                // This should be the client id of the issuer of the
                // search request.
                ClientID l_ClientID = new ClientID(a_Buffer.ReadUInt128ME());

                if (false == l_ClientID.f_Val.Equals(a_Contact.f_ClientID.f_Val))
                {
                    // This is not a terrible error, no
                    // need to abort.

                    f_Kad.f_Stats.NotifyOddPublishRequest();
                }
            }

            f_Map.SendNoteResult(l_FileID, a_Src);
        }

        // KADEMLIA_PUB_NOTES_REQ
        private void ProcessPublishNotesRequest(BufferIO a_Buffer,
                                                Contact a_Contact,
                                                I2PAddress a_Src)
        {
            f_Kad.f_Stats.f_NotesPublishRequestsRcvd++;

            if (false == f_Kad.VerifyRemainingMin(a_Buffer, (2 * UInt128.c_ByteSize * 2) + sizeof(byte)))
            {
                return;
            }

            if (true == f_Kad.f_KadPrefs.f_PublAcceptOnlyTrustedUsers &&
                false == f_Kad.f_Inst.f_UserMngr.MayDatagramDestinationPublish(a_Src))
            {
                f_Kad.f_Stats.NotifyUntrustedPublishRequest();

                return;
            }

            // The file id of the file the notes belong to.
            UInt128 l_FileID = a_Buffer.ReadUInt128ME();
            Distance l_Dist = f_Kad.CalcDistance(new ClientID(l_FileID));

            if (false == l_Dist.HasMinFitness(2))
            {
                f_Kad.f_Stats.NotifyOddPublishRequest();

                return;
            }

            // The client id of the publisher.
            ClientID l_ClientID = new ClientID(a_Buffer.ReadUInt128ME());

            if (null != a_Contact &&
                false == a_Contact.HasClientID(l_ClientID))
            {
                f_Kad.f_Stats.NotifyCorruptValues();

                return;
            }

            if (true == f_Kad.f_Self.HasClientID(l_ClientID))
            {
                f_Kad.f_Stats.NotifyCorruptValues();

                // An impostor.

                return;
            }

            TagList l_TagList = TagList.ReadListUntrusted(a_Buffer, a_Buffer.ReadByte(), c_ExpectNormalStrsAsUTF8);

            if (null == l_TagList)
            {
                f_Kad.f_Stats.NotifyCorruptValues();

                return;
            }

            NoteEntry l_Entry = NoteEntry.FromTagList(f_Kad, new FileID(l_FileID), l_ClientID, l_TagList);

            if (null == l_Entry)
            {
                f_Kad.f_Stats.NotifyCorruptValues();

                return;
            }

            int l_Load;

            if (true == f_Kad.f_Map.AddNote(l_Entry, out l_Load))
            {
                MemoryIO l_Answer = MemoryIO.CreateWritable(UInt128.c_ByteSize + sizeof(byte));
                l_Answer.WriteUInt128ME(l_Entry.f_KeyID.f_Val);
                l_Answer.WriteByte((byte)l_Load);

                SendPacket(KadOpcodes.KADEMLIA_PUBLISH_RES, a_Src, l_Answer);
            }

            f_Kad.VerifyEndReached(a_Buffer);
        }
    }
}

// End of file.
