/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.27-src\Nachtblitz\Language.cs
 * PURPOSE: Translation handling
 * VERSION: 1.4.27
 * DATE: March 23, 2016
 * COPYRIGHT: Copyright (C) 2012-2016 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using Support;
using Support.Encodings;

namespace NB.Language
{
    class TextSet
    {
        private const ushort c_ItemCount = 810;
        private const ushort c_MaxTextLen = 2048;
        private const uint c_MaxTextSetFileSize = (750 * 1024);

        private const char c_CommentChar = ';';
        private const char c_AssignmentChar = '=';
        private const char c_StringLiteralChar = '"';
        private const char c_EscapeChar = '\\';
        private const byte c_ShortHexCharLen = 2;
        private const byte c_LongHexCharLen = 4;

        private const string c_TimeFormat = "G";
        private const string c_DateFormat = "d";
        private const string c_IntegerFormat = "N0";
        // For number strings without thousand separators:
        private const string c_PlainIntegerFormat = "F0";
        private const string c_PercentageFormat = "P1";
        private const string c_FloatFormat = "N3";
        private const string c_ShortFloatFormat = "0.#";

        public readonly CultureInfo f_LanguageCulture;
        public readonly CultureInfo f_FormatCulture;
        private readonly Dictionary<TextKeys, string> f_Texts;

        private readonly bool f_IsEnglish;
        public readonly string f_AppName;
        public readonly string f_AppNameWithVer;

        private readonly string f_KeyValuePairFormat;
        private readonly string f_KeyFormat;
        private readonly string f_CntrlsInputCaptionFormat;
        private readonly string f_CntrlsOpenDialogCaptionFormat;
        private readonly string f_CntrlsExpandFormButtonFormat;
        private readonly string f_CntrlsCollapseFormButtonFormat;

        private NB.Gui.TimeStrHelper f_TimeStrHlpr;

        private TextSet(CultureInfo a_LanguageCulture,
                        CultureInfo a_FormatCulture,
                        Dictionary<TextKeys, string> a_Texts,
                        bool a_Intern)
        {
            if (true == a_Intern)
            {
                Dictionary<TextKeys, string> l_Clone = new Dictionary<TextKeys, string>(a_Texts);

                foreach (KeyValuePair<TextKeys, string> l_Pair in l_Clone)
                {
                    string l_NewStr = string.Intern(l_Pair.Value);

                    if (false == object.ReferenceEquals(l_NewStr, l_Pair.Value))
                    {
                        a_Texts[l_Pair.Key] = l_NewStr;
                    }
                }
            }

            f_LanguageCulture = a_LanguageCulture;
            f_FormatCulture = a_FormatCulture;
            f_Texts = a_Texts;

            f_IsEnglish = Util.StartsWith(a_LanguageCulture.EnglishName, "English");
            f_AppName = GetString(TextKeys.AppName);
            f_AppNameWithVer = string.Format(f_FormatCulture, f_Texts[TextKeys.ClientNameWithVersion], f_AppName, Software.s_ReleaseInfo.f_VersionStr);

            f_KeyValuePairFormat = GetString(TextKeys.KeyValuePairFormat);
            f_KeyFormat = GetString(TextKeys.KeyFormat);
            f_CntrlsInputCaptionFormat = GetString(TextKeys.CntrlsInputCaptionFormat);
            f_CntrlsOpenDialogCaptionFormat = GetString(TextKeys.CntrlsOpenDialogCaptionFormat);
            f_CntrlsExpandFormButtonFormat = GetString(TextKeys.CntrlsExpandFormButtonFormat);
            f_CntrlsCollapseFormButtonFormat = GetString(TextKeys.CntrlsCollapseFormButtonFormat);

            NB.Gui.TimeStrHelper.Create(out f_TimeStrHlpr);
        }

        public void ApplyCulture()
        {
            System.Threading.Thread.CurrentThread.CurrentCulture = f_FormatCulture;
            System.Windows.Forms.Application.CurrentCulture = f_FormatCulture;

            if (f_FormatCulture.LCID != System.Windows.Forms.Application.CurrentInputLanguage.Culture.LCID)
            {
                System.Windows.Forms.Application.CurrentInputLanguage = System.Windows.Forms.InputLanguage.FromCulture(f_FormatCulture);
            }
        }

        public static TextSet LoadTemplate(CultureInfo a_LanguageCulture,
                                           CultureInfo a_FormatCulture,
                                           BinaryReader a_Rdr)
        {
            uint l_Count = a_Rdr.ReadUInt32();
            Dictionary<TextKeys, string> l_Texts = new Dictionary<TextKeys, string>((int)l_Count);

            for (uint i = 0; i != l_Count; i++)
            {
                TextKeys l_Key = (TextKeys)a_Rdr.ReadUInt16();
                string l_Text = a_Rdr.ReadString();

                l_Texts.Add(l_Key, l_Text);
            }

            PrepareFormats(l_Texts);

            // The application name is the only string we expect to be
            // missing even in the (English) default text set.
            if (false == l_Texts.ContainsKey(TextKeys.AppName))
            {
                l_Texts[TextKeys.AppName] = Software.c_AppName;
            }

            return new TextSet(a_LanguageCulture, a_FormatCulture, l_Texts, true);
        }

#if REFERENCIAL_SOURCE_CODE

        public static byte[] TextSetToBinaryStream(Dictionary<TextKeys, string> a_Texts)
        {
            // This routine can be used to generate the default text set
            // resource that's embedded into the release excutable.
            // The input is expected to originate from ParseTextSet().
            // NOTE: This code is never used in release builds and
            // included only for referencial purposes.

            byte[] l_ResArr;

            using (MemoryStream l_MemStrm = new MemoryStream())
            {
                using (BinaryWriter l_Wtr = new BinaryWriter(l_MemStrm))
                {
                    l_Wtr.Write((uint)a_Texts.Count);

                    foreach (KeyValuePair<TextKeys, string> l_Pair in a_Texts)
                    {
                        string l_Text = l_Pair.Value;

                        l_Wtr.Write((ushort)l_Pair.Key);
                        l_Wtr.Write(l_Text);
                    }

                    l_Wtr.Flush();
                    l_ResArr = l_MemStrm.ToArray();
                }
            }

            ByteArray l_Data = new ByteArray(l_ResArr);

            return Util.DeflateBytes(ref l_Data);
        }

#endif

        public static TextSet ChangeFormatCulture(TextSet a_TemplateTextSet,
                                                  CultureInfo a_FormatCulture)
        {
            // Recreates the text set with the passed format culture.

            if (true == BootConfig.AreSameCultures(a_TemplateTextSet.f_FormatCulture, a_FormatCulture))
            {
                return a_TemplateTextSet;
            }

            return new TextSet(a_TemplateTextSet.f_LanguageCulture, a_FormatCulture, a_TemplateTextSet.f_Texts, false);
        }

        public static TextSet LoadCustom(TextSet a_TemplateTextSet,
                                         CultureInfo a_LanguageCulture,
                                         CultureInfo a_FormatCulture,
                                         bool a_Intern,
                                         out string a_Summary)
        {
            // Return can be null. a_Summary will contain information
            // on any problems in any case.

            string l_LangFile = BootConfig.BuildTranslationFileName(BootConfig.BuildTranslationFolderName(), a_LanguageCulture);
            ulong l_FileSize;
            StreamReader l_Rdr = Util.OpenTextFile(l_LangFile, out l_FileSize);

            string l_SubErrMsg;

            if (null == l_Rdr)
            {
                l_SubErrMsg = "Failed to open the file.";
            }
            else
            {
                if (c_MaxTextSetFileSize < l_FileSize)
                {
                    l_Rdr.Close();

                    l_SubErrMsg = "The file is larger than expected.";
                }
                else
                {
                    TextSet l_TextSet = LoadCustom(a_TemplateTextSet, a_LanguageCulture, a_FormatCulture, a_Intern, l_Rdr, out a_Summary);
                    l_Rdr.Close();

                    if (null != l_TextSet)
                    {
                        return l_TextSet;
                    }

                    l_SubErrMsg = a_Summary;
                }
            }

            a_Summary = string.Format(a_TemplateTextSet.f_FormatCulture, "An error occurred while trying to access the file \"{0}\":\n\n{1}", l_LangFile, l_SubErrMsg);

            return null;
        }

        private static TextSet LoadCustom(TextSet a_TemplateTextSet,
                                          CultureInfo a_LanguageCulture,
                                          CultureInfo a_FormatCulture,
                                          bool a_Intern,
                                          StreamReader a_Rdr,
                                          out string a_Summary)
        {
            StringBuilder l_Bld = new StringBuilder();
            CultureInfo l_InvarCulture = CultureInfo.InvariantCulture;

            string l_ParsingSummary;
            Dictionary<TextKeys, string> l_CustTexts = ParseTextSet(a_Rdr, out l_ParsingSummary);

            if (null != l_ParsingSummary)
            {
                l_Bld.AppendFormat(l_InvarCulture, "\nThe following errors occurred during parsing:\n{0}", l_ParsingSummary);
            }

            if (null == l_CustTexts)
            {
                a_Summary = l_Bld.ToString();

                return null;
            }

            StringBuilder l_MalformedBld = new StringBuilder();
            StringBuilder l_MissingBld = new StringBuilder();

            foreach (KeyValuePair<TextKeys, string> l_VanillaPair in a_TemplateTextSet.f_Texts)
            {
                string l_CustText;

                if (false == l_CustTexts.TryGetValue(l_VanillaPair.Key, out l_CustText))
                {
                    l_CustTexts[l_VanillaPair.Key] = l_VanillaPair.Value;

                    if (TextKeys.AppName != l_VanillaPair.Key)
                    {
                        // It's okay not to have the application name.

                        l_MissingBld.AppendLine(((ushort)l_VanillaPair.Key).ToString(l_InvarCulture));
                    }
                }
                else
                {
                    uint l_VanillaCount;

                    if (false == CountPlaceholders(true, l_VanillaPair.Value, out l_VanillaCount))
                    {
                        throw (new ApplicationException());
                    }

                    string l_ErrMsg = null;

                    uint l_CustCount;
                    if (0 >= l_CustText.Length &&
                        0 < l_VanillaPair.Value.Length)
                    {
                        l_ErrMsg = string.Format(l_InvarCulture, "Error in key {0}: Empty value.", (ushort)l_VanillaPair.Key);
                    }
                    else if (false == CountPlaceholders(false, l_CustText, out l_CustCount))
                    {
                        l_ErrMsg = string.Format(l_InvarCulture, "Error in key {0}: Illegal placeholders.", (ushort)l_VanillaPair.Key);
                    }
                    else if (l_CustCount != l_VanillaCount)
                    {
                        l_ErrMsg = string.Format(l_InvarCulture, "Key {0}: Illegal number of placeholders ({1} expected, {2} found).", (ushort)l_VanillaPair.Key, l_VanillaCount, l_CustCount);
                    }

                    if (null != l_ErrMsg)
                    {
                        l_MalformedBld.AppendLine(l_ErrMsg);
                        l_CustTexts[l_VanillaPair.Key] = l_VanillaPair.Value;
                    }
                }
            }

            List<TextKeys> l_SurplusLs = new List<TextKeys>();

            foreach (TextKeys l_CustKey in l_CustTexts.Keys)
            {
                if (false == a_TemplateTextSet.f_Texts.ContainsKey(l_CustKey))
                {
                    l_SurplusLs.Add(l_CustKey);
                }
            }

            StringBuilder l_SurplusBld = new StringBuilder();

            foreach (TextKeys l_Key in l_SurplusLs)
            {
                l_SurplusBld.AppendLine(((ushort)l_Key).ToString(l_InvarCulture));
                l_CustTexts.Remove(l_Key);
            }

            if (7 == a_LanguageCulture.LCID &&
                (2055 == Software.s_StartupCulture.LCID ||
                 5127 == Software.s_StartupCulture.LCID))
            {
                // Substitute the eszett with double s if the general
                // German text set was loaded and the startup culture
                // was German of Switzerland or Liechtenstein.

                SubstituteEszett(l_CustTexts);
            }

            // We have now finished loading.

            PrepareFormats(l_CustTexts);

            if (0 != l_MalformedBld.Length)
            {
                l_Bld.AppendFormat(l_InvarCulture, "\nSome texts could not be used due to problems regarding placeholders (default text is used instead):\n{0}", l_MalformedBld.ToString());
            }

            if (0 != l_MissingBld.Length)
            {
                l_Bld.AppendFormat(l_InvarCulture, "\nThe following text-keys are missing/invalid (default text is used instead):\n{0}", l_MissingBld.ToString());
            }

            if (0 != l_SurplusBld.Length)
            {
                l_Bld.AppendFormat(l_InvarCulture, "\nThe following text-keys are unknown and were ignored:\n{0}", l_SurplusBld.ToString());
            }

            a_Summary = l_Bld.ToString();

            return new TextSet(a_LanguageCulture, a_FormatCulture, l_CustTexts, a_Intern);
        }

        private static void SubstituteEszett(Dictionary<TextKeys, string> a_Dic)
        {
            TextKeys[] l_KeyArr = new TextKeys[a_Dic.Count];
            a_Dic.Keys.CopyTo(l_KeyArr, 0);
            StringBuilder l_Bld = new StringBuilder();

            foreach (TextKeys l_Key in l_KeyArr)
            {
                string l_Str = a_Dic[l_Key];
                int l_StrLen = l_Str.Length;
                int l_ScanIdx = 0;

                while (true)
                {
                    int l_Idx = l_Str.IndexOf('\x00DF', l_ScanIdx);

                    if (0 <= l_Idx ||
                        0 < l_ScanIdx)
                    {
                        int l_PortionLen = ((0 > l_Idx ? l_StrLen : l_Idx) - l_ScanIdx);

                        if (0 < l_PortionLen)
                        {
                            l_Bld.Append(l_Str, l_ScanIdx, l_PortionLen);
                        }
                    }

                    if (0 > l_Idx)
                    {
                        break;
                    }

                    bool l_Uppercase = ((0 < l_Idx &&
                                         true == char.IsUpper(l_Str[l_Idx - 1])) &&
                                        ((l_StrLen - 1) > l_Idx &&
                                         false == char.IsLower(l_Str[l_Idx + 1])));
                    char l_Char = (false == l_Uppercase ? 's' : 'S');

                    l_Bld.Append(l_Char);
                    l_Bld.Append(l_Char);
                    l_ScanIdx = (l_Idx + 1);
                }

                if (0 < l_ScanIdx)
                {
                    // The text was modified.

                    a_Dic[l_Key] = l_Bld.ToString();
                    l_Bld.Length = 0;
                }
            }
        }

        public string JoinAllStringsForSpellChecking(TextSet a_VanillaTextSet)
        {
            // a_VanillaTextSet can be null. Otherwise each
            // key-value-pair in this text set that already occurres
            // in the vanilla text set is not added to the output.

            SortedDictionary<TextKeys, string> l_SortedTexts = new SortedDictionary<TextKeys, string>(this.f_Texts);

            StringBuilder l_Bld = new StringBuilder();

            foreach (KeyValuePair<TextKeys, string> l_Pair in l_SortedTexts)
            {
                string l_CurText = l_Pair.Value;

                if (0 == l_CurText.Trim().Length)
                {
                    continue;
                }

                string l_VanillaText;
                if (null != a_VanillaTextSet &&
                    true == a_VanillaTextSet.f_Texts.TryGetValue(l_Pair.Key, out l_VanillaText) &&
                    l_CurText == l_VanillaText)
                {
                    // This text is the same as in the vanilla set, so
                    // skip it.

                    continue;
                }

                l_Bld.AppendLine(l_CurText);
            }

            return l_Bld.ToString();
        }

        public string GetString(TextKeys a_Key)
        {
            string l_Text;
            if (true == f_Texts.TryGetValue(a_Key, out l_Text))
            {
                return l_Text;
            }

            return string.Empty;
        }

        public string GetAsKeyPortion(TextKeys a_Key)
        {
            return Format(f_KeyFormat, GetString(a_Key));
        }

        public string GetAsInputControlCaption(TextKeys a_Key)
        {
            return Format(f_CntrlsInputCaptionFormat, GetString(a_Key));
        }

        public string GetAsDialogOpener(TextKeys a_Key)
        {
            return GetAsDialogOpener(GetString(a_Key));
        }

        public string GetAsDialogOpener(string a_Str)
        {
            return Format(f_CntrlsOpenDialogCaptionFormat, a_Str);
        }

        public string GetAsExpandButtonCaption(TextKeys a_Key)
        {
            return Format(f_CntrlsExpandFormButtonFormat, GetString(a_Key));
        }

        public string GetAsCollapseButtonCaption(TextKeys a_Key)
        {
            return Format(f_CntrlsCollapseFormButtonFormat, GetString(a_Key));
        }

        public string[] ReadArray(TextKeys a_InitialKey,
                                  int a_Count)
        {
            string[] l_Arr = new string[a_Count];

            TextKeys l_Keys = a_InitialKey;
            for (int i = 0; i < a_Count; i++, l_Keys++)
            {
                l_Arr[i] = GetString(l_Keys);
            }

            return l_Arr;
        }

        public string MakeTitlecasedIfEn(string a_Str)
        {
            if (false == f_IsEnglish)
            {
                return a_Str;
            }

            StringBuilder Bld = new StringBuilder(a_Str);

            bool HadWhite = true;

            for (int i = 0, lim = a_Str.Length; i < lim; i++)
            {
                if (true == char.IsLetter(Bld[i]))
                {
                    if (true == HadWhite)
                    {
                        Bld[i] = char.ToUpper(Bld[i]);

                        HadWhite = false;
                    }
                }
                else if (true == char.IsWhiteSpace(Bld[i]))
                {
                    HadWhite = true;
                }
            }

            return Bld.ToString();
        }

        public static string EnsureHasLineBreak(string a_Text,
                                                int a_BreakIfLongerThanChars)
        {
            // Ensures that the return value has at least one line
            // break if a_Text is longer than a_BreakIfLongerThanChars
            // and a suitable position for line breaking is found.

            const char l_LF = '\n';
            const char l_Space = (char)0x20;
            int l_Len = a_Text.Length;

            if (0 >= a_BreakIfLongerThanChars ||
                l_Len <= a_BreakIfLongerThanChars ||
                0 <= a_Text.IndexOf(l_LF))
            {
                return a_Text;
            }

            // Find a space char near the middle.

            int l_Mid = (l_Len / 2);
            int l_Idx0 = a_Text.LastIndexOf(l_Space, l_Mid, l_Mid);
            int l_Idx1 = a_Text.IndexOf(l_Space, l_Mid, l_Len - l_Mid);
            int l_TrgtIdx;

            if (0 > l_Idx0)
            {
                if (0 > l_Idx1)
                {
                    return a_Text;
                }

                l_TrgtIdx = l_Idx1;
            }
            else
            {
                if (0 > l_Idx1)
                {
                    l_TrgtIdx = l_Idx0;
                }
                else
                {
                    l_TrgtIdx = ((l_Mid - l_Idx0) <= (l_Idx1 - l_Mid) ? l_Idx0 : l_Idx1);
                }
            }

            // Replace the space char with a line feed char. (We can't
            // do this in one call.)

            string l_TempStr = a_Text.Remove(l_TrgtIdx, 1);

            return l_TempStr.Insert(l_TrgtIdx, l_LF.ToString());
        }

        public string Format(string a_Format,
                             params object[] a_Args)
        {
            // Try to use this routine instead of string.Format(),
            // since it uses the proper culture.

#if !DEBUG
            try
            {
#endif
                return string.Format(f_FormatCulture, a_Format, a_Args);
#if !DEBUG
            }
            catch (FormatException)
            {
                return string.Empty;
            }
#endif
        }

        public string Format(TextKeys a_FormatKey,
                             params object[] a_Args)
        {
            return Format(GetString(a_FormatKey), a_Args);
        }

        public string FormatWithAppName(TextKeys a_Key)
        {
            return Format(a_Key, f_AppName);
        }

        public string BuildKeyValuePair(TextKeys a_Key,
                                        string a_Value)
        {
            return string.Format(f_FormatCulture, f_KeyValuePairFormat, GetString(a_Key), a_Value);
        }

        public string this[TextKeys a_Key]
        {
            get { return GetString(a_Key); }
        }

        public void AddText(StringBuilder a_Bld,
                            TextKeys a_Key,
                            string a_Value)
        {
            a_Bld.AppendFormat(f_FormatCulture, f_KeyValuePairFormat, GetString(a_Key), a_Value);
            a_Bld.AppendLine();
        }

        public void AddInteger(StringBuilder a_Bld,
                               TextKeys a_Key,
                               long a_Value)
        {
            AddText(a_Bld, a_Key, NumStr(a_Value));
        }

        public string TimeStr(long a_Secs)
        {
            return f_TimeStrHlpr.GetTimeString(a_Secs);
        }

        public string TimeStr(Time a_Time,
                              string a_TextIfNotSet)
        {
            // a_TextIfNotSet may be null.

            if (false == a_Time.IsSet())
            {
                return (null == a_TextIfNotSet ? string.Empty : a_TextIfNotSet);
            }

            DateTime l_DateTime = a_Time.ToDateTime();

            return TimeStr(l_DateTime);
        }

        public string TimeStr(DateTime a_DateTime)
        {
            return a_DateTime.ToLocalTime().ToString(c_TimeFormat, f_FormatCulture);
        }

        public string NumStr(long a_Number)
        {
            return a_Number.ToString(c_IntegerFormat, f_FormatCulture);
        }

        public string NumStr(ulong a_Number)
        {
            return a_Number.ToString(c_IntegerFormat, f_FormatCulture);
        }

        public string PercStr(float a_Float)
        {
            return a_Float.ToString(c_PercentageFormat, f_FormatCulture);
        }

        public string FloatStr(float a_Float)
        {
            return a_Float.ToString(c_FloatFormat, f_FormatCulture);
        }

        public string FloatStr(double a_Float,
                               int a_Precision)
        {
            return a_Float.ToString("N" + a_Precision, f_FormatCulture);
        }

        public string ShortFloatStr(float a_Float)
        {
            return a_Float.ToString(c_ShortFloatFormat, f_FormatCulture);
        }

        public bool TryParseFloat(string a_Str,
                                  float a_Min,
                                  float a_Max,
                                  out float a_Float,
                                  out string a_ErrMsg)
        {
            if (0 == a_Str.Trim().Length)
            {
                a_Float = 0f;
                a_ErrMsg = GetString(TextKeys.MsgNumberParsingNoInput);

                return false;
            }

            if (false == float.TryParse(a_Str, NumberStyles.Float, f_FormatCulture, out a_Float))
            {
                a_ErrMsg = Format(TextKeys.MsgNumberParsingInvalidInput, a_Str);

                return false;
            }

            if (a_Min > a_Float)
            {
                a_ErrMsg = ToMinMaxErrMsg(ShortFloatStr(a_Min), false);

                return false;
            }

            if (a_Max < a_Float)
            {
                a_ErrMsg = ToMinMaxErrMsg(ShortFloatStr(a_Max), true);

                return false;
            }

            a_ErrMsg = null;

            return true;
        }

        private string ToMinMaxErrMsg(string a_LimitStr,
                                      bool a_Maximum)
        {
            TextKeys l_Key = (false == a_Maximum ? TextKeys.MsgNumberParsingTooLow : TextKeys.MsgNumberParsingTooHigh);

            return Format(l_Key, a_LimitStr);
        }

        public string BuildFileErrMsg(TextKeys a_MainMsg,
                                      string a_FileName,
                                      TextKeys a_SubMsg)
        {
            return BuildFileErrMsg(a_MainMsg, a_FileName, GetString(a_SubMsg));
        }

        public string BuildFileErrMsg(TextKeys a_MainMsg,
                                      string a_FileName,
                                      string a_SubErrMsg)
        {
            return Format(a_MainMsg, a_FileName, a_SubErrMsg);
        }

        private static Dictionary<TextKeys, string> ParseTextSet(StreamReader a_Rdr,
                                                                 out string a_Summary)
        {
            // Return can be null. a_Summary can be null on return.
            // a_Rdr won't be closed in any case.

            const char l_UnicodeReplacementChar = (char)0xFFFD;
            StringBuilder l_SummaryBld = new StringBuilder();
            CultureInfo l_InvarCulture = CultureInfo.InvariantCulture;
            Dictionary<TextKeys, string> l_Texts = new Dictionary<TextKeys, string>(c_ItemCount);

            StringBuilder l_TempBld = new StringBuilder();
            for (int i = 1; true; i++)
            {
                string l_Line;

                try
                {
                    l_Line = a_Rdr.ReadLine();
                }
                catch (Exception l_Exc)
                {
                    a_Summary = l_Exc.Message;

                    return null;
                }

                if (null == l_Line)
                {
                    // EOF

                    break;
                }

                int l_Len = l_Line.Length;
                for (int k = 0; k < l_Len; k++)
                {
                    char l_CurChar = l_Line[k];

                    if (true == char.IsWhiteSpace(l_CurChar))
                    {
                        continue;
                    }

                    if (c_CommentChar == l_CurChar)
                    {
                        break;
                    }

                    if (false == char.IsDigit(l_CurChar))
                    {
                        l_SummaryBld.AppendFormat(l_InvarCulture, "Unrecognized sequence beginning at column {1} of line #{0}.\n", i, k + 1);
                    }
                    else
                    {
                        ParseLineContext l_Cntxt = new ParseLineContext(l_Line, k);
                        if (false == ParseKeyAndValue(ref l_Cntxt, l_TempBld))
                        {
                            l_SummaryBld.AppendFormat(l_InvarCulture, "An error occurred while trying to interpret line #{0}: {1}\n", i, l_Cntxt.f_ErrMsg);
                        }
                        else if (0 <= l_Cntxt.f_Value.IndexOf(l_UnicodeReplacementChar))
                        {
                            l_SummaryBld.AppendFormat(l_InvarCulture, "The value of key {0} in line #{1} could not be fully interpreted due to encoding problems. File encoded in UTF-8? Used a text editor that doesn't support UTF-8?\n", l_Cntxt.f_Key, i);
                        }
                        else
                        {
                            try
                            {
                                l_Texts.Add((TextKeys)l_Cntxt.f_Key, l_Cntxt.f_Value);
                            }
                            catch (ArgumentException)
                            {
                                // Duplicate key.

                                l_SummaryBld.AppendFormat(l_InvarCulture, "The key {0} in line #{1} occurs after being used already earlier.\n", (ushort)l_Cntxt.f_Key, i);
                            }
                        }
                    }

                    break;
                }
            }

            a_Summary = (0 == l_SummaryBld.Length ? null : l_SummaryBld.ToString());

            return l_Texts;
        }

        private static bool ParseKeyAndValue(ref ParseLineContext a_Cntxt,
                                             StringBuilder a_TempBld)
        {
            int l_AssignmentIdx = a_Cntxt.GetAssignmentIndex();

            if (0 > l_AssignmentIdx)
            {
                a_Cntxt.f_ErrMsg = "Key and value expected.";

                return false;
            }

            string l_Str = a_Cntxt.f_Line;

            if (false == Util.TryParseUInt16Inv(l_Str, a_Cntxt.f_StartingIdx, l_AssignmentIdx - a_Cntxt.f_StartingIdx, out a_Cntxt.f_Key))
            {
                a_Cntxt.f_ErrMsg = "The sequence '" + a_Cntxt.GetKeyString(l_AssignmentIdx) + "' is not a valid key.";

                return false;
            }

            a_TempBld.Length = 0;

            bool l_InLiteral = false;
            for (int i = (l_AssignmentIdx + 1), lim = l_Str.Length; i < lim; i++)
            {
                char l_CurChar = l_Str[i];

                if (false == l_InLiteral)
                {
                    if (c_StringLiteralChar == l_CurChar)
                    {
                        l_InLiteral = true;

                        continue;
                    }

                    if (true == char.IsWhiteSpace(l_CurChar))
                    {
                        continue;
                    }

                    a_Cntxt.f_ErrMsg = "Unrecognized value.";

                    return false;
                }

                if (c_StringLiteralChar == l_CurChar)
                {
                    // Check if there is only space behind the value.

                    for (i++; i < lim; i++)
                    {
                        if (false == char.IsWhiteSpace(l_Str[i]))
                        {
                            a_Cntxt.f_ErrMsg = "Illegal value.";

                            return false;
                        }
                    }

                    if (c_MaxTextLen < a_TempBld.Length)
                    {
                        a_Cntxt.f_ErrMsg = "Value too long.";

                        return false;
                    }

                    a_Cntxt.f_Value = a_TempBld.ToString();

                    return true;
                }

                if (c_EscapeChar == l_CurChar)
                {
                    i++;
                    if (false == InterpretEscapeSequence(l_Str, ref i, a_TempBld))
                    {
                        a_Cntxt.f_ErrMsg = "Unrecognized escape sequence.";

                        return false;
                    }

                    continue;
                }

                a_TempBld.Append(l_CurChar);
            }

            a_Cntxt.f_ErrMsg = (false == l_InLiteral ? "No value." : "End of value marker missing.");

            return false;
        }

        private static bool InterpretEscapeSequence(string a_Str,
                                                    ref int a_Idx,
                                                    StringBuilder a_DestBld)
        {
            // At successful return, a_Idx will reference the last
            // character of a_Str that is part of the escape
            // sequence.

            int l_Len = (a_Str.Length - a_Idx);

            if (0 >= l_Len)
            {
                return false;
            }

            char l_Char = a_Str[a_Idx];

            switch (l_Char)
            {
                case '"':
                    break;

                case 'n':
                    l_Char = '\n';
                    break;

                case 't':
                    l_Char = '\t';
                    break;

                case '\\':
                    l_Char = '\\';
                    break;

                case 'r':
                    l_Char = '\r';
                    break;

                case 'k':
                    // The environment specific line terminator.
                    a_DestBld.AppendLine();
                    return true;

                case 'x':
                    {
                        // We expect two or four hexadecimal
                        // characters.

                        uint l_Num;

                        if ((c_LongHexCharLen + 1) <= l_Len)
                        {
                            if (true == HexEncoding.TryParseUInt(a_Str, a_Idx + 1, c_LongHexCharLen, out l_Num))
                            {
                                a_Idx += c_LongHexCharLen;
                                l_Char = (char)l_Num;
                                break;
                            }
                        }

                        if ((c_ShortHexCharLen + 1) <= l_Len)
                        {
                            if (true == HexEncoding.TryParseUInt(a_Str, a_Idx + 1, c_ShortHexCharLen, out l_Num))
                            {
                                a_Idx += c_ShortHexCharLen;
                                l_Char = (char)l_Num;
                                break;
                            }
                        }

                        return false;
                    }

                default:
                    return false;
            }

            a_DestBld.Append(l_Char);

            return true;
        }

        private static bool CountPlaceholders(bool a_AllowFormatInstructions,
                                              string a_Str,
                                              out uint a_Count)
        {
            // This routine will return true if all placeholders
            // between zero and a_Count occur exactly once. It
            // also verifies escape sequences like '{{' and '}}'.

            a_Count = 0;

            uint l_Mask = 0;

            for (int i = 0, lim = a_Str.Length; i < lim; i++)
            {
                char l_Char = a_Str[i];

                if ('{' == l_Char)
                {
                    i++;
                    uint l_Number;
                    switch (InterpretPlaceholder(a_AllowFormatInstructions, a_Str, ref i, out l_Number))
                    {
                        case ReadPlaceholderResults.Success:
                            break;

                        case ReadPlaceholderResults.Escaped:
                            continue;

                        case ReadPlaceholderResults.Error:
                            return false;

                        default: throw (new ApplicationException());
                    }

                    if (31 < l_Number ||
                        31 < a_Count)
                    {
                        Util.DebugBreak();

                        return false;
                    }

                    a_Count++;
                    l_Mask |= (1u << (int)l_Number);
                }
                else if ('}' == l_Char)
                {
                    // Look for a '}}' escape sequence.

                    if (++i == lim ||
                        '}' != a_Str[i])
                    {
                        return false;
                    }
                }
            }

            return (0 == a_Count ||
                    ((1u << (int)a_Count) - 1) == l_Mask);
        }

        private static ReadPlaceholderResults InterpretPlaceholder(bool a_AllowFormatInstructions,
                                                                   string a_Str,
                                                                   ref int a_Idx,
                                                                   out uint a_Number)
        {
            // If no error occurrs, a_Idx will reference the last
            // character of the processed sequence on return.

            for (int i = a_Idx, lim = a_Str.Length; i < lim; i++)
            {
                char l_Char = a_Str[i];

                if (true == Util.IsDigitInvariant(l_Char))
                {
                    continue;
                }

                if ('}' == l_Char)
                {
                    if (true == Util.TryParseUInt32Inv(a_Str, a_Idx, i - a_Idx, out a_Number))
                    {
                        a_Idx = i;

                        return ReadPlaceholderResults.Success;
                    }
                }
                else if (':' == l_Char)
                {
                    // There seems to be a format instruction.

                    if (true == a_AllowFormatInstructions)
                    {
                        // Fast forward to the end.
                        int l_ClosingIdx = a_Str.IndexOf('}', i + 1, lim - (i + 1));

                        if (0 <= l_ClosingIdx &&
                            true == Util.TryParseUInt32Inv(a_Str, a_Idx, i - a_Idx, out a_Number))
                        {
                            a_Idx = l_ClosingIdx;

                            return ReadPlaceholderResults.Success;
                        }
                    }
                }
                else if ('{' == l_Char)
                {
                    if (i == a_Idx)
                    {
                        a_Number = 0;

                        return ReadPlaceholderResults.Escaped;
                    }
                }

                break;
            }

            a_Number = 0;

            return ReadPlaceholderResults.Error;
        }

        private static void PrepareFormats(Dictionary<TextKeys, string> a_Texts)
        {
            PrepareFormat(a_Texts, TextKeys.TimeStrFormat, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.SourcesLong, c_IntegerFormat, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.SourcesShort, c_IntegerFormat, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.CompletenessPercFormat, c_PercentageFormat);
            PrepareFormat(a_Texts, TextKeys.NameWithCountFormat, null, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.NameWithTwoCountsFormat, null, c_IntegerFormat, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.FileDeletionFailedMsgWithCounters, c_IntegerFormat, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.IPEndPointFormat, null, c_PlainIntegerFormat);
            PrepareFormat(a_Texts, TextKeys.RsaKeyYesFormat, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.RsaKeyLongFormat, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.DlStatusRangesQueried, c_IntegerFormat, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.CompletenessPercFormat, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.OptionsDlgAddressExistingAddrAge, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.OptionsDlgAddressChangeAfterNDays, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.AboutDlgAppBuildDate, c_DateFormat);
            PrepareFormat(a_Texts, TextKeys.AboutDlgAppBuildAgeFormat, null, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.OpenCollDlgFailedToTakeOver, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.StatsTabFormatVerifiedNum, c_IntegerFormat, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.StatsTabPubFormatActiveNDone, c_PercentageFormat);
            PrepareFormat(a_Texts, TextKeys.KademliaTabContactCountFormat, null, c_IntegerFormat, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.PeersTabCountFormat, c_IntegerFormat, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.PeersTabInConnCountFormat, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.PeersTabOutConnCountFormat, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.FileSearchTabNFilesWereTakenOver, c_IntegerFormat, c_IntegerFormat);
            PrepareFormat(a_Texts, TextKeys.MoveDataFilesDlgProgressFormat, c_PercentageFormat);
        }

        private static void PrepareFormat(Dictionary<TextKeys, string> a_Texts,
                                          TextKeys a_Key,
                                          params string[] a_Formats)
        {
            // Elements in a_Formats may be null.

            string l_Str;

            if (false == a_Texts.TryGetValue(a_Key, out l_Str))
            {
                return;
            }

            StringBuilder l_Bld = new StringBuilder(l_Str, l_Str.Length + 20);

            for (int i = (a_Formats.Length - 1); 0 <= i; i--)
            {
                string l_Format = a_Formats[i];

                if (null != l_Format)
                {
                    l_Bld.Replace("{" + i + "}", "{" + i + ":" + l_Format + "}");
                }
            }

            a_Texts[a_Key] = l_Bld.ToString();
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct ParseLineContext
        {
            public readonly string f_Line;
            public readonly int f_StartingIdx;

            public ushort f_Key;
            public string f_Value;
            public string f_ErrMsg;

            public ParseLineContext(string a_Line,
                                    int a_StartingIdx)
            {
                f_Line = a_Line;
                f_StartingIdx = a_StartingIdx;

                f_Key = 0;
                f_Value = null;
                f_ErrMsg = null;
            }

            public int GetAssignmentIndex()
            {
                return f_Line.IndexOf(c_AssignmentChar, f_StartingIdx);
            }

            public string GetKeyString(int a_AssignmentIdx)
            {
                return f_Line.Substring(f_StartingIdx, a_AssignmentIdx - f_StartingIdx);
            }
        }

        private enum ReadPlaceholderResults : byte
        {
            Success = 0,
            Escaped = 1,
            Error = 2,
        }
    }

    enum TextKeys : ushort
    {
        // Empty is a key that can be used with GetString(), but it's
        // not really part of the text set.
        Empty = 0,

        AppName = 1,
        ClientNameWithVersion = 20,

        ButtonTextOK = 30,
        ButtonTextCancel = 31,
        ButtonTextApply = 32,
        ButtonTextYes = 33,
        ButtonTextNo = 34,
        ButtonTextAccept = 35,
        ButtonTextClose = 36,
        ButtonTextAbort = 37,
        ButtonTextTest = 38,

        NotAvailable = 40,
        UnknownValue = 41,
        KeyValuePairFormat = 42,
        KeyFormat = 43,
        CntrlsInputCaptionFormat = 44,
        CntrlsOpenDialogCaptionFormat = 45,
        CntrlsExpandFormButtonFormat = 46,
        CntrlsCollapseFormButtonFormat = 47,
        CntrlsBrowseFileSystemButton = 48,
        CntrlsOpenDirectoryButton = 49,

        SpeedUnknown = 50,
        SpeedUnlimited = 51,
        SpeedNoLimit = 52,
        SpeedUnitFormat = 53,
        SpeedI2PBusy = 54,
        SpeedLimitFailedToParse = 55,
        SpeedDontDl = 56,
        SpeedDontUl = 57,

        TimeStrSecs = 60,
        TimeStrMins = 61,
        TimeStrHours = 62,
        TimeStrDays = 63,
        TimeStrWeeks = 64,
        TimeStrNever = 70,
        TimeStrNow = 71,
        TimeStrLongerThanMonth = 72,
        TimeStrNegFormat = 73,
        TimeStrFormat = 74,
        TimeStrSpace = 75,

        ByteSingStr = 80,
        BytePlurStr = 81,
        DefaultSizeFormat = 82,
        LongSizeFormat = 83,

        UlDlRatioFormat = 90,
        OnlyDlStr = 91,
        OnlyUlStr = 92,

        FileTypeAny = 100,
        FileTypeAudio = 101,
        FileTypeVideo = 102,
        FileTypeImage = 103,
        FileTypeProgram = 104,
        FileTypeDocument = 105,
        FileTypeArchive = 106,
        FileTypeCDImage = 107,
        FileTypeCollection = 108,

        FileRatingNotRated = 120,
        FileRatingInvalidCorruptFake = 121,
        FileRatingPoor = 122,
        FileRatingFair = 123,
        FileRatingGood = 124,
        FileRatingExcellent = 125,

        PriorityVeryLow = 130,
        PriorityLow = 131,
        PriorityNormal = 132,
        PriorityHigh = 133,
        PriorityVeryHigh = 134,
        PriorityNoSuchErrMsg = 139,

        SourcesNone = 140,
        SourcesShort = 141,
        SourcesLong = 142,

        NameWithCountFormat = 150,
        NameWithTwoCountsFormat = 151,

        ParseIPAddrErrMsgFormat = 160,
        IPEndPointFormat = 161,

        DirectionDown = 170,
        DirectionUp = 171,
        DirectionUpAndDown = 172,

        TransportStylePlain = 180,
        TransportStyleCompressed = 181,

        FileStatusIncomplDisabled = 190,
        FileStatusIncomplEnabled = 191,
        FileStatusMissing = 192,
        FileStatusComplete = 193,

        RsaKeyNo = 200,
        RsaKeyYesFormat = 201,
        RsaKeyLongFormat = 202,

        CompletenessEverything = 210,
        CompletenessPercFormat = 211,

        DlStatusWaitingForFileStatus = 221,
        DlStatusUploadRequested = 222,
        DlStatusRangesQueried = 223,
        DlStatusContentReceived = 224,

        ColumnsReset = 230,
        ColumnsMsgReallyReset = 231,
        ColumnsReorder = 232,

        FileDlgAllFiles = 240,
        FileDlgNodeFiles = 241,
        FileDlgNBCollections = 242,

        AppTitleWithSpeedFormat = 250,
        AppTitleNotConn = 251,
        AppTitleConnecting = 252,
        AppTitleSpeedFormat = 253,

        LogEntryFormat = 260,

        MsgPreviewFileName = 270,

        MsgOperationFailed = 280,
        QuestOperationReallyAbort = 281,
        MsgOperationAbortedByUser = 282,

        BorderStylesNoBorder = 290,
        BorderStylesHorizontal = 291,
        BorderStylesVertical = 292,
        BorderStylesBoth = 293,

        DirectoryOpen = 298,
        DirectoryFailedSinceEmptyString = 299,
        DirectoryFailedToOpen = 300,
        FileDeletionFailedMsgWithCounters = 301,
        MsgErrWhileWritingBootConfig = 302,

        IconColorPickerCaption = 310,
        IconColorPickerDefaultCol = 311,
        IconColorPickerCustomCol = 312,

        FilesTabDetailsUnknownValue = 13430,

        // Application items.
        MainMenuApplication = 1100,
        MainMenuOptions = 1101,
        MainMenuExit = 1102,
        // Network items.
        MainMenuNetwork = 1200,
        MainMenuConnect = 1201,
        MainMenuStopConnecting = 1202,
        MainMenuDisconnect = 1203,
        MainMenuReconnect = 1204,
        MainMenuShowConnSecInfo = 1205,
        MainMenuCopyOwnStrmAddrToClipboard = 1206,
        MainMenuResetAllBans = 1207,
        MainMenuKademlia = 1230,
        MainMenuRepublishOwnFiles = 1231,
        MainMenuImportNodes = 1232,
        MainMenuExportNodes = 1233,
        // Files items.
        MainMenuFiles = 1300,
        MainMenuPasteLinkFromClipboard = 1301,
        MainMenuOpenCollection = 1302,
        MainMenuRescanSharedDirs = 1303,
        MainMenuImportPartialDlsFromiMule = 1304,
        // Info items,
        MainMenuInfo = 1400,
        MainMenuCopyStatsToClipboard = 1401,
        MainMenuCopyLogToClipboard = 1402,
        MainMenuShowLicense = 1403,
        MainMenuAbout = 1404,
        MainMenuViewReleaseNotes = 1405,

        MsgAppAlreadyUp = 3000,
        MsgUnrecognizedCmdLineArgs = 3001,
        MsgFailedProcessCmdLineArgs = 3002,
        QuestionReallyExitApp = 3003,
        QuestionReallyCloseConn = 3004,
        MsgWelcomeToApp = 3005,
        MsgNewerVersionDetected = 3006,

        MsgSAMBridgeConnecting = 3010,
        MsgSAMBridgeConnAttemptFailed = 3011,
        MsgSAMBridgeConnAbortedAsDemanded = 3012,
        MsgSAMBridgeConnSuccess = 3013,
        MsgSAMBridgeConnLost = 3014,
        MsgSAMBridgeConnClosedAsDemanded = 3015,

        MsgHardMemErrFormat = 3030,
        MsgAppDataErrFromNewerVersion = 3031,
        QuestAppDataDirIsMissingWantReset = 3032,

        MsgFileErrWrongFileSize = 3040,
        MsgFileErrFileTooLarge = 3041,
        MsgFileErrAccessFailed = 3042,
        MsgFileErrReadFailed = 3043,
        MsgFileErrWriteFailed = 3044,
        MsgFileErrCorruption = 3045,
        MsgFileErrSetMismatch = 3046,
        MsgFileErrCreationFailed = 3047,

        MsgLinkNotMagnet = 3050,
        MsgLinkSeveralFilesNotAllowed = 3053,
        MsgLinkNoteD2k = 3054,
        MsgLinkOnlyFileAllowed = 3055,
        MsgLinkNotAccepted = 3063,

        MsgDlFileExistsAlready = 3070,
        TitleDlFinished = 3071,
        MsgFileDlSucceeded = 3072,
        TitleMissingOutDir = 3073,
        MsgCantFinishDlDueToMissingOutDir = 3074,
        MsgCantFinishDlDueToLackOfSpace = 3075,
        TitleDlHashingSuspended = 3076,
        MsgDlHashingSuspendedDueToLackOfSpace = 3077,
        MsgDlHashingResumed = 3078,
        TitleSAMConnBroken = 3080,
        MsgSAMConnBroken = 3081,
        TitleMissingKadNodes = 3082,
        MsgMissingKadNodes = 3083,
        TitleReleaseOlderThanSixMonth = 3084,
        MsgReleaseOlderThanSixMonth = 3085,
        TitleListOfSharedDirsAutoModified = 3086,
        MsgListOfSharedDirsAutoModified = 3087,

        MsgPreviewProcessDidntStart = 3110,
        MsgPreviewGeneralErr = 3111,

        MsgNodesErrLoadingFromFile = 3120,
        MsgNodesErrImportFailed = 3121,
        MsgNodesImportWorked = 3122,

        MsgNodesErrNothingToExport = 3130,
        MsgNodesErrExportFailed = 3131,
        MsgNodesExportWorked = 3132,

        MsgCollErrLoadingFromFile = 3140,
        MsgCollErrCantInterpretFile = 3141,
        MsgCollErrTooAdvancedFile = 3142,

        MsgNumberParsingNoInput = 3150,
        MsgNumberParsingInvalidInput = 3151,
        MsgNumberParsingTooLow = 3152,
        MsgNumberParsingTooHigh = 3153,

        MsgTakeOverRequiresReconnect = 3170,
        MsgTakeOverRequiresAppRestart = 3171,

        MsgDestBanned = 3180,
        MsgYourClientID = 3181,

        LinksMagnetLink = 3190,
        LinkseD2kLink = 3191,
        LinksClipboardContent = 3192,
        LinksShellInput = 3193,
        LinksMsgFailedToInterpretFormat = 3200,
        LinksMsgInputNotRecognized = 3201,
        LinksMsgFileTooLarge = 3202,
        LinksMsgFileIDConflict = 3203,
        LinksQuestionWantToAddFile = 3204,
        LinksMsgFailedToAdd = 3205,

        // Kademlia tab.
        KademliaTabTitle = 11000,
        KademliaTabNotConn = 11100,
        KademliaTabTryingToConn = 11101,
        KademliaTabConn = 11102,
        KademliaTabContactCountFormat = 11103,
        KademliaTabColumnTitles = 11900,
        // File search tab.
        FileSearchTabTitle = 12000,
        FileSearchTabRemoteCaption = 12010,
        FileSearchTabRemoteFileNameAND = 12011,
        FileSearchTabRemoteFileNameOR = 12012,
        FileSearchTabRemoteFileNameNOT = 12013,
        FileSearchTabRemoteFileType = 12014,
        FileSearchTabRemoteFileExtension = 12015,
        FileSearchTabRemoteFileMinSize = 12020,
        FileSearchTabRemoteFileMaxSize = 12021,
        FileSearchTabLocalCaption = 12030,
        FileSearchTabLocalTextFilter = 12031,
        FileSearchTabLocalApply = 12032,
        FileSearchTabLocalApplyInverted = 12033,
        FileSearchTabLocalEndFiltering = 12034,
        FileSearchTabLocalHideKnown = 12035,
        FileSearchTabLocalShowSizeCntrls = 12040,
        FileSearchTabLocalHideSizeCntrls = 12041,
        FileSearchTabLocalShowFilterCntrls = 12042,
        FileSearchTabLocalHideFilterCntrls = 12043,
        FileSearchTabLocalReset = 12050,
        FileSearchTabLocalResetForm = 12051,
        FileSearchTabLocalDelKeywordHist = 12052,
        FileSearchTabLocalStopAll = 12053,
        FileSearchTabLocalStart = 12054,
        FileSearchTabCntxtMenuRestoreSearchParams = 12100,
        FileSearchTabCntxtMenuClearSearchResults = 12101,
        FileSearchTabCntxtMenuRerunSearch = 12102,
        FileSearchTabCntxtMenuStopSearch = 12103,
        FileSearchTabCntxtMenuCloseSearch = 12104,
        FileSearchTabCntxtMenuStopAllSearches = 12105,
        FileSearchTabCntxtMenuCloseAllSearches = 12106,
        FileSearchTabCntxtMenuReorderSearches = 12107,
        FileSearchTabReallyDeleteKeywordHist = 12200,
        FileSearchTabKadConnRequired = 12201,
        FileSearchTabMaxSearchTabCountReached = 12202,
        FileSearchTabSearchStartFailed = 12203,
        FileSearchTabReallyCloseAllSearches = 12204,
        FileSearchTabResultsCntxtMenuAddToIncompl = 12300,
        FileSearchTabResultsCntxtMenuCopySelItemsToClipboardAsText = 12301,
        FileSearchTabResultsCntxtMenuAddAndAssignToCategory = 12302,
        FileSearchTabErrMustHaveOneKeyword = 12400,
        FileSearchTabErrProblemWithNOT = 12401,
        FileSearchTabErrProblemWithSizes = 12402,
        FileSearchTabRunningFormat = 12410,
        FileSearchTabFileWasTakenOver = 12420,
        FileSearchTabFileWasNotTakenOver = 12421,
        FileSearchTabNFilesWereTakenOver = 12422,
        FileSearchTabColumnTitles = 12900,
        // Files tab.
        FilesTabTitle = 13000,
        FilesTabSubTabAll = 13010,
        FilesTabSubTabIncomplete = 13011,
        FilesTabSubTabMissing = 13012,
        FilesTabSubTabComplete = 13013,
        FilesTabSubTabActive = 13014,
        FilesTabSubTabDownloads = 13015,
        FilesTabSubTabUploads = 13016,

        FilesTabCntxtMenuHideTab = 13050,
        FilesTabCntxtMenuDeleteTab = 13051,
        FilesTabCntxtMenuModifyTab = 13052,
        FilesTabCntxtMenuCreateNewTab = 13053,
        FilesTabCntxtMenuUnhideTab = 13054,
        FilesTabCntxtMenuReorderCustomTabs = 13055,

        FilesTabCntxtMenuQuestionReallyDeleteTab = 13060,
        FilesTabCntxtMenuMsgMaxTabCountReached = 13061,

        FilesTabCustomTabDlgTitleCreate = 13070,
        FilesTabCustomTabDlgTitleRename = 13071,
        FilesTabCustomTabDlgTabName = 13072,
        FilesTabCustomTabDlgGroupBoxCaption = 13073,
        FilesTabCustomTabDlgMsgBlockEmptyName = 13080,
        FilesTabCustomTabDlgMsgBlockLongName = 13081,
        FilesTabCustomTabDlgMsgBlockExistingName = 13082,
        FilesTabCustomTabDlgIcon = 13090,
        FilesTabCustomTabDlgPickIconDlgCaption = 13091,
        FilesTabCustomTabDlgPickIconDlgNoIcon = 13092,

        // Files context menu.
        FilesTabListCntxtMenuOpen = 13100,
        FilesTabListCntxtMenuPreview = 13101,
        FilesTabListCntxtMenuOpenFolder = 13102,
        FilesTabListCntxtMenuDisableDl = 13103,
        FilesTabListCntxtMenuEnableDl = 13104,
        FilesTabListCntxtMenuDlBandwidth = 13105,
        FilesTabListCntxtMenuUlBandwidth = 13106,
        FilesTabListCntxtMenuAdjustSpeedAndPriority = 13107,
        FilesTabListCntxtMenuAssignToCustomTab = 13108,
        FilesTabListCntxtMenuNoCustomTab = 13109,
        FilesTabListCntxtMenuAdvanced = 13110,
        FilesTabListCntxtMenuRequerySources = 13111,
        FilesTabListCntxtMenuManuallyAddAddr = 13112,
        FilesTabListCntxtMenuResetPeerStates = 13113,
        FilesTabListCntxtMenuLinks = 13114,
        FilesTabListCntxtMenuCopyeD2kLink = 13115,
        FilesTabListCntxtMenuCopyeD2kLinkPlusHtml = 13116,
        FilesTabListCntxtMenuCopyMagnetLink = 13117,
        FilesTabListCntxtMenuCopyMagnetLinkPlusHtml = 13118,
        FilesTabListCntxtMenuCreateColl = 13119,
        FilesTabListCntxtMenuRemoveIncomplOrMissing = 13120,
        FilesTabListCntxtMenuRemoveAnyFiles = 13121,
        FilesTabListCntxtMenuRemoveTurnMissingIntoIncompl = 13122,
        FilesTabListCntxtMenuRemoveForceRecheckOfCompl = 13123,

        // Open or preview file.
        FilesTabListLaunchQuestReallyExecuteStrangeFile = 13130,
        FilesTabListLaunchMsgFailedToLaunch = 13131,
        FilesTabListLaunchMsgFailedToPreview = 13132,
        FilesTabListLaunchMsgBeginningNotDled = 13133,
        FilesTabListLaunchMsgNotFitForPreviewing = 13134,
        FilesTabListLaunchMsgLackOfFreeSpace = 13135,
        FilesTabListLaunchMsgLowFreeSpace = 13136,
        FilesTabListLaunchMsgErrAccessingTempFile = 13137,
        FilesTabListLaunchMsgUnspecifiedErr = 13138,
        FilesTabListLaunchMsgPreviewFolderMissing = 13139,

        // Copy link to clipboard.
        FilesTabListCopyLinkMsgFileNotHashedYet = 13150,
        FilesTabListCopyLinkMsgMissingParams = 13151,
        FilesTabListCopyLinkMsgFailedToCopy = 13152,

        // Create collection.
        FilesTabListCreateCollMsgNoFilesSelected = 13160,
        FilesTabListCreateCollMsgTooManyFiles = 13161,
        FilesTabListCreateCollMsgFilesNotHashedYet = 13162,

        // Drop files.
        FilesTabDropFilesQuestionReally = 13170,
        FilesTabDropFilesSubMsgCompleteFilesWillBeMovedToRecycleBin = 13171,
        FilesTabDropFilesSubMsgCompleteFilesWillBeDeleted = 13172,

        FilesTabQuestionReallyTurnMissingIntoIncomplFiles = 13180,
        FilesTabQuestionReallyRecheckComplFiles = 13190,

        FilesTabColumnTitles = 13300,

        FilesTabDetailsTitle = 13400,

        // File Details.
        FilesTabDetailsAvailability = 13410,
        FilesTabDetailsDone = 13411,
        FilesTabDetailsFileName = 13412,
        FilesTabDetailsLocation = 13413,
        FilesTabDetailsFileSize = 13414,
        FilesTabDetailsFileID = 13415,
        FilesTabDetailsAICHHash = 13416,
        FilesTabDetailsTAICHHash = 13417,
        FilesTabDetailsSources = 13418,
        FilesTabDetailsIncompleteLoc = 13430,
        FilesTabDetailsMissingLoc = 13431,

        FilesTabPeersTitle = 13500,
        FilesTabPeersColumnTitles = 13550,

        FilesTabTransfersTitle = 13600,
        FilesTabTransfersColumnTitles = 13650,

        FilesTabFileNamesTitle = 13700,
        FilesTabFileNamesCleanup = 13710,
        FilesTabFileNamesTakeOver = 13711,
        FilesTabFileNamesMsgNeedFileName = 13720,
        FilesTabFileNamesMsgNameHasIllegalChars = 13721,
        FilesTabFileNamesMsgFailedToRename = 13722,
        FilesTabFileNamesColumnTitles = 13750,

        FilesTabCommentsTitle = 13800,
        FilesTabCommentsSetOwn = 13810,
        FilesTabCommentsEditFilter = 13811,
        FilesTabCommentsQueryFromKad = 13812,
        FilesTabCommentsSetDlgTitle = 13820,
        FilesTabCommentsSetDlgGroupBoxCaption = 13821,
        FilesTabCommentsSetDlgDescription = 13822,
        FilesTabCommentsSetDlgRating = 13823,
        FilesTabCommentsSetDlgWarnMsgFilter = 13824,
        FilesTabCommentsColumnTitles = 13850,

        // Manual source drop dialog.
        FilesTabAddrDlgTitle = 14000,
        FilesTabAddrDlgGroupBoxCaption = 14010,
        FilesTabAddrDlgStreamAddr = 14011,
        FilesTabAddrDlgTextNotRecognizedAsAddr = 14020,

        // File speed and priority limit dialog.
        FilesTabBwLimitDlgTitle = 14100,
        FilesTabBwLimitDlgDlCaption = 14110,
        FilesTabBwLimitDlgDlLimitSpeed = 14111,
        FilesTabBwLimitDlgDlPriority = 14112,
        FilesTabBwLimitDlgUlCaption = 14120,
        FilesTabBwLimitDlgUlLimitSpeed = 14121,
        FilesTabBwLimitDlgUlPriority = 14122,

        // Peers tab.
        PeersTabTitle = 15000,
        PeersTabCountFormat = 15100,
        PeersTabInConnCountFormat = 15101,
        PeersTabOutConnCountFormat = 15102,
        PeersTabConnCountJoinedFormat = 15103,
        PeersTabQuestionReallyBanPeers = 15105,

        // Peers context menu.
        PeersTabCntxtMenuCopyStrmAddr = 15200,
        PeersTabCntxtMenuMarkAsFriend = 15201,
        PeersTabCntxtMenuResetFriendStatus = 15202,
        PeersTabCntxtMenuParoleBannedPeers = 15203,
        PeersTabCntxtMenuAdvanced = 15204,
        PeersTabCntxtMenuBan = 15205,
        PeersTabCntxtMenuAdjustReputation = 15206,
        PeersTabCntxtMenuSetNegRepToZero = 15207,

        // Set reputation dialog.
        PeersTabSetRepDlgTitle = 15250,
        PeersTabSetRepDlgGroupBox = 15251,
        PeersTabSetRepDlgInfoLabel = 15252,
        PeersTabSetRepDlgNumBox = 15253,
        PeersTabSetRepDlgUnits = 15254,

        PeersTabColumnTitles = 15900,
        // Stats tab.
        StatsTabTitle = 16000,
        StatsTabFormatVerifiedNum = 16010,
        StatsTabApp = 16040,
        StatsTabAppUptime = 16041,
        StatsTabAppMemAllocated = 16042,
        StatsTabAppCPUTime = 16043,
        StatsTabAppCPUUsage = 16044,
        StatsTabI2P = 16060,
        StatsTabI2PConnUptime = 16061,
        StatsTabI2PUnrecognizedDgsRcvd = 16062,
        StatsTabI2PRecognizedDgsRcvd = 16063,
        StatsTabI2PDgsSent = 16064,
        StatsTabI2PDgBytesRcvd = 16065,
        StatsTabI2PDgBytesSent = 16066,
        StatsTabI2PTotalBytesRcvd = 16067,
        StatsTabI2PTotalBytesSent = 16068,
        StatsTabI2PInternedDests = 16069,
        StatsTabI2PBannedDests = 16070,
        StatsTabStreams = 16080,
        StatsTabStreamsSuccOutAttempts = 16081,
        StatsTabStreamsAcceptedInConns = 16082,
        StatsTabStreamsCurOutAttempts = 16083,
        StatsTabStreamsCurEstablishedConns = 16084,
        StatsTabContacts = 16100,
        StatsTabContactsInContactMngr = 16101,
        StatsTabKadRouting = 16120,
        StatsTabKadRoutingHighestLevel = 16121,
        StatsTabKadRoutingRoutingBins = 16122,
        StatsTabKadRoutingContactsUsed = 16123,
        StatsTabKadRoutingEstimatedNodes = 16124,
        StatsTabPub = 16140,
        StatsTabPubStatus = 16141,
        StatsTabPubPassive = 16142,
        StatsTabPubNextPublishing = 16143,
        StatsTabPubWhenPossible = 16144,
        StatsTabPubFormatActiveNDone = 16145,
        StatsTabIndex = 16160,
        StatsTabIndexFiles = 16161,
        StatsTabIndexKeywords = 16162,
        StatsTabIndexNotes = 16163,
        StatsTabIndexNetSize = 16164,
        // Log tab.
        LogTabTitle = 17000,
        // Options dialog.
        OptionsDlgTitle = 20000,
        OptionsDlgResetToDefaults = 20010,
        OptionsDlgQuestionReallyReset = 20011,
        // General page.
        OptionsDlgGeneralTitle = 20100,
        OptionsDlgGeneralComment = 20101,
        OptionsDlgGeneralAppTitle = 20110,
        OptionsDlgGeneralAppPromptOnExit = 20111,
        OptionsDlgGeneralAppBeepWhenDlIsDone = 20112,
        OptionsDlgGeneralAppAutoStartDl = 20113,
        OptionsDlgGeneralAppSaveSearchHist = 20114,
        OptionsDlgGeneralConnMode = 20120,
        OptionsDlgGeneralConnManual = 20121,
        OptionsDlgGeneralConnManualButRepair = 20122,
        OptionsDlgGeneralConnAuto = 20123,
        OptionsDlgGeneralUIMode = 20130,
        OptionsDlgGeneralUIModeSimple = 20131,
        OptionsDlgGeneralUIModeAdvanced = 20132,
        // Language page.
        OptionsDlgLangTitle = 20200,
        OptionsDlgLangComment = 20201,
        OptionsDlgLangLocalization = 20210,
        OptionsDlgLangTextLang = 20211,
        OptionsDlgLangVerifyTextSet = 20212,
        OptionsDlgLangFormatCulture = 20213,
        OptionsDlgLangCopyTextsToClipboard = 20214,
        OptionsDlgLangBuiltInEng = 20220,
        OptionsDlgLangOSSetting = 20221,
        OptionsDlgLangMsgNoErrorsInTextSet = 20230,
        OptionsDlgLangMsgErrorsDetectedSeeLog = 20231,
        // OS integration page.
        OptionsDlgOSIntgrTitle = 20300,
        OptionsDlgOSIntgrComment = 20301,
        OptionsDlgOSIntgrWinTitle = 20310,
        OptionsDlgOSIntgrStartOnLogon = 20311,
        OptionsDlgOSIntgrAssoceD2kURIs = 20312,
        OptionsDlgOSIntgrAssocMagnetURIs = 20313,
        OptionsDlgOSIntgrAssocWithCollections = 20314,
        // Display page.
        OptionsDlgDisplTitle = 20400,
        OptionsDlgDisplComment = 20401,
        OptionsDlgDisplGeneralTitle = 20410,
        OptionsDlgDisplGeneralShowSpeedOnTitle = 20411,
        OptionsDlgDisplGeneralShowTrayIcon = 20412,
        OptionsDlgDisplGeneralShowBalloons = 20413,
        OptionsDlgDisplGeneralMinimizeToTray = 20414,
        OptionsDlgDisplDataFormatKindTitle = 20430,
        OptionsDlgDisplDataFormatIEC = 20431,
        OptionsDlgDisplDataFormatSI = 20432,
        OptionsDlgDisplComplBarBgImgsTitle = 20435,
        OptionsDlgDisplComplBarBgImgsNo = 20436,
        OptionsDlgDisplComplBarBgImgsYes = 20437,
        OptionsDlgDisplProgressBarTitle = 20440,
        OptionsDlgDisplProgressBarVerifiedOnly = 20441,
        OptionsDlgDisplProgressBarInclUnverified = 20442,
        OptionsDlgDisplListViewCaption = 20450,
        OptionsDlgDisplListViewTextSize = 20451,
        OptionsDlgDisplListViewTextSizeUnit = 20452,
        OptionsDlgDisplListViewBorderStyle = 20453,
        // Tunnels page.
        OptionsDlgTunnelsTitle = 20500,
        OptionsDlgTunnelsComment = 20501,
        OptionsDlgTunnelsInboundSettings = 20510,
        OptionsDlgTunnelsOutboundSettings = 20511,
        OptionsDlgTunnelsUseInbound = 20520,
        OptionsDlgTunnelsDatagramQuant = 20521,
        OptionsDlgTunnelsStreamQuant = 20522,
        OptionsDlgTunnelsMinHops = 20523,
        OptionsDlgTunnelsRandExtraHops = 20524,
        OptionsDlgTunnelsNoHops = 20525,
        OptionsDlgTunnelsPosNumFormat = 20526,
        OptionsDlgTunnelsMsgDoNotReveal = 20530,
        OptionsDlgTunnelsMsgWarnShortTunnels = 20531,
        // Addresses page.
        OptionsDlgAddressTitle = 20600,
        OptionsDlgAddressComment = 20601,
        OptionsDlgAddressYourDatagramAddr = 20610,
        OptionsDlgAddressYourStreamAddr = 20611,
        OptionsDlgAddressNewAddrWhenConnecting = 20620,
        OptionsDlgAddressExistingAddrAge = 20621,
        OptionsDlgAddressDeleteAddr = 20622,
        OptionsDlgAddressNeverChange = 20623,
        OptionsDlgAddressChangeAfterNDays = 20624,
        OptionsDlgAddressReallyDeleteAddr = 20630,
        // Bandwidth page.
        OptionsDlgBandwidthTitle = 20700,
        OptionsDlgBandwidthComment = 20701,
        OptionsDlgBandwidthDlLimits = 20710,
        OptionsDlgBandwidthDlGlobalSpeedLimit = 20711,
        OptionsDlgBandwidthDlNumOfFiles = 20712,
        OptionsDlgBandwidthDlNumOfSlotsPerFile = 20713,
        OptionsDlgBandwidthUlLimits = 20720,
        OptionsDlgBandwidthUlGlobalSpeedLimit = 20721,
        OptionsDlgBandwidthUlNumOfGlobalSlots = 20722,
        OptionsDlgBandwidthConnsLimits = 20730,
        OptionsDlgBandwidthConnsTotal = 20731,
        OptionsDlgBandwidthConnsInPerAddr = 20732,
        OptionsDlgBandwidthConnsOutPerAddr = 20733,
        // Kademlia page.
        OptionsDlgKademliaTitle = 20800,
        OptionsDlgKademliaComment = 20801,
        OptionsDlgKademliaCommentsCaption = 20810,
        OptionsDlgKademliaFilter = 20811,
        OptionsDlgKademliaMiscCaption = 20820,
        OptionsDlgKademliaMiscChangeClientID = 20821,
        OptionsDlgKademliaMiscMaxNetSizeOfKadIndex = 20822,
        OptionsDlgKademliaPublCaption = 20830,
        OptionsDlgKademliaPublSource = 20831,
        OptionsDlgKademliaPublKeywords = 20832,
        OptionsDlgKademliaPublNotes = 20833,
        // eD2k page.
        OptionsDlgeD2kTitle = 20900,
        OptionsDlgeD2kComments = 20901,
        OptionsDlgeD2kUserCaption = 20910,
        OptionsDlgeD2kUserHash = 20911,
        OptionsDlgeD2kUserGenerateNew = 20912,
        OptionsDlgeD2kUserMsgWarnChange = 20913,
        OptionsDlgeD2kAppearance = 20920,
        OptionsDlgeD2kAppearanceNickname = 20921,
        // Shared directories page.
        OptionsDlgSharedDirsTitle = 21000,
        OptionsDlgSharedDirsComments = 21001,
        OptionsDlgSharedDirsWatchDirs = 21010,
        OptionsDlgSharedDirsOpenFilesWithSharedAccess = 21011,
        OptionsDlgSharedDirsAdd = 21020,
        OptionsDlgSharedDirsRemove = 21021,
        OptionsDlgSharedDirsModify = 21022,
        OptionsDlgSharedDirsSetOutputDir = 21023,
        OptionsDlgSharedDirsImport = ImportPartialDlsDlgImport,
        OptionsDlgSharedDirsCntxtMenuSetOutputDir = 21030,
        OptionsDlgSharedDirsPickDlgTitle = 21050,
        OptionsDlgSharedDirsPickDlgSharedDir = 21051,
        OptionsDlgSharedDirsPickDlgPath = 21052,
        OptionsDlgSharedDirsPickDlgIncludeSubDirs = 21053,
        OptionsDlgSharedDirsPickDlgFriendsOnly = 21054,
        OptionsDlgSharedDirsPickDlgPickADir = 21060,
        OptionsDlgSharedDirsMissingOutputDir = 21070,
        OptionsDlgSharedDirsMaxNumberOfDirsReached = 21071,
        OptionsDlgSharedDirsNothingSelectedForRemoval = 21072,
        OptionsDlgSharedDirsReallyRemove = 21073,
        OptionsDlgSharedDirsSelectOneItemForModification = 21074,
        OptionsDlgSharedDirsNoDir = 21080,
        OptionsDlgSharedDirsDirNotFoundFormat = 21081,
        OptionsDlgSharedDirsWarnNoLongPathSupport = 21082,
        OptionsDlgSharedDirsConflictOfSharedDirsFormat = 21083,
        OptionsDlgSharedDirsDlgColumnTitles = 21100,
        // Advanced options page.
        OptionsDlgAdvancedTitle = 21200,
        OptionsDlgAdvancedComment = 21201,
        OptionsDlgAdvancedSamCaption = 21210,
        OptionsDlgAdvancedSamWarnText = 21211,
        OptionsDlgAdvancedSamTcpHost = 21212,
        OptionsDlgAdvancedSamTcpPort = 21213,
        OptionsDlgAdvancedMisc = 21220,
        OptionsDlgAdvancedMiscAvoidCompression = 21221,
        OptionsDlgAdvancedMiscRequireHighRepForAcceptingPubRequests = 21222,
        OptionsDlgAdvancedMiscDontCompact = 21223,
        OptionsDlgAdvancedMiscAllowComplFilesDeletion = 21224,
        OptionsDlgAdvancedMiscPublUnicodeTextsInNBStyleOnly = 21225,
        OptionsDlgAdvancedMiscShowAdvPeersPanelSubMenu = 21226,
        OptionsDlgAdvancedMiscRunFilesOnDblClick = 21227,
        OptionsDlgAdvancedMiscTreatFileNamesCaseSenstive = 21228,
        // Application dirs options page.
        OptionsDlgAppDirsTitle = 21300,
        OptionsDlgAppDirsComment = 21301,
        OptionsDlgAppDataGroupBox = 21310,
        OptionsDlgAppDataDir = 21311,
        OptionsDlgAppDataDirPickADir = 21312,
        OptionsDlgAppDataMsgPromptLater = 21313,
        OptionsDlgTempDataGroupBox = 21320,
        OptionsDlgTempDataPreviewFolder = 21321,
        OptionsDlgTempDataPreviewFolderPickADir = 21322,
        OptionsDlgTempDataDelPreviewFiles = 21323,
        OptionsDlgTempDataTempFolder = 21324,
        OptionsDlgTempDataTempFolderPickADir = 21325,
        OptionsDlgMsgNoFilesInDir = 21330,
        OptionsDlgQuestReallyDelFilesInDir = 21331,
        // List views options page.
        OptionsDlgListViewsTitle = 21400,
        OptionsDlgListViewsComment = 21401,
        OptionsDlgListViewsPerfGroupBox = 21410,
        OptionsDlgListViewsAutoSortThreshold = 21411,
        OptionsDlgListViewsItems = 21412,
        // Shared files options page.
        OptionsDlgSharedFilesTitle = 21500,
        OptionsDlgSharedFilesComment = 21501,
        OptionsDlgSharedFilesAccessGroupBox = 21510,
        OptionsDlgSharedFilesAttribsGroupBox = 21520,
        OptionsDlgSharedFilesAttribsInclHidden = 21521,
        OptionsDlgSharedFilesAttribsInclSystem = 21522,
        // Kademlia filter options page.
        OptionsDlgKadFilterTitle = 21600,
        OptionsDlgKadFilterComment = 21601,
        OptionsDlgKadFilterTestDlgTitle = 21610,
        OptionsDlgKadFilterTestDlgGroupBox = 21611,
        OptionsDlgKadFilterTestDlgLabel = 21612,
        OptionsDlgKadFilterTestDlgTextBox = 21613,
        // About dialog.
        AboutDlgTitle = 30000,
        AboutDlgAppDescription = 30001,
        AboutDlgAppCredits = 30002,
        AboutDlgAppVersion = 30003,
        AboutDlgAppBuildDate = 30004,
        AboutDlgProvideFeedback = 30005,
        AboutDlgCheckForUpdates = 30006,
        AboutDlgAppBuildAgeFormat = 30007,
        AboutDlgMailAddrCopiedToClipboardMsg = 30020,
        // License dialog.
        LicenseDlgTitle = 30100,
        LicenseDlgComment = 30101,
        LicenseDlgIAccept = 30102,
        LicenseDlgCopyLicenseToClipboard = 30103,
        // Create collection dialog.
        CreateCollDlgTitle = 30200,
        CreateCollDlgGroupBoxCaption = 30210,
        CreateCollDlgTitleLabel = 30211,
        CreateCollDlgNumOfFiles = 30212,
        CreateCollDlgTotalSize = 30213,
        CreateCollDlgMsgInsertTitle = 30220,
        CreateCollDlgMsgFailedToCreate = 30221,
        CreateCollDlgMsgFailedToWriteToFile = 30222,
        // Open collection dialog.
        OpenCollDlgTitle = 30300,
        // The following three are copied from CreateCollDlg.
        OpenCollDlgGroupBoxCaption = CreateCollDlgGroupBoxCaption,
        OpenCollDlgTitleLabel = CreateCollDlgTitleLabel,
        OpenCollDlgNumOfFiles = CreateCollDlgNumOfFiles,
        OpenCollDlgSelectedSize = 30312,
        OpenCollDlgFormatXOfN = 30313,
        OpenCollDlgAddToCustomTab = 30314,
        OpenCollDlgSelectNone = 30315,
        OpenCollDlgSelectAll = 30316,
        OpenCollDlgFailedToTakeOver = 30320,
        // Import partial downloads dialog.
        ImportPartialDlsDlgTitle = 30330,
        ImportPartialDlsDlgSourceFolder = 30331,
        ImportPartialDlsDlgCopyAttributes = 30332,
        ImportPartialDlsDlgImport = 30333,
        // Open collection column titles.
        OpenCollDlgColumnTitles = 30350,
        // Kademlia filter rule columns.
        OptionsDlgKadFilterRuleColumnEnabled = 30360,
        OptionsDlgKadFilterRuleColumnTitles = 30361,
        // Text items dialog column.
        TextItemsDlgColumnText = 30370,
        // Key generation dialog.
        KeyGenDlgTitle = 30400,
        KeyGenDlgMsgPleaseWait = 30401,
        // Import partial downloads list view columns.
        ImportPartialDlsDlgColumnTitles = 30450,
        // Move data files dialog.
        MoveDataFilesDlgTitle = 30500,
        MoveDataFilesDlgGroupBoxCaption = 30501,
        MoveDataFilesDlgSrcDir = 30502,
        MoveDataFilesDlgDestDir = 30503,
        MoveDataFilesDlgProgress = 30504,
        MoveDataFilesDlgProgressFormat = 30505,
        MoveDataFilesDlgQuestReallyMoveFiles = 30510,
        MoveDataFilesDlgQuestionReallyAbort = 30511,
        MoveDataFilesDlgMsgErrPreparingFiles = 30512,
        MoveDataFilesDlgMsgErrWhileMoving = 30513,
        MoveDataFilesDlgMsgErrInsuffSpace = 30514,
        MoveDataFilesDlgMsgMoveSuccess = 30515,
        MoveDataFilesDlgMsgDirChangeSuccess = 30516,
        // Removing files dialog.
        RemovingFilesDlgTitle = 30600,
        RemovingFilesDlgMsgMovingToTrash = 30601,
        RemovingFilesDlgMsgInGeneral = 30602,
        // Shutdown dialog.
        ShutdownDlgTitle = 30610,
        ShutdownDlgMsg = 30611,
        // Options take over dialog.
        OptionsTakeOverDlgTitle = 30620,
        OptionsTakeOverDlgMsg = 30621,
        // Loading application data dialog.
        LoadingAppDataDlgTitle = 30630,
        LoadingAppDataDlgMsg = 30631,
        // Import files dialog, load phase.
        ImportPartialDlsLoadingDlgTitle = 30640,
        ImportPartialDlsLoadingDlgMsg = 30641,
        // Import files dialog, import phase.
        ImportPartialDlsImportingDlgTitle = 30645,
        ImportPartialDlsImportingDlgMsg = 30646,
        // Import shared dirs dialog.
        ImportSharedDirsDlgTitle = 30650,
        ImportSharedDirsDlgMsg = 30651,
        // Moving complete files.
        MoveCompleteFilesDlgTitle = 30660,
        MoveCompleteFilesDlgMsg = 30661,
        // Importing Kademlia filter rules.
        ImportKadFilterRulesDlgTitle = 30670,
        ImportKadFilterRulesDlgMsg = 30671,

        // Reorder items dialog.
        ReorderItemsDlgGroupBoxCaption = 30801,
        ReorderItemsDlgListViewColCaption = 30802,
        ReorderItemsDlgTop = 30805,
        ReorderItemsDlgUp = 30806,
        ReorderItemsDlgDown = 30807,
        ReorderItemsDlgBottom = 30808,

        // Pick iMule folder dialog.
        PickiMuleFolderDialogMsg = 30850,
        PickiMuleFolderDialogTempDirExtraMsg = 30851,

        // iMule related messages.
        KeepiMuleClosedDuringOpMsg = 30860,
        iMuleProcessDetected = 30861,
        UnableToDetectiMuleCnfgMsg = 30862,

        // Sam bridge connectivity test dialog
        SamConnTestDlgTitle = 30900,
        SamConnTestDlgGroupBox = 30901,
        SamConnTestDlgInfoText = 30902,
        SamConnTestDlgTestResultLabel = 30903,
        SamConnTestDlgTestButton = 30904,
        SamConnTestDlgWaitDlgTitle = 30910,
        SamConnTestDlgSuccessMsg = 30912,
    }
}

// End of file.
