/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.27-src\Nachtblitz\Tabism.cs
 * PURPOSE: Custom tab control implementation
 * VERSION: 1.4.27
 * DATE: March 23, 2016
 * COPYRIGHT: Copyright (C) 2012-2016 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;
using Support;

namespace NB.Gui
{
    class GuiToolBarSet : IDisposable
    {
        public readonly Size f_IconSize;
        private readonly Bitmap f_Icons;
        private readonly Item[] f_Items;

        public uint p_Count
        {
            get { return (uint)f_Items.Length; }
        }

        public GuiToolBarSet(Size a_IconSize,
                             Point a_SrcPos,
                             Bitmap a_Icons,
                             string[] a_ToolTipTexts,
                             uint a_ToolTipArrIdx,
                             uint a_Count)
        {
            // The caller may dispose a_Icons immediately after
            // the call.

            const byte l_OriginalIconSize = 32;

            f_IconSize = a_IconSize;
            Rectangle l_DestRect = new Rectangle(0, 0, (int)a_Count * f_IconSize.Width, f_IconSize.Height);
            f_Icons = new Bitmap(l_DestRect.Width, l_DestRect.Height);

            Graphics l_G = Graphics.FromImage(f_Icons);
            {
                Rectangle l_SrcRect = new Rectangle(a_SrcPos.X, a_SrcPos.Y, (int)a_Count * l_OriginalIconSize, l_OriginalIconSize);
                l_G.DrawImage(a_Icons, l_DestRect, l_SrcRect, GraphicsUnit.Pixel);
            }
            l_G.Dispose();

            f_Items = new Item[a_Count];

            for (uint i = 0, k = a_ToolTipArrIdx; i != a_Count; i++, k++)
            {
                f_Items[i] = new Item(a_ToolTipTexts[k]);
            }
        }

        public void Dispose()
        {
            f_Icons.Dispose();
        }

        public Bitmap ExposeIconsAndItems(out Item[] a_Items)
        {
            a_Items = f_Items;

            return f_Icons;
        }

        public string GetToolTipText(uint a_Index)
        {
            return f_Items[a_Index].f_ToolTipText;
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct Item
        {
            public readonly string f_ToolTipText;

            public Item(string a_ToolTipText)
            {
                f_ToolTipText = a_ToolTipText;
            }
        }
    }

    class GuiToolBar : Control
    {
        private delegate void OnFlagChangedHandler(uint a_ItemIndex);

        public delegate void ItemEventHandler(GuiToolBar a_Sender,
                                              uint a_ItemIndex);

        public delegate void GetItemToolTipHandler(GuiToolBar a_Sender,
                                                   uint a_ItemIndex,
                                                   ref string a_ToolTipText);

        private static readonly Pen s_FramePen;
        private static readonly Brush s_ClickBrush;

        private readonly GuiToolBarSet f_Set;
        private readonly ButtonFlags[] f_Flags;

        private uint f_VisibleCount;
        private uint f_EnabledCount;

        private bool f_ShowToolTips;
        private int f_Space;
        private Rectangle f_PreferredRect;
        private Rectangle f_IconsRect;

        private MouseState f_MouseState;
        private uint f_MousedItemIndex;
        private uint f_MousedDisplayIndex;

        public event ItemEventHandler e_ItemVisibleChanged;
        public event ItemEventHandler e_ItemEnabledChanged;
        public event ItemEventHandler e_ItemMouseDown;
        public event ItemEventHandler e_ItemMouseUp;
        public event ItemEventHandler e_ItemMouseClicked;
        public event GetItemToolTipHandler e_GetItemToolTip;

        public uint p_Count
        {
            get { return f_Set.p_Count; }
        }

        public bool p_ShowToolTips
        {
            get { return f_ShowToolTips; }
            set
            {
                if (value != f_ShowToolTips)
                {
                    if (false == value)
                    {
                        ClearToolTip();
                    }

                    f_ShowToolTips = value;
                }
            }
        }

        static GuiToolBar()
        {
            s_FramePen = new Pen(Color.FromArgb(239, 188, 11));
            s_ClickBrush = new SolidBrush(Color.FromArgb(70, 250, 228, 150));
        }

        public GuiToolBar(GuiToolBarSet a_Set,
                          Control a_Owner) :
                     this(a_Set, a_Owner, false, false)
        {
        }

        public GuiToolBar(GuiToolBarSet a_Set,
                          Control a_Owner,
                          bool a_ShowAll,
                          bool a_EnableAll)
        {
            f_Set = a_Set;
            uint l_Count = p_Count;
            f_Flags = new ButtonFlags[l_Count];

            a_Owner.Controls.Add(this);

            this.SetStyle(ControlStyles.SupportsTransparentBackColor, true);
            this.BackColor = Color.Transparent;

            f_ShowToolTips = true;
            f_Space = Lyt.s_TinySpace2;
            UpdateSizes(false);

            if (true == a_EnableAll)
            {
                AdjustEnabledRange(0, l_Count, true);
            }

            if (true == a_ShowAll)
            {
                AdjustVisibilityRange(0, l_Count, true);
            }
        }

        protected override void Dispose(bool a_Disposing)
        {
            if (true == a_Disposing)
            {
                ClearToolTip();
            }

            base.Dispose(a_Disposing);
        }

        private ButtonFlags GetItemFlags(uint a_ItemIndex)
        {
            return f_Flags[a_ItemIndex];
        }

        public bool IsVisible(uint a_ItemIndex)
        {
            return (0 != (GetItemFlags(a_ItemIndex) & ButtonFlags.Visible));
        }

        public bool IsEnabled(uint a_ItemIndex)
        {
            return (0 != (GetItemFlags(a_ItemIndex) & ButtonFlags.Enabled));
        }

        public bool IsVisibleAndEnabled(uint a_ItemIndex)
        {
            return (ButtonFlags.VisibleAndEnabled == (GetItemFlags(a_ItemIndex) & ButtonFlags.VisibleAndEnabled));
        }

        public void AdjustVisibilityAndEnabledRange(uint a_ItemIdx,
                                                    uint a_Len,
                                                    bool a_Visible,
                                                    bool a_Enabled)
        {
            // This routine updates in the fashion (order of calls)
            // that is generally most efficient.

            if (true == a_Visible)
            {
                // Enable before (possibly) getting visible.

                AdjustEnabledRange(a_ItemIdx, a_Len, a_Enabled);
            }

            AdjustVisibilityRange(a_ItemIdx, a_Len, a_Visible);

            if (false == a_Visible)
            {
                // Disable after getting invisible.

                AdjustEnabledRange(a_ItemIdx, a_Len, a_Enabled);
            }
        }

        public void AdjustVisibility(uint a_ItemIdx,
                                     bool a_Visible)
        {
            AdjustVisibilityRange(a_ItemIdx, 1, a_Visible);
        }

        public void AdjustVisibilityRange(uint a_ItemIdx,
                                          uint a_Len,
                                          bool a_Visible)
        {
            if (0 == a_Len)
            {
                return;
            }

            // Before changing anything, check if there will be changes
            // that affect at least the position of the moused item and
            // if so reset the mouse status.

            if (MouseState.None != f_MouseState &&
                a_ItemIdx <= f_MousedItemIndex)
            {
                for (uint i = a_ItemIdx, lim = (a_ItemIdx + a_Len); i != lim; i++)
                {
                    if (IsVisible(i) != a_Visible)
                    {
                        UpdateMousePosition();

                        break;
                    }

                    if (i >= f_MousedItemIndex)
                    {
                        break;
                    }
                }
            }

            bool l_Change = AdjustFlagsCore(a_ItemIdx, a_Len, OnItemVisibleChanged, ButtonFlags.Visible, a_Visible, ref f_VisibleCount);

            if (true == l_Change)
            {
                UpdateSizes(true);

                uint l_DisplIdx;

                if (0 != f_VisibleCount &&
                    0 != a_ItemIdx &&
                    true == ItemIndexToDisplayIndex(a_ItemIdx - 1, true, out l_DisplIdx))
                {
                    Rectangle l_InvalRect;
                    CalcItemRectangle(l_DisplIdx, 1, out l_InvalRect);
                    l_InvalRect.X += l_InvalRect.Width;
                    l_InvalRect.Width = (int.MaxValue - l_InvalRect.X);
                    this.Invalidate(l_InvalRect);
                }
                else
                {
                    this.Invalidate();
                }
            }
        }

        public void AdjustEnabled(uint a_ItemIdx,
                                  bool a_Enabled)
        {
            AdjustEnabledRange(a_ItemIdx, 1, a_Enabled);
        }

        public void AdjustEnabledRange(uint a_ItemIdx,
                                       uint a_Len,
                                       bool a_Enabled)
        {
            if (0 == a_Len)
            {
                return;
            }

            if (false == a_Enabled &&
                MouseState.None != f_MouseState &&
                a_ItemIdx <= f_MousedItemIndex &&
                (a_ItemIdx + a_Len) > f_MousedItemIndex)
            {
                // Reset the mouse status.

                UpdateMousePosition();
            }

            bool l_Change = AdjustFlagsCore(a_ItemIdx, a_Len, OnItemEnabledChanged, ButtonFlags.Enabled, a_Enabled, ref f_EnabledCount);
            uint l_LowDisplIdx;
            uint l_HighDisplIdx;

            if (true == l_Change &&
                0 != f_VisibleCount &&
                true == ItemIndexToDisplayIndex(a_ItemIdx, false, out l_LowDisplIdx) &&
                true == ItemIndexToDisplayIndex(a_ItemIdx + a_Len - 1, true, out l_HighDisplIdx) &&
                l_LowDisplIdx <= l_HighDisplIdx)
            {
                InvalidateItems(l_LowDisplIdx, l_HighDisplIdx - l_LowDisplIdx + 1);
            }
        }

        private bool AdjustFlagsCore(uint a_ItemIdx,
                                     uint a_Len,
                                     OnFlagChangedHandler a_Handler,
                                     ButtonFlags a_Flag,
                                     bool a_Value,
                                     ref uint a_Count)
        {
            // This routine doesn't invalidate the affected display
            // regions. Returns nonzero if there was a change.

            bool l_Change = false;

            for (uint i = a_ItemIdx, lim = (a_ItemIdx + a_Len); i != lim; i++)
            {
                if (a_Value == (0 != (f_Flags[i] & a_Flag)))
                {
                    continue;
                }

                l_Change = true;

                if (false == a_Value)
                {
                    f_Flags[i] &= ~a_Flag;
                    a_Count--;
                }
                else
                {
                    f_Flags[i] |= a_Flag;
                    a_Count++;
                }

                a_Handler(i);
            }

            return l_Change;
        }

        private void CalcSpaces(out int a_InitialSpace,
                                out int a_EndingSpace)
        {
            a_InitialSpace = (f_Space / 2);
            a_EndingSpace = (f_Space - a_InitialSpace);
        }

        private void DeflateItemRectToDisplayRect(ref Rectangle a_Rect)
        {
            InflateRect(-(f_Space / 2), -f_Space, ref a_Rect);
        }

        private void UpdateSizes(bool a_PaddingOrVisibleCountChanged)
        {
            Padding l_Padding = this.Padding;

            if (true == a_PaddingOrVisibleCountChanged)
            {
                f_PreferredRect = new Rectangle(l_Padding.Left, l_Padding.Top, 0, 0);

                if (0 != f_VisibleCount)
                {
                    Size l_Size;
                    CalcSizeCore(f_VisibleCount, out l_Size);
                    f_PreferredRect.Size = l_Size;
                }
            }

            f_IconsRect = f_PreferredRect;
            f_IconsRect.Width = Math.Min(f_IconsRect.Width, Math.Max(0, this.Width - l_Padding.Horizontal));
            f_IconsRect.Height = Math.Min(f_IconsRect.Height, Math.Max(0, this.Height - l_Padding.Vertical));
        }

        public override Size GetPreferredSize(Size a_ProposedSize)
        {
            // We'll ignore a_ProposedSize.

            Size l_Size;
            GetPreferredSize(out l_Size);

            return l_Size;
        }

        private void CalcSizeCore(uint a_VisibleCount,
                                  out Size a_Size)
        {
            Size l_IconSize = f_Set.f_IconSize;
            a_Size = new Size((int)a_VisibleCount * (l_IconSize.Width + f_Space), l_IconSize.Height + f_Space);
        }

        public void GetPreferredSize(uint a_VisibleCount,
                                     out Size a_Size)
        {
            CalcSizeCore(a_VisibleCount, out a_Size);
            a_Size.Width += this.Padding.Horizontal;
            a_Size.Height += this.Padding.Vertical;
        }

        public void GetPreferredSize(out Size a_Size)
        {
            a_Size = new Size(f_PreferredRect.Right + this.Padding.Right, f_PreferredRect.Bottom + this.Padding.Bottom);
        }

        protected override void OnPaddingChanged(EventArgs e)
        {
            // Need to call base class implementation.

            UpdateSizes(true);

            this.Invalidate();

            base.OnPaddingChanged(e);
        }

        protected override void OnResize(EventArgs e)
        {
            // Need to call base class implementation.

            UpdateSizes(false);

            base.OnResize(e);
        }

        private void InvalidateItem(uint a_DisplayIndex)
        {
            InvalidateItems(a_DisplayIndex, 1);
        }

        private void InvalidateItems(uint a_DisplayIndex,
                                     uint a_Count)
        {
            Rectangle l_Rect;
            CalcItemRectangle(a_DisplayIndex, a_Count, out l_Rect);
            this.Invalidate(l_Rect);
        }

        private void CalcItemRectangle(uint a_DisplayIndex,
                                       uint a_Count,
                                       out Rectangle a_Rect)
        {
            Size l_IconSize = f_Set.f_IconSize;
            a_Rect = new Rectangle(f_IconsRect.X, f_IconsRect.Y, l_IconSize.Width + f_Space, l_IconSize.Height + f_Space);
            a_Rect.X += ((int)a_DisplayIndex * a_Rect.Width);
            a_Rect.Width *= (int)a_Count;
        }

        private bool HitTest(Point a_Pos,
                             out uint a_DisplayIndex)
        {
            if (false == f_IconsRect.Contains(a_Pos))
            {
                a_DisplayIndex = Util.c_NoIndex;

                return false;
            }

            a_DisplayIndex = (uint)((a_Pos.X - f_IconsRect.X) / (f_Set.f_IconSize.Width + f_Space));

            return true;
        }

        private void DrawItem(Graphics a_G,
                              Bitmap a_SrcImage,
                              ref Rectangle a_SrcRect,
                              ref Rectangle a_DestRect,
                              bool a_Enabled)
        {
            if (true == a_Enabled)
            {
                a_G.DrawImage(a_SrcImage, a_DestRect, a_SrcRect, GraphicsUnit.Pixel);

                return;
            }

            bool l_UseOriginalImage = (0 == a_SrcRect.X &&
                                       0 == a_SrcRect.Y &&
                                       a_SrcImage.Size == a_SrcRect.Size);
            Bitmap l_Map;

            if (false == l_UseOriginalImage)
            {
                // Create a temporary image, since the interface isn't
                // powerful enough to support drawing of partial images.

                l_Map = new Bitmap(a_DestRect.Width, a_DestRect.Height);

                Graphics l_TempG = Graphics.FromImage(l_Map);
                {
                    l_TempG.DrawImage(a_SrcImage, new Rectangle(Point.Empty, a_DestRect.Size), a_SrcRect, GraphicsUnit.Pixel);
                }
                l_TempG.Dispose();
            }
            else
            {
                l_Map = a_SrcImage;
            }
            {
                ControlPaint.DrawImageDisabled(a_G, l_Map, a_DestRect.X, a_DestRect.Y, Color.Transparent);
            }
            if (false == l_UseOriginalImage)
            {
                l_Map.Dispose();
            }
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            // Need to call base class implementation.

            Graphics l_G = e.Graphics;
            RectangleF l_StoredClip = l_G.ClipBounds;
            Rectangle l_NewClip = e.ClipRectangle;
            l_NewClip.Intersect(f_IconsRect);

            if (true == MainForm.HasPositiveArea(ref l_NewClip))
            {
                l_G.SetClip(l_NewClip);
                {
                    PaintCore(l_G, ref l_NewClip);
                }
                l_G.SetClip(l_StoredClip);
            }

            base.OnPaint(e);
        }

        private void PaintCore(Graphics a_G,
                               ref Rectangle a_InvalidRect)
        {
            GuiToolBarSet.Item[] l_Items;
            Bitmap l_Icons = f_Set.ExposeIconsAndItems(out l_Items);

            Rectangle l_SrcRect = new Rectangle(Point.Empty, f_Set.f_IconSize);
            Rectangle l_ItemRect;
            CalcItemRectangle(0, 1, out l_ItemRect);
            int l_InvalRight = a_InvalidRect.Right;

            for (uint i = 0, k = 0, lim = p_Count; i != lim; i++)
            {
                if (true == IsVisible(i))
                {
                    if (MouseState.HoveringEnabled <= f_MouseState &&
                        k == f_MousedDisplayIndex)
                    {
                        Rectangle l_MouseRect = l_ItemRect;
                        // Reduce the rect to make the result of
                        // FillRectangle() really fit the item rect.
                        l_MouseRect.Width--;
                        l_MouseRect.Height--;

                        a_G.DrawRectangle(s_FramePen, l_MouseRect);

                        if (MouseState.Clicking == f_MouseState)
                        {
                            a_G.FillRectangle(s_ClickBrush, l_MouseRect);
                        }
                    }

                    int l_DestRight = l_ItemRect.Right;

                    if (l_DestRight >= a_InvalidRect.X)
                    {
                        // This item is possibly visible.

                        Rectangle l_DestRect = l_ItemRect;
                        DeflateItemRectToDisplayRect(ref l_DestRect);
                        DrawItem(a_G, l_Icons, ref l_SrcRect, ref l_DestRect, IsEnabled(i));
                    }

                    if (l_DestRight >= l_InvalRight)
                    {
                        // End reached.

                        break;
                    }

                    l_ItemRect.X = l_DestRight;
                    k++;
                }

                l_SrcRect.X += l_SrcRect.Width;
            }
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            // Need to call base class implementation.

            if (MouseState.Clicking != f_MouseState)
            {
                UpdateMousePosition(true, e.Location, false);
            }

            base.OnMouseMove(e);
        }

        protected override void OnMouseLeave(EventArgs e)
        {
            // Need to call base class implementation.

            if (MouseState.Clicking != f_MouseState)
            {
                UpdateMousePosition();
            }

            base.OnMouseLeave(e);
        }

        protected override void OnMouseDown(MouseEventArgs e)
        {
            // Need to call base class implementation.

            UpdateMousePosition(true, e.Location, MouseButtons.Left == e.Button);

            base.OnMouseDown(e);
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            // Need to call base class implementation.

            bool l_WasClicked = (MouseState.Clicking == f_MouseState);
            uint l_ItemIndex = f_MousedItemIndex;

            UpdateMousePosition(true, e.Location, false);

            // See if a click was completed.

            if (true == l_WasClicked &&
                MouseState.HoveringEnabled == f_MouseState &&
                l_ItemIndex == f_MousedItemIndex &&
                true == IsVisibleAndEnabled(l_ItemIndex))
            {
                OnItemMouseClicked(l_ItemIndex);
            }

            base.OnMouseUp(e);
        }

        public bool TryRegisterKey(KeyEventArgs e)
        {
            if (MouseState.Clicking == f_MouseState &&
                Keys.Escape == e.KeyCode)
            {
                UpdateMousePosition();
                e.Handled = true;

                return true;
            }

            return false;
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            // Need to call base class implementation.

            bool l_Handled = TryRegisterKey(e);

            base.OnKeyDown(e);
        }

        protected virtual void OnItemVisibleChanged(uint a_ItemIdx)
        {
            if (null != e_ItemVisibleChanged)
            {
                e_ItemVisibleChanged(this, a_ItemIdx);
            }
        }

        protected virtual void OnItemEnabledChanged(uint a_ItemIdx)
        {
            if (null != e_ItemEnabledChanged)
            {
                e_ItemEnabledChanged(this, a_ItemIdx);
            }
        }

        protected virtual void OnItemMouseDown(uint a_ItemIdx)
        {
            // NOTE: Try not to block (e.g. by showing a new modal
            // dialog) while this handler runs, since this control
            // might otherwise miss the mouse up event.

            if (null != e_ItemMouseDown)
            {
                e_ItemMouseDown(this, a_ItemIdx);
            }
        }

        private void OnItemMouseUpCore(uint a_ItemIdx)
        {
            OnItemMouseUp(a_ItemIdx);
        }

        protected virtual void OnItemMouseUp(uint a_ItemIdx)
        {
            if (null != e_ItemMouseUp)
            {
                e_ItemMouseUp(this, a_ItemIdx);
            }
        }

        protected virtual void OnItemMouseClicked(uint a_ItemIdx)
        {
            if (null != e_ItemMouseClicked)
            {
                e_ItemMouseClicked(this, a_ItemIdx);
            }
        }

        protected void OnGetItemToolTip(uint a_ItemIndex,
                                        ref string a_ToolTipText)
        {
            if (null != e_GetItemToolTip)
            {
                e_GetItemToolTip(this, a_ItemIndex, ref a_ToolTipText);
            }
        }

        private void ClearToolTip()
        {
            Software.s_ToolTip.ClearToolTip(this);
        }

        private void UpdateMousePosition()
        {
            UpdateMousePosition(false, Point.Empty, false);
        }

        private void UpdateMousePosition(bool a_PosIsValid,
                                         Point a_Pos,
                                         bool a_MouseDown)
        {
            bool l_Hit = false;
            bool l_Enabled = false;
            uint l_ItemIdx = Util.c_NoIndex;
            uint l_DisplIdx = Util.c_NoIndex;

            if (true == a_PosIsValid &&
                true == HitTest(a_Pos, out l_DisplIdx))
            {
                l_Hit = true;
                l_ItemIdx = DisplayIndexToItemIndexCore(l_DisplIdx);
                l_Enabled = IsVisibleAndEnabled(l_ItemIdx);
            }

            if (false == l_Hit)
            {
                if (MouseState.HoveringEnabled <= f_MouseState)
                {
                    if (MouseState.Clicking == f_MouseState)
                    {
                        OnItemMouseUpCore(f_MousedItemIndex);
                    }

                    InvalidateItem(f_MousedDisplayIndex);
                }

                f_MouseState = MouseState.None;
                ClearToolTip();

                return;
            }

            MouseState l_NewState = (false == l_Enabled ? MouseState.HoveringDisabled : (false == a_MouseDown ? MouseState.HoveringEnabled : MouseState.Clicking));

            if (l_NewState != f_MouseState ||
                l_ItemIdx != f_MousedItemIndex)
            {
                if (MouseState.Clicking == f_MouseState)
                {
                    OnItemMouseUpCore(f_MousedItemIndex);
                }

                bool l_InvalidateNewSlot = (MouseState.HoveringEnabled <= l_NewState);

                if (l_DisplIdx != f_MousedDisplayIndex)
                {
                    if (MouseState.HoveringEnabled <= f_MouseState)
                    {
                        // Invalidate the old display slot.
                        InvalidateItem(f_MousedDisplayIndex);
                    }
                }
                else if (MouseState.HoveringEnabled <= f_MouseState)
                {
                    l_InvalidateNewSlot = true;
                }

                f_MouseState = l_NewState;
                f_MousedItemIndex = l_ItemIdx;
                f_MousedDisplayIndex = l_DisplIdx;

                if (true == l_InvalidateNewSlot)
                {
                    InvalidateItem(l_DisplIdx);
                }

                // Set up tool tip.

                if (MouseState.Clicking == l_NewState)
                {
                    // Do not show the tool tip while the item is being
                    // clicked on.

                    ClearToolTip();

                    OnItemMouseDown(l_ItemIdx);
                }
                else
                {
                    if (true == f_ShowToolTips)
                    {
                        string l_ToolTip = f_Set.GetToolTipText(l_ItemIdx);

                        OnGetItemToolTip(l_ItemIdx, ref l_ToolTip);

                        Software.s_ToolTip.SetToolTip(this, l_ToolTip);
                    }
                }
            }
        }

        private uint DisplayIndexToItemIndexCore(uint a_DisplayIndex)
        {
            for (uint i = 0, k = 0; true; i++)
            {
                if (false == IsVisible(i))
                {
                    continue;
                }

                if (k == a_DisplayIndex)
                {
                    return i;
                }

                k++;
            }
        }

        private bool ItemIndexToDisplayIndex(uint a_ItemIndex,
                                             bool a_RoundDown,
                                             out uint a_DisplayIndex)
        {
            // Returns nonzero if there was a valid display index
            // found in the specified search direction starting at
            // the passed item index.

            if (0 != f_VisibleCount)
            {
                uint l_LastValidIdx = (p_Count - 1);

                for (uint i = 0, l_LastDisplIdx = Util.c_NoIndex; true; i++)
                {
                    if (true == IsVisible(i))
                    {
                        if (Util.c_NoIndex == l_LastDisplIdx)
                        {
                            l_LastDisplIdx = 0;
                        }
                        else
                        {
                            l_LastDisplIdx++;
                        }

                        if (false == a_RoundDown &&
                            i >= a_ItemIndex)
                        {
                            a_DisplayIndex = l_LastDisplIdx;

                            return true;
                        }
                    }

                    if (false == a_RoundDown)
                    {
                        if (i == l_LastValidIdx)
                        {
                            break;
                        }
                    }
                    else
                    {
                        if (i == a_ItemIndex)
                        {
                            if (Util.c_NoIndex == l_LastDisplIdx)
                            {
                                break;
                            }

                            a_DisplayIndex = l_LastDisplIdx;

                            return true;
                        }
                    }
                }
            }

            a_DisplayIndex = Util.c_NoIndex;

            return false;
        }

        private static void InflateRect(int a_XY,
                                        int a_WidthHeight,
                                        ref Rectangle a_Rect)
        {
            a_Rect.X -= a_XY;
            a_Rect.Y -= a_XY;
            a_Rect.Width += a_WidthHeight;
            a_Rect.Height += a_WidthHeight;
        }

        [Flags]
        private enum ButtonFlags : byte
        {
            None = 0,

            Visible = 0x1,
            Enabled = 0x2,

            VisibleAndEnabled = (Visible | Enabled),
        }

        private enum MouseState : byte
        {
            None = 0,

            // Hovering over a disabled item (mainly needed for the
            // tool tip).
            HoveringDisabled = 1,
            // The following two values have an visual indicator each.
            // Hovering over an enabled item.
            HoveringEnabled = 2,
            // Clicking on an enabled item.
            Clicking = 3,
        }
    }

    class AdvancedTabControl : Control, System.Collections.IEnumerable, IEnumerable<AdvancedTabPage>
    {
        private const byte c_MinFrameSize = 3;
        // These spaces should be 2 (pixels) at least to match the
        // appearance of the frame.
        private const byte c_TabBorderSpaceStart = 4;
        private const byte c_TabBorderSpaceEnd = 6;
        private const byte c_TabBaseHeight = 2;

        public delegate void SelectedPageChangedHandler(AdvancedTabControl a_Sender);

        public delegate void SelectedPageChangedExHandler(AdvancedTabControl a_Sender,
                                                          AdvancedTabPage a_OldSelPage,
                                                          SelectionChangeOrigins a_OriginDeselect,
                                                          SelectionChangeOrigins a_OriginSelect);

        public delegate void AbortablePageHandler(AdvancedTabControl a_Sender,
                                                  AdvancedTabPage a_Page,
                                                  ref bool a_Handled);

        public delegate void AbortablePageHandlerEx(AdvancedTabControl a_Sender,
                                                    AdvancedTabPage a_Page,
                                                    ref bool a_Handled,
                                                    ref bool a_DisposePage);

        public delegate void DraggingTabHandler(AdvancedTabControl a_Sender,
                                                AdvancedTabPage a_Page,
                                                ref uint a_NewIndex);

        public delegate void AfterDraggingTabHandler(AdvancedTabControl a_Sender,
                                                     AdvancedTabPage a_Page,
                                                     bool a_DidDrag,
                                                     bool a_Commited);

        public delegate void PrepareListingPagesHandler(AdvancedTabControl a_Sender,
                                                        ref PageListItem[] a_Items);

        public delegate void ItemSelectedFromPageListHandler(AdvancedTabControl a_Sender,
                                                             uint a_Index,
                                                             ref bool a_Handled);

        // s_Brushes is mutable and shared (single threaded) by all
        // instances of this class.
        private static TabBrushes s_Brushes;
        private static readonly Pen s_OuterPen;
        private static readonly Pen s_LightFramePen;
        private static readonly Brush s_FrameBrush;

        private static readonly GuiToolBarSet s_ToolBarSet;

        private GuiToolBar f_ToolBar;
        private FlexArray<AdvancedTabPage> f_Pages;
        // f_VisiblePages is the count of pages that have the visible
        // flag set.
        private uint f_VisiblePages;

        private int f_ReservedToolBarWidth;
        private int f_TextHeight;
        private Size f_IconSize;
        private int f_TabHeight;
        private int f_HiddenWidth;
        private int f_TextY;
        private int f_ImageY;

        private Rectangle f_HeaderRect;
        private Rectangle f_InnerRect;
        private Rectangle f_FrameRect;

        // The following two indices can be invalid (use
        // AreAnyTabsVisible() to find out). Both are inclusive and in
        // item index notation.
        private uint f_DisplStartIdx;
        private uint f_DisplEndIdx;
        // f_DisplOverflow indicates that there was not enough space
        // to display further existing tabs (that have their
        // visibility flag set).
        private bool f_DisplOverflow;
        private FlexArray<TabPosition> f_TabPositions;

        // The following four references can be null.
        private AdvancedTabPage f_SelPage;
        private AdvancedTabPage f_MousedPage;
        private AdvancedTabPage f_DraggedPage;
        private DragContext f_DragCntxt;

        // The following reference can be null.
        private DynamicScrollContext f_DynScrollCntxt;

        // f_TabCntxtMenuStrip can be null. This instance does not
        // own the tab menu (if present).
        private ContextMenuStrip f_TabCntxtMenuStrip;

        // The following three references can be null.
        private ContextMenuStrip f_PageListingMenu;
        private Font f_PageListingBoldFont;
        private Font f_PageListingItalicFont;

        private BitField f_Flags;

        public event SelectedPageChangedHandler e_SelectedPageChanged;
        public event SelectedPageChangedExHandler e_SelectedPageChangedEx;
        public event AbortablePageHandler e_BeforeDraggingTab;
        public event DraggingTabHandler e_DraggingTab;
        public event AfterDraggingTabHandler e_AfterDraggingTab;
        public event AbortablePageHandler e_HidingSelPageThroughUI;
        public event AbortablePageHandlerEx e_ClosingSelPageThroughUI;
        public event PrepareListingPagesHandler e_PrepareListingPages;
        public event ItemSelectedFromPageListHandler e_ItemSelectedFromPageList;

        public uint p_PageCount
        {
            get { return (uint)f_Pages.p_Count; }
        }

        public uint p_VisiblePageCount
        {
            get { return f_VisiblePages; }
        }

        public AdvancedTabPage p_SelectedPage
        {
            get { return f_SelPage; }
        }

        public uint p_SelectedPageIndex
        {
            // The return value can be Util.c_NoIndex if nothing is selected.
            get { return (null == f_SelPage ? Util.c_NoIndex : GetIndexOfPageCore(f_SelPage)); }
        }

        public int p_TabHeight
        {
            get { return f_TabHeight; }
        }

        public bool p_DisplayImages
        {
            get { return f_Flags.CheckFlag(FlagsToUInt(TabControlFlags.DisplayImages)); }
            set { AdjustVisibilityOfImages(value); }
        }

        public bool p_AllowSelectingThroughUI
        {
            get { return f_Flags.CheckFlag(FlagsToUInt(TabControlFlags.AllowSelectingThroughUI)); }
            set { f_Flags.AdjustFlag(FlagsToUInt(TabControlFlags.AllowSelectingThroughUI), value); }
        }

        public bool p_AllowScrollingThroughButtons
        {
            get { return f_Flags.CheckFlag(FlagsToUInt(TabControlFlags.AllowScrollingThroughToolBar)); }
            set { AdjustToolBar(TabControlFlags.AllowScrollingThroughToolBar, value); }
        }

        public bool p_AllowHidingThroughUI
        {
            get { return f_Flags.CheckFlag(FlagsToUInt(TabControlFlags.AllowHidingThroughUI)); }
            set { AdjustToolBar(TabControlFlags.AllowHidingThroughUI, value); }
        }

        public bool p_AllowClosingThroughUI
        {
            get { return f_Flags.CheckFlag(FlagsToUInt(TabControlFlags.AllowClosingThroughUI)); }
            set { AdjustToolBar(TabControlFlags.AllowClosingThroughUI, value); }
        }

        // This property can be used if the list is overriden and shows
        // custom data.
        public bool p_AlwaysShowPageListingButton
        {
            get { return f_Flags.CheckFlag(FlagsToUInt(TabControlFlags.AlwaysShowPageListingToolBarButton)); }
            set { AdjustToolBar(TabControlFlags.AlwaysShowPageListingToolBarButton, value); }
        }

        public bool p_IncludeHiddenPagesInAutoPageListing
        {
            get { return f_Flags.CheckFlag(FlagsToUInt(TabControlFlags.IncludeHiddenPagesInAutoPageListing)); }
            set { AdjustToolBar(TabControlFlags.IncludeHiddenPagesInAutoPageListing, value); }
        }

        public bool p_AllowReorderingThroughUI
        {
            get { return f_Flags.CheckFlag(FlagsToUInt(TabControlFlags.AllowReorderingThroughUI)); }
            set
            {
                if (true == f_Flags.AdjustFlagWithReturn(FlagsToUInt(TabControlFlags.AllowReorderingThroughUI), value) &&
                    false == value)
                {
                    AbortAnyDragging(SelectionChangeOrigins.Programmatic);
                }
            }
        }

        public ContextMenuStrip p_TabContextMenuStrip
        {
            get { return f_TabCntxtMenuStrip; }
            set { f_TabCntxtMenuStrip = value; }
        }

        private int p_FrameCore
        {
            get { return Lyt.s_TinySpace; }
        }

        private int p_FrameSize
        {
            get { return (c_MinFrameSize + p_FrameCore); }
        }

        static AdvancedTabControl()
        {
            s_Brushes = new TabBrushes();
            s_OuterPen = new Pen(Color.FromArgb(183, 170, 97));
            s_LightFramePen = Pens.White;
            s_FrameBrush = new SolidBrush(Color.FromArgb(245, 250, 208));

            System.IO.Stream l_Strm = Software.OpenResource("TabControlToolBar.png", false);
            Bitmap l_ToolBarIcons = new Bitmap(l_Strm);
            l_Strm.Dispose();
            {
                Size l_IconSize = Lyt.s_SmallIconSize;
                // Keep the tool bar somewhat away from the page frame
                // of the tab control, since it looks better that way.
                l_IconSize.Width -= Lyt.CalcX(1);
                l_IconSize.Height -= Lyt.CalcY(1);
                s_ToolBarSet = new GuiToolBarSet(l_IconSize, Point.Empty, l_ToolBarIcons, new string[] { "Scroll Left", "Scroll Right", "Hide Tab", "Close Tab", "List All Pages" }, 0, 5);
            }
            l_ToolBarIcons.Dispose();
        }

        public AdvancedTabControl(Control a_Owner)
        {
            a_Owner.Controls.Add(this);

            f_ToolBar = new GuiToolBar(s_ToolBarSet, this);

            FlexArray<AdvancedTabPage>.Create(out f_Pages);

            f_IconSize = Lyt.s_SmallIconSize;
            f_DisplEndIdx = f_DisplStartIdx = Util.c_NoIndex;
            FlexArray<TabPosition>.Create(out f_TabPositions);

            const TabControlFlags l_InitialFlags = (TabControlFlags.DisplayImages | TabControlFlags.AllowSelectingThroughUI | TabControlFlags.AllowScrollingThroughToolBar | TabControlFlags.IncludeHiddenPagesInAutoPageListing);
            f_Flags.SetCombination(FlagsToUInt(l_InitialFlags));

            this.DoubleBuffered = true;
            this.SetStyle(ControlStyles.SupportsTransparentBackColor, true);
            this.BackColor = Color.Transparent;
            int l_HorPadding = Lyt.s_TinySpace;
            f_ToolBar.Padding = new Padding(l_HorPadding, 0, l_HorPadding, 0);
            // The following button is useable whenever visible.
            f_ToolBar.AdjustEnabled((uint)ToolBarButtons.ListAllPages, true);

            NotifyFontUpdated();
            UpdateToolBar();
            UpdateRectangles();

            f_ToolBar.e_ItemMouseDown += ToolBar_ItemMouseDown;
            f_ToolBar.e_ItemMouseUp += ToolBar_ItemMouseUp;
            f_ToolBar.e_ItemMouseClicked += ToolBar_ItemMouseClicked;
        }

        protected override void Dispose(bool a_Disposing)
        {
            if (true == a_Disposing)
            {
                AbortAnyDragging(SelectionChangeOrigins.Programmatic);
                StopAnyDynamicScrolling();

                Util.DisposeAndZero(ref f_PageListingMenu);
                DisposePageListingFonts();

                RemoveAllPages(true, SelectionChangeOrigins.Programmatic);

                f_Pages.Clear(false);
                f_TabPositions.Clear(false);

                f_ToolBar.Dispose();
            }

            base.Dispose(a_Disposing);
        }

        public void SetDefaultPadding(bool a_Top)
        {
            int l_Val = (Lyt.s_TinySpace + 1);
            this.Padding = new Padding(l_Val, false == a_Top ? 0 : l_Val, l_Val, false == a_Top ? l_Val : 0);
        }

        private void AdjustVisibilityOfImages(bool a_Value)
        {
            if (false == f_Flags.AdjustFlagWithReturn(FlagsToUInt(TabControlFlags.DisplayImages), a_Value))
            {
                return;
            }

            CalculateVisibleTabsAndKeepSelDisplayedIfAlready();
        }

        protected override void OnFontChanged(EventArgs e)
        {
            // Need to call base class implementation.

            NotifyFontUpdated();

            base.OnFontChanged(e);
        }

        private void NotifyFontUpdated()
        {
            AdvancedTabPage[] l_Arr = f_Pages.p_Arr;

            for (uint i = p_PageCount; 0 != i--;)
            {
                l_Arr[i].InvalidateCaptionWidth();
            }

            NotifySizesUpdated();
        }

        private void NotifySizesUpdated()
        {
            f_TextHeight = (this.Font.Height + 1);
            int l_AlwaysVisibleWidth = Lyt.s_SmallSpace;

            f_TabHeight = (Lyt.s_SmallSpace + Math.Max(f_TextHeight, f_IconSize.Height));
            f_HiddenWidth = Math.Max(0, f_TabHeight - l_AlwaysVisibleWidth);
            f_TextY = (((f_TabHeight - f_TextHeight) / 2) + 1);
            f_ImageY = (((f_TabHeight - f_IconSize.Height) / 2) + 1);

            UpdateRectangles();

            if (null != f_SelPage)
            {
                PlaceSelectedPageIntoInnerRect();
            }
        }

        private void UpdateRectangles()
        {
            f_FrameRect = new Rectangle(this.Padding.Left, this.Padding.Top + f_TabHeight, this.Width - this.Padding.Horizontal, this.Height - this.Padding.Vertical - f_TabHeight);
            f_HeaderRect = new Rectangle(f_FrameRect.X + c_TabBorderSpaceStart, this.Padding.Top, f_FrameRect.Width - c_TabBorderSpaceStart - c_TabBorderSpaceEnd - f_ReservedToolBarWidth, f_TabHeight);
            f_InnerRect = f_FrameRect;
            int l_FrameSize = p_FrameSize;
            f_InnerRect.Inflate(-l_FrameSize, -l_FrameSize);

            MainForm.ZeroImpossibleSize(ref f_HeaderRect);
            MainForm.ZeroImpossibleSize(ref f_InnerRect);
            MainForm.ZeroImpossibleSize(ref f_FrameRect);

            this.Invalidate();
        }

        private void BuildWeakDragRectForScrolling(out Rectangle a_Rect)
        {
            int l_DisplCount = f_TabPositions.p_Count;
            TabPosition[] l_Arr = f_TabPositions.p_Arr;
            int l_ActWidth = (l_Arr[l_DisplCount - 1].CalcRight(this) - f_HeaderRect.X);

            a_Rect = f_HeaderRect;
            a_Rect.Width = (l_ActWidth + 1);
        }

        private void AdjustToolBar(TabControlFlags a_Flag,
                                   bool a_Value)
        {
            if (false == f_Flags.AdjustFlagWithReturn(FlagsToUInt(a_Flag), a_Value))
            {
                return;
            }

            uint l_ButtonIdx;
            bool l_Enabled;

            switch (a_Flag)
            {
                case TabControlFlags.AllowHidingThroughUI:
                    l_ButtonIdx = (uint)ToolBarButtons.HideTab;
                    l_Enabled = (true == a_Value &&
                                 CanHideSelPagePerUI());
                    break;

                case TabControlFlags.AllowClosingThroughUI:
                    l_ButtonIdx = (uint)ToolBarButtons.CloseTab;
                    l_Enabled = (true == a_Value &&
                                 CanCloseSelPagePerUI());
                    break;

                default:
                    l_ButtonIdx = Util.c_NoIndex;
                    l_Enabled = false;
                    break;
            }

            if (Util.c_NoIndex != l_ButtonIdx)
            {
                f_ToolBar.AdjustVisibilityAndEnabledRange(l_ButtonIdx, 1, a_Value, l_Enabled);
            }

            UpdateToolBar();
        }

        private void UpdateToolBar()
        {
            // The one stands for the page listing button.
            uint l_MaxButtons = 1;

            if (true == f_Flags.CheckFlag(FlagsToUInt(TabControlFlags.AllowScrollingThroughToolBar)))
            {
                l_MaxButtons += 2;
            }

            if (true == f_Flags.CheckFlag(FlagsToUInt(TabControlFlags.AllowHidingThroughUI)))
            {
                l_MaxButtons++;
            }

            if (true == f_Flags.CheckFlag(FlagsToUInt(TabControlFlags.AllowClosingThroughUI)))
            {
                l_MaxButtons++;
            }

            Size l_ToolBarSize;
            f_ToolBar.GetPreferredSize(l_MaxButtons, out l_ToolBarSize);
            int l_NewWidth = l_ToolBarSize.Width;

            if (l_NewWidth == f_ReservedToolBarWidth)
            {
                return;
            }

            f_ReservedToolBarWidth = l_NewWidth;

            UpdateRectangles();
            CalculateVisibleTabsAndKeepSelDisplayedIfAlready();
        }

        private int CalcTabWidth(AdvancedTabPage a_Page,
                                 bool a_IncludeImage)
        {
            int l_TabWidth = a_Page.GetCaptionWidth();

            if (true == a_IncludeImage &&
                true == a_Page.p_DisplayImage &&
                null != a_Page.p_Image)
            {
                l_TabWidth += (Lyt.s_TinySpace2 + f_IconSize.Width);
            }

            return Math.Max(l_TabWidth, Lyt.s_TinySpace);
        }

        private int CalcTabBaseWidth(int a_TabWidth,
                                     bool a_Full)
        {
            int l_Width = (f_TabHeight + Lyt.s_TinySpace + a_TabWidth);

            if (false == a_Full)
            {
                l_Width -= f_HiddenWidth;
            }

            return l_Width;
        }

        private bool TabHitTestCore(int a_X,
                                    int a_Y,
                                    int a_TabWidth,
                                    bool a_Full)
        {
            int l_TriangleX = f_TabHeight;

            if (false == a_Full)
            {
                l_TriangleX -= f_HiddenWidth;
            }

            if (l_TriangleX <= a_X)
            {
                // Check if we are on the rectangular shape.

                return ((l_TriangleX + Lyt.s_TinySpace + a_TabWidth) > a_X);
            }

            // Check if we are on the triangular shape.

            return ((l_TriangleX - a_Y) < a_X);
        }

        public AdvancedTabPage TabHitTest(Point a_Pos)
        {
            bool l_Junk;

            return TabHitTest(a_Pos, out l_Junk);
        }

        public AdvancedTabPage TabHitTest(Point a_Pos,
                                          out bool a_RatherLeft)
        {
            // Return can be null. On successful return, a_RatherLeft
            // indicates if the mouse was on the left half of the tab
            // (including any triangle).

            uint l_Count = (uint)f_TabPositions.p_Count;

            if (0 == l_Count ||
                false == f_HeaderRect.Contains(a_Pos))
            {
                goto FAIL;
            }

            int l_X = (a_Pos.X - 1);
            int l_Y = (a_Pos.Y - f_HeaderRect.Y);
            TabPosition[] l_Arr = f_TabPositions.p_Arr;

            // We've got to do this from end to start, due
            // to the way how tabs can overlap each other.

            for (uint i = l_Count; 0 != i--;)
            {
                if (l_X < l_Arr[i].f_X)
                {
                    continue;
                }

                int l_CurX = (l_X - l_Arr[i].f_X);
                int l_TabWidth = l_Arr[i].f_TabWidth;
                bool l_Full = l_Arr[i].f_Full;

                if (true == TabHitTestCore(l_CurX, l_Y, l_TabWidth, l_Full))
                {
                    int l_BaseWidth = CalcTabBaseWidth(l_TabWidth, l_Full);

                    a_RatherLeft = ((l_BaseWidth / 2) >= l_CurX);

                    return l_Arr[i].f_Page;
                }

                if (false == l_Full)
                {
                    break;
                }
            }

            FAIL:

            a_RatherLeft = false;

            return null;
        }

        public bool GetTabRect(uint a_Index,
                               bool a_IncludeTriangle,
                               out Rectangle a_TabRect)
        {
            // Returns zero if the tab with that index isn't visible.
            // On success, the returned rectangle is just an
            // approximation of the real shape of the tab.

            if (true == AreAnyTabsVisible() &&
                a_Index >= f_DisplStartIdx &&
                a_Index <= f_DisplEndIdx)
            {
                TabPosition[] l_Arr = f_TabPositions.p_Arr;
                uint l_ArrIdx = (a_Index - f_DisplStartIdx);

                int l_X = l_Arr[l_ArrIdx].f_X;
                int l_Right = Math.Min(l_Arr[l_ArrIdx].CalcRight(this), f_HeaderRect.Right);

                if (false == a_IncludeTriangle &&
                    true == l_Arr[l_ArrIdx].f_Full)
                {
                    l_X += f_HiddenWidth;
                }

                a_TabRect = new Rectangle(l_X, f_HeaderRect.Y, l_Right - l_X, f_HeaderRect.Height);

                return true;
            }

            a_TabRect = new Rectangle();

            return false;
        }

        private bool SelectViaMouse(Point a_Pos)
        {
            // Returns nonzero if a tab was clicked on and that
            // tab is the selected tab on return.

            AdvancedTabPage l_Page = TabHitTest(a_Pos);

            if (null != l_Page)
            {
                if (l_Page == f_SelPage)
                {
                    return true;
                }

                if (true == p_AllowSelectingThroughUI)
                {
                    bool l_Cancel = false;
                    OnBeforeSelectingThroughUI(l_Page, ref l_Cancel);

                    if (false == l_Cancel)
                    {
                        ChangeSelectedPageCore(l_Page, null == f_SelPage ? SelectionChangeOrigins.None : SelectionChangeOrigins.UserInteraction, SelectionChangeOrigins.UserInteraction);

                        return true;
                    }
                }
            }

            return false;
        }

        protected override void OnMouseClick(MouseEventArgs e)
        {
            // Need to call base class implementation.

            bool l_Handled = false;

            if (MouseButtons.Right == e.Button &&
                null != f_TabCntxtMenuStrip)
            {
                bool l_Selected = SelectViaMouse(e.Location);

                if (true == l_Selected)
                {
                    l_Handled = true;
                    f_TabCntxtMenuStrip.Show(this, e.Location);
                }
            }

            if (false == l_Handled)
            {
                base.OnMouseClick(e);
            }
        }

        private bool IsDragging()
        {
            return (null != f_DragCntxt);
        }

        private void StartDragPage()
        {
            StopAnyDynamicScrolling();

            f_MousedPage = null;
            f_DraggedPage = f_SelPage;
            f_DragCntxt = new DragContext(this, true, GetIndexOfPageCore(f_DraggedPage));
            f_DragCntxt.e_DragOutside += DragCntxt_DragOutside;

            f_DragCntxt.StartDrag();
        }

        private void EndDragPage(bool a_Commit,
                                 SelectionChangeOrigins a_Origin)
        {
            // This routine can change the index of the dragged page.

            if (false == a_Commit)
            {
                uint l_FallbackIndex = Math.Min(f_DragCntxt.f_OriginIndex, p_PageCount - 1);
                DragPageCore(GetIndexOfPageCore(f_DraggedPage), l_FallbackIndex, a_Origin);
            }

            f_DragCntxt.EndDrag();

            AdvancedTabPage l_Page = f_DraggedPage;
            bool l_DidDrag = f_DragCntxt.DidDrag();

            f_DraggedPage = null;
            f_DragCntxt = null;

            InvalidateAllTabs();

            // l_DidDrag indicates of an index change actually occurred,
            // regardless of this change was temporary or permanent.
            // a_Commit indicates if possible changes were allowed to be
            // taken over (not if changes actually happened).
            OnAfterDraggingTab(l_Page, l_DidDrag, a_Commit);
        }

        private void AbortAnyDragging(SelectionChangeOrigins a_Origin)
        {
            if (true == IsDragging())
            {
                EndDragPage(false, a_Origin);
            }
        }

        private void DragPageCore(uint a_OldIdx,
                                  uint a_NewIdx,
                                  SelectionChangeOrigins a_Origin)
        {
            if (a_OldIdx == a_NewIdx)
            {
                return;
            }

            uint l_EffNewIdx = a_NewIdx;
            OnDraggingTab(f_DraggedPage, ref l_EffNewIdx);

            if (l_EffNewIdx == a_OldIdx)
            {
                return;
            }

            f_DragCntxt.NotifyDragging();
            ChangePageIndex(f_DraggedPage, l_EffNewIdx, a_Origin);
            EnsurePageVisibilityCore(f_DraggedPage, l_EffNewIdx, a_Origin);
        }

        private void DragCntxt_DragOutside(bool a_Near)
        {
            uint l_CurIdx = GetIndexOfPageCore(f_DraggedPage);
            uint l_NewIdx;

            if (false == a_Near)
            {
                if (l_CurIdx > f_DisplEndIdx)
                {
                    return;
                }

                l_NewIdx = ((f_DisplEndIdx + 1) >= p_PageCount ? f_DisplEndIdx : (f_DisplEndIdx + 1));
            }
            else
            {
                if (l_CurIdx < f_DisplStartIdx)
                {
                    return;
                }

                l_NewIdx = (0 == f_DisplStartIdx ? 0 : (f_DisplStartIdx - 1));
            }

            if (l_CurIdx != l_NewIdx)
            {
                DragPageCore(l_CurIdx, l_NewIdx, SelectionChangeOrigins.UserInteraction);
            }
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            if (false == f_ToolBar.TryRegisterKey(e))
            {
                // The tool bar didn't want it.

                if (Keys.Escape == e.KeyCode &&
                    true == IsDragging())
                {
                    EndDragPage(false, SelectionChangeOrigins.UserInteraction);
                    e.Handled = true;
                }
            }

            base.OnKeyDown(e);
        }

        protected override void OnMouseDown(MouseEventArgs e)
        {
            // Need to call base class implementation.

            bool l_Handled = false;

            if (false == IsDragging() &&
                MouseButtons.Left == e.Button &&
                true == SelectViaMouse(e.Location) &&
                // There must be at least two displayed tabs currently.
                f_DisplStartIdx != f_DisplEndIdx)
            {
                if (true == p_AllowReorderingThroughUI &&
                    true == f_SelPage.p_AllowReorderingThroughUI)
                {
                    OnBeforeDraggingTab(f_SelPage, ref l_Handled);

                    if (false == l_Handled)
                    {
                        l_Handled = true;
                        StartDragPage();
                        InvalidateAllTabs();
                    }
                }
            }

            if (false == l_Handled)
            {
                base.OnMouseDown(e);
            }
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            // Need to call base class implementation.

            if (true == IsDragging())
            {
                EndDragPage(true, SelectionChangeOrigins.UserInteraction);
            }

            base.OnMouseUp(e);
        }

        protected override void OnMouseDoubleClick(MouseEventArgs e)
        {
            // Need to call base class implementation.

            AdvancedTabPage l_Page = TabHitTest(e.Location);

            if (null != l_Page)
            {
                OnTabDoubleClick(e, l_Page);
            }

            base.OnMouseDoubleClick(e);
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            // Need to call base class implementation.

            OnMouseMoveCore(e);

            base.OnMouseMove(e);
        }

        private void OnMouseMoveCore(MouseEventArgs e)
        {
            Point l_Location = e.Location;
            AdvancedTabPage l_Page;

            if (false == IsDragging())
            {
                l_Page = TabHitTest(l_Location);

                if (l_Page != f_MousedPage)
                {
                    f_MousedPage = l_Page;
                    InvalidateAllTabs();
                }

                return;
            }

            Rectangle l_WeakRect;
            BuildWeakDragRectForScrolling(out l_WeakRect);
            bool l_Outside = f_DragCntxt.SetMousePos(Form.MousePosition, l_Location, ref l_WeakRect, ref f_HeaderRect);

            if (true == l_Outside)
            {
                return;
            }

            // Check if the mouse rests on any tab.
            bool l_RatherLeft;
            l_Page = TabHitTest(l_Location, out l_RatherLeft);

            if (null == l_Page ||
                l_Page == f_DraggedPage)
            {
                return;
            }

            uint l_OldIdx = GetIndexOfPageCore(f_DraggedPage);
            uint l_NewIdx = GetIndexOfPageCore(l_Page);

            if ((l_OldIdx + 1) == l_NewIdx ||
                (l_NewIdx + 1) == l_OldIdx)
            {
                // The page being dragged is next to the page the
                // mouse cursor is on.

                int l_WidthDiff = (CalcTabWidth(l_Page, true) - CalcTabWidth(f_DraggedPage, true));

                if (0 < l_WidthDiff)
                {
                    // The page that is being dragged is more narrow
                    // than the one that it's dragged on.

                    Rectangle l_Rect;

                    if (true == GetTabRect(l_OldIdx, false, out l_Rect))
                    {
                        if (l_NewIdx < l_OldIdx)
                        {
                            // Dragging to the left.

                            if ((l_Rect.X - l_Location.X) <= l_WidthDiff)
                            {
                                return;
                            }
                        }
                        else
                        {
                            // Draging to the right.

                            if ((l_Location.X - l_Rect.Right) <= (l_WidthDiff + Lyt.s_TinySpace))
                            {
                                return;
                            }
                        }
                    }
                }
            }
            else
            {
                if (l_OldIdx > l_NewIdx)
                {
                    if (false == l_RatherLeft &&
                        (l_NewIdx + 1) < p_PageCount)
                    {
                        l_NewIdx++;
                    }
                }
                else
                {
                    if (true == l_RatherLeft &&
                        0 != l_NewIdx)
                    {
                        l_NewIdx--;
                    }
                }
            }

            DragPageCore(l_OldIdx, l_NewIdx, SelectionChangeOrigins.UserInteraction);
        }

        protected override void OnMouseLeave(EventArgs e)
        {
            // Need to call base class implementation.

            if (null != f_MousedPage)
            {
                f_MousedPage = null;
                InvalidateAllTabs();
            }

            base.OnMouseLeave(e);
        }

        private void StopAnyDynamicScrolling()
        {
            if (null != f_DynScrollCntxt)
            {
                f_DynScrollCntxt.Dispose();
                f_DynScrollCntxt = null;
            }
        }

        private void ToolBar_ItemMouseDown(GuiToolBar a_Sender,
                                           uint a_ItemIndex)
        {
            ToolBarButtons l_Button = (ToolBarButtons)a_ItemIndex;

            if (ToolBarButtons.HideTab <= l_Button)
            {
                return;
            }

            bool l_ScrollNear = (0 == a_ItemIndex);

            if (null == f_DynScrollCntxt)
            {
                AbortAnyDragging(SelectionChangeOrigins.UserInteraction);
            }
            else
            {
                if (f_DynScrollCntxt.f_ScrollNear == l_ScrollNear)
                {
                    return;
                }

                StopAnyDynamicScrolling();
            }

            f_DynScrollCntxt = new DynamicScrollContext(this, l_ScrollNear);
        }

        private void ToolBar_ItemMouseUp(GuiToolBar a_Sender,
                                         uint a_ItemIndex)
        {
            ToolBarButtons l_Button = (ToolBarButtons)a_ItemIndex;

            if (ToolBarButtons.HideTab <= l_Button)
            {
                return;
            }

            StopAnyDynamicScrolling();
        }

        private void ToolBar_ItemMouseClicked(GuiToolBar a_Sender,
                                              uint a_ItemIndex)
        {
            ToolBarButtons l_Button = (ToolBarButtons)a_ItemIndex;

            if (ToolBarButtons.HideTab > l_Button)
            {
                return;
            }

            switch (l_Button)
            {
                case ToolBarButtons.HideTab:
                    if (true == CanHideSelPagePerUI())
                    {
                        bool l_Handled = false;
                        OnHidingSelPageThroughUI(ref l_Handled);

                        if (false == l_Handled)
                        {
                            AdjustPageVisibility(f_SelPage, false, SelectionChangeOrigins.UserInteraction);
                        }
                    }
                    break;

                case ToolBarButtons.CloseTab:
                    if (true == CanCloseSelPagePerUI())
                    {
                        bool l_Handled = false;
                        bool l_DisposePage = true;
                        OnClosingSelPageThroughUI(ref l_Handled, ref l_DisposePage);

                        if (false == l_Handled)
                        {
                            RemovePage(f_SelPage, l_DisposePage, SelectionChangeOrigins.UserInteraction);
                        }
                    }
                    break;

                case ToolBarButtons.ListAllPages:
                    ListAllPages();
                    break;

                default: throw (new UnexpectedStateException());
            }
        }

        private void ListAllPages()
        {
            if (null != f_PageListingMenu)
            {
                f_PageListingMenu.Close();
            }

            PageListItem[] l_Items = null;
            OnPrepareListingPages(ref l_Items);

            // l_Items can have more items than there are pages
            // now.

            uint l_ItemCount;

            if (null == l_Items)
            {
                // Create an automatic page listing.

                uint l_TotalPageCount = p_PageCount;
                bool l_IncludeHiddenPages = p_IncludeHiddenPagesInAutoPageListing;
                l_ItemCount = (false == l_IncludeHiddenPages ? f_VisiblePages : l_TotalPageCount);

                l_Items = new PageListItem[l_ItemCount];
                AdvancedTabPage[] l_Arr = f_Pages.p_Arr;
                uint k = 0;

                for (uint i = 0; i != l_TotalPageCount; i++)
                {
                    bool l_Visible = l_Arr[i].p_PageVisible;

                    if (false == l_IncludeHiddenPages &&
                        false == l_Visible)
                    {
                        continue;
                    }

                    ListItemState l_ItemState = (false == l_Visible ? ListItemState.HiddenPage : (l_Arr[i] == f_SelPage ? ListItemState.SelectedPage : ListItemState.VisiblePage));
                    l_Items[k++] = new PageListItem(i, l_ItemState, l_Arr[i].p_Image, l_Arr[i].p_Caption);
                }
            }
            else
            {
                l_ItemCount = (uint)l_Items.Length;
            }

            if (0 == l_ItemCount)
            {
                // The list is not needed.

                return;
            }

            if (null == f_PageListingMenu)
            {
                // Initial menu use, so create.

                f_PageListingMenu = new ContextMenuStrip();
                f_PageListingMenu.ImageScalingSize = Lyt.s_SmallIconSize;
                f_PageListingMenu.Closed += PageListingMenu_Closed;
            }
            else
            {
                MainForm.DisposeMenuItems(f_PageListingMenu);
            }

            for (uint i = 0; i != l_ItemCount; i++)
            {
                ListItemState l_ItemState = l_Items[i].f_ItemState;
                ToolStripMenuItem l_MenuItem = l_Items[i].CreateMenuItem(f_PageListingMenu, PageListingMenu_ItemClicked);

                // Apply the appropriate font.

                if (ListItemState.HiddenPage == l_ItemState)
                {
                    ApplyPageListingMenuFont(l_MenuItem, FontStyle.Italic, ref f_PageListingItalicFont);
                }
                else if (ListItemState.SelectedPage == l_ItemState)
                {
                    ApplyPageListingMenuFont(l_MenuItem, FontStyle.Bold, ref f_PageListingBoldFont);
                }
            }

            f_Flags.ClearFlag(FlagsToUInt(TabControlFlags.PageListingContextMenuObsolete));

            Point l_Pos = f_ToolBar.PointToClient(Form.MousePosition);
            l_Pos.Y = f_ToolBar.Height;
            f_PageListingMenu.Show(f_ToolBar, l_Pos);
        }

        private void ApplyPageListingMenuFont(ToolStripMenuItem a_MenuItem,
                                              FontStyle a_Style,
                                              ref Font a_Font)
        {
            // If a_Font is zero on entry, a new font will be created
            // and referenced.

            if (null == a_Font)
            {
                a_Font = MainForm.RecreateFont(f_PageListingMenu.Font, 0f, a_Style);
            }

            a_MenuItem.Font = a_Font;
        }

        private void DisposePageListingFonts()
        {
            Util.DisposeAndZero(ref f_PageListingBoldFont);
            Util.DisposeAndZero(ref f_PageListingItalicFont);
        }

        private void PageListingMenu_Closed(object a_Sender,
                                            ToolStripDropDownClosedEventArgs e)
        {
            // Release some memory by disposing the extra fonts.
            // Before doing this, make sure they are not referenced
            // anymore.

            Font l_FallbackFont = f_PageListingMenu.Font;

            foreach (ToolStripMenuItem l_MenuItem in f_PageListingMenu.Items)
            {
                l_MenuItem.Font = l_FallbackFont;
            }

            DisposePageListingFonts();
        }

        private void PageListingMenu_ItemClicked(object a_Sender,
                                                 EventArgs e)
        {
            if (true == f_Flags.CheckFlag(FlagsToUInt(TabControlFlags.PageListingContextMenuObsolete)))
            {
                // Ignore this click.

                return;
            }

            // The index can be larger than the index of the last page,
            // since custom items can have higher indices.

            uint l_PageIndex = ((GenericMenuItem<uint>)a_Sender).f_GenericTag;
            bool l_Handled = false;
            OnItemSelectedFromPageList(l_PageIndex, ref l_Handled);

            if (false == l_Handled)
            {
                if (p_PageCount > l_PageIndex)
                {
                    SetSelectedPageIndex(l_PageIndex, SelectionChangeOrigins.UserInteraction);
                }
            }
        }

        private void InvalidatePageListingContextMenu()
        {
            f_Flags.SetFlag(FlagsToUInt(TabControlFlags.PageListingContextMenuObsolete));
        }

        protected virtual void OnPrepareListingPages(ref PageListItem[] a_Items)
        {
            // a_Items is zero on entry and may remain so, otherwise
            // the callee overrides the default items.

            if (null != e_PrepareListingPages)
            {
                e_PrepareListingPages(this, ref a_Items);
            }
        }

        protected virtual void OnItemSelectedFromPageList(uint a_Index,
                                                          ref bool a_Handled)
        {
            // a_Index can be equal to or larger than the page count
            // if custom items were listed.

            if (null != e_ItemSelectedFromPageList)
            {
                e_ItemSelectedFromPageList(this, a_Index, ref a_Handled);
            }
        }

        private void SetHeaderRectPlusBase(out Rectangle a_Rect)
        {
            a_Rect = f_HeaderRect;
            a_Rect.Height += c_TabBaseHeight;
        }

        public void UpdateImagesOfAllPages()
        {
            if (false == p_DisplayImages ||
                false == AreAnyTabsVisible())
            {
                return;
            }

            AdvancedTabPage[] l_Arr = f_Pages.p_Arr;

            for (uint i = f_DisplStartIdx, lim = (f_DisplEndIdx + 1); lim != i; i++)
            {
                AdvancedTabPage l_Page = l_Arr[i];

                if (null != l_Page.p_Image &&
                    true == l_Page.p_PageVisible &&
                    true == l_Page.p_DisplayImage)
                {
                    // There is at least one tab that really has
                    // an image.

                    InvalidateAllTabs();

                    return;
                }
            }
        }

        private void InvalidateAllTabs()
        {
            Rectangle l_InvalidRect;
            SetHeaderRectPlusBase(out l_InvalidRect);
            this.Invalidate(l_InvalidRect);
        }

        private void ResetDisplayFields()
        {
            if (0 >= f_TabPositions.p_Count)
            {
                return;
            }

            f_DisplEndIdx = f_DisplStartIdx = Util.c_NoIndex;
            f_DisplOverflow = false;
            f_TabPositions.Clear();

            InvalidateAllTabs();
        }

        private void CalculateVisibleTabsAndKeepSelDisplayedIfAlready()
        {
            CalculateVisibleTabs(true, true);
        }

        private void CalculateVisibleTabs(bool a_EnsureSelPageIsDisplayed,
                                          bool a_EnsureOnlyIfAlreadySo)
        {
            // f_DisplStartIdx and related may already be invalid
            // while this routine gets called. a_EnsureOnlyIfAlreadySo
            // is ignored if a_EnsureSelPageIsDisplayed is zero.

            uint l_Count = p_PageCount;

            if (0 == l_Count)
            {
                SetVisibleTabs(Util.c_NoIndex, false);

                return;
            }

            uint l_Idx = Math.Min(f_DisplStartIdx, l_Count - 1);

            if (true == a_EnsureSelPageIsDisplayed &&
                null != f_SelPage)
            {
                uint l_SelIdx = p_SelectedPageIndex;

                if (false == a_EnsureOnlyIfAlreadySo ||
                    true == IsPageInDisplayRange(l_SelIdx))
                {
                    // Try to preserve the old display adjustment.

                    SetVisibleTabs(l_Idx, false);

                    if (true == IsPageInDisplayRange(l_SelIdx))
                    {
                        // Worked.

                        return;
                    }

                    l_Idx = l_SelIdx;
                }
            }

            bool l_FromFar = (true == AreAnyTabsVisible() &&
                              l_Idx > f_DisplEndIdx);
            SetVisibleTabs(l_Idx, l_FromFar);
        }

        private void SetVisibleTabs(uint a_MustBeIncludedPageIndex,
                                    bool a_FromFar)
        {
            // a_MustBeIncludedPageIndex may be invalid if there
            // are no pages. If it's valid, the page hinted by
            // a_MustBeIncludedPageIndex may be hidden and will
            // not be included in that case.

            ResetDisplayFields();

            if (0 == f_VisiblePages)
            {
                DeactivateToolBar();

                return;
            }

            uint l_Count = p_PageCount;
            int l_Right = f_HeaderRect.X;
            int l_AvailWidth = (f_HeaderRect.Width - f_HiddenWidth);
            AdvancedTabPage[] l_PagesArr = f_Pages.p_Arr;

            uint l_DisplCount = 0;
            TabPosition[] l_Arr = f_TabPositions.p_Arr;

            // The following two indicate that there are no more
            // visible pages left on the corresponding end.
            bool l_NearEndReached = false;
            bool l_FarEndReached = false;
            bool l_UseImages = p_DisplayImages;
            bool l_Turned = false;
            bool l_GoNear = a_FromFar;
            uint l_NearIdx = a_MustBeIncludedPageIndex;
            uint l_FarIdx = a_MustBeIncludedPageIndex;

            for (uint l_TryPageIdx = a_MustBeIncludedPageIndex, lim = (l_Count - 1); true; )
            {
                AdvancedTabPage l_Page = l_PagesArr[l_TryPageIdx];

                if (true == l_Page.p_PageVisible)
                {
                    int l_TabWidth = CalcTabWidth(l_Page, l_UseImages);
                    bool l_Selected = (l_Page == f_SelPage);
                    int l_ExpectedBaseWidth = this.CalcTabBaseWidth(l_TabWidth, false);
                    bool l_JustTurned = false;

                    if (0 != l_DisplCount &&
                        l_ExpectedBaseWidth > l_AvailWidth)
                    {
                        if (true == l_Turned)
                        {
                            break;
                        }

                        // We clearly haven't reached the end, but
                        // there is no more pixel space left to include
                        // another (whole) page on this end.

                        l_Turned = true;
                        l_GoNear = !l_GoNear;
                        l_TryPageIdx = (false == l_GoNear ? l_FarIdx : l_NearIdx);
                        l_JustTurned = true;
                    }

                    if (false == l_JustTurned)
                    {
                        bool l_UpdateStartIdx;
                        bool l_UpdateEndIdx;
                        uint l_TrgtIdx;
                        int l_X;

                        if (false == l_GoNear)
                        {
                            l_UpdateStartIdx = (Util.c_NoIndex == f_DisplStartIdx);
                            l_UpdateEndIdx = true;

                            l_TrgtIdx = (uint)f_TabPositions.ReserveAtEnd();
                            l_X = l_Right;
                        }
                        else
                        {
                            l_UpdateStartIdx = true;
                            l_UpdateEndIdx = (Util.c_NoIndex == f_DisplEndIdx);

                            // Reserve pixel space for the new item first.
                            TabPosition.Shift(l_Arr, 0, l_DisplCount, l_ExpectedBaseWidth, false);

                            f_TabPositions.ReserveAt(0);
                            l_TrgtIdx = 0;
                            l_X = f_HeaderRect.X;
                        }

                        if (true == l_UpdateStartIdx)
                        {
                            f_DisplStartIdx = l_TryPageIdx;
                        }

                        if (true == l_UpdateEndIdx)
                        {
                            f_DisplEndIdx = l_TryPageIdx;
                        }

                        l_DisplCount++;
                        l_Arr = f_TabPositions.p_Arr;
                        l_Arr[l_TrgtIdx] = new TabPosition(l_Page, l_TabWidth, l_X, l_Selected);

                        l_Right += l_ExpectedBaseWidth;
                        l_AvailWidth -= l_ExpectedBaseWidth;
                    }
                }

                if (false == l_GoNear)
                {
                    if (lim == l_FarIdx)
                    {
                        l_FarEndReached = true;

                        if (true == l_Turned ||
                            0 == l_NearIdx)
                        {
                            break;
                        }

                        l_Turned = true;
                        l_GoNear = true;
                    }
                }
                else if (0 == l_NearIdx)
                {
                    l_NearEndReached = true;

                    if (true == l_Turned ||
                        lim == l_FarIdx)
                    {
                        break;
                    }

                    l_Turned = true;
                    l_GoNear = false;
                }

                if (false == l_GoNear)
                {
                    l_TryPageIdx = ++l_FarIdx;
                }
                else
                {
                    l_TryPageIdx = --l_NearIdx;
                }
            }

            if (0 == l_DisplCount)
            {
                // Should never happen.

                DeactivateToolBar();

                return;
            }

            if (l_DisplCount < f_VisiblePages)
            {
                f_DisplOverflow = true;
            }

            // Turn the first item into a full tab.
            TabPosition.Shift(l_Arr, 1, l_DisplCount - 1, f_HiddenWidth, true);
            l_Arr[0].f_Full = true;

            if (false == f_DisplOverflow ||
                false == p_AllowScrollingThroughButtons)
            {
                f_ToolBar.AdjustVisibilityRange((uint)ToolBarButtons.ScrollLeft, 2, false);
            }
            else
            {
                uint l_Junk;

                if (false == l_NearEndReached)
                {
                    l_NearEndReached = (0 == f_DisplStartIdx ||
                                        false == FindIndexOfVisiblePage(f_DisplStartIdx - 1, true, out l_Junk));
                }

                if (false == l_FarEndReached)
                {
                    l_FarEndReached = ((l_Count - 1) == f_DisplEndIdx ||
                                       false == FindIndexOfVisiblePage(f_DisplEndIdx + 1, false, out l_Junk));
                }

                if (l_NearEndReached == l_FarEndReached)
                {
                    f_ToolBar.AdjustEnabledRange((uint)ToolBarButtons.ScrollLeft, 2, !l_NearEndReached);
                }
                else
                {
                    f_ToolBar.AdjustEnabled((uint)ToolBarButtons.ScrollLeft, !l_NearEndReached);
                    f_ToolBar.AdjustEnabled((uint)ToolBarButtons.ScrollRight, !l_FarEndReached);
                }

                f_ToolBar.AdjustVisibilityRange((uint)ToolBarButtons.ScrollLeft, 2, true);
            }

            bool l_ShowListingButton = (l_DisplCount != (false == p_IncludeHiddenPagesInAutoPageListing ? f_VisiblePages : l_Count) ||
                                        true == p_AlwaysShowPageListingButton);
            f_ToolBar.AdjustVisibilityAndEnabledRange((uint)ToolBarButtons.ListAllPages, 1, l_ShowListingButton, l_ShowListingButton);

            UpdateToolBarSizeAndPosition();

            if (true == IsDragging())
            {
                Point l_ScreenMousePos = Form.MousePosition;
                Point l_CntrlMousePos = this.PointToClient(l_ScreenMousePos);
                Rectangle l_WeakRect;
                BuildWeakDragRectForScrolling(out l_WeakRect);
                bool l_Outside = f_DragCntxt.SetMousePos(l_ScreenMousePos, l_CntrlMousePos, ref l_WeakRect, ref f_HeaderRect);
            }
        }

        private void DeactivateToolBar()
        {
            // This is a helper routine for SetVisibleTabs().

            f_ToolBar.AdjustVisibilityRange((uint)ToolBarButtons.ScrollLeft, 2, false);
            f_ToolBar.AdjustVisibilityAndEnabledRange((uint)ToolBarButtons.ListAllPages, 1, p_AlwaysShowPageListingButton, false);

            UpdateToolBarSizeAndPosition();
        }

        private void UpdateToolBarSizeAndPosition()
        {
            // This is a helper routine for SetVisibleTabs().

            Size l_ToolBarSize;
            f_ToolBar.GetPreferredSize(out l_ToolBarSize);
            f_ToolBar.Location = new Point(f_FrameRect.Right - l_ToolBarSize.Width, f_HeaderRect.Top);
            f_ToolBar.Size = l_ToolBarSize;
        }

        public void ScrollNear(uint a_Count)
        {
            // Scroll left.

            uint l_DisplIdx;

            if (0 != a_Count &&
                true == AreAnyTabsVisible() &&
                0 != f_DisplStartIdx &&
                true == ItemIndexToDisplayIndex(f_DisplStartIdx, out l_DisplIdx) &&
                0 != l_DisplIdx)
            {
                l_DisplIdx = (a_Count >= l_DisplIdx ? 0 : (l_DisplIdx - a_Count));
                uint l_ItemIdx;

                if (true == DisplayIndexToItemIndex(l_DisplIdx, out l_ItemIdx))
                {
                    SetVisibleTabs(l_ItemIdx, false);
                }
            }
        }

        public void ScrollFar(uint a_Count)
        {
            // Scroll right.

            uint l_DisplIdx;

            if (0 != a_Count &&
                true == AreAnyTabsVisible() &&
                (p_PageCount - 1) != f_DisplEndIdx &&
                true == ItemIndexToDisplayIndex(f_DisplEndIdx, out l_DisplIdx))
            {
                uint l_MaxDisplIdx = (f_VisiblePages - 1);

                if (l_MaxDisplIdx != l_DisplIdx)
                {
                    l_DisplIdx += a_Count;

                    if (l_DisplIdx < a_Count ||
                        l_DisplIdx > l_MaxDisplIdx)
                    {
                        // Overflow or out of bounds.

                        l_DisplIdx = l_MaxDisplIdx;
                    }

                    uint l_ItemIdx;

                    if (true == DisplayIndexToItemIndex(l_DisplIdx, out l_ItemIdx))
                    {
                        SetVisibleTabs(l_ItemIdx, true);
                    }
                }
            }
        }

        public void EnsurePageVisibility(AdvancedTabPage a_Page,
                                         SelectionChangeOrigins a_Origin)
        {
            EnsurePageVisibilityCore(a_Page, GetIndexOfPageCore(a_Page), a_Origin);
        }

        private void EnsurePageVisibility(uint a_Index,
                                          SelectionChangeOrigins a_Origin)
        {
            EnsurePageVisibilityCore(GetPageByIndex(a_Index), a_Index, a_Origin);
        }

        private void EnsurePageVisibilityCore(AdvancedTabPage a_Page,
                                              uint a_Index,
                                              SelectionChangeOrigins a_Origin)
        {
            // This routine also unhides a_Page if it's not visible
            // yet.

            if (false == a_Page.p_PageVisible)
            {
                MakePageVisibleWithoutDefaultHandler(a_Page, a_Origin);
            }

            if (a_Index < f_DisplStartIdx)
            {
                SetVisibleTabs(a_Index, false);
            }

            if (a_Index > f_DisplEndIdx)
            {
                SetVisibleTabs(a_Index, true);
            }
        }

        private bool AreAnyTabsVisible()
        {
            return (Util.c_NoIndex != f_DisplStartIdx);
        }

        public bool IsPageInDisplayRange(AdvancedTabPage a_Page)
        {
            return IsPageInDisplayRange(GetIndexOfPageCore(a_Page));
        }

        public bool IsPageInDisplayRange(uint a_Index)
        {
            // This routine ignores the visibility state
            // (AdvancedTabPage.p_PageVisible) of the page.

            return (true == AreAnyTabsVisible() &&
                    f_DisplStartIdx <= a_Index &&
                    f_DisplEndIdx >= a_Index);
        }

        private int DrawTabCore(Graphics a_G,
                                Point a_DestPos,
                                int a_TabWidth,
                                bool a_Full,
                                Brush a_Brush)
        {
            int l_X0 = a_DestPos.X;

            if (false == a_Full)
            {
                l_X0 -= f_HiddenWidth;
            }

            int l_X3 = (l_X0 + (f_TabHeight + Lyt.s_TinySpace));
            int l_X2 = (l_X3 - Lyt.s_TinySpace2);
            int l_X5 = (l_X3 + a_TabWidth);
            int l_X4 = (l_X5 - Lyt.s_TinySpace);
            int l_X1;

            int l_Y0 = f_HeaderRect.Bottom;
            int l_Y3 = a_DestPos.Y;
            int l_Y2 = (l_Y3 + Lyt.s_TinySpace);
            int l_Y1;

            if (false == a_Full)
            {
                l_X1 = (l_X0 + f_HiddenWidth);
                l_X0 = l_X1;
                l_Y1 = (l_Y0 - f_HiddenWidth);
            }
            else
            {
                l_X1 = l_X0;
                l_Y1 = l_Y0;
            }

            Point[] l_Points = new Point[7] { new Point(l_X0, l_Y0),
                                              new Point(l_X1, l_Y1),
                                              new Point(l_X2, l_Y2),
                                              new Point(l_X3, l_Y3),
                                              new Point(l_X4, l_Y3),
                                              new Point(l_X5, l_Y2),
                                              new Point(l_X5, l_Y0) };

            a_G.DrawPolygon(s_OuterPen, l_Points);

            l_Points[0].X++;
            l_Points[1].X++;
            l_Points[2].Y++;
            l_Points[3].Y++;
            l_Points[4].Y++;
            l_Points[4].X++;

            a_G.FillPolygon(a_Brush, l_Points);

            int l_BaseWidth = (l_X5 - l_X0);

            return l_BaseWidth;
        }

        private void DrawTab(Graphics a_G,
                             Point a_DestPos,
                             int a_TabWidth,
                             string a_Caption,
                             Image a_Img,
                             bool a_AlignImgRight,
                             bool a_Full,
                             PageActionState a_PageState)
        {
            // a_Img may be null.

            // Draw the tab.

            Brush l_Br = s_Brushes.GetBrush(a_PageState);
            int l_BaseWidth = DrawTabCore(a_G, a_DestPos, a_TabWidth, a_Full, l_Br);

            // Connect the frame directly with this tab if it
            // is selected.

            if (PageActionState.Selected == a_PageState ||
                PageActionState.Dragged == a_PageState)
            {
                Rectangle l_BaseRect = new Rectangle(a_DestPos.X + 1, f_HeaderRect.Bottom, l_BaseWidth - 1, c_TabBaseHeight);
                a_G.FillRectangle(s_FrameBrush, l_BaseRect);
            }

            // Draw the caption and the image (if any).

            Point l_Pos = new Point(a_DestPos.X + f_TabHeight, a_DestPos.Y);

            if (false == a_Full)
            {
                l_Pos.X -= f_HiddenWidth;
            }

            int l_TextRight = (a_DestPos.X + l_BaseWidth);

            if (null != a_Img)
            {
                if (false == a_AlignImgRight)
                {
                    l_Pos.X += Lyt.s_TinySpace;
                    a_G.DrawImage(a_Img, new Rectangle(new Point(l_Pos.X, a_DestPos.Y + f_ImageY), f_IconSize), new Rectangle(Point.Empty, a_Img.Size), GraphicsUnit.Pixel);
                    l_Pos.X += (Lyt.s_TinySpace + f_IconSize.Width);
                }
                else
                {
                    l_TextRight -= (Lyt.s_TinySpace2 + f_IconSize.Width);
                }
            }

            Rectangle l_CaptionRect = new Rectangle(l_Pos.X, l_Pos.Y + f_TextY, l_TextRight - l_Pos.X, f_TextHeight);
            a_G.DrawString(a_Caption, this.Font, Brushes.Black, l_CaptionRect);

            if (null != a_Img &&
                true == a_AlignImgRight)
            {
                a_G.DrawImage(a_Img, new Rectangle(l_TextRight + Lyt.s_TinySpace, a_DestPos.Y + f_ImageY, f_IconSize.Width, f_IconSize.Height), new Rectangle(Point.Empty, a_Img.Size), GraphicsUnit.Pixel);
            }
        }

        private void DrawFrame(Graphics a_G)
        {
            Rectangle l_Rect = f_FrameRect;
            l_Rect.Width--;
            l_Rect.Height--;

            int l_Right = l_Rect.Right;
            int l_Bottom = l_Rect.Bottom;
            Point[] l_Path = new Point[8] { new Point(l_Rect.X + 2, l_Rect.Y),
                                            new Point(l_Right - 2, l_Rect.Y),
                                            new Point(l_Right, l_Rect.Y + 2),
                                            new Point(l_Right, l_Bottom - 2),
                                            new Point(l_Right - 2, l_Bottom),
                                            new Point(l_Rect.X + 2, l_Bottom),
                                            new Point(l_Rect.X, l_Bottom - 2),
                                            new Point(l_Rect.X, l_Rect.Y + 2) };

            // Draw light frame (before outer frame).
            l_Rect.Inflate(-1, -1);
            a_G.DrawRectangle(s_LightFramePen, l_Rect);

            // Draw outer frame.
            a_G.DrawPolygon(s_OuterPen, l_Path);
            int l_FrameCore = p_FrameCore;

            // Draw frame core (variable width).
            l_Rect.Inflate(-1, -1);
            {
                int l_CoreWidth = (l_Rect.Width + 1);
                int l_CoreHeight = (l_Rect.Height - (2 * l_FrameCore) + 1);
                int l_CoreY1 = (l_Rect.Y + l_FrameCore);
                a_G.FillRectangle(s_FrameBrush, l_Rect.X, l_Rect.Y, l_CoreWidth, l_FrameCore);
                a_G.FillRectangle(s_FrameBrush, l_Rect.X, l_Rect.Bottom - l_FrameCore + 1, l_CoreWidth, l_FrameCore);
                a_G.FillRectangle(s_FrameBrush, l_Rect.X, l_CoreY1, l_FrameCore, l_CoreHeight);
                a_G.FillRectangle(s_FrameBrush, l_Rect.Right - l_FrameCore + 1, l_CoreY1, l_FrameCore, l_CoreHeight);
            }

            // Draw inner frame.
            l_Rect.Inflate(-l_FrameCore, -l_FrameCore);
            a_G.DrawRectangle(s_OuterPen, l_Rect);
        }

        protected override void OnResize(EventArgs e)
        {
            // Need to call base class implementation.

            UpdateRectangles();
            CalculateVisibleTabsAndKeepSelDisplayedIfAlready();

            if (null != f_SelPage)
            {
                f_SelPage.Size = f_InnerRect.Size;
            }

            base.OnResize(e);
        }

        protected override void OnPaddingChanged(EventArgs e)
        {
            // Need to call base class implementation.

            UpdateRectangles();
            CalculateVisibleTabsAndKeepSelDisplayedIfAlready();

            base.OnPaddingChanged(e);
        }

        public IEnumerator<AdvancedTabPage> GetEnumerator()
        {
            return GetEnumerator(false);
        }

        public IEnumerator<AdvancedTabPage> GetEnumerator(bool a_VisibleOnly)
        {
            return GetEnumerator(a_VisibleOnly, 0, p_PageCount);
        }

        public IEnumerator<AdvancedTabPage> GetEnumerator(bool a_VisibleOnly,
                                                          uint a_StartingItemIdx,
                                                          uint a_Len)
        {
            // This routine returns a value type.

            Enum l_Enum = new Enum(a_VisibleOnly, a_StartingItemIdx, a_Len, f_Pages.p_Arr);

            return l_Enum;
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            int l_MinFrameSpace = ((2 * p_FrameSize) + 1);

            if (l_MinFrameSpace <= f_FrameRect.Width &&
                l_MinFrameSpace <= f_FrameRect.Height)
            {
                Graphics l_G = e.Graphics;
                System.Drawing.Text.TextRenderingHint l_StoredHint = l_G.TextRenderingHint;
                {
                    l_G.TextRenderingHint = System.Drawing.Text.TextRenderingHint.ClearTypeGridFit;
                    PaintCore(l_G);
                }
                l_G.TextRenderingHint = l_StoredHint;
            }

            base.OnPaint(e);
        }

        private PageActionState DeterminePageState(AdvancedTabPage a_Page)
        {
            if (a_Page == f_SelPage)
            {
                if (a_Page == f_DraggedPage)
                {
                    return PageActionState.Dragged;
                }

                return PageActionState.Selected;
            }

            if (a_Page == f_MousedPage)
            {
                return PageActionState.Moused;
            }

            return PageActionState.None;
        }

        private void PaintCore(Graphics a_G)
        {
            DrawFrame(a_G);

            uint l_Count = (uint)f_TabPositions.p_Count;

            if (0 == l_Count ||
                false == MainForm.HasPositiveArea(ref f_HeaderRect))
            {
                return;
            }

            s_Brushes.SetParameters(this.Padding.Top, f_TabHeight);

            bool l_NeedClip = (true == f_DisplOverflow &&
                               1 == l_Count);
            RectangleF l_StoredClip = a_G.ClipBounds;

            if (true == l_NeedClip)
            {
                Rectangle l_ClipRect;
                SetHeaderRectPlusBase(out l_ClipRect);
                a_G.SetClip(l_ClipRect);
            }

            TabPosition[] l_Arr = f_TabPositions.p_Arr;
            bool l_FromRight = (RightToLeft.Yes == this.RightToLeft);
            bool l_UseImages = p_DisplayImages;

            for (uint i = 0; i != l_Count; i++)
            {
                AdvancedTabPage l_Page = l_Arr[i].f_Page;
                bool l_AllowImg = (true == l_UseImages &&
                                   true == l_Page.p_DisplayImage);
                bool l_AlignImgRight = (l_FromRight != l_Page.p_DisplayImageOnFarSide);
                DrawTab(a_G, new Point(l_Arr[i].f_X, f_HeaderRect.Y), l_Arr[i].f_TabWidth, l_Page.p_Caption, false == l_AllowImg ? null : l_Page.p_Image, l_AlignImgRight, l_Arr[i].f_Full, DeterminePageState(l_Page));
            }

            if (true == l_NeedClip)
            {
                a_G.SetClip(l_StoredClip);
            }
        }

        public void NotifyTabParamsChanged(AdvancedTabPage a_Page,
                                           bool a_VisibilityChanged,
                                           SelectionChangeOrigins a_Origin)
        {
            // This routine may only get called by the AdvancedTabPage
            // class.

            HandleChangeOfTabParams(a_Page, a_VisibilityChanged, a_Origin, true);
        }

        private void HandleChangeOfTabParams(AdvancedTabPage a_Page,
                                             bool a_VisibilityChanged,
                                             SelectionChangeOrigins a_Origin,
                                             bool a_KeepSelPageDisplayedIfAlreadySo)
        {
            if (false == a_VisibilityChanged)
            {
                if (false == a_Page.p_PageVisible ||
                    false == IsPageInDisplayRange(a_Page))
                {
                    return;
                }
            }
            else
            {
                InvalidatePageListingContextMenu();

                if (false == a_Page.p_PageVisible)
                {
                    // Page gets hidden.

                    f_VisiblePages--;
                    DeselectPageIfSelected(a_Page);

                    if (f_MousedPage == f_SelPage)
                    {
                        f_MousedPage = null;
                    }

                    if (false == IsPageInDisplayRange(a_Page))
                    {
                        return;
                    }
                }
                else
                {
                    // Page gets visible.

                    f_VisiblePages++;
                    uint l_Idx = GetIndexOfPageCore(a_Page);

                    if (true == f_DisplOverflow &&
                        l_Idx > f_DisplEndIdx)
                    {
                        return;
                    }
                }
            }

            CalculateVisibleTabs(a_KeepSelPageDisplayedIfAlreadySo, true);

            if (true == a_VisibilityChanged)
            {
                OnPageVisibilityChanged(a_Page, a_Origin);
            }
        }

        private void DeselectPageIfSelected(AdvancedTabPage a_Page)
        {
            if (a_Page == f_SelPage)
            {
                AdvancedTabPage l_NewPage = FindNextSelectablePage(a_Page);
                ChangeSelectedPageCore(l_NewPage, SelectionChangeOrigins.ItemRemoval, null == l_NewPage ? SelectionChangeOrigins.None : SelectionChangeOrigins.AutoSelect);
            }
        }

        private void InsertPageCore(AdvancedTabPage a_Page,
                                    uint a_NewIndex)
        {
            a_Page.Dock = DockStyle.None;
            a_Page.Size = new Size();
            a_Page.Visible = false;

            base.Controls.Add(a_Page);

            f_Pages.ReserveAt((int)a_NewIndex);
            f_Pages.p_Arr[a_NewIndex] = a_Page;

            if (true == a_Page.p_PageVisible)
            {
                f_VisiblePages++;
            }
        }

        public AdvancedTabPage AddPage(string a_Caption,
                                       IconTypes a_IconType)
        {
            AdvancedTabPage l_Page = new AdvancedTabPage(a_Caption, a_IconType);
            AddPage(l_Page);

            return l_Page;
        }

        public void AddPageWithoutSelecting(AdvancedTabPage a_Page,
                                            IconTypes a_IconType)
        {
            a_Page.SetImage(a_IconType);
            AddPageWithoutSelecting(a_Page);
        }

        public void AddPageWithoutSelecting(AdvancedTabPage a_Page)
        {
            InsertPage(a_Page, p_PageCount, false);
        }

        public void AddPage(AdvancedTabPage a_Page)
        {
            InsertPage(a_Page, p_PageCount, true);
        }

        public void AddPageRange(IEnumerable<AdvancedTabPage> a_Coll)
        {
            uint l_Idx = p_PageCount;

            foreach (AdvancedTabPage l_Page in a_Coll)
            {
                InsertPage(l_Page, l_Idx++, true);
            }
        }

        public void InsertPage(AdvancedTabPage a_Page,
                               uint a_Index,
                               bool a_AllowAutoSelect)
        {
            InvalidatePageListingContextMenu();

            InsertPageCore(a_Page, a_Index);

            if (false == f_DisplOverflow ||
                a_Index <= f_DisplEndIdx)
            {
                CalculateVisibleTabsAndKeepSelDisplayedIfAlready();
            }

            a_Page.OnPageAddedOrRemoved(true, SelectionChangeOrigins.Programmatic);
            OnPageAddedOrRemoved(a_Page, true, SelectionChangeOrigins.Programmatic);

            if (true == a_AllowAutoSelect &&
                null == f_SelPage &&
                true == a_Page.p_PageVisible)
            {
                ChangeSelectedPageCore(a_Page, SelectionChangeOrigins.None, SelectionChangeOrigins.AutoSelect);
            }
        }

        public void RemoveAllPages(bool a_DisposePages,
                                   SelectionChangeOrigins a_Origin)
        {
            RemovePageRange(0, p_PageCount, a_DisposePages, a_Origin);
        }

        public void RemovePageRange(uint a_Idx,
                                    uint a_Len,
                                    bool a_DisposePages,
                                    SelectionChangeOrigins a_Origin)
        {
            if (0 == a_Len)
            {
                return;
            }

            InvalidatePageListingContextMenu();

            AdvancedTabPage[] l_PagesArr = f_Pages.p_Arr;
            bool l_All = (p_PageCount == a_Len);

            if (false == l_All)
            {
                // l_SelIdx can be Util.c_NoIndex.
                uint l_SelIdx = p_SelectedPageIndex;

                if (Util.c_NoIndex != l_SelIdx)
                {
                    uint l_FinalIdx = (a_Idx + a_Len);

                    if (l_SelIdx >= a_Idx &&
                        l_SelIdx < l_FinalIdx)
                    {
                        AdvancedTabPage l_NewSelPage = FindNextSelectablePage(a_Idx, a_Len);
                        SetSelectedPage(l_NewSelPage, a_Origin);
                    }
                }
            }
            else
            {
                ClearSelection(a_Origin);
            }

            AdvancedTabPage[] l_RemArr = new AdvancedTabPage[a_Len];
            Array.Copy(l_PagesArr, (int)a_Idx, l_RemArr, 0, (int)a_Len);

            // Remove pages and also clear the moused-page reference
            // if affected (do not use Controls.Clear() even if l_All
            // is nonzero, since it would remove the tool bar).

            foreach (AdvancedTabPage l_Page in l_RemArr)
            {
                base.Controls.Remove(l_Page);

                if (l_Page == f_MousedPage)
                {
                    f_MousedPage = null;
                }

                if (false == l_All &&
                    true == l_Page.p_PageVisible)
                {
                    f_VisiblePages--;
                }
            }

            // Columnated ruins domino.

            if (false == l_All)
            {
                f_Pages.DeleteRange((int)a_Idx, (int)a_Len);
            }
            else
            {
                f_Pages.Clear();
                f_VisiblePages = 0;
            }

            if (true == AreAnyTabsVisible() &&
                a_Idx <= f_DisplEndIdx)
            {
                CalculateVisibleTabsAndKeepSelDisplayedIfAlready();
            }

            foreach (AdvancedTabPage l_Page in l_RemArr)
            {
                l_Page.OnPageAddedOrRemoved(false, a_Origin);
                OnPageAddedOrRemoved(l_Page, false, a_Origin);
            }

            if (true == a_DisposePages)
            {
                Util.DisposeArray(l_RemArr);
            }
        }

        public void RemovePageAt(uint a_Index,
                                 bool a_DisposePage,
                                 SelectionChangeOrigins a_Origin)
        {
            RemovePageRange(a_Index, 1, a_DisposePage, a_Origin);
        }

        public void RemovePage(AdvancedTabPage a_Page,
                               bool a_DisposePage,
                               SelectionChangeOrigins a_Origin)
        {
            RemovePageRange(GetIndexOfPageCore(a_Page), 1, a_DisposePage, a_Origin);
        }

        private bool CanHideSelPagePerUI()
        {
            return (null != f_SelPage &&
                    true == this.p_AllowHidingThroughUI &&
                    true == f_SelPage.p_AllowHidingThroughUI);
        }

        private bool CanCloseSelPagePerUI()
        {
            return (null != f_SelPage &&
                    true == this.p_AllowClosingThroughUI &&
                    true == f_SelPage.p_AllowClosingThroughUI);
        }

        public AdvancedTabPage GetPageByIndex(uint a_Index)
        {
            return (AdvancedTabPage)f_Pages.p_Arr[a_Index];
        }

        private uint GetIndexOfPageCore(AdvancedTabPage a_Page)
        {
            // a_Page must belong to this tab control.

            uint l_Idx;

            if (false == GetIndexOfPage(a_Page, out l_Idx))
            {
                throw (new ArgumentException());
            }

            return l_Idx;
        }

        public bool GetIndexOfPage(AdvancedTabPage a_Page,
                                   out uint a_Index)
        {
            // Returns the (item) index of a page.

            AdvancedTabPage[] l_Arr = f_Pages.p_Arr;

            for (uint i = (uint)f_Pages.p_Count; 0 != i--; )
            {
                if (a_Page == l_Arr[i])
                {
                    a_Index = i;

                    return true;
                }
            }

            a_Index = Util.c_NoIndex;

            return false;
        }

        public void ChangePageIndex(AdvancedTabPage a_Page,
                                    uint a_NewIndex,
                                    SelectionChangeOrigins a_Origin)
        {
            uint l_OldIdx = GetIndexOfPageCore(a_Page);

            if (l_OldIdx == a_NewIndex)
            {
                return;
            }

            InvalidatePageListingContextMenu();

            bool l_Simple = (false == AreAnyTabsVisible() ||
                             ((l_OldIdx < f_DisplStartIdx &&
                               a_NewIndex < f_DisplStartIdx) ||
                              (l_OldIdx > f_DisplEndIdx &&
                               a_NewIndex > f_DisplEndIdx)));
            bool l_IsSelPageDisplayed = false;

            if (false == l_Simple)
            {
                // Query this piece of information before the change
                // of index.
                l_IsSelPageDisplayed = IsPageInDisplayRange(f_SelPage);
            }

            f_Pages.DeleteAt((int)l_OldIdx);
            f_Pages.ReserveAt((int)a_NewIndex);
            f_Pages.p_Arr[a_NewIndex] = a_Page;

            if (false == l_Simple)
            {
               CalculateVisibleTabs(l_IsSelPageDisplayed, false);
            }

            a_Page.OnPageIndexChanged(l_OldIdx, a_NewIndex, a_Origin);
            OnPageIndexChanged(a_Page, l_OldIdx, a_NewIndex, a_Origin);
        }

        public bool IsPageVisible(uint a_Index)
        {
            return GetPageByIndex(a_Index).p_PageVisible;
        }

        public void AdjustPageVisibility(AdvancedTabPage a_Page,
                                         bool a_NewVisibility,
                                         SelectionChangeOrigins a_Origin)
        {
            a_Page.AdjustPageVisibilityCore(a_NewVisibility, a_Origin, true);
        }

        private void MakePageVisibleWithoutDefaultHandler(AdvancedTabPage a_Page,
                                                          SelectionChangeOrigins a_Origin)
        {
            // This routine prevents NotifyTabParamsChanged() from
            // getting called.

            a_Page.AdjustPageVisibilityCore(true, a_Origin, false);
            HandleChangeOfTabParams(a_Page, true, a_Origin, false);
        }

        private AdvancedTabPage FindNextSelectablePage(AdvancedTabPage a_StartPage)
        {
            return FindNextSelectablePage(GetIndexOfPageCore(a_StartPage), 1);
        }

        private AdvancedTabPage FindNextSelectablePage(uint a_StartIndex,
                                                       uint a_Count)
        {
            // a_Count must not be zero. Return can be null.

            uint l_ResIdx;
            uint l_FinalIdx = (a_StartIndex + a_Count);

            // First try out the pages before the passed range,
            // then the ones behind it.

            if ((0 != a_StartIndex &&
                 true == FindIndexOfVisiblePage(a_StartIndex - 1, true, out l_ResIdx)) ||
                (p_PageCount > l_FinalIdx &&
                 true == FindIndexOfVisiblePage(l_FinalIdx, false, out l_ResIdx)))
            {
                return GetPageByIndex(l_ResIdx);
            }

            return null;
        }

        private bool FindIndexOfVisiblePage(uint a_StartItemIndex,
                                            bool a_RoundDown,
                                            out uint a_ItemIndex)
        {
            uint l_Count = p_PageCount;
            AdvancedTabPage[] l_Arr = f_Pages.p_Arr;

            for (uint i = a_StartItemIndex; true; )
            {
                if (true == l_Arr[i].p_PageVisible)
                {
                    a_ItemIndex = i;

                    return true;
                }

                if (false == a_RoundDown)
                {
                    i++;

                    if (l_Count == i)
                    {
                        break;
                    }
                }
                else
                {
                    if (0 == i)
                    {
                        break;
                    }

                    i--;
                }
            }

            a_ItemIndex = Util.c_NoIndex;

            return false;
        }

        public bool ItemIndexToDisplayIndex(uint a_ItemIndex,
                                            out uint a_DisplayIndex)
        {
            uint l_Count = p_PageCount;
            AdvancedTabPage[] l_Arr = f_Pages.p_Arr;

            if (true == l_Arr[a_ItemIndex].p_PageVisible)
            {
                for (uint i = 0, k = 0; i != l_Count; i++)
                {
                    if (i == a_ItemIndex)
                    {
                        a_DisplayIndex = k;

                        return true;
                    }

                    if (true == l_Arr[i].p_PageVisible)
                    {
                        k++;
                    }
                }
            }

            a_DisplayIndex = Util.c_NoIndex;

            return false;
        }

        public bool DisplayIndexToItemIndex(uint a_DisplayIndex,
                                            out uint a_ItemIndex)
        {
            uint l_Count = p_PageCount;
            AdvancedTabPage[] l_Arr = f_Pages.p_Arr;

            for (uint i = 0, k = 0; i != l_Count; i++)
            {
                if (false == l_Arr[i].p_PageVisible)
                {
                    continue;
                }

                if (k++ == a_DisplayIndex)
                {
                    a_ItemIndex = i;

                    return true;
                }
            }

            a_ItemIndex = Util.c_NoIndex;

            return false;
        }

        public void ClearSelection(SelectionChangeOrigins a_Origin)
        {
            if (null != f_SelPage)
            {
                SetSelectedPage(null, a_Origin);
            }
        }

        public bool TrySelectNearbyVisiblePage(AdvancedTabPage a_PreferredPage,
                                               SelectionChangeOrigins a_Origin)
        {
            // This routine won't change the visiblity of a_PreferredPage.
            // Returns zero if no page is visible (and consequently
            // none got selected).

            if (0 == f_VisiblePages)
            {
                return false;
            }

            AdvancedTabPage l_TargetPage;

            if (false == a_PreferredPage.p_PageVisible)
            {
                l_TargetPage = FindNextSelectablePage(a_PreferredPage);
            }
            else
            {
                l_TargetPage = a_PreferredPage;
            }

            SetSelectedPage(l_TargetPage, a_Origin);

            return true;
        }

        public void SetSelectedPageIndex(uint a_PageIndex,
                                         SelectionChangeOrigins a_Origin)
        {
            SetSelectedPage(GetPageByIndex(a_PageIndex), a_Origin);
        }

        public void SetSelectedPage(AdvancedTabPage a_NewSelPage,
                                    SelectionChangeOrigins a_Origin)
        {
            // a_NewSelPage may be null.

            if (null != a_NewSelPage)
            {
                if (false == a_NewSelPage.p_PageVisible)
                {
                    MakePageVisibleWithoutDefaultHandler(a_NewSelPage, a_Origin);
                }
            }

            ChangeSelectedPageCore(a_NewSelPage, null == f_SelPage ? SelectionChangeOrigins.None : a_Origin, null == a_NewSelPage ? SelectionChangeOrigins.None : a_Origin);
        }

        private void PlaceSelectedPageIntoInnerRect()
        {
            f_SelPage.Location = f_InnerRect.Location;
            f_SelPage.Size = f_InnerRect.Size;
        }

        public bool HasSelectedPage()
        {
            return (null == f_SelPage);
        }

        private void ChangeSelectedPageCore(AdvancedTabPage a_NewSelPage,
                                            SelectionChangeOrigins a_OriginDeselect,
                                            SelectionChangeOrigins a_OriginSelect)
        {
            // a_NewSelPage may be null.

            if (a_NewSelPage == f_SelPage)
            {
                // Make sure the page is displayed.

                CalculateVisibleTabs(true, false);

                return;
            }

            // First run handlers of the affected pages.

            if (null != f_SelPage)
            {
                if (f_DraggedPage == f_SelPage)
                {
                    // NOTE: EndDragPage() can change the index of
                    // f_SelPage.

                    EndDragPage(false, a_OriginDeselect);
                }

                f_SelPage.Size = new Size();
                f_SelPage.Visible = false;

                if (null == a_NewSelPage)
                {
                    f_ToolBar.AdjustEnabledRange((uint)ToolBarButtons.HideTab, 2, false);

                    if (true == IsPageInDisplayRange(f_SelPage))
                    {
                        CalculateVisibleTabs(false, false);
                    }
                }

                f_SelPage.TabControl_SelectionStateChanged(false, a_OriginSelect);
            }

            AdvancedTabPage l_OldSelPage = f_SelPage;
            f_SelPage = a_NewSelPage;

            if (null != f_SelPage)
            {
                // Update tool bar buttons.

                bool l_CanHide = CanHideSelPagePerUI();
                bool l_CanClose = CanCloseSelPagePerUI();

                if (l_CanHide == l_CanClose)
                {
                    f_ToolBar.AdjustEnabledRange((uint)ToolBarButtons.HideTab, 2, l_CanHide);
                }
                else
                {
                    f_ToolBar.AdjustEnabled((uint)ToolBarButtons.HideTab, l_CanHide);
                    f_ToolBar.AdjustEnabled((uint)ToolBarButtons.CloseTab, l_CanClose);
                }

                // Make page visible and run page handler.

                PlaceSelectedPageIntoInnerRect();
                f_SelPage.Visible = true;

                CalculateVisibleTabs(true, false);
                EnsurePageVisibility(a_NewSelPage, a_OriginSelect);

                f_SelPage.TabControl_SelectionStateChanged(true, a_OriginSelect);
            }

            // Now run the handler of this tab control.

            OnSelectedPageChanged(l_OldSelPage, a_OriginDeselect, a_OriginSelect);
        }

        protected virtual void OnSelectedPageChanged(AdvancedTabPage a_OldSelPage,
                                                     SelectionChangeOrigins a_OriginDeselect,
                                                     SelectionChangeOrigins a_OriginSelect)
        {
            // a_OldSelPage may be null.

            if (null != e_SelectedPageChanged)
            {
                e_SelectedPageChanged(this);
            }

            if (null != e_SelectedPageChangedEx)
            {
                e_SelectedPageChangedEx(this, a_OldSelPage, a_OriginDeselect, a_OriginSelect);
            }
        }

        protected virtual void OnPageAddedOrRemoved(AdvancedTabPage a_Page,
                                                    bool a_Added,
                                                    SelectionChangeOrigins a_Origin)
        {
            // This handler (as the identifier suggests) gets called
            // after a page was added and after it was removed.
        }

        protected virtual void OnPageIndexChanged(AdvancedTabPage a_Page,
                                                  uint a_OldIdx,
                                                  uint a_NewIdx,
                                                  SelectionChangeOrigins a_Origin)
        {
            // There is currently no event for this.
        }

        protected virtual void OnPageVisibilityChanged(AdvancedTabPage a_Page,
                                                       SelectionChangeOrigins a_Origin)
        {
            // There is currently no event for this.
        }

        protected virtual void OnTabDoubleClick(MouseEventArgs e,
                                                AdvancedTabPage a_Page)
        {
            // There is currently no event for this.
        }

        protected virtual void OnBeforeSelectingThroughUI(AdvancedTabPage a_Page,
                                                          ref bool a_Cancel)
        {
            // There is currently no event for this.
        }

        protected virtual void OnBeforeDraggingTab(AdvancedTabPage a_Page,
                                                   ref bool a_Handled)
        {
            if (null != e_BeforeDraggingTab)
            {
                e_BeforeDraggingTab(this, a_Page, ref a_Handled);
            }
        }

        protected virtual void OnDraggingTab(AdvancedTabPage a_Page,
                                             ref uint a_NewIndex)
        {
            if (null != e_DraggingTab)
            {
                e_DraggingTab(this, a_Page, ref a_NewIndex);
            }
        }

        protected virtual void OnAfterDraggingTab(AdvancedTabPage a_Page,
                                                  bool a_DidDrag,
                                                  bool a_Commited)
        {
            if (null != e_AfterDraggingTab)
            {
                e_AfterDraggingTab(this, a_Page, a_DidDrag, a_Commited);
            }
        }

        protected virtual void OnHidingSelPageThroughUI(ref bool a_Handled)
        {
            if (null != e_HidingSelPageThroughUI)
            {
                e_HidingSelPageThroughUI(this, f_SelPage, ref a_Handled);
            }
        }

        protected virtual void OnClosingSelPageThroughUI(ref bool a_Handled,
                                                         ref bool a_DisposePage)
        {
            // a_DisposePage is only relevant if a_Handled is zero
            // on return.

            if (null != e_ClosingSelPageThroughUI)
            {
                e_ClosingSelPageThroughUI(this, f_SelPage, ref a_Handled, ref a_DisposePage);
            }
        }

        public int MeasureCaptionWidth(string a_Caption)
        {
            return MainForm.MeasureTextWidth(a_Caption, this.Font);
        }

        private static uint FlagsToUInt(TabControlFlags a_Flags)
        {
            return (uint)a_Flags;
        }

        [Flags]
        private enum TabControlFlags : uint
        {
            DisplayImages = 0x1,
            AllowSelectingThroughUI = 0x2,
            AllowScrollingThroughToolBar = 0x4,
            AllowHidingThroughUI = 0x8,
            AllowClosingThroughUI = 0x10,
            AlwaysShowPageListingToolBarButton = 0x20,
            IncludeHiddenPagesInAutoPageListing = 0x40,
            AllowReorderingThroughUI = 0x80,
            PageListingContextMenuObsolete = 0x100,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct TabPosition
        {
            public readonly AdvancedTabPage f_Page;
            // f_TabWidth is the result of CalcTabWidth(), not the
            // result of CalcTabBaseWidth().
            public readonly int f_TabWidth;
            public int f_X;
            public bool f_Full;

            public TabPosition(AdvancedTabPage a_Page,
                               int a_TabWidth,
                               int a_X,
                               bool a_Full)
            {
                // a_TabWidth must be the result of CalcTabWidth().

                f_Page = a_Page;
                f_TabWidth = a_TabWidth;
                f_X = a_X;
                f_Full = a_Full;
            }

            public int CalcRight(AdvancedTabControl a_Parent)
            {
                return (f_X + a_Parent.CalcTabBaseWidth(f_TabWidth, f_Full));
            }

            public static void Shift(TabPosition[] a_Arr,
                                     uint a_Idx,
                                     uint a_Len,
                                     int a_XDelta,
                                     bool a_OnlyIfFullFlagIsClear)
            {
                for (uint i = a_Idx, lim = (a_Idx + a_Len); i != lim; i++)
                {
                    if (false == a_OnlyIfFullFlagIsClear ||
                        false == a_Arr[i].f_Full)
                    {
                        a_Arr[i].f_X += a_XDelta;
                    }
                }
            }
        }

        private enum ToolBarButtons : byte
        {
            ScrollLeft = 0,
            ScrollRight = 1,
            HideTab = 2,
            CloseTab = 3,
            ListAllPages = 4,
        }

        private enum PageActionState : byte
        {
            None = 0,

            Selected = 1,
            Moused = 2,
            Dragged = 3,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct TabBrushes : IDisposable
        {
            private static readonly Color s_UnselInactUpperCol;
            private static readonly Color s_UnselActUpperCol;
            private static readonly Color s_SelUpperCol;
            private static readonly Color s_DragUpperCol;

            private static readonly Color s_UnselInactLowerCol;
            private static readonly Color s_UnselActLowerCol;
            private static readonly Color s_SelLowerCol;
            private static readonly Color s_DragLowerCol;

            private bool f_Initialized;
            private int f_TopPadding;
            private int f_TabHeight;

            private LinearGradientBrush f_UnselInactGradient;
            private LinearGradientBrush f_UnselActGradient;
            private LinearGradientBrush f_SelGradient;
            private LinearGradientBrush f_DragGradient;

            public LinearGradientBrush p_UnselInactGradient
            {
                get { return f_UnselInactGradient; }
            }

            public LinearGradientBrush p_UnselActGradient
            {
                get { return f_UnselActGradient; }
            }

            public LinearGradientBrush p_SelGradient
            {
                get { return f_SelGradient; }
            }

            public LinearGradientBrush p_DragGradient
            {
                get { return f_DragGradient; }
            }

            static TabBrushes()
            {
                s_UnselInactUpperCol = Color.FromArgb(247, 243, 213);
                s_UnselActUpperCol = Color.FromArgb(254, 254, 246);
                s_SelUpperCol = Color.FromArgb(254, 254, 250);
                s_DragUpperCol = Color.FromArgb(228, 184, 90);

                s_UnselInactLowerCol = Color.FromArgb(241, 224, 129);
                s_UnselActLowerCol = Color.FromArgb(243, 237, 168);
                s_SelLowerCol = Color.FromArgb(245, 250, 208);
                s_DragLowerCol = Color.FromArgb(245, 250, 208);
            }

            public void Dispose()
            {
                Util.DisposeAndZero(ref f_UnselInactGradient);
                Util.DisposeAndZero(ref f_UnselActGradient);
                Util.DisposeAndZero(ref f_SelGradient);
                Util.DisposeAndZero(ref f_DragGradient);
            }

            public void SetParameters(int a_TopPadding,
                                      int a_TabHeight)
            {
                if (false == f_Initialized)
                {
                    f_Initialized = true;
                }
                else
                {
                    if (a_TopPadding == f_TopPadding &&
                        a_TabHeight == f_TabHeight)
                    {
                        // Nothing to do.

                        return;
                    }

                    this.Dispose();
                }

                f_TopPadding = a_TopPadding;
                f_TabHeight = a_TabHeight;

                const LinearGradientMode l_Mode = LinearGradientMode.Vertical;
                Rectangle l_Rect = new Rectangle(0, f_TopPadding, int.MaxValue, f_TabHeight);
                f_UnselInactGradient = new LinearGradientBrush(l_Rect, s_UnselInactUpperCol, s_UnselInactLowerCol, l_Mode);
                f_UnselActGradient = new LinearGradientBrush(l_Rect, s_UnselActUpperCol, s_UnselActLowerCol, l_Mode);
                f_SelGradient = new LinearGradientBrush(l_Rect, s_SelUpperCol, s_SelLowerCol, l_Mode);
                f_DragGradient = new LinearGradientBrush(l_Rect, s_DragUpperCol, s_DragLowerCol, l_Mode);
            }

            public Brush GetBrush(PageActionState a_PageState)
            {
                switch (a_PageState)
                {
                    case PageActionState.None:
                        return f_UnselInactGradient;

                    case PageActionState.Selected:
                        return f_SelGradient;

                    case PageActionState.Moused:
                        return f_UnselActGradient;

                    case PageActionState.Dragged:
                        return f_DragGradient;
                }

                throw (new UnexpectedStateException());
            }
        }

        private class DragContext
        {
            public delegate void DragOutsideHandler(bool a_Near);

            private static readonly uint s_MinExtent = Lyt.s_LargeSpace;
            private static readonly uint s_MaxExtent = (8 * (uint)Lyt.s_LargeSpace);
            private static readonly uint s_ThresholdExtent = (8 * (uint)Lyt.s_LargeSpace);

            private readonly Control f_Cntrl;
            private readonly bool f_DragHorizontal;
            public readonly uint f_OriginIndex;

            private readonly int f_MinMousePos;
            private readonly int f_MaxMousePos;

            private GuiTimer f_Timer;
            private bool f_DragNear;
            private uint f_DragExtent;
            private uint f_DragPoints;

            private bool f_ActuallyDragged;

            // This is a mandatory event.
            public event DragOutsideHandler e_DragOutside;

            public DragContext(Control a_Cntrl,
                               bool a_DragHorizontal,
                               uint a_OriginIndex)
            {
                f_Cntrl = a_Cntrl;
                f_DragHorizontal = a_DragHorizontal;
                f_OriginIndex = a_OriginIndex;

                Rectangle l_ScreenRect = Screen.GetBounds(f_Cntrl);
                f_MinMousePos = (false == f_DragHorizontal ? l_ScreenRect.Y : l_ScreenRect.X);
                f_MaxMousePos = Math.Max(f_MinMousePos, (false == f_DragHorizontal ? l_ScreenRect.Bottom : l_ScreenRect.Right) - 1);
            }

            public bool IsDragging()
            {
                return (null != f_Timer);
            }

            public void StartDrag()
            {
                const ushort c_IntervalInMs = 50;

                f_Cntrl.Capture = true;
                f_DragExtent = 0;
                f_DragPoints = 0;

                f_Timer = new GuiTimer();
                f_Timer.Interval = c_IntervalInMs;
                f_Timer.e_AdvancedTick += DragTimer_AdvancedTick;
                f_Timer.Start();
            }

            public void EndDrag()
            {
                f_Timer.Stop();
                f_Timer.Dispose();
                f_Timer = null;

                f_Cntrl.Capture = false;
            }

            public void NotifyDragging()
            {
                f_ActuallyDragged = true;
            }

            public bool DidDrag()
            {
                return f_ActuallyDragged;
            }

            private bool SetDragPos(int a_ScreenPos,
                                    int a_CntrlPos,
                                    int a_NearLimWeak,
                                    int a_FarLimWeak,
                                    int a_NearLimStrong,
                                    int a_FarLimStrong)
            {
                // Returns true if the position is beyond any near or
                // far limit.

                bool l_Min = false;
                bool l_Max = false;

                if (a_CntrlPos < a_NearLimWeak)
                {
                    f_DragNear = true;

                    if (a_CntrlPos >= a_NearLimStrong)
                    {
                        l_Min = true;
                    }
                    else if (f_MinMousePos >= a_ScreenPos)
                    {
                        // Mouse is at end of sceen.

                        l_Max = true;
                    }
                    else
                    {
                        f_DragExtent = (uint)(a_NearLimStrong - a_CntrlPos);
                    }
                }
                else if (a_CntrlPos >= a_FarLimWeak)
                {
                    f_DragNear = false;

                    if (a_CntrlPos < a_FarLimStrong)
                    {
                        l_Min = true;
                    }
                    else if (f_MaxMousePos <= a_ScreenPos)
                    {
                        // Mouse is at end of sceen.

                        l_Max = true;
                    }
                    else
                    {
                        f_DragExtent = ((uint)(a_CntrlPos - a_FarLimStrong) + 1);
                    }
                }
                else
                {
                    f_DragExtent = 0;

                    return false;
                }

                if (true == l_Min)
                {
                    f_DragExtent = s_MinExtent;
                }
                else if (true == l_Max)
                {
                    f_DragExtent = s_MaxExtent;
                }
                else
                {
                    f_DragExtent = Util.MinMax(s_MinExtent, s_MaxExtent, f_DragExtent);
                }

                return true;
            }

            public bool SetMousePos(Point a_ScreenMousePos,
                                    Point a_CntrlMousePos,
                                    ref Rectangle a_WeakRect,
                                    ref Rectangle a_StrongRect)
            {
                bool l_Outside;

                if (false == f_DragHorizontal)
                {
                    l_Outside = SetDragPos(a_ScreenMousePos.Y, a_CntrlMousePos.Y, a_WeakRect.Y, a_WeakRect.Bottom, a_StrongRect.Y, a_StrongRect.Bottom);
                }
                else
                {
                    l_Outside = SetDragPos(a_ScreenMousePos.X, a_CntrlMousePos.X, a_WeakRect.X, a_WeakRect.Right, a_StrongRect.X, a_StrongRect.Right);
                }

                return l_Outside;
            }

            private void DragTimer_AdvancedTick(GuiTimer a_Sender,
                                                uint a_Times)
            {
                ulong l_NewPoints = f_DragExtent;

                if (1 != a_Times)
                {
                    l_NewPoints *= a_Times;
                }

                l_NewPoints += f_DragPoints;

                if (s_ThresholdExtent > l_NewPoints)
                {
                    // Insufficient points.

                    f_DragPoints = (uint)l_NewPoints;

                    return;
                }

                f_DragPoints = 0;
                e_DragOutside(f_DragNear);
            }
        }

        private sealed class DynamicScrollContext : IDisposable
        {
            private readonly AdvancedTabControl f_TabCntrl;
            private readonly GuiTimer f_Timer;
            public readonly bool f_ScrollNear;

            private byte f_Step;
            private byte f_ScrollCount;

            public DynamicScrollContext(AdvancedTabControl a_TabCntrl,
                                        bool a_ScrollNear)
            {
                const ushort c_IntervalInMs = 100;

                f_TabCntrl = a_TabCntrl;
                f_Timer = new GuiTimer(true);
                f_ScrollNear = a_ScrollNear;
                f_ScrollCount = 1;

                f_Timer.Interval = c_IntervalInMs;
                f_Timer.e_AdvancedTick += ScrollTimer_AdvancedTick;
                f_Timer.Start();
            }

            public void Dispose()
            {
                f_Timer.Stop();
                f_Timer.Dispose();
            }

            private void ScrollTimer_AdvancedTick(GuiTimer a_Sender,
                                                  uint a_Times)
            {
                // Increase the speed each fifth tick until a speed
                // of ten tabs per tick is reached.
                const byte l_MaxSteps = 5;
                const byte l_MaxScrollCount = 10;

                uint l_EffScrollCount = 0;

                for (uint i = a_Times; 0 != i; i--)
                {
                    if (l_MaxScrollCount > f_ScrollCount &&
                        l_MaxSteps == ++f_Step)
                    {
                        // Double the speed.

                        f_Step = 0;
                        f_ScrollCount = (byte)Math.Min(l_MaxScrollCount, f_ScrollCount * 2u);
                    }

                    l_EffScrollCount += f_ScrollCount;
                }

                Scroll(l_EffScrollCount);
            }

            private void Scroll(uint a_Count)
            {
                if (false == f_ScrollNear)
                {
                    f_TabCntrl.ScrollFar(a_Count);
                }
                else
                {
                    f_TabCntrl.ScrollNear(a_Count);
                }
            }
        }

        public enum ListItemState : byte
        {
            HiddenPage = 0,
            VisiblePage = 1,
            SelectedPage = 2,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct PageListItem
        {
            public readonly uint f_PageIndex;
            public readonly ListItemState f_ItemState;
            // f_Icon is not owned by this instance. It can be null.
            public readonly Image f_Icon;
            public readonly string f_Text;

            public PageListItem(uint a_PageIndex,
                                ListItemState a_ItemState,
                                Image a_Icon,
                                string a_Text)
            {
                // a_Icon may be null.

                f_PageIndex = a_PageIndex;
                f_ItemState = a_ItemState;
                f_Icon = a_Icon;
                f_Text = a_Text;
            }

            public GenericMenuItem<uint> CreateMenuItem(ContextMenuStrip a_Menu,
                                                        EventHandler a_ClickHandler)
            {
                GenericMenuItem<uint> l_Item = GenericMenuItem<uint>.CreateItem(a_Menu, f_Text, a_ClickHandler, f_Icon, f_PageIndex);

                return l_Item;
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct Enum : System.Collections.IEnumerator, IEnumerator<AdvancedTabPage>
        {
            // Don't use the struct default constructor.
            // The array's content should not be changed while
            // enumerating.

            private readonly bool f_VisibleOnly;
            private readonly uint f_StartingIdx;
            private readonly uint f_FinalIdx;
            // f_Arr can be null.
            private readonly AdvancedTabPage[] f_Arr;
            private uint f_Index;

            public AdvancedTabPage Current
            {
                // We might throw an InvalidOperationException here if
                // the index is out of range (as the manual suggests).

                get
                {
                    return f_Arr[f_Index];
                }
            }

            object System.Collections.IEnumerator.Current
            {
                get { return Current; }
            }

            public Enum(bool a_VisibleOnly,
                        uint a_StartingIdx,
                        uint a_Len,
                        AdvancedTabPage[] a_Arr)
            {
                // a_Len may be zero. a_Arr may be null.

                f_VisibleOnly = a_VisibleOnly;
                f_StartingIdx = a_StartingIdx;
                f_FinalIdx = (f_StartingIdx + a_Len);
                f_Arr = a_Arr;
                f_Index = Util.c_NoIndex;
            }

            void IDisposable.Dispose()
            {
            }

            public void Reset()
            {
                f_Index = Util.c_NoIndex;
            }

            public bool MoveNext()
            {
                if (Util.c_NoIndex == f_Index)
                {
                    // A reset just happened.

                    f_Index = f_StartingIdx;
                }
                else
                {
                    f_Index++;
                }

                for (; f_FinalIdx > f_Index; f_Index++)
                {
                    if (false == f_VisibleOnly ||
                        true == f_Arr[f_Index].p_PageVisible)
                    {
                        return true;
                    }
                }

                f_Index = f_FinalIdx;

                return false;
            }
        }
    }

    class AdvancedTabPage : GuiAutoPanel
    {
        private AdvancedTabControl f_TabCntrl;

        // f_Caption will never be null.
        private string f_Caption;
        // This object does not own f_Image. f_Image can be null.
        private Image f_Image;

        private bool f_TextWidthDirty;
        private int f_TextWidth;

        private BitField f_Flags;

        public AdvancedTabControl p_TabControl
        {
            // The return value can be null if the page is not
            // accociated currently.
            get { return f_TabCntrl; }
        }

        public string p_Caption
        {
            get { return f_Caption; }
            set { SetCaptionCore(null == value ? string.Empty : value); }
        }

        public Image p_Image
        {
            get { return f_Image; }
            set { SetImage(value, true); }
        }

        public bool p_DisplayImage
        {
            get { return f_Flags.CheckFlag(FlagsToUInt(TabPageFlags.DisplayImage)); }
            set { ChangeImageRelatedFlag(TabPageFlags.DisplayImage, value); }
        }

        public bool p_DisplayImageOnFarSide
        {
            get { return f_Flags.CheckFlag(FlagsToUInt(TabPageFlags.DisplayImageOnFarSide)); }
            set { ChangeImageRelatedFlag(TabPageFlags.DisplayImageOnFarSide, value); }
        }

        public bool p_AllowHidingThroughUI
        {
            get { return f_Flags.CheckFlag(FlagsToUInt(TabPageFlags.AllowHidingThroughUI)); }
            set { f_Flags.AdjustFlag(FlagsToUInt(TabPageFlags.AllowHidingThroughUI), value); }
        }

        public bool p_AllowClosingThroughUI
        {
            get { return f_Flags.CheckFlag(FlagsToUInt(TabPageFlags.AllowClosingThroughUI)); }
            set { f_Flags.AdjustFlag(FlagsToUInt(TabPageFlags.AllowClosingThroughUI), value); }
        }

        public bool p_AllowReorderingThroughUI
        {
            get { return f_Flags.CheckFlag(FlagsToUInt(TabPageFlags.AllowReorderingThroughUI)); }
            set { f_Flags.AdjustFlag(FlagsToUInt(TabPageFlags.AllowReorderingThroughUI), value); }
        }

        public bool p_PageVisible
        {
            get { return f_Flags.CheckFlag(FlagsToUInt(TabPageFlags.PageVisible)); }
            set { AdjustPageVisibilityCore(value, SelectionChangeOrigins.Programmatic, true); }
        }

        public uint p_PageIndex
        {
            // This property may only be used while the page
            // is associated with a tab control.
            get { return GetPageIndex(); }
            set { f_TabCntrl.ChangePageIndex(this, value, SelectionChangeOrigins.Programmatic); }
        }

        public AdvancedTabPage() :
                          this(null, IconTypes.NoIcon)
        {
        }

        public AdvancedTabPage(string a_Caption,
                               IconTypes a_IconType) :
                          base(false)
        {
            // a_Caption may be null.

            f_Caption = (null == a_Caption ? string.Empty : a_Caption);
            f_Image = Software.s_IconStorage.GetDefaultIcon(a_IconType);

            const TabPageFlags l_InitialFlags = (TabPageFlags.PageVisible | TabPageFlags.DisplayImage);
            f_Flags.SetCombination(FlagsToUInt(l_InitialFlags));
        }

        protected override void Dispose(bool a_Disposing)
        {
            if (true == a_Disposing)
            {
                if (null != f_TabCntrl)
                {
                    f_TabCntrl.RemovePage(this, false, SelectionChangeOrigins.Programmatic);
                }
            }

            base.Dispose(a_Disposing);
        }

        protected override void OnParentChanged(EventArgs e)
        {
            // Need to call base class implementation.

            f_TabCntrl = (AdvancedTabControl)this.Parent;
            InvalidateCaptionWidth();

            base.OnParentChanged(e);
        }

        public void AdjustPageVisibilityCore(bool a_NewVisibility,
                                             SelectionChangeOrigins a_Origin,
                                             bool a_RunTabControlHandler)
        {
            // May only get called by the AdvancedTabControl class and
            // this class.

            bool l_Change = f_Flags.AdjustFlagWithReturn(FlagsToUInt(TabPageFlags.PageVisible), a_NewVisibility);

            if (true == l_Change)
            {
                OnPageVisibilityChanged(a_Origin);

                if (true == a_RunTabControlHandler &&
                    null != f_TabCntrl)
                {
                    NotifyParentOfChange(true, a_Origin);
                }
            }
        }

        protected virtual void OnPageVisibilityChanged(SelectionChangeOrigins a_Origin)
        {
        }

        public virtual void OnPageAddedOrRemoved(bool a_Added,
                                                 SelectionChangeOrigins a_Origin)
        {
            // May only get called by the AdvancedTabControl class.
        }

        public virtual void OnPageIndexChanged(uint a_OldIdx,
                                               uint a_NewIdx,
                                               SelectionChangeOrigins a_Origin)
        {
            // May only get called by the AdvancedTabControl class.
        }

        public void TabControl_SelectionStateChanged(bool a_IsSelected,
                                                     SelectionChangeOrigins a_Origin)
        {
            // May only get called by the AdvancedTabControl class.

            if (false == a_IsSelected)
            {
                OnDeselectPage(a_Origin);
            }
            else
            {
                OnSelectPage(a_Origin);
            }
        }

        protected virtual void OnSelectPage(SelectionChangeOrigins a_Origin)
        {
        }

        protected virtual void OnDeselectPage(SelectionChangeOrigins a_Origin)
        {
        }

        private void SetCaptionCore(string a_NewCaption)
        {
            if (a_NewCaption == f_Caption)
            {
                return;
            }

            f_Caption = a_NewCaption;
            NotifyPropertiesChanged(true);
        }

        public void SetImage(IconTypes a_IconType)
        {
            SetImage(Software.s_IconStorage.GetDefaultIcon(a_IconType), false);
        }

        public void SetImage(Image a_Image,
                             bool a_ForceUpdate)
        {
            // a_Image may be null.

            if (a_Image == f_Image)
            {
                if (null == f_Image ||
                    false == a_ForceUpdate)
                {
                    return;
                }
            }
            else
            {
                f_Image = a_Image;
            }

            NotifyPropertiesChanged(false);
        }

        public int GetCaptionWidth()
        {
            // May only get called while the page is associated
            // with a tab control

            if (true == f_TextWidthDirty)
            {
                f_TextWidthDirty = false;
                f_TextWidth = f_TabCntrl.MeasureCaptionWidth(f_Caption);
            }

            return f_TextWidth;
        }

        public void InvalidateCaptionWidth()
        {
            // May only get called by the this class and the
            // AdvancedTabControl class. Even call this if
            // the tab page is not visible.

            if (false == f_TextWidthDirty)
            {
                f_TextWidthDirty = (0 < f_Caption.Length);
                f_TextWidth = 0;
            }
        }

        public void SelectThisPage()
        {
            f_TabCntrl.SetSelectedPage(this, SelectionChangeOrigins.Programmatic);
        }

        private void NotifyPropertiesChanged(bool a_TextChanged)
        {
            if (null == f_TabCntrl)
            {
                // The width of the caption will be invalidated when
                // the page is attached to a tab control for the next
                // time.

                return;
            }

            if (true == a_TextChanged)
            {
                InvalidateCaptionWidth();
            }

            if (true == this.p_PageVisible)
            {
                NotifyParentOfChange(false, SelectionChangeOrigins.Programmatic);
            }
        }

        private void ChangeImageRelatedFlag(TabPageFlags a_Flag,
                                            bool a_State)
        {
            bool l_Change = f_Flags.AdjustFlagWithReturn(FlagsToUInt(a_Flag), a_State);

            if (true == l_Change &&
                null != f_Image)
            {
                NotifyPropertiesChanged(false);
            }
        }

        private uint GetPageIndex()
        {
            // The page must belong to a tab control when
            // this routine gets called.

            uint l_Idx;
            bool l_OK = f_TabCntrl.GetIndexOfPage(this, out l_Idx);

            return l_Idx;
        }

        private void NotifyParentOfChange(bool a_VisibilityChanged,
                                          SelectionChangeOrigins a_Origin)
        {
            // The page must be attached to a tab control
            // for this routine to get called.

            f_TabCntrl.NotifyTabParamsChanged(this, a_VisibilityChanged, a_Origin);
        }

        public bool IsPageSelected()
        {
            return (null != f_TabCntrl &&
                    this == f_TabCntrl.p_SelectedPage);
        }

        private static uint FlagsToUInt(TabPageFlags a_Flags)
        {
            return (uint)a_Flags;
        }

        [Flags]
        private enum TabPageFlags : uint
        {
            // This flag tells if the page is within the displayable
            // set of pages. It does not indicate if the page is
            // acutally visible on screen or selected.
            PageVisible = 0x1,
            DisplayImage = 0x2,
            DisplayImageOnFarSide = 0x4,
            AllowHidingThroughUI = 0x8,
            AllowClosingThroughUI = 0x10,
            AllowReorderingThroughUI = 0x20,
        }
    }
}

// End of file.
