/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.27-src\Native\CsRandom.c
 * PURPOSE: Secure Random Number Generator ('Cowstay')
 * VERSION: 1.4.27
 * DATE: March 23, 2016
 * COPYRIGHT: Copyright (C) 2012-2016 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

#include <malloc.h>

#define INCLUDE_WINDOWS
#include "general.h"

#include <psapi.h>
#include <Iphlpapi.h>

// Prototypes.

void
SHA256MultiProcessInternal(IN OUT UINT32 * const a_HashValues,
                           IN UINT32 const * const a_InputBlockBuf,
                           IN UINT32 const a_InputBlockCount,
                           IN BOOLEAN const a_WipeStack);

void
TigerMultiCompressInternal(IN OUT UINT64 * const a_States,
                           IN UINT64 const * const a_InputBlockBuf,
                           IN UINT32 const a_InputBlockCount,
                           IN BOOLEAN const a_WipeStack);

// Prototypes of operating system routines that are not supported on
// NT 5.0.

typedef BOOL WINAPI GetPerformanceInfoFunc(METAIN OUT PPERFORMANCE_INFORMATION pPerformanceInformation,
                                           IN DWORD cb);

typedef BOOL WINAPI GetProcessHandleCountFunc(IN HANDLE hProcess,
                                              METAIN OUT PDWORD pdwHandleCount);

typedef BOOL WINAPI QueryProcessCycleTimeFunc(IN HANDLE ProcessHandle,
                                              METAIN OUT PULONG64 CycleTime);

typedef BOOL WINAPI QueryThreadCycleTimeFunc(IN HANDLE ThreadHandle,
                                             METAIN OUT PULONG64 CycleTime);

// Use this preprocessor definition to deactivate the use of
// uninitialized memory as a source of entropy. Since uninitialized
// memory can be a good source of entropy, it is not recommended to
// deactivate it.
// #define DISALLOW_UNINITIALIZED_MEM_AS_ENTROPY         1

// Each buffer is 64 bytes long.
#define BUF0_CORE_ELEMENTS      (8)
#define BUF1_CORE_ELEMENTS      (6)
#define BUF_TOTAL_ELEMENTS      (16)
#define HASH_INPUT_BYTES        (BUF_TOTAL_ELEMENTS * sizeof(UINT32))

// One block is 20 bytes long.
#define ELEMENTS_PER_BLOCK      (5)
#define BYTES_PER_BLOCK         (ELEMENTS_PER_BLOCK * sizeof(UINT32))

#define RAND_BUFFER0            (g_EffRandContextPtr->f_Buf0)
#define RAND_BUFFER1            (g_EffRandContextPtr->f_Buf1)

typedef struct _RANDOM_CONTEXT
{
    ALIGN64 UINT32 f_Buf0[BUF_TOTAL_ELEMENTS];
    ALIGN64 UINT32 f_Buf1[BUF_TOTAL_ELEMENTS];
    UINT32 f_RotSeed;
    UINT32 f_ScrambleMethod;
}
RANDOM_CONTEXT;

static volatile UINT32 g_RandLock = 0;
// g_EffRandContextPtr will either reference g_HeapAutoRandContext
// or contain the same value as g_PhysicalRandContextPtr.
static RANDOM_CONTEXT * g_EffRandContextPtr = NULL;

static RANDOM_CONTEXT g_HeapAutoRandContext;
static RANDOM_CONTEXT * g_PhysicalRandContextPtr = NULL;

static HMODULE g_Kernel32Module = NULL;
static GetProcessHandleCountFunc * g_GetHandles = NULL;
static QueryProcessCycleTimeFunc * g_QueryProcessTime = NULL;
static QueryThreadCycleTimeFunc * g_QueryThreadTime = NULL;

static HMODULE g_PsapiMod = NULL;
static GetPerformanceInfoFunc * g_GetPerfInfo = NULL;

void
RandInitInternalWithoutSeeding()
{
    RANDOM_CONTEXT * l_PhysPtr;

    ASSERT(0 == g_RandLock &&
           NULL == g_EffRandContextPtr &&
           NULL == g_PhysicalRandContextPtr);

    l_PhysPtr = VirtualAlloc(NULL, sizeof(RANDOM_CONTEXT), MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE);

    if (NULL != l_PhysPtr)
    {
        BOOL l_Succ = VirtualLock(l_PhysPtr, sizeof(RANDOM_CONTEXT));

        if (FALSE == l_Succ)
        {
            l_Succ = VirtualFree(l_PhysPtr, 0, MEM_RELEASE);

            ASSERT(FALSE != l_Succ);

            l_PhysPtr = NULL;
        }
    }

    if (NULL == l_PhysPtr)
    {
        // Use the auto variable on the heap.

        g_EffRandContextPtr = &g_HeapAutoRandContext;
    }
    else
    {
        // Use the buffer in physical memory we've just allocated.

        g_EffRandContextPtr = g_PhysicalRandContextPtr = l_PhysPtr;
    }

    UtilSecWipePointer(&l_PhysPtr);

    // NOTE: We just allocated the buffer of g_EffRandContextPtr, we
    // don't initialize it yet.

    // Dynamically prepare optional function pointers.

    g_Kernel32Module = LoadLibrary(L"Kernel32.dll");

    if (NULL != g_Kernel32Module)
    {
        g_GetHandles = (GetProcessHandleCountFunc *)GetProcAddress(g_Kernel32Module, "GetProcessHandleCount");
        g_QueryProcessTime = (QueryProcessCycleTimeFunc *)GetProcAddress(g_Kernel32Module, "QueryProcessCycleTime");
        g_QueryThreadTime = (QueryThreadCycleTimeFunc *)GetProcAddress(g_Kernel32Module, "QueryThreadCycleTime");
    }

    g_PsapiMod = LoadLibrary(L"Psapi.dll");

    if (NULL != g_PsapiMod)
    {
        g_GetPerfInfo = (GetPerformanceInfoFunc *)GetProcAddress(g_PsapiMod, "GetPerformanceInfo");
    }
}

void
RandCleanUpInternal()
{
    ASSERT(0 == g_RandLock &&
           NULL != g_EffRandContextPtr);

    UtilSecWipeBuffer(g_EffRandContextPtr, sizeof(RANDOM_CONTEXT));

    if (NULL == g_PhysicalRandContextPtr)
    {
        ASSERT(&g_HeapAutoRandContext == g_EffRandContextPtr);
    }
    else
    {
        BOOL l_Succ;

        ASSERT(g_PhysicalRandContextPtr == g_EffRandContextPtr);

        l_Succ = VirtualUnlock(g_PhysicalRandContextPtr, sizeof(RANDOM_CONTEXT));

        ASSERT(FALSE != l_Succ);

        l_Succ = VirtualFree(g_PhysicalRandContextPtr, 0, MEM_RELEASE);

        ASSERT(FALSE != l_Succ);

        UtilSecWipePointer(&g_PhysicalRandContextPtr);
    }

    UtilSecWipePointer(&g_EffRandContextPtr);

    // Release dynamically loaded modules.

    if (NULL != g_Kernel32Module)
    {
        if (FALSE == FreeLibrary(g_Kernel32Module))
        {
            ASSERT(FALSE);
        }

        g_Kernel32Module = NULL;
        g_GetHandles = NULL;
        g_QueryThreadTime = g_QueryProcessTime = NULL;
    }

    if (NULL != g_PsapiMod)
    {
        if (FALSE == FreeLibrary(g_PsapiMod))
        {
            ASSERT(FALSE);
        }

        g_PsapiMod = NULL;
        g_GetPerfInfo = NULL;
    }
}

static UINT64
RandGetProcessorTimeStampInternal()
{
    // Returns the processor time stamp, the number of clock cycles
    // since the last CPU reset.

    return __rdtsc();
}

static void
RandHashInternal(IN BOOLEAN const a_WipeStack)
{
    UINT32 * const l_Buf0 = RAND_BUFFER0;
    UINT64 * const l_Buf1 = (UINT64 *)RAND_BUFFER1;

    SHA256MultiProcessInternal(l_Buf0, l_Buf0, 1, a_WipeStack);
    TigerMultiCompressInternal(l_Buf1, l_Buf1, 1, a_WipeStack);
}

static void
RandScrambleBufferInternal(IN BOOLEAN const a_BufferOne,
                           IN UINT32 const a_Delta)
{
    UINT32 * const l_Buf = (FALSE == a_BufferOne ? RAND_BUFFER0 : RAND_BUFFER1);
    UINT32 * l_Ptr = (l_Buf + (FALSE == a_BufferOne ? BUF0_CORE_ELEMENTS : BUF1_CORE_ELEMENTS));
    UINT32 const * const l_Final = &l_Buf[BUF_TOTAL_ELEMENTS];

    for (; l_Final != l_Ptr; l_Ptr++)
    {
        if (FALSE == a_Delta)
        {
            *l_Ptr += a_Delta;
        }
        else
        {
            *l_Ptr -= a_Delta;
        }
    }
}

static UINT32
RandScrambleInternal(IN BOOLEAN const a_WipeStack,
                     METAIN OUT UINT32 * const a_SecondRand)
{
    // Returns random data. a_SecondRand will also contain random data
    // on return, however of lower quality than the return value.

    UINT32 const l_Ret = RAND_BUFFER0[ELEMENTS_PER_BLOCK];
    UINT32 l_Delta = RAND_BUFFER0[ELEMENTS_PER_BLOCK + 1];
    // The following twist is intentional.
    UINT32 l_Extra1 = RAND_BUFFER0[ELEMENTS_PER_BLOCK + 2];
    UINT32 l_Extra0 = RAND_BUFFER1[ELEMENTS_PER_BLOCK];
    DWORD l_ThreadID = GetCurrentThreadId();
    UINT64 l_CPUTimeStamp = RandGetProcessorTimeStampInternal();
    UINT32 l_BaseIdx0;
    UINT32 l_BaseIdx1;

    *a_SecondRand = (UtilUInt64ToUInt32AndRotate(l_CPUTimeStamp, 13) + (ROT(l_Extra0, 7) ^ ROT(l_Extra1, 29)));

    // Reseed the buffers with thread-dependent values.

    RandScrambleBufferInternal(FALSE, l_Delta - UtilPtrToUInt32(&l_Delta, 21));
    RandScrambleBufferInternal(TRUE, l_Delta + ROT(l_ThreadID, 4) - ROT((UINT32)(l_CPUTimeStamp & 0xF), 23));

    // Do some more scrambling in order to make the mathematical
    // properties of the function crazier.

    l_BaseIdx0 = ROT(l_ThreadID ^ l_Delta, 1);
    l_BaseIdx1 = ROT(l_ThreadID + l_Delta, 29);

// Define temporary helper macros.
#define BUF0_ELM(x)            RAND_BUFFER0[(l_BaseIdx0 + (x)) % BUF_TOTAL_ELEMENTS]
#define BUF1_ELM(x)            RAND_BUFFER1[(l_BaseIdx1 + (x)) % BUF_TOTAL_ELEMENTS]

    switch (g_EffRandContextPtr->f_ScrambleMethod++ % 7)
    {
        case 0:
            BUF0_ELM(0) ^= l_Extra0;
            BUF1_ELM(0) ^= ROT(BUF1_ELM(6), l_Extra1);
            break;

        case 1:
            BUF0_ELM(1) += l_Extra0;
            BUF1_ELM(1) ^= l_Extra1;
            break;

        case 2:
            BUF0_ELM(2) ^= ROT(BUF0_ELM(6), l_Extra0);
            BUF1_ELM(2) += l_Extra1;
            break;

        case 3:
            BUF0_ELM(3) -= l_Extra0;
            BUF1_ELM(3) ^= ROT(BUF1_ELM(7), l_Extra1);
            break;

        case 4:
            BUF0_ELM(4) ^= ROT(BUF0_ELM(7), l_Extra0);
            BUF1_ELM(4) -= l_Extra1;
            break;

        case 5:
            BUF0_ELM(5) += ROT(l_Extra0, BUF0_ELM(8));
            BUF1_ELM(5) ^= ROT(l_Extra1, BUF1_ELM(8));
            break;

        case 6:
            BUF0_ELM(6) ^= l_Extra0;
            BUF1_ELM(6) -= l_Extra1;
            break;

        default: ASSERT(FALSE);
    }

#undef BUF0_ELM
#undef BUF1_ELM

    if (FALSE != a_WipeStack)
    {
        UtilSecWipeUInt32(&l_Delta);
        UtilSecWipeUInt32(&l_Extra0);
        UtilSecWipeUInt32(&l_Extra1);
        UtilSecWipeUInt32((UINT32 *)&l_ThreadID);
        UtilSecWipeUInt64(&l_CPUTimeStamp);
        UtilSecWipeUInt32(&l_BaseIdx0);
        UtilSecWipeUInt32(&l_BaseIdx1);
    }

    return l_Ret;
}

static void
RandBuildNextBlockInternal(METAIN OUT UINT32 * const a_Dest,
                           IN BOOLEAN const a_WipeStack)
{
    UINT32 l_Rot;
    UINT32 l_BaseIdx0;
    UINT32 l_BaseIdx1;
    UINT32 l_BaseIdx2;
    ALIGN64 UINT32 l_EndBuf[BUF_TOTAL_ELEMENTS];
    UINT32 const * const l_Buf0 = RAND_BUFFER0;
    UINT32 const * const l_Buf1 = RAND_BUFFER1;

    l_Rot = RandScrambleInternal(a_WipeStack, &l_BaseIdx2);
    RandHashInternal(a_WipeStack);

    l_BaseIdx0 = (UtilPtrToUInt32(&l_BaseIdx0, 4) + (l_BaseIdx2 & 0xF) + ROT(GetCurrentThreadId(), 3));
    l_BaseIdx1 = ((l_BaseIdx2 >> 4) & 0xF);
    l_BaseIdx2 >>= 8;

// Define temporary macros.
#define IDX0(x)         ((l_BaseIdx0 + (x)) % ELEMENTS_PER_BLOCK)
#define IDX1(x)         ((l_BaseIdx1 + (x)) % ELEMENTS_PER_BLOCK)
#define IDX2(x)         ((l_BaseIdx2 + (x)) % ELEMENTS_PER_BLOCK)

    // Take the 5 input uints of each buffer and mix them to form 16
    // unique pairs. We consider the input uints to be good random,
    // but they form our key at the same time, so we won't expose them
    // directly.

    l_EndBuf[IDX2(0)] =  (ROT(l_Buf0[IDX0(0)], l_Rot) +       ROT(l_Buf1[IDX1(0)], g_EffRandContextPtr->f_RotSeed));
    l_EndBuf[IDX2(1)] =  (ROT(l_Buf0[IDX0(1)], l_Rot >> 5) ^  ROT(l_Buf1[IDX1(1)], g_EffRandContextPtr->f_RotSeed >> 5));
    l_EndBuf[IDX2(2)] =  (ROT(l_Buf0[IDX0(2)], l_Rot >> 10) - ROT(l_Buf1[IDX1(2)], g_EffRandContextPtr->f_RotSeed >> 10));
    l_EndBuf[IDX2(3)] =  (ROT(l_Buf0[IDX0(3)], l_Rot >> 15) + ROT(l_Buf1[IDX1(3)], g_EffRandContextPtr->f_RotSeed >> 15));
    l_EndBuf[IDX2(4)] =  (ROT(l_Buf0[IDX0(4)], l_Rot >> 20) ^ ROT(l_Buf1[IDX1(4)], g_EffRandContextPtr->f_RotSeed >> 20));
    l_EndBuf[IDX2(5)] =  (ROT(l_Buf0[IDX0(1)], l_Rot >> 25) + ROT(l_Buf1[IDX1(0)], g_EffRandContextPtr->f_RotSeed >> 25));
    l_EndBuf[IDX2(6)] =  (ROT(l_Buf0[IDX0(2)], l_Rot >> 1) ^  ROT(l_Buf1[IDX1(1)], g_EffRandContextPtr->f_RotSeed >> 1));
    l_EndBuf[IDX2(7)] =  (ROT(l_Buf0[IDX0(3)], l_Rot >> 6) -  ROT(l_Buf1[IDX1(2)], g_EffRandContextPtr->f_RotSeed >> 6));
    l_EndBuf[IDX2(8)] =  (ROT(l_Buf0[IDX0(4)], l_Rot >> 11) + ROT(l_Buf1[IDX1(3)], g_EffRandContextPtr->f_RotSeed >> 11));
    l_EndBuf[IDX2(9)] =  (ROT(l_Buf0[IDX0(0)], l_Rot >> 16) ^ ROT(l_Buf1[IDX1(4)], g_EffRandContextPtr->f_RotSeed >> 16));
    l_EndBuf[IDX2(10)] = (ROT(l_Buf0[IDX0(2)], l_Rot >> 21) + ROT(l_Buf1[IDX1(0)], g_EffRandContextPtr->f_RotSeed >> 21));
    l_EndBuf[IDX2(11)] = (ROT(l_Buf0[IDX0(3)], l_Rot >> 26) ^ ROT(l_Buf1[IDX1(1)], g_EffRandContextPtr->f_RotSeed >> 26));
    l_EndBuf[IDX2(12)] = (ROT(l_Buf0[IDX0(4)], l_Rot >> 2) -  ROT(l_Buf1[IDX1(2)], g_EffRandContextPtr->f_RotSeed >> 2));
    l_EndBuf[IDX2(13)] = (ROT(l_Buf0[IDX0(0)], l_Rot >> 7) +  ROT(l_Buf1[IDX1(3)], g_EffRandContextPtr->f_RotSeed >> 7));
    l_EndBuf[IDX2(14)] = (ROT(l_Buf0[IDX0(1)], l_Rot >> 12) ^ ROT(l_Buf1[IDX1(4)], g_EffRandContextPtr->f_RotSeed >> 12));
    l_EndBuf[IDX2(15)] = (ROT(l_Buf0[IDX0(3)], l_Rot >> 17) + ROT(l_Buf1[IDX1(0)], g_EffRandContextPtr->f_RotSeed >> 17));

    TigerMultiCompressInternal((UINT64 *)l_EndBuf, (UINT64 *)l_EndBuf, 1, a_WipeStack);

#undef IDX1
#define IDX1(x)         ((l_BaseIdx1 + (x)) % BUF1_CORE_ELEMENTS)

    l_Rot += (IDX1(5) & (MAX_UINT32 >> 7));
    l_BaseIdx0 += ((l_BaseIdx2 >> 4) + (IDX1(5) >> 25));

    a_Dest[IDX0(0)] = ROT(l_EndBuf[IDX1(0)], l_Rot);
    a_Dest[IDX0(1)] = ROT(l_EndBuf[IDX1(1)], l_Rot >> 5);
    a_Dest[IDX0(2)] = ROT(l_EndBuf[IDX1(2)], l_Rot >> 10);
    a_Dest[IDX0(3)] = ROT(l_EndBuf[IDX1(3)], l_Rot >> 15);
    a_Dest[IDX0(4)] = ROT(l_EndBuf[IDX1(4)], l_Rot >> 20);

#undef IDX0
#undef IDX1
#undef IDX2

    if (FALSE != a_WipeStack)
    {
        UtilSecWipeUInt32(&l_Rot);
        UtilSecWipeUInt32(&l_BaseIdx0);
        UtilSecWipeUInt32(&l_BaseIdx1);
        UtilSecWipeBuffer(l_EndBuf, sizeof(l_EndBuf));
    }
}

static UINT32 *
RandQueryBlocksInternal(METAIN OUT UINT32 * const a_Dest,
                        IN CPUWORD const a_BlockCount,
                        IN BOOLEAN const a_DisallowZeros,
                        IN BOOLEAN const a_AllowSlackWiping)
{
    // Fills a CPUWORD-aligned buffer with a_BlockCount blocks of random
    // data.

    UINT32 * l_Dest = a_Dest;
    CPUWORD i;

    ASSERT(0 == ((CPUWORD)l_Dest % sizeof(CPUWORD)));

    for (i = 0; a_BlockCount != i++;)
    {
        BOOLEAN const l_WipeSlack = (FALSE != a_AllowSlackWiping &&
                                     a_BlockCount == i);
        UINT32 * const l_Final = (l_Dest + ELEMENTS_PER_BLOCK);

        for (;;)
        {
            BOOLEAN l_FoundZero;
            BYTE const * l_Ptr;

            RandBuildNextBlockInternal(l_Dest, l_WipeSlack);

            if (FALSE == a_DisallowZeros)
            {
                break;
            }

            // Check for a zero byte in the new block.

            l_FoundZero = FALSE;

            for (l_Ptr = (BYTE *)l_Dest; (void *)l_Final != (void *)l_Ptr; l_Ptr++)
            {
                if (0 == *l_Ptr)
                {
                    l_FoundZero = TRUE;

                    break;
                }
            }

            if (FALSE == l_FoundZero)
            {
                break;
            }

            // The block contains a zero byte. We will land here with a
            // probability of roughly eight percent. Now generate a new
            // block.
        }

        l_Dest = l_Final;
    }

    return l_Dest;
}

#if defined(DISALLOW_UNINITIALIZED_MEM_AS_ENTROPY)

static void
RandInitUninitializedMemory(void * const a_Ptr,
                            CPUWORD const a_Size)
{
    // We don't just zero the buffer, but write some pseudo-random
    // data depending on the addresses of a_Ptr and a_Size.

    UINT32 l_Rand = (UtilPtrToUInt32(a_Ptr, 11) + UtilPtrToUInt32(&a_Size, 26));
    BYTE * l_Ptr = (BYTE *)a_Ptr;
    BYTE const * const l_FinalPtr = &l_Ptr[a_Size];

    for (; l_FinalPtr != l_Ptr; l_Ptr++)
    {
        *l_Ptr = (l_Rand & MAX_BYTE);
        l_Rand = ROT_L(l_Rand, 1);
    }
}

#else // defined(DISALLOW_UNINITIALIZED_MEM_AS_ENTROPY)

// Disable warnings about variables used without initialization
// while compiling RandReseedInternal(), since we will use stack garbage
// as one source of entropy.
#pragma warning(disable: 4700)

#endif // defined(DISALLOW_UNINITIALIZED_MEM_AS_ENTROPY)

void
RandReseedInternal()
{
    // This routine is thread-safe.

    RANDOM_CONTEXT l_NewCntxt;
    UINT32 l_Buf0BaseIndex;
    UINT32 l_Buf1BaseIndex;
    UINT32 l_TempAccessableSize;
    UINT32 l_InitTurns;
    LARGE_INTEGER l_LargeInt;
    FILETIME l_FileTime0;
    FILETIME l_FileTime1;
    FILETIME l_FileTime2;
    FILETIME l_FileTimeJunk;
    HANDLE const l_CurProcess = GetCurrentProcess();
    HANDLE const l_CurThread = GetCurrentThread();

#if defined(DISALLOW_UNINITIALIZED_MEM_AS_ENTROPY)

    RandInitUninitializedMemory(l_NewCntxt.f_Buf0, sizeof(l_NewCntxt.f_Buf0));
    RandInitUninitializedMemory(l_NewCntxt.f_Buf1, sizeof(l_NewCntxt.f_Buf1));
    RandInitUninitializedMemory(&l_Buf0BaseIndex, sizeof(l_Buf0BaseIndex));
    RandInitUninitializedMemory(&l_Buf1BaseIndex, sizeof(l_Buf1BaseIndex));

#endif // defined(DISALLOW_UNINITIALIZED_MEM_AS_ENTROPY)

// Define helper macros for buffer access.
#define BUF0_ELM(x)            l_NewCntxt.f_Buf0[(l_Buf0BaseIndex + (x)) % BUF_TOTAL_ELEMENTS]
#define BUF1_ELM(x)            l_NewCntxt.f_Buf1[(l_Buf1BaseIndex + (x)) % BUF_TOTAL_ELEMENTS]

    {
        // We do not trust the OS random number generator, but we
        // still take four bytes from it, since the benefit should
        // exceed the potential harm.

        HCRYPTPROV l_Prov;

        if (FALSE != CryptAcquireContext(&l_Prov, (LPCWSTR)NULL, (LPCWSTR)NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT | CRYPT_SILENT))
        {
            if (FALSE != CryptGenRandom(l_Prov, sizeof(DWORD), (BYTE *)&l_LargeInt.LowPart))
            {
                BUF0_ELM(0) += l_LargeInt.LowPart;
            }

            if (FALSE == CryptReleaseContext(l_Prov, 0))
            {
                ASSERT(FALSE);
            }
        }

        UtilSecWipeBuffer(&l_Prov, sizeof(l_Prov));
    }

    // Query the time the process has spent in kernel and user mode.

    if (FALSE != GetProcessTimes(l_CurProcess, &l_FileTime2, &l_FileTimeJunk, &l_FileTime0, &l_FileTime1))
    {
        BUF0_ELM(1) -= UtilUInt64ToUInt32AndRotate(*(UINT64 *)&l_FileTime0, 13);
        BUF1_ELM(0) += UtilUInt64ToUInt32AndRotate(*(UINT64 *)&l_FileTime1, 18);
    }

    // Query statistics about the network (system-wide).

    {
        // We see this as optional and do nothing if we don't get that
        // stats.

        MIB_IPSTATS l_Stats;

        if (NO_ERROR == GetIpStatistics(&l_Stats))
        {
            // NOTE: These are the stats for IPv4 only.

            BUF0_ELM(2) ^= ROT(l_Stats.dwOutRequests, 27);
            BUF1_ELM(1) -= ROT(l_Stats.dwOutNoRoutes, 23);
            BUF0_ELM(3) += ROT(l_Stats.dwInReceives, 12);
            BUF1_ELM(2) ^= ROT(l_Stats.dwInDiscards, 15);
            BUF0_ELM(4) -= ROT(l_Stats.dwInDelivers, 22);
            BUF1_ELM(3) ^= ROT(l_Stats.dwOutDiscards, 30);
            BUF0_ELM(5) += ROT(l_Stats.dwReasmOks, 13);
            BUF1_ELM(4) -= ROT(l_Stats.dwReasmFails, 29);
            BUF0_ELM(6) ^= ROT(l_Stats.dwFragOks, 5);
            BUF1_ELM(5) -= ROT(l_Stats.dwFragFails, 11);
        }

        UtilSecWipeBuffer(&l_Stats, sizeof(l_Stats));
    }

    // Query the time that was elapsed since the start of the system.

    if (FALSE == QueryPerformanceCounter(&l_LargeInt))
    {
        // We don't really expect to land here.

        // NOTE: Do not use the 64-bit variant here, since it's not
        // available before NT 6.0 (although the headers don't seem
        // to care).
        BUF1_ELM(6) += ROT(GetTickCount(), 3);
    }
    else
    {
        // The lowest few bits are usually clear.

        BUF1_ELM(6) = ROT(l_LargeInt.LowPart, 19);
    }

    // Query the time since the last terminal input.

    {
        LASTINPUTINFO l_InputInfo;
        l_InputInfo.cbSize = sizeof(LASTINPUTINFO);

        if (FALSE != GetLastInputInfo(&l_InputInfo))
        {
            // l_InputInfo.dwTime is a tick count.

            BUF0_ELM(7) -= ROT(l_InputInfo.dwTime, 11);
        }

        UtilSecWipeBuffer(&l_InputInfo, sizeof(l_InputInfo));
    }

    if (FALSE == GetThreadTimes(l_CurThread, &l_FileTime0, &l_FileTimeJunk, &l_FileTime1, &l_FileTime2))
    {
        // We don't really expect to land here.

        WCHAR l_NoBuf[1];
        DWORD l_Size = 0;

        // Query the UTF-16 character length of the current dir and
        // the user name.

        BUF0_ELM(8) += ROT(GetCurrentDirectory(0, l_NoBuf), 14);

        if (FALSE == GetUserName(l_NoBuf, &l_Size) &&
            ERROR_INSUFFICIENT_BUFFER == GetLastError())
        {
            l_Size--;
            BUF1_ELM(7) -= ROT(l_Size, 29);
        }
    }
    else
    {
        BUF0_ELM(8) ^= ROT(l_FileTime0.dwLowDateTime & 0x3FF, 7);
        BUF1_ELM(7) += UtilUInt64ToUInt32AndRotate(*(UINT64 *)&l_FileTime1, 18);
        BUF1_ELM(8) -= UtilUInt64ToUInt32AndRotate(*(UINT64 *)&l_FileTime2, 11);
    }

    {
        HWND const l_Window = GetForegroundWindow();

        if (NULL != l_Window)
        {
            BUF0_ELM(9) -= UtilPtrToUInt32(GetParent(l_Window), 9);
            BUF1_ELM(9) += UtilPtrToUInt32(GetNextWindow(l_Window, GW_HWNDNEXT), 27);

            if (FALSE == IsZoomed(l_Window))
            {
                RECT l_Rect;

                if (FALSE != GetWindowRect(l_Window, &l_Rect))
                {
                    BUF0_ELM(10) ^= ROT(l_Rect.top, 7);
                    BUF1_ELM(10) += ROT(l_Rect.right, 31);
                    BUF0_ELM(11) ^= ROT(l_Rect.left, 29);
                    BUF1_ELM(11) -= ROT(l_Rect.bottom, 14);
                }

                UtilSecWipeBuffer(&l_Rect, sizeof(l_Rect));
            }

            UtilSecWipePointer(&(void *)l_Window);
        }
    }

    SHA256MultiProcessInternal(l_NewCntxt.f_Buf0, l_NewCntxt.f_Buf0, 1, FALSE);
    TigerMultiCompressInternal((UINT64 *)l_NewCntxt.f_Buf1, (UINT64 *)l_NewCntxt.f_Buf1, 1, FALSE);

    // We now have a first set of secure random data that we can use
    // in the following steps.

    l_Buf0BaseIndex = (BUF0_ELM(0) ^ BUF1_ELM(0));
    l_Buf1BaseIndex = (BUF0_ELM(1) ^ BUF1_ELM(1));

    // Buffer access is now more random.

    l_TempAccessableSize = (BUF0_ELM(2) ^ BUF1_ELM(2));
    l_InitTurns = (BUF0_ELM(3) ^ BUF1_ELM(3));
    l_NewCntxt.f_RotSeed = (BUF0_ELM(4) ^ BUF1_ELM(4));
    l_NewCntxt.f_ScrambleMethod = (BUF0_ELM(5) ^ BUF1_ELM(5));

    // Generate another set of random.

    {
        ALIGN64 BYTE l_TempBuf[HASH_INPUT_BYTES];

        memcpy(l_TempBuf, l_NewCntxt.f_Buf0, HASH_INPUT_BYTES);
        SHA256MultiProcessInternal(l_NewCntxt.f_Buf0, l_NewCntxt.f_Buf1, 1, FALSE);
        TigerMultiCompressInternal((UINT64 *)l_NewCntxt.f_Buf1, (UINT64 *)l_TempBuf, 1, FALSE);

        UtilSecWipeBuffer(l_TempBuf, sizeof(l_TempBuf));
    }

    // We will now try to hash some random memory location in order
    // to add entropy that's really hard to predict.

    l_TempAccessableSize %= (999 * 1024);

    {
        BYTE * const l_TempBuf = (BYTE *)_aligned_malloc(HASH_INPUT_BYTES + l_TempAccessableSize, sizeof(UINT64));
        void const * l_Ptr0;
        void const * l_Ptr1;
        UINT32 i;

        if (NULL == l_TempBuf)
        {
            // Very low available memory. Resort to stack.

            l_Ptr0 = l_NewCntxt.f_Buf0;
            l_Ptr1 = l_NewCntxt.f_Buf1;
        }
        else
        {
            // Hash the content of a random memory location in the
            // buffer we've just allocated. The longer our process
            // has run already, the more likely we will find some
            // nonzero content. The mask aligns pointers for us.

            UINT32 const l_AlignMask = ~(UINT32)(sizeof(UINT64) - 1);

            l_Ptr0 = &l_TempBuf[(l_TempAccessableSize / 2) & l_AlignMask];
            l_Ptr1 = &l_TempBuf[(l_TempAccessableSize / 3) & l_AlignMask];

#if defined(DISALLOW_UNINITIALIZED_MEM_AS_ENTROPY)

            // In this case we may not use uninitialized memory.
            RandInitUninitializedMemory((void *)l_Ptr0, HASH_INPUT_BYTES);
            RandInitUninitializedMemory((void *)l_Ptr1, HASH_INPUT_BYTES);

#endif // defined(DISALLOW_UNINITIALIZED_MEM_AS_ENTROPY)
        }

        // Make the 32-bit and the 64-bit version differ slightly in
        // the hope that an attacker is not always aware of the
        // architecture we're on.

#if _M_X64
        BUF1_ELM(1) += (UtilPtrToUInt32(GetProcessHeap(), 15) ^ 0x33A7968D);
#else // _M_X64
        BUF1_ELM(7) -= (ROT(GetCurrentProcessId(), 27) ^ 0x98E2E9DA);
#endif // _M_X64

        // Hash several times, depending on the CPU we're running on (the
        // return value of GetCurrentProcessorNumber() is zero based).

        for (i = (1 + min(13, GetCurrentProcessorNumber())); 0 != i; i--)
        {
            BOOLEAN const a_LastRound = (0 == i);

            SHA256MultiProcessInternal(l_NewCntxt.f_Buf0, (UINT32 const *)l_Ptr0, 1, a_LastRound);
            TigerMultiCompressInternal((UINT64 *)l_NewCntxt.f_Buf1, (UINT64 const *)l_Ptr1, 1, a_LastRound);
        }

        if (NULL != l_TempBuf)
        {
            _aligned_free(l_TempBuf);
        }

        UtilSecWipePointer(&(void *)l_TempBuf);
        UtilSecWipePointer(&l_Ptr0);
        UtilSecWipePointer(&l_Ptr1);
    }

    // Add final entropy to the buffers to reach operational status.

    // Retrieve the cursor position.

    {
        POINT l_Point;

        if (FALSE != GetCursorPos(&l_Point))
        {
            BUF0_ELM(8) += ROT(l_Point.x, 17);
            BUF1_ELM(6) ^= ROT(l_Point.y, 23);
        }

        UtilSecWipeBuffer(&l_Point, sizeof(l_Point));
    }

    // Retrieve the number of GUI objects.

    BUF0_ELM(9) -= ROT(GetGuiResources(l_CurProcess, GR_GDIOBJECTS), 15);
    BUF1_ELM(7) += ROT(GetGuiResources(l_CurProcess, GR_USEROBJECTS), 27);

// Define a temporary helper macro.
#define PROCESS_SIZE_T(x, rot)          UtilPtrToUInt32((void const *)(x), (rot))

    // Query the number of page faults that happened so far and the size
    // of the working set.

    {
        PROCESS_MEMORY_COUNTERS l_MemCounters;
        l_MemCounters.cb = sizeof(PROCESS_MEMORY_COUNTERS);

        if (FALSE != GetProcessMemoryInfo(l_CurProcess, &l_MemCounters, sizeof(PROCESS_MEMORY_COUNTERS)))
        {
            BUF0_ELM(10) ^= ROT(l_MemCounters.PageFaultCount, 25);
            BUF1_ELM(8) += PROCESS_SIZE_T(l_MemCounters.WorkingSetSize, 13);
            BUF0_ELM(11) -= PROCESS_SIZE_T(l_MemCounters.QuotaNonPagedPoolUsage, 5);
            BUF1_ELM(9) += PROCESS_SIZE_T(l_MemCounters.QuotaPagedPoolUsage, 17);
        }

        UtilSecWipeBuffer(&l_MemCounters, sizeof(l_MemCounters));
    }

    // Query memory status (system-wide).

    {
        MEMORYSTATUSEX l_MemStatus;
        l_MemStatus.dwLength = sizeof(MEMORYSTATUSEX);

        if (FALSE != GlobalMemoryStatusEx(&l_MemStatus))
        {
            BUF0_ELM(12) += UtilUInt64ToUInt32AndRotate(l_MemStatus.ullAvailPhys, 10);
            BUF1_ELM(10) ^= UtilUInt64ToUInt32AndRotate(l_MemStatus.ullAvailPageFile, 7);
            BUF1_ELM(11) -= (UtilUInt64ToUInt32AndRotate(l_MemStatus.ullAvailVirtual, 12) ^ UtilUInt64ToUInt32AndRotate(l_MemStatus.ullAvailExtendedVirtual, 23));
        }

        UtilSecWipeBuffer(&l_MemStatus, sizeof(l_MemStatus));
    }

    {
        IO_COUNTERS l_IoCounters;

        if (FALSE != GetProcessIoCounters(l_CurProcess, &l_IoCounters))
        {
            BUF0_ELM(7) += UtilUInt64ToUInt32AndRotate(l_IoCounters.ReadOperationCount, 17);
            BUF1_ELM(12) ^= UtilUInt64ToUInt32AndRotate(l_IoCounters.WriteOperationCount, 20);
            BUF1_ELM(13) -= UtilUInt64ToUInt32AndRotate(l_IoCounters.OtherOperationCount, 22);
            BUF0_ELM(13) += UtilUInt64ToUInt32AndRotate(l_IoCounters.ReadTransferCount, 9);
            BUF0_ELM(14) ^= UtilUInt64ToUInt32AndRotate(l_IoCounters.WriteTransferCount, 10);
            BUF1_ELM(14) -= UtilUInt64ToUInt32AndRotate(l_IoCounters.OtherTransferCount, 25);
        }

        UtilSecWipeBuffer(&l_IoCounters, sizeof(l_IoCounters));
    }

    // The following code tries to call NT 6.0 routines that may
    // or may not be available.

    if (NULL != g_Kernel32Module)
    {
        // GetProcessHandleCount().

        if (NULL != g_GetHandles &&
            FALSE != g_GetHandles(l_CurProcess, &l_LargeInt.LowPart))
        {
            BUF0_ELM(15) -= ROT(l_LargeInt.LowPart, 12);
        }

        // QueryProcessCycleTime().

        if (NULL != g_QueryProcessTime &&
            FALSE != g_QueryProcessTime(l_CurProcess, (PULONG64)&l_LargeInt.QuadPart))
        {
            BUF1_ELM(15) -= UtilUInt64ToUInt32AndRotate(l_LargeInt.QuadPart, 27);
        }

        // QueryThreadCycleTime().

        if (NULL != g_QueryThreadTime &&
            FALSE != g_QueryThreadTime(l_CurThread, (PULONG64)&l_LargeInt.QuadPart))
        {
            BUF0_ELM(6) += UtilUInt64ToUInt32AndRotate(l_LargeInt.QuadPart, 5);
        }

        // Another option would be GetSystemTimes() on NT 5.1 and later.
    }

    // GetPerformanceInfo() was introduced with NT 5.1 and can be found
    // in Psapi.dll (as well as Kernel32.dll from NT 6.1 on). It's only
    // used as an extra, so do nothing if we aren't able to employ it.

    if (NULL != g_GetPerfInfo)
    {
        PERFORMANCE_INFORMATION l_PerfInfo;

        ASSERT(NULL != g_PsapiMod);

        l_PerfInfo.cb = sizeof(PERFORMANCE_INFORMATION);

        if (FALSE != g_GetPerfInfo(&l_PerfInfo, sizeof(PERFORMANCE_INFORMATION)))
        {
            BUF0_ELM(0) += PROCESS_SIZE_T(l_PerfInfo.CommitTotal, 3);
            BUF1_ELM(0) ^= PROCESS_SIZE_T(l_PerfInfo.CommitLimit, 28);
            BUF0_ELM(1) -= PROCESS_SIZE_T(l_PerfInfo.CommitPeak, 23);
            BUF1_ELM(1) ^= PROCESS_SIZE_T(l_PerfInfo.PhysicalAvailable, 11);
            BUF0_ELM(2) += PROCESS_SIZE_T(l_PerfInfo.SystemCache, 20);
            BUF1_ELM(2) ^= PROCESS_SIZE_T(l_PerfInfo.KernelPaged, 19);
            BUF0_ELM(3) -= PROCESS_SIZE_T(l_PerfInfo.KernelNonpaged, 22);
            // The next two line don't provide much randomness, but
            // still include them.
            BUF1_ELM(3) += PROCESS_SIZE_T(l_PerfInfo.PageSize, 15);
            BUF0_ELM(4) -= PROCESS_SIZE_T(l_PerfInfo.PhysicalTotal, 17);
            BUF1_ELM(4) ^= ROT(l_PerfInfo.HandleCount, 3);
            BUF0_ELM(5) += ROT(l_PerfInfo.ProcessCount, 14);
            BUF1_ELM(5) -= ROT(l_PerfInfo.ThreadCount, 18);
        }

        UtilSecWipeBuffer(&l_PerfInfo, sizeof(l_PerfInfo));
    }

#undef GET_AUXIL_UINT
#undef BUF0_ELM
#undef BUF1_ELM
#undef PROCESS_SIZET

    // We are done collecting entropy. Turn the new context into the
    // global one and retrieve some random data.

    {
        ALIGN64 UINT32 l_JunkBuf[4 * ELEMENTS_PER_BLOCK];

        UtilAquireLock(&g_RandLock);
        {
            *g_EffRandContextPtr = l_NewCntxt;

            // Query up to four blocks of random data that will be
            // discarded mostly.

            RandQueryBlocksInternal(l_JunkBuf, 1 + (l_InitTurns & 0x3), FALSE, TRUE);

            g_EffRandContextPtr->f_RotSeed = l_JunkBuf[0];
            g_EffRandContextPtr->f_ScrambleMethod = l_JunkBuf[1];
        }
        UtilReleaseLock(&g_RandLock);

        UtilSecWipeBuffer(l_JunkBuf, sizeof(l_JunkBuf));
    }

    UtilSecWipeBuffer(&l_NewCntxt, sizeof(l_NewCntxt));
    UtilSecWipeUInt32(&l_Buf0BaseIndex);
    UtilSecWipeUInt32(&l_Buf1BaseIndex);
    UtilSecWipeUInt32(&l_TempAccessableSize);
    UtilSecWipeUInt32(&l_InitTurns);
    UtilSecWipeBuffer(&l_LargeInt, sizeof(l_LargeInt));
    UtilSecWipeBuffer(&l_FileTime0, sizeof(l_FileTime0));
    UtilSecWipeBuffer(&l_FileTime1, sizeof(l_FileTime1));
    UtilSecWipeBuffer(&l_FileTime2, sizeof(l_FileTime2));
    UtilSecWipeBuffer(&l_FileTimeJunk, sizeof(l_FileTimeJunk));
    UtilSecWipePointer(&(void *)l_CurProcess);
    UtilSecWipePointer(&(void *)l_CurThread);
}

#if !defined(DISALLOW_UNINITIALIZED_MEM_AS_ENTROPY)

// Restore warnings.
#pragma warning(default: 4700)

#endif // !defined(DISALLOW_UNINITIALIZED_MEM_AS_ENTROPY)

void
RandQueryInternal(METAIN OUT void * const a_Dest,
                  IN CPUWORD const a_Length,
                  IN BOOLEAN const a_DisallowZeros)
{
    // This routine is thread-safe. a_Dest does not have to meet any
    // alignment requirements. a_Length is the size in bytes.
    // a_DisallowZeros can be used to request that no output byte may
    // be zero.

    UINT32 * l_Dest = a_Dest;
    CPUWORD l_WholeBlocks = 0;
    CPUWORD l_StartingSlack = 0;
    CPUWORD l_EndingBytes;

    if (0 == a_Length)
    {
        return;
    }

    if (BYTES_PER_BLOCK > a_Length)
    {
        l_EndingBytes = a_Length;
    }
    else
    {
        CPUWORD const l_Rem = ((CPUWORD)l_Dest % sizeof(CPUWORD));
        CPUWORD l_EffLen = a_Length;

        if (0 != l_Rem)
        {
            l_StartingSlack = (sizeof(CPUWORD) - l_Rem);
            l_Dest = (UINT32 *)((BYTE *)l_Dest + l_StartingSlack);
            l_EffLen -= l_StartingSlack;
        }

        if (BYTES_PER_BLOCK > l_EffLen)
        {
            l_EndingBytes = l_EffLen;
        }
        else
        {
            l_WholeBlocks = (l_EffLen / BYTES_PER_BLOCK);
            l_EndingBytes = (l_EffLen % BYTES_PER_BLOCK);
        }
    }

    ASSERT(BYTES_PER_BLOCK > l_StartingSlack &&
           BYTES_PER_BLOCK > l_EndingBytes &&
           a_Length >= (l_StartingSlack + l_EndingBytes));

    {
        BOOLEAN const l_JustWholeBlocks = (0 == l_StartingSlack &&
                                           0 == l_EndingBytes);
        CPUWORD const l_BlockRem = (BYTES_PER_BLOCK - l_StartingSlack);
        BOOLEAN const l_OneStackBlockSuffices = (l_BlockRem >= l_EndingBytes);
        BOOLEAN l_Exl = TRUE;

        // We will keep the time that we're holding the lock as short
        // as possible.

        UtilAquireLock(&g_RandLock);

        if (0 != l_WholeBlocks)
        {
            l_Dest = RandQueryBlocksInternal(l_Dest, l_WholeBlocks, a_DisallowZeros, l_JustWholeBlocks);
        }

        if (FALSE == l_JustWholeBlocks)
        {
            ALIGN64 UINT32 l_StackBlock[2 * ELEMENTS_PER_BLOCK];

            RandQueryBlocksInternal(l_StackBlock, FALSE == l_OneStackBlockSuffices ? 2 : 1, a_DisallowZeros, TRUE);

            UtilReleaseLock(&g_RandLock);
            l_Exl = FALSE;

            if (0 != l_StartingSlack)
            {
                BYTE const * const l_EffSrcBlock = (BYTE *)(FALSE == l_OneStackBlockSuffices ? &l_StackBlock[ELEMENTS_PER_BLOCK] : l_StackBlock);

                // Handle the starting slack first.
                memcpy(a_Dest, l_EffSrcBlock, l_StartingSlack);

                if (FALSE == l_OneStackBlockSuffices)
                {
                    // Copy the remaining part of the second stack
                    // block to the start of the ending slack.

                    CPUWORD const l_FirstPortion = (l_EndingBytes / 2);

                    ASSERT(l_BlockRem >= l_FirstPortion);

                    memcpy(l_Dest, &l_EffSrcBlock[l_StartingSlack], l_FirstPortion);

                    l_Dest = (UINT32 *)((BYTE *)l_Dest + l_FirstPortion);
                    l_EndingBytes -= l_FirstPortion;
                }
            }

            if (0 != l_EndingBytes)
            {
                memcpy(l_Dest, l_StackBlock, l_EndingBytes);
            }

            UtilSecWipeBuffer(l_StackBlock, sizeof(l_StackBlock));
        }

        if (FALSE != l_Exl)
        {
            UtilReleaseLock(&g_RandLock);
        }
    }
}

// Exports (thread-safe).

EXPORT void CALLCONV
RandomReseed_EXPORT()
{
    RandReseedInternal();
}

EXPORT void CALLCONV
RandomQuery_EXPORT(METAIN OUT void * const a_Dest,
                   IN CPUWORD const a_Length,
                   IN BOOL const a_DisallowZeros)
{
    // NOTE: It's often misquoted, but when Bruce Schneier killed
    // Julius Caesar for promoting weak cryptography he actually said,
    // "Et tu, Bruce?"

    RandQueryInternal(a_Dest, a_Length, 0 != a_DisallowZeros);
}

// End of file.
