/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.27-src\Native\Socket.c
 * PURPOSE: Stream-oriented connection status checking
 * VERSION: 1.4.27
 * DATE: March 23, 2016
 * COPYRIGHT: Copyright (C) 2012-2016 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

#define INCLUDE_WINDOWS
#include "general.h"

#include <Winsock2.h>

// This interface allows to check if a connection-oriented socket
// is (still) connected.

// Result values used by this interface.

typedef UINT32 CHECK_CONN_RESULT;

// Failed to retrieve the status of the connection.
#define CONN_STAT_ERR               (0)
// The connection was closed.
#define CONN_STAT_DEAD              (1)
// The connection might be usable.
#define CONN_STAT_POSSIBLY_ALIVE    (2)

// Macros and structures used internally.

// Inofficial success macros for winsock.
#define SOCKET_SUCCESS              (0)
#define WSASUCCESS                  (0)

typedef struct _timeval_custom
{
    long tv_sec;
    long tv_usec;
}
timeval_custom;

typedef struct _fd_set_custom
{
    u_int fd_count;
    SOCKET fd_array[1];
}
fd_set_custom;

typedef struct _WSAPOLLFD_CUSTOM
{
  SOCKET fd;
  SHORT events;
  SHORT revents;
}
WSAPOLLFD_CUSTOM;

// Prototypes.

typedef INT WSAAPI WSAPollFunc(IN OUT WSAPOLLFD_CUSTOM * fdarray,
                               IN ULONG nfds,
                               IN INT timeout);

// Globals (all can be zero).

static BOOLEAN g_WSAStartupSucc = FALSE;
static HMODULE g_Ws2_32Module = NULL;
static WSAPollFunc * g_WSAPoll = NULL;

BOOLEAN
SocketInitOrShutdownInternal(IN BOOLEAN const a_Init)
{
    // This routine is used for initialization and shutdown, regarding
    // global data. Returns a nonzero value if the desired final state
    // was reached.

    if (FALSE == a_Init)
    {
        if (FALSE == g_WSAStartupSucc &&
            NULL == g_Ws2_32Module)
        {
            return TRUE;
        }
    }
    else
    {
        // We always require g_WSAStartupSucc to become nonzero
        // for a successful initilization, but we don't require
        // g_Ws2_32Module to become nonzero.

        if (FALSE == g_WSAStartupSucc)
        {
            // Version 2.2 dates back to May 1996.
            WORD const l_ReqVer = MAKEWORD(2, 2);
            WSADATA l_Data;
            INT32 l_StartupRes = WSAStartup(l_ReqVer, &l_Data);

            if (WSASUCCESS != l_StartupRes ||
                l_ReqVer != l_Data.wVersion)
            {
                return FALSE;
            }

            g_WSAStartupSucc = TRUE;
        }

        if (NULL != g_Ws2_32Module)
        {
            // We're already up.

            goto INITIALIZED;
        }

        g_Ws2_32Module = LoadLibrary(L"Ws2_32.dll");

        if (NULL == g_Ws2_32Module)
        {
            goto INITIALIZED;
        }

        // The following function is exported by Ws2_32.dll
        // on NT 6.0 and later.
        g_WSAPoll = (WSAPollFunc *)GetProcAddress(g_Ws2_32Module, "WSAPoll");

        if (NULL == g_WSAPoll)
        {
            // Nice try. We will still return a nonzero value.

            BOOL const l_Succ = FreeLibrary(g_Ws2_32Module);

            ASSERT(FALSE != l_Succ);
            UNUSED_STACK_ITEM(l_Succ);

            g_Ws2_32Module = NULL;
        }

        INITIALIZED:

        return TRUE;
    }

    ASSERT(FALSE != g_WSAStartupSucc ||
           NULL != g_Ws2_32Module);

    if (FALSE != g_WSAStartupSucc)
    {
        // Terminate the use of services.

        INT32 l_CleanupRes = WSACleanup();

        ASSERT(WSASUCCESS == l_CleanupRes);
        UNUSED_STACK_ITEM(l_CleanupRes);

        g_WSAStartupSucc = FALSE;
    }

    if (NULL != g_Ws2_32Module)
    {
        // Free the library.

        BOOL const l_Succ = FreeLibrary(g_Ws2_32Module);

        ASSERT(FALSE != l_Succ);
        UNUSED_STACK_ITEM(l_Succ);

        g_Ws2_32Module = NULL;
        g_WSAPoll = NULL;
    }

    return !a_Init;
}

static CHECK_CONN_RESULT
TranslateErrorCode()
{
    // We assumed an error occurred before this routine gets called,
    // so this routine uses WSAGetLastError() right away.

    INT32 l_LastErr = WSAGetLastError();

    if (WSAENOTSOCK == l_LastErr)
    {
        // a_Socket is not a valid socket.

        return CONN_STAT_DEAD;
    }

    // Any other error.

    return CONN_STAT_ERR;
}

static CHECK_CONN_RESULT
SocketCheckConnectionStatusNT6Internal(IN SOCKET const a_Socket)
{
    // This routine is thread-safe.

    SHORT const l_PollInAndOut = 0x310;
    SHORT const l_QueryFailed = 0x5;
    SHORT const l_Disconnected = 0x2;
    WSAPOLLFD_CUSTOM l_Struct;
    INT32 l_PollRes;

    l_Struct.fd = a_Socket;
    l_Struct.events = l_PollInAndOut;
    l_Struct.revents = 0;

    // A return value of 1 indicates that the l_Struct.revents was set.

    l_PollRes = g_WSAPoll(&l_Struct, 1, 0);

    if (1 != l_PollRes)
    {
        if (0 == l_PollRes)
        {
            // The socket seems to be fine.

            return CONN_STAT_POSSIBLY_ALIVE;
        }

        // Although the documentation doesn't mention this, the extended
        // error code can be WSAENOTSOCK.

        return TranslateErrorCode();
    }

    if (0 != (l_Struct.revents & l_QueryFailed))
    {
        // POLLERR and/or POLLNVAL.

        return CONN_STAT_DEAD;
    }

    // Check for POLLHUP. POLLHUP can become set already while reading
    // is still legal and possible.

    return (0 == (l_Struct.revents & l_Disconnected) ? CONN_STAT_POSSIBLY_ALIVE : CONN_STAT_DEAD);
}

EXPORT CHECK_CONN_RESULT CALLCONV
SocketCheckConnectionStatus_EXPORT(IN SOCKET const a_Socket)
{
    // This routine is not thread-safe, in the sense that this
    // routine may be safely invoked while the passed socket is also
    // used by other threads in parallel, but the return value of
    // this routine will not be reliable in that case.

    fd_set_custom l_SocketArr;
    timeval_custom l_Timeout;
    INT32 l_SelectedSockets;
    ULONG l_ReadableBytes;

    if (FALSE == g_WSAStartupSucc)
    {
        return CONN_STAT_ERR;
    }

    if (NULL != g_WSAPoll)
    {
        // We're on NT 6.0 or later it seems. The advantage of
        // SocketCheckConnectionStatusNT6Internal() is that it's fully
        // thread-safe and requires only one syscall. However, this
        // gives us slightly different results, since WSAPoll() can
        // report a socket as disconnected/aborted several milliseconds
        // earlier than select() and ioctlsocket().

        UINT32 l_Res = SocketCheckConnectionStatusNT6Internal(a_Socket);

        if (CONN_STAT_ERR != l_Res)
        {
            return l_Res;
        }
    }

    l_SocketArr.fd_count = 1;
    l_SocketArr.fd_array[0] = a_Socket;
    l_Timeout.tv_usec = l_Timeout.tv_sec = 0;

    l_SelectedSockets = select(0, (fd_set *)&l_SocketArr, NULL, NULL, (struct timeval const *)&l_Timeout);

    if (1 != l_SelectedSockets)
    {
        // The socket is not ready for reading.

        if (0 == l_SelectedSockets)
        {
            // The connection might be alive however.

            return CONN_STAT_POSSIBLY_ALIVE;
        }

        return TranslateErrorCode();
    }

    if (SOCKET_SUCCESS != ioctlsocket(a_Socket, FIONREAD, &l_ReadableBytes))
    {
        // We weren't able to retrieve the number of bytes ready
        // for reading.

        return TranslateErrorCode();
    }

    // If there are no bytes ready for reading, then this connection
    // is dead by now. Otherwise it should be alive.

    return (0 == l_ReadableBytes ? CONN_STAT_DEAD : CONN_STAT_POSSIBLY_ALIVE);
}

// End of file.
