/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.27-src\Support\LargeUInt.cs
 * PURPOSE: 128 bit integer
 * VERSION: 1.4.27
 * DATE: March 23, 2016
 * COPYRIGHT: Copyright (C) 2012-2016 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Text;
using Support.Encodings;

namespace Support
{
#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    public struct UInt128Value
    {
        public readonly ulong f_ULong0;
        public readonly ulong f_ULong1;

        public UInt128Value(ulong a_ULong0,
                            ulong a_ULong1)
        {
            f_ULong0 = a_ULong0;
            f_ULong1 = a_ULong1;
        }

        public int CalcHashCode()
        {
            return UInt128.CalcHashCode((uint)(f_ULong0 & uint.MaxValue),
                                        (uint)((f_ULong0 >> Util.c_BitsPerUInt) & uint.MaxValue),
                                        (uint)(f_ULong1 & uint.MaxValue),
                                        (uint)((f_ULong1 >> Util.c_BitsPerUInt) & uint.MaxValue));
        }

        public override bool Equals(object obj)
        {
            // Warn us if someone operates non-generic.
            throw (new ApplicationException());
        }

        public bool ValueEquals(ref UInt128Value a_Other)
        {
            return (this.f_ULong0 == a_Other.f_ULong0 &&
                    this.f_ULong1 == a_Other.f_ULong1);
        }

        public override int GetHashCode()
        {
            return CalcHashCode();
        }

        public void ToBufferLE(BufferIO a_Buffer)
        {
            a_Buffer.WriteUInt64LE(f_ULong0);
            a_Buffer.WriteUInt64LE(f_ULong1);
        }

        public static void FromBufferLE(BufferIO a_Buffer,
                                        out UInt128Value a_Dest)
        {
            a_Dest = new UInt128Value(a_Buffer.ReadUInt64LE(), a_Buffer.ReadUInt64LE());
        }

        public UInt128 ToRef()
        {
            return new UInt128((uint)f_ULong0,
                               (uint)(f_ULong0 >> Util.c_BitsPerUInt),
                               (uint)f_ULong1,
                               (uint)(f_ULong1 >> Util.c_BitsPerUInt));
        }
    }

    public class UInt128 : IEquatable<UInt128>, IComparable, IComparable<UInt128>
    {
        public const byte c_BitSize = 128;
        public const byte c_ByteSize = 16;
        public const byte c_UIntSize = 4;

        private const byte c_BitsPerUInt = Util.c_BitsPerUInt;
        private const uint c_Bit31Mask = (1u << (c_BitsPerUInt - 1));

        private static readonly byte s_HashRotateBits0;
        private static readonly byte s_HashRotateBits1;
        private static readonly byte s_HashRotateBits2;
        private static readonly byte s_HashRotateBits3;
        public static readonly uint s_HashSeedUInt0;
        public static readonly uint s_HashSeedUInt1;
        public static readonly uint s_HashSeedUInt2;
        public static readonly uint s_HashSeedUInt3;

        public static readonly UInt128 s_Zero;

        // NOTE: This is a little endian data structure,
        // so f_UInt0 is the lowest and f_UInt3 the highest
        // uint.
        private readonly uint f_UInt0;
        private readonly uint f_UInt1;
        private readonly uint f_UInt2;
        private readonly uint f_UInt3;
        private readonly int f_HashCode;

        static UInt128()
        {
            ulong l_RotSeed = Util.GetRandomUInt();
            const byte l_RotateMask = (c_BitsPerUInt - 1);
            s_HashRotateBits0 = (byte)(l_RotSeed & l_RotateMask);
            s_HashRotateBits1 = (byte)((l_RotSeed >> 8) & l_RotateMask);
            s_HashRotateBits2 = (byte)((l_RotSeed >> 16) & l_RotateMask);
            s_HashRotateBits3 = (byte)((l_RotSeed >> 24) & l_RotateMask);

            ulong l_Seed0 = Util.GetRandomULong();
            ulong l_Seed1 = Util.GetRandomULong();
            s_HashSeedUInt0 = (uint)(l_Seed0 & uint.MaxValue);
            s_HashSeedUInt1 = (uint)((l_Seed0 >> c_BitsPerUInt) & uint.MaxValue);
            s_HashSeedUInt2 = (uint)(l_Seed1 & uint.MaxValue);
            s_HashSeedUInt3 = (uint)((l_Seed1 >> c_BitsPerUInt) & uint.MaxValue);

            s_Zero = new UInt128(0, 0, 0, 0);
        }

        public UInt128(byte[] a_Src) :
                  this(a_Src, 0)
        {
        }

        public UInt128(byte[] a_Src,
                       int a_SrcIndex) :
                  this(Util.ReadUInt32LE(a_Src, a_SrcIndex), Util.ReadUInt32LE(a_Src, a_SrcIndex + 4), Util.ReadUInt32LE(a_Src, a_SrcIndex + 8), Util.ReadUInt32LE(a_Src, a_SrcIndex + 12))
        {
        }

        public UInt128(uint[] a_Src,
                       int a_SrcIndex) :
                  this(a_Src[a_SrcIndex], a_Src[a_SrcIndex + 1], a_Src[a_SrcIndex + 2], a_Src[a_SrcIndex + 3])
        {
            // Input must be in little endian notation.
        }

        public UInt128(uint a_UInt0,
                       uint a_UInt1,
                       uint a_UInt2,
                       uint a_UInt3)
        {
            f_UInt0 = a_UInt0;
            f_UInt1 = a_UInt1;
            f_UInt2 = a_UInt2;
            f_UInt3 = a_UInt3;

            f_HashCode = CalcHashCode(f_UInt0, f_UInt1, f_UInt2, f_UInt3);
        }

        public UInt128(uint a_Value) :
                  this(a_Value, 0, 0, 0)
        {
        }

        private UInt128(uint[] a_UInts) :
                   this(a_UInts[0], a_UInts[1], a_UInts[2], a_UInts[3])
        {
        }

        public static int CalcHashCode(uint a_UInt0,
                                       uint a_UInt1,
                                       uint a_UInt2,
                                       uint a_UInt3)
        {
            // This is a simplistic hashing routine.

            uint l_Code = (Util.RotateLeft(a_UInt0, s_HashRotateBits0) - s_HashSeedUInt0);
            l_Code ^= (Util.RotateLeft(a_UInt1, s_HashRotateBits1) + s_HashSeedUInt1);
            l_Code -= (Util.RotateLeft(a_UInt2, s_HashRotateBits2) ^ s_HashSeedUInt2);
            l_Code ^= (Util.RotateLeft(a_UInt3, s_HashRotateBits3) + s_HashSeedUInt3);

            return (int)l_Code;
        }

        public override bool Equals(object obj)
        {
            // Warn us if someone operates non-generic.
            throw (new ApplicationException());
        }

        public bool Equals(UInt128 a_Other)
        {
            // Tests for value equality.

            return (this.f_UInt0 == a_Other.f_UInt0 &&
                    this.f_UInt1 == a_Other.f_UInt1 &&
                    this.f_UInt2 == a_Other.f_UInt2 &&
                    this.f_UInt3 == a_Other.f_UInt3);
        }

        public bool ValueEquals(UInt128 a_Other)
        {
            return Equals(a_Other);
        }

        public bool ValueEquals(ref UInt128Value a_Other)
        {
            return (this.f_UInt0 == (uint)a_Other.f_ULong0 &&
                    this.f_UInt1 == (a_Other.f_ULong0 >> c_BitsPerUInt) &&
                    this.f_UInt2 == (uint)a_Other.f_ULong1 &&
                    this.f_UInt3 == (a_Other.f_ULong1 >> c_BitsPerUInt));
        }

        public override int GetHashCode()
        {
            return f_HashCode;
        }

        public int CompareTo(object a_Other)
        {
            return CompareTo((UInt128)a_Other);
        }

        public int CompareTo(UInt128 a_Other)
        {
            for (int i = (c_UIntSize - 1); 0 <= i; i--)
            {
                int c = this.Get32BitChunk(i).CompareTo(a_Other.Get32BitChunk(i));

                if (0 != c)
                {
                    return c;
                }
            }

            return 0;
        }

        public int CompareTo(uint a_UInt)
        {
            if (0 != f_UInt1 ||
                0 != f_UInt2 ||
                0 != f_UInt3)
            {
                return 1;
            }

            return f_UInt0.CompareTo(a_UInt);
        }

        public void ToValue(out UInt128Value a_Dest)
        {
            a_Dest = new UInt128Value(this.f_UInt0 | ((ulong)this.f_UInt1 << c_BitsPerUInt),
                                      this.f_UInt2 | ((ulong)this.f_UInt3 << c_BitsPerUInt));
        }

        public static UInt128 BuildRandom()
        {
            return Util.GetRandomUInt128();
        }

        public static UInt128 CopyPartial(UInt128 a_LowSrc,
                                          UInt128 a_HighSrc,
                                          uint a_LowBits)
        {
            // This routine will copy the lowest a_LowBits bits
            // from LowSrc and (128 - a_LowBits) highest bits
            // from HighSrc into a new UInt128.

            if (0 == a_LowBits)
            {
                return a_HighSrc;
            }

            if (c_BitSize <= a_LowBits)
            {
                return a_LowSrc;
            }

            uint l_UInt0 = MergeUInts(a_LowSrc.f_UInt0, a_HighSrc.f_UInt0, (int)a_LowBits);
            uint l_UInt1 = MergeUInts(a_LowSrc.f_UInt1, a_HighSrc.f_UInt1, (int)a_LowBits - c_BitsPerUInt);
            uint l_UInt2 = MergeUInts(a_LowSrc.f_UInt2, a_HighSrc.f_UInt2, (int)a_LowBits - Util.c_BitsPerULong);
            uint l_UInt3 = MergeUInts(a_LowSrc.f_UInt3, a_HighSrc.f_UInt3, (int)a_LowBits - 96);

            return new UInt128(l_UInt0, l_UInt1, l_UInt2, l_UInt3);
        }

        private static uint MergeUInts(uint a_LowSrc,
                                       uint a_HighSrc,
                                       int a_LowBits)
        {
            // Special helper routine for CopyPartial().

            if (0 >= a_LowBits)
            {
                return a_HighSrc;
            }

            if (c_BitsPerUInt <= a_LowBits)
            {
                return a_LowSrc;
            }

            uint l_LowMask = ((1u << a_LowBits) - 1);
            uint l_Val = ((a_LowSrc & l_LowMask) |
                          (a_HighSrc & (~l_LowMask)));

            return l_Val;
        }

        public static bool AreHighBitsEqual(UInt128 a_A,
                                            UInt128 a_B,
                                            uint a_HighBits)
        {
            if (0 == a_HighBits)
            {
                return true;
            }

            if (c_BitSize <= a_HighBits)
            {
                return a_A.Equals(a_B);
            }

            bool l_Equal = (true == SameHighBits(a_A.f_UInt3, a_B.f_UInt3, (int)a_HighBits) &&
                            true == SameHighBits(a_A.f_UInt2, a_B.f_UInt2, (int)a_HighBits - c_BitsPerUInt) &&
                            true == SameHighBits(a_A.f_UInt1, a_B.f_UInt1, (int)a_HighBits - Util.c_BitsPerULong) &&
                            true == SameHighBits(a_A.f_UInt0, a_B.f_UInt0, (int)a_HighBits - 96));

            return l_Equal;
        }

        private static bool SameHighBits(uint a_A,
                                         uint a_B,
                                         int a_HighBits)
        {
            // Special helper routine for AreHighBitsEqual().

            if (0 >= a_HighBits)
            {
                return true;
            }

            if (c_BitsPerUInt <= a_HighBits)
            {
                return (a_A == a_B);
            }

            int l_Shr = (c_BitsPerUInt - a_HighBits);
            bool l_Equal = ((a_A >> l_Shr) == (a_B >> l_Shr));

            return l_Equal;
        }

        public uint GetClearHighBitCount()
        {
            uint Count = 0;

            for (int i = (c_UIntSize - 1); 0 <= i; i--)
            {
                uint CurUInt = Get32BitChunk(i);

                if (0 == CurUInt)
                {
                    Count += c_BitsPerUInt;
                }
                else
                {
                    Count += Util.GetClearHighBits(CurUInt);

                    break;
                }
            }

            return Count;
        }

        public uint HighBitsToNum(uint a_Bits)
        {
            // Returns the n highest bits as number.

            return (f_UInt3 >> (c_BitsPerUInt - (int)a_Bits));
        }

        public bool AreHighBitsClear(uint a_Bits)
        {
            // Checks if all n high bits are clear.

            return (0 == HighBitsToNum(a_Bits));
        }

        public bool BelongsToUpperHalf()
        {
            // Checks if the highest bit is set.

            return (0 != (f_UInt3 & c_Bit31Mask));
        }

        public static UInt128 CreateForPublishing(uint a_Num,
                                                  uint a_Bits)
        {
            // Places a_Num into the highest n bits and fills the
            // remaining bits with the sequence 10000...

            uint l_UInt0 = 0;
            uint l_UInt1 = 0;
            uint l_UInt2 = 0;
            uint l_UInt3 = ((c_Bit31Mask >> (int)a_Bits) | (a_Num << (c_BitsPerUInt - (int)a_Bits)));

            return new UInt128(l_UInt0, l_UInt1, l_UInt2, l_UInt3);
        }

        public uint GetHighUInt()
        {
            return f_UInt3;
        }

        public uint Get32BitChunk(int a_Index)
        {
            switch (a_Index)
            {
                case 0: return f_UInt0;
                case 1: return f_UInt1;
                case 2: return f_UInt2;
                case 3: return f_UInt3;
                default: throw (new IndexOutOfRangeException());
            }
        }

        private bool GetBit(uint a_Index)
        {
            // 0 indicates the lowest, 127 the highest bit.

            int l_UIntIdx = (int)(a_Index / c_BitsPerUInt);
            uint l_Mask = (1u << (int)(a_Index % c_BitsPerUInt));

            return (0 != (Get32BitChunk(l_UIntIdx) & l_Mask));
        }

        private UInt128 SetBit(uint a_Index,
                               bool a_State)
        {
            // 0 indicates the lowest, 127 the highest bit.

            int l_UIntIdx = (int)(a_Index / c_BitsPerUInt);
            uint l_Mask = (1u << (int)(a_Index % c_BitsPerUInt));

            uint l_UInt0 = ManipulateUInt(this.f_UInt0, 0 == l_UIntIdx, a_State, l_Mask);
            uint l_UInt1 = ManipulateUInt(this.f_UInt1, 1 == l_UIntIdx, a_State, l_Mask);
            uint l_UInt2 = ManipulateUInt(this.f_UInt2, 2 == l_UIntIdx, a_State, l_Mask);
            uint l_UInt3 = ManipulateUInt(this.f_UInt3, 3 == l_UIntIdx, a_State, l_Mask);

            return new UInt128(l_UInt0, l_UInt1, l_UInt2, l_UInt3);
        }

        private static uint ManipulateUInt(uint a_UInt,
                                           bool a_Manipulate,
                                           bool a_NewState,
                                           uint a_Mask)
        {
            // Special helper routine for SetBit().

            if (true == a_Manipulate)
            {
                if (false == a_NewState)
                {
                    a_UInt &= ~a_Mask;
                }
                else
                {
                    a_UInt |= a_Mask;
                }
            }

            return a_UInt;
        }

        public bool GetHighBit(uint a_Index)
        {
            return GetBit(c_BitSize - 1 - a_Index);
        }

        public UInt128 SetHighBit(uint a_Index,
                                  bool a_State)
        {
            return SetBit(c_BitSize - 1 - a_Index, a_State);
        }

        public UInt128 Xor(UInt128 a_Other)
        {
            uint l_UInt0 = (this.f_UInt0 ^ a_Other.f_UInt0);
            uint l_UInt1 = (this.f_UInt1 ^ a_Other.f_UInt1);
            uint l_UInt2 = (this.f_UInt2 ^ a_Other.f_UInt2);
            uint l_UInt3 = (this.f_UInt3 ^ a_Other.f_UInt3);

            return new UInt128(l_UInt0, l_UInt1, l_UInt2, l_UInt3);
        }

        public static UInt128 FromByteArrayBE(byte[] a_Src,
                                              int a_DestIndex)
        {
            uint l_UInt0 = Util.ByteSwapUInt32(Util.ReadUInt32LE(a_Src, a_DestIndex + 12));
            uint l_UInt1 = Util.ByteSwapUInt32(Util.ReadUInt32LE(a_Src, a_DestIndex + 8));
            uint l_UInt2 = Util.ByteSwapUInt32(Util.ReadUInt32LE(a_Src, a_DestIndex + 4));
            uint l_UInt3 = Util.ByteSwapUInt32(Util.ReadUInt32LE(a_Src, a_DestIndex));

            return new UInt128(l_UInt0, l_UInt1, l_UInt2, l_UInt3);
        }

        public static UInt128 FromByteArrayME(byte[] a_Src,
                                              int a_DestIndex)
        {
            uint l_UInt0 = Util.ReadUInt32LE(a_Src, a_DestIndex + 12);
            uint l_UInt1 = Util.ReadUInt32LE(a_Src, a_DestIndex + 8);
            uint l_UInt2 = Util.ReadUInt32LE(a_Src, a_DestIndex + 4);
            uint l_UInt3 = Util.ReadUInt32LE(a_Src, a_DestIndex);

            return new UInt128(l_UInt0, l_UInt1, l_UInt2, l_UInt3);
        }

        public void ToByteArrayLE(byte[] a_Dest,
                                  int a_DestIndex)
        {
            // Little endian output.

            Util.WriteUInt32LE(f_UInt0, a_Dest, a_DestIndex);
            Util.WriteUInt32LE(f_UInt1, a_Dest, a_DestIndex + 4);
            Util.WriteUInt32LE(f_UInt2, a_Dest, a_DestIndex + 8);
            Util.WriteUInt32LE(f_UInt3, a_Dest, a_DestIndex + 12);
        }

        public void ToByteArrayBE(byte[] a_Dest,
                                  int a_DestIndex)
        {
            // Big endian output.

            ToByteArrayLE(a_Dest, a_DestIndex);
            ReverseArrayBE(a_Dest, a_DestIndex);
        }

        public void ToByteArrayME(byte[] a_Dest,
                                  int a_DestIndex)
        {
            // Mixed endian output.

            ToByteArrayLE(a_Dest, a_DestIndex);
            ReverseArrayME(a_Dest, a_DestIndex);
        }

        public byte[] ToByteArrayLE()
        {
            byte[] l_LE = new byte[c_ByteSize];

            ToByteArrayLE(l_LE, 0);

            return l_LE;
        }

        public byte[] ToByteArrayBE()
        {
            byte[] l_BE = new byte[c_ByteSize];

            ToByteArrayBE(l_BE, 0);

            return l_BE;
        }

        public byte[] ToByteArrayME()
        {
            byte[] l_ME = new byte[c_ByteSize];

            ToByteArrayME(l_ME, 0);

            return l_ME;
        }

        public uint[] AsUIntArray()
        {
            uint[] l_UInts = new uint[c_UIntSize];

            l_UInts[0] = f_UInt0;
            l_UInts[1] = f_UInt1;
            l_UInts[2] = f_UInt2;
            l_UInts[3] = f_UInt3;

            return l_UInts;
        }

        public string ToHexString()
        {
            // Output is in big endian form.

            StringBuilder Bld = new StringBuilder(c_ByteSize * Util.c_HexCharsPerByte);

            for (int i = (c_UIntSize - 1); 0 <= i; i--)
            {
                Bld.Append(Get32BitChunk(i).ToString("X8"));
            }

            return Bld.ToString();
        }

        public string ToBinaryString()
        {
            return ToBinaryString(c_BitSize);
        }

        public string ToBinaryString(uint a_HighBits)
        {
            // Writes the highest n bits into a string, with the most
            // significant bit left (big endian).

            int l_Bits = (int)a_HighBits;
            bool l_Short = (c_BitSize > l_Bits);
            char[] l_Chars = new char[l_Bits + (false == l_Short ? 0 : 3)];

            if (true == l_Short)
            {
                l_Chars[l_Bits] = '.';
                l_Chars[l_Bits + 1] = '.';
                l_Chars[l_Bits + 2] = '.';
            }

            for (int l_ChunkIdx = (c_UIntSize - 1), l_CharIdx = 0; 0 < l_Bits; l_ChunkIdx--)
            {
                uint l_Val = Get32BitChunk(l_ChunkIdx);
                int l_Rounds = Math.Min(c_BitsPerUInt, l_Bits);

                l_Bits -= l_Rounds;

                for (uint l_Mask = c_Bit31Mask; 0 < l_Rounds; l_Rounds--, l_Mask >>= 1)
                {
                    l_Chars[l_CharIdx++] = (0 == (l_Val & l_Mask) ? '0' : '1');
                }
            }

            return new string(l_Chars);
        }

        public static void ReverseArrayBE(byte[] a_Buffer,
                                          int a_Index)
        {
            Util.ReverseArray<byte>(a_Buffer, a_Index, c_ByteSize);
        }

        public static
#if FAST
        unsafe
#endif
        void ReverseArrayME(byte[] a_Buffer,
                            int a_Index)
        {
            // iMule uses a strange way to represent UInt128's
            // when exchanging Kademlia datagrams. Little endian
            // notation is used within each uint, but big endian
            // notation when it comes to the inter-relation of
            // the uints - so we might call it mixed endian (ME).

#if FAST
            if (0 == (a_Index % sizeof(uint)))
            {
                Util.VerifyRangeOrFail(a_Index, c_ByteSize, a_Buffer.Length);

                fixed (byte* l_BytePtr = a_Buffer)
                {
                    uint* l_Ptr = (uint*)&l_BytePtr[a_Index];

                    uint l_SwapTemp = l_Ptr[0];
                    l_Ptr[0] = l_Ptr[3];
                    l_Ptr[3] = l_SwapTemp;

                    l_SwapTemp = l_Ptr[1];
                    l_Ptr[1] = l_Ptr[2];
                    l_Ptr[2] = l_SwapTemp;
                }

                return;
            }
#endif

            uint l_Temp = Util.ReadUInt32LE(a_Buffer, a_Index);
            Util.WriteUInt32LE(Util.ReadUInt32LE(a_Buffer, a_Index + 12), a_Buffer, a_Index);
            Util.WriteUInt32LE(l_Temp, a_Buffer, a_Index + 12);

            l_Temp = Util.ReadUInt32LE(a_Buffer, a_Index + 4);
            Util.WriteUInt32LE(Util.ReadUInt32LE(a_Buffer, a_Index + 8), a_Buffer, a_Index + 4);
            Util.WriteUInt32LE(l_Temp, a_Buffer, a_Index + 8);
        }
    }
}

// End of file.
