﻿// Library.cs
// ------------------------------------------------------------------
//
// Copyright (C) 2012-2015 Stephan A. Röger. All rights reserved.
//
// ------------------------------------------------------------------
//
// This file is licensed under the Microsoft Public License.
// See the file License.txt for the license details.
//

using System;
using System.Diagnostics;
using IonicZlib;

namespace zlibManaged
{
    // This is provided for legacy reasons and will be removed in
    // the future.
    public enum CompressionLevels : byte
    {
        Z_NO_COMPRESSION = 0,
        Z_BEST_SPEED = 1,
        Z_DEFAULT_COMPRESSION = 6,
        Z_BEST_COMPRESSION = 9,
    }

    public enum CompressionLevel : byte
    {
        Z_NO_COMPRESSION = 0,
        Z_BEST_SPEED = 1,
        Z_DEFAULT_COMPRESSION = 6,
        Z_BEST_COMPRESSION = 9,
    }

    public static class zlib
    {
        // These are arbitrary limits.
        public const int c_MaxUncompressedSize = (2 * 512 * 1024 * 1024);
        public const int c_MaxCompressedSize = 0x4005002D;

        private const byte c_WindowSize = 15;

        public static int compressBound(int a_UncompressedByteLen)
        {
            // Returns the maximum output size of the compress2()
            // routine for a certain input size.

            uint l_Len = (uint)a_UncompressedByteLen;

            if (c_MaxUncompressedSize < l_Len)
            {
                throw (new ArgumentOutOfRangeException(string.Format(SharedUtils.invariantCulture, "The input parameter ({0}) exceeds the limit of this library.", a_UncompressedByteLen)));
            }

            // See zlib v1.2.6 compress.c.
            // Copyright (C) 1995-2005 Jean-loup Gailly.

            // If the default memLevel or windowBits for
            // deflateInit() is changed, then this function
            // needs to be updated.

            int l_MaxComprSize = (int)(l_Len + (l_Len >> 12) + (l_Len >> 14) + (l_Len >> 25) + 13);

            Debug.Assert(c_MaxCompressedSize >= l_MaxComprSize);

            return l_MaxComprSize;
        }

        public static bool TryCompress(ref ArraySegment<byte> a_Src,
                                       ref ArraySegment<byte> a_Dest,
                                       CompressionLevel a_Level)
        {
            // a_Src is passed by reference but will not get modified.
            // a_Dest.Array may be zero on input, in which case a buffer
            // will be created. If a_Src.Array is the same as a_Dest.Array,
            // the segments of a_Src and a_Dest must not overlap.

            // We don't need to catch any ZlibException here, since
            // a_Dest.Array cannot contain invalid data.
            return CompressCore(ref a_Src, ref a_Dest, a_Level);
        }

        public static bool TryUncompress(ref ArraySegment<byte> a_Src,
                                         ref ArraySegment<byte> a_Dest,
                                         int a_MaxExpectedDestSize)
        {
            // a_Src is passed by reference but will not get modified.
            // a_Dest.Array may be zero on input, in which case a buffer
            // will be created. If a_Src.Array is the same as a_Dest.Array,
            // the segments of a_Src and a_Dest must not overlap. This
            // routine returns zero if the a_Src segment contains invalid
            // data for decompression.
            
            try
            {
                return UncompressCore(ref a_Src, ref a_Dest, a_MaxExpectedDestSize);
            }
            catch (ZlibException)
            {
            }

            a_Dest = new ArraySegment<byte>();

            return false;
        }
                
        public static bool compress2(out ArraySegment<byte> a_Dest,
                                     ArraySegment<byte> a_Src,
                                     CompressionLevels a_Level)
        {
            // This is a legacy routine.

            a_Dest = new ArraySegment<byte>();

            return CompressCore(ref a_Src, ref a_Dest, (CompressionLevel)a_Level);
        }

        public static bool uncompress(out ArraySegment<byte> a_Dest,
                                      ArraySegment<byte> a_Src,
                                      int a_MaxExpectedDestSize)
        {
            // This is a legacy routine.

            a_Dest = new ArraySegment<byte>();

            return UncompressCore(ref a_Src, ref a_Dest, a_MaxExpectedDestSize);
        }

        private static bool CompressCore(ref ArraySegment<byte> a_Src,
                                         ref ArraySegment<byte> a_Dest,
                                         CompressionLevel a_Level)
        {
            ZlibCodec l_Codec = CreateCodec(CompressionMode.Compress, ref a_Src, ref a_Dest, -1);
            int l_DestIdx = l_Codec.NextOut;

            DeflateManager l_DM = new DeflateManager();

            if (ZlibConstants.Z_OK == l_DM.Initialize(l_Codec, (IonicZlib.CompressionLevel)a_Level) &&
                ZlibConstants.Z_STREAM_END == l_DM.Deflate(FlushType.Finish) &&
                ZlibConstants.Z_OK == l_DM.End())
            {
                a_Dest = new ArraySegment<byte>(l_Codec.OutputBuffer, l_DestIdx, (int)l_Codec.TotalBytesOut);

                return true;
            }

            a_Dest = new ArraySegment<byte>();

            return false;
        }

        private static bool UncompressCore(ref ArraySegment<byte> a_Src,
                                           ref ArraySegment<byte> a_Dest,
                                           int a_MaxExpectedDestSize)
        {
            // a_MaxExpectedDestSize must not be negative.

            if (c_MaxCompressedSize < (uint)a_MaxExpectedDestSize)
            {
                if (0 > a_MaxExpectedDestSize)
                {
                    throw (new ArgumentException("Maximum expected output size must not be negative."));
                }

                throw (new ArgumentException("Maximum expected output size too large."));
            }

            ZlibCodec l_Codec = CreateCodec(CompressionMode.Decompress, ref a_Src, ref a_Dest, a_MaxExpectedDestSize);
            int l_DestIdx = l_Codec.NextOut;

            InflateManager l_IM = new InflateManager();

            if (ZlibConstants.Z_OK == l_IM.Initialize(l_Codec, c_WindowSize) &&
                ZlibConstants.Z_STREAM_END == l_IM.Inflate() &&
                ZlibConstants.Z_OK == l_IM.End())
            {
                a_Dest = new ArraySegment<byte>(l_Codec.OutputBuffer, l_DestIdx, (int)l_Codec.TotalBytesOut);

                return true;    
            }

            a_Dest = new ArraySegment<byte>();

            return false;
        }

        private static ZlibCodec CreateCodec(CompressionMode a_Mode,
                                             ref ArraySegment<byte> a_Src,
                                             ref ArraySegment<byte> a_Dest,
                                             int a_MaxExpectedDestSize)
        {
            // If a_Mode is CompressionMode.Compress, a_MaxExpectedDestSize
            // may be negative to indicate that this routine should pick
            // a suitable value. a_Src and a_Dest are treated as volatile.

            if (CompressionMode.Compress == a_Mode)
            {
                Debug.Assert(0 > a_MaxExpectedDestSize ||
                             c_MaxUncompressedSize >= a_MaxExpectedDestSize);

            }
            else
            {
                Debug.Assert(CompressionMode.Decompress == a_Mode &&
                             0 <= a_MaxExpectedDestSize &&
                             c_MaxCompressedSize >= a_MaxExpectedDestSize);
            }
            
            ZlibCodec l_Codec = new ZlibCodec(a_Mode);

            // l_Codec.InputBuffer may be null if l_Codec.AvailableBytesIn
            // is zero.
            l_Codec.InputBuffer = a_Src.Array;
            l_Codec.NextIn = a_Src.Offset;
            l_Codec.AvailableBytesIn = a_Src.Count;
            
            l_Codec.OutputBuffer = a_Dest.Array;
            l_Codec.NextOut = a_Dest.Offset;
            l_Codec.AvailableBytesOut = a_Dest.Count;
            
            System.Threading.Thread.MemoryBarrier();

            // NOTE: The ArraySegment struct already does a lot of
            // checking, so that we only have to check for too large
            // sizes and for input/output overlap.

            if ((CompressionMode.Compress == a_Mode ? c_MaxUncompressedSize : c_MaxCompressedSize) < l_Codec.AvailableBytesIn)
            {
                throw (new ArgumentException("Maximum input size exceeded."));
            }
            
            if (null == l_Codec.OutputBuffer)
            {
                // Create a destination buffer.

                int l_MaxArrLen = a_MaxExpectedDestSize;

                if (0 > l_MaxArrLen)
                {
                    Debug.Assert(CompressionMode.Compress == a_Mode);

                    l_MaxArrLen = compressBound(l_Codec.AvailableBytesIn);
                }

                l_Codec.OutputBuffer = new byte[l_MaxArrLen];
                l_Codec.NextOut = 0;
                l_Codec.AvailableBytesOut = l_MaxArrLen;
            }
            else
            {
                if (0 <= a_MaxExpectedDestSize)
                {
                    l_Codec.AvailableBytesOut = Math.Min(l_Codec.AvailableBytesOut, a_MaxExpectedDestSize);
                }
                
                if (l_Codec.InputBuffer == l_Codec.OutputBuffer &&
                    (l_Codec.NextIn <= l_Codec.NextOut ?
                     (l_Codec.NextIn + l_Codec.AvailableBytesIn) > l_Codec.NextOut :
                     (l_Codec.NextOut + l_Codec.AvailableBytesOut) > l_Codec.NextIn))
                {
                    // Overlap.

                    throw (new ArgumentException("Input array segment overlaps with output array segment."));
                }
            }

            return l_Codec;
        }
    }
}
