/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.27-src\Nachtblitz\Network2.cs
 * PURPOSE: Network related code (2)
 * VERSION: 1.4.27
 * DATE: March 23, 2016
 * COPYRIGHT: Copyright (C) 2012-2016 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using NB.Compression;
using NB.Content;
using NB.Gui;
using NB.I2P;
using NB.Network;
using NB.Network.eD2k;
using NB.Language;
using Support;

namespace NB.Network
{
    abstract class PeerContext : ITransferable
    {
        // May not be smaller than DonkeyClientBase.c_KillConnectionAfterIdleSecs.
        protected const uint c_CompletionBitmapUpdateIntervalSecs =
            (10 * Time.c_SecsPerMinute);

        public readonly KnownFile f_Known;
        // Always use this.f_FileID instead of KnownFile.f_FileID.
        public readonly FileID f_FileID;
        public readonly DonkeyClient f_Clnt;
        protected SourceCookie f_SrcCookie;
        private Queue<HashQueryCookie> f_HashQueryCookies;

        private bool f_Destroyed;

        protected Time f_NextComplUpdateTime;
        private Time f_LastComplUpdateTime;
        // Only valid if f_LastComplUpdateTime is not zero.
        protected BitMap f_eD2kCompl;
        // f_NativeNBCompl can be null.
        protected BitMap f_NativeNBCompl;
        private bool f_RegisteredFileAvail;
        // Always valid, updated by the eD2k thread.
        private uint f_eD2kComplParts;
        public float f_eD2kComplNum;
        public OwnerDrawContext f_AvailBarImgDrawCntxt;

        private TransferVolume f_TransferVol;
        private SpeedOMeter f_SpeedOMeter;
        private Speed f_DlSpeed;
        private Speed f_UlSpeed;

        protected DestructionInfo f_DestructionInfo;
        protected bool f_NeedToReleaseRemoteUlResources;

        public string f_PeersFileName;
        public FileRatings f_PeersRating;
        // f_PeersDescription can be null.
        public string f_PeersDescription;

        protected bool f_PeerHasNoContent;

        private ushort f_PrimaryTransferID;
        private ushort f_AuxiliaryTransferID;

        private Time f_MinNextSourcesQueryTime;

        protected uint p_TransferID
        {
            get { return f_PrimaryTransferID; }
        }

        protected uint p_AuxiliaryTransferID
        {
            get { return f_AuxiliaryTransferID; }
        }

        protected bool p_IsDl
        {
            get { return (this is DlPeerContext); }
        }

        public bool p_Destroyed
        {
            get { return f_Destroyed; }
        }

        public ContentManager p_ContentMngr
        {
            get { return f_Clnt.f_Inst.f_ContentMngr; }
        }

        protected PeerContext(KnownFile a_Known,
                              FileID a_FileID,
                              DonkeyClient a_Clnt)
        {
            f_Known = a_Known;
            f_FileID = a_FileID;
            f_Clnt = a_Clnt;
            f_HashQueryCookies = new Queue<HashQueryCookie>();
            SpeedOMeter.Create(out f_SpeedOMeter);
        }

        protected void InitializeBase()
        {
            if (ComplStatus.Incomplete == f_Known.p_StatusVolatile)
            {
                f_SrcCookie = new SourceCookie(f_Clnt.p_UserCntxt.f_UserNum);

                if (false == p_ContentMngr.RegisterSourceAsync(f_Known, f_SrcCookie))
                {
                    f_SrcCookie = null;
                }
            }
        }

        public void Destroy()
        {
            // May only be used by the KnownFile class.

            DestroyChild();

            ClearFileStatus();

            if (null != f_SrcCookie)
            {
                f_SrcCookie.SetPeerGone();
                p_ContentMngr.UnregisterSourceAsync(f_Known, f_SrcCookie);
            }

            f_Destroyed = true;
        }

        protected void ClearTransferID()
        {
            f_AuxiliaryTransferID = f_PrimaryTransferID = DonkeyClient.c_NoTransferID;
        }

        protected bool TrySetNewTransferID()
        {
            // Returns zero if the peer does not support transfer ids
            // (if GetNextTransferID() failed).

            f_PrimaryTransferID = f_Clnt.GetNextTransferID();

            if (DonkeyClient.c_NoTransferID == f_PrimaryTransferID)
            {
                f_AuxiliaryTransferID = DonkeyClient.c_NoTransferID;

                return false;
            }

            f_AuxiliaryTransferID = f_Clnt.GetNextTransferID();

            return true;
        }

        protected void SetTransferID(uint a_PrimTransferID,
                                     uint a_AuxTransferID)
        {
            f_PrimaryTransferID = (ushort)a_PrimTransferID;
            f_AuxiliaryTransferID = (ushort)a_AuxTransferID;
        }

        protected void MutateTransferID(bool a_ActiveRole)
        {
            f_PrimaryTransferID = f_AuxiliaryTransferID;
            f_AuxiliaryTransferID = (false == a_ActiveRole ? DonkeyClient.c_NoTransferID : f_Clnt.GetNextTransferID());
        }

        protected bool HasTransferID(uint a_TransferID)
        {
            return (a_TransferID == f_PrimaryTransferID);
        }

        public ObtainCookie CreateObtainCookie(int a_MaxPiecesPerGroup,
                                               out BitMap a_ComplMapClone)
        {
            // Return can be null.

            if (null == f_SrcCookie ||
                null == f_eD2kCompl)
            {
                a_ComplMapClone = null;

                return null;
            }

            a_ComplMapClone = (null == f_NativeNBCompl ? f_eD2kCompl : f_NativeNBCompl);

            if (true == a_ComplMapClone.p_AllClear)
            {
                return null;
            }

            a_ComplMapClone = a_ComplMapClone.Clone();

            return new ObtainCookie(f_SrcCookie, a_MaxPiecesPerGroup);
        }

        public bool HasFileID(FileID a_Other)
        {
            return f_FileID.f_Val.Equals(a_Other.f_Val);
        }

        public ReadCookie ReadAsync(KnownFile a_Known,
                                    ref FileRange a_Range)
        {
            // Return can be null.

            return f_Clnt.f_Inst.f_ContentMngr.ReadAsync(a_Known, (null != f_SrcCookie), ShareKinds.eD2k_on_I2P, a_Range.f_StartingOffset, (int)a_Range.f_Length);
        }

        protected abstract void DestroyChild();

        public bool Process(Time a_Time)
        {
            // Returns false if the client should be closed.

            if (DestructionInfo.Keep != f_DestructionInfo)
            {
                return true;
            }

            if (null != f_SrcCookie)
            {
                switch (f_SrcCookie.p_ContentMngrStateVolatile)
                {
                    case CookieStates.Success:
                    case CookieStates.WaitingForHashes:
                        if (CookieStates.Pending == f_SrcCookie.p_PeerState)
                        {
                            f_SrcCookie.SetPeerAlive();
                        }
                        break;

                    case CookieStates.Pending:
                        return true;

                    case CookieStates.Failure:
                        if (false == p_IsDl &&
                            true == f_Known.p_eD2kCompleted)
                        {
                            // This file got completed just now,
                            // (probably) thus our source cookie was
                            // rejected.

                            f_SrcCookie = null;
                            break;
                        }
                        MarkAsRemovable(SourceStates.UnspecifiedContentMngrProblem, DestructionInfo.RemoveAtOurRequestPeerNeedsInfo);
                        return true;

                    default: throw (new ApplicationException());
                }
            }

            bool l_IsDl = p_IsDl;

            // Check access to uploadable content.

            if (false == l_IsDl)
            {
                if (false == f_Known.IsUlWorthy() ||
                    ContentManager.AccessLevel.AccessGranted != f_Clnt.f_Inst.f_ContentMngr.CheckUlAccess(f_Known, ShareKinds.eD2k_on_I2P, f_Clnt.p_UserCntxt))
                {
                    MarkAsRemovable(SourceStates.Undefined, DestructionInfo.RemoveAtOurRequestPeerNeedsInfo);

                    return true;
                }
            }
            else
            {
                if (DownloadControl.DlEnabled != f_Known.f_DlControl)
                {
                    MarkAsRemovable(SourceStates.AllowedBandwidthTooLow, DestructionInfo.RemoveAtOurRequestPeerNeedsInfo);

                    return true;
                }

                if (true == f_PeerHasNoContent &&
                    0 == f_HashQueryCookies.Count)
                {
                    MarkAsRemovable(SourceStates.NoRelevantRange, DestructionInfo.RemoveAtOurRequestPeerNeedsInfo);

                    return true;
                }
            }

            // Try to update the file status of the peer.

            if (f_NextComplUpdateTime <= a_Time)
            {
                if (true == l_IsDl ||
                    (false == f_NeedToReleaseRemoteUlResources &&
                     false == f_Clnt.p_UserCntxt.HasA4AFLimitation() &&
                     false == f_Known.p_ActCntxt.HasDlPeerContext(f_Clnt.p_Addr)))
                {
                    if (false == f_Clnt.RequestFileStatusOrComplBitmap(this, !l_IsDl))
                    {
                        return false;
                    }

                    if (false == l_IsDl)
                    {
                        f_NeedToReleaseRemoteUlResources = true;
                    }
                }

                f_NextComplUpdateTime = a_Time.AddSeconds(c_CompletionBitmapUpdateIntervalSecs);
            }

            if (false == HandleHashQueries())
            {
                return false;
            }

            return ProcessChild(a_Time);
        }

        public bool ShouldBeRemoved()
        {
            return (DestructionInfo.Keep != f_DestructionInfo);
        }

        // This routine may be called at any time (even while
        // processing).
        public abstract void MarkAsRemovable(SourceStates a_SourceState,
                                             DestructionInfo a_DestructionInfo);

        protected bool PrepareDestruction(DestructionInfo a_DestructionInfo)
        {
            // Returns a nonzero value if the destruction info changed.

            if (DestructionInfo.Keep != f_DestructionInfo &&
                DestructionInfo.RemoveAtOurRequestPeerNeedsInfo != f_DestructionInfo &&
                a_DestructionInfo == f_DestructionInfo)
            {
                return false;
            }

            f_DestructionInfo = a_DestructionInfo;

            return true;
        }

        public bool AddHashQuery(HashQueryCookie a_Cookie)
        {
            if (HashQueryCookie.c_SuggestedMaxSize < a_Cookie.f_EstimatedResultSize)
            {
                return false;
            }

            bool l_Succ;
            P2PHashSets l_HashType = a_Cookie.f_HashType;

            if (P2PHashSets.eD2kBitmap == l_HashType ||
                P2PHashSets.ClientNativeBitmap == l_HashType)
            {
                // Handle this right here, since the content manager
                // isn't necessary for it.

                bool l_Native = (P2PHashSets.ClientNativeBitmap == l_HashType);
                BitMap l_Map = f_Known.p_ActCntxt.ExposeOwnComplBitmap(l_Native);
                byte[] l_Data = null;
                l_Succ = l_Map.CopyToBufferWithCheck(a_Cookie.f_StartingIndex, a_Cookie.f_Count, ref l_Data, false == l_Native ? 0 : sizeof(uint));

                if (true == l_Succ)
                {
                    if (true == l_Native)
                    {
                        // Prepend the granularity.

                        Util.WriteUInt32LE(l_Map.f_Granularity, l_Data, 0);
                    }

                    a_Cookie.Succeed(l_Data);
                }
            }
            else
            {
                l_Succ = p_ContentMngr.NetworkQueryHashes(a_Cookie);
            }

            if (true == l_Succ)
            {
                f_HashQueryCookies.Enqueue(a_Cookie);
            }

            return l_Succ;
        }

        private bool HandleHashQueries()
        {
            while (0 < f_HashQueryCookies.Count)
            {
                HashQueryCookie l_Cookie = f_HashQueryCookies.Peek();
                CookieStates l_State = l_Cookie.p_StateVolatile;

                if (CookieStates.Pending == l_State)
                {
                    break;
                }

                f_HashQueryCookies.Dequeue();

                // Regardless of success or failure, pass the cookie
                // down to the client.

                return f_Clnt.ProcessHashQuery(this, l_Cookie);
            }

            return true;
        }

        protected abstract bool ProcessChild(Time a_Time);

        private void ClearFileStatus()
        {
            if (true == f_LastComplUpdateTime.IsSet())
            {
                UnregisterFromFileAvail();

                f_LastComplUpdateTime = new Time();
                f_eD2kCompl = null;
                f_NativeNBCompl = null;

                f_eD2kComplParts = 0;
                f_eD2kComplNum = 0;

                f_AvailBarImgDrawCntxt = null;
            }
        }

        private void UnregisterFromFileAvail()
        {
            if (true == f_RegisteredFileAvail)
            {
                f_Known.p_ActCntxt.ChangeAvailability(f_eD2kCompl, null);
                f_RegisteredFileAvail = false;
            }
        }

        protected bool NotifyFileStatusBase(Time a_Time,
                                            BitMap a_Compl)
        {
            f_PeerHasNoContent = a_Compl.p_AllClear;

            KnownFile.ActivityContext l_ActCntxt = f_Known.p_ActCntxt;

            if (true == f_NeedToReleaseRemoteUlResources)
            {
                f_NeedToReleaseRemoteUlResources = false;

                if (false == l_ActCntxt.HasDlPeerContext(f_Clnt.p_Addr))
                {
                    // We queried the file status of the peer and don't
                    // plan to download anything, so tell the peer that
                    // resources can now be released.

                    if (false == f_Clnt.SendDeselectDlFileMessage(f_FileID))
                    {
                        return false;
                    }

                    f_Clnt.SpareIfDlContextIsMissing();
                }
            }

            bool l_NativeInput = (BitMap.c_eD2kGranularity != a_Compl.f_Granularity);
            BitMap l_NeweD2kMap;

            if (false == l_NativeInput)
            {
                f_NativeNBCompl = null;
                l_NeweD2kMap = a_Compl;
            }
            else
            {
                f_NativeNBCompl = a_Compl;
                l_NeweD2kMap = new BitMap(BitMap.c_eD2kGranularity, f_Known.f_PartCount);
                l_NeweD2kMap.TakeOver(f_NativeNBCompl, BitMap.TakeOverMode.RoundDown);
            }

            if (false == f_RegisteredFileAvail)
            {
                // If there is a dl and and ul context, the dl context
                // is responsible alone.

                bool l_Register;

                if (false == p_IsDl)
                {
                    l_Register = !l_ActCntxt.HasDlPeerContext(f_Clnt.p_Addr);
                }
                else
                {
                    UlPeerContext l_UlPeerCntxt = l_ActCntxt.TryGetUlPeerContext(f_Clnt.p_Addr);

                    if (null != l_UlPeerCntxt)
                    {
                        l_UlPeerCntxt.UnregisterFromFileAvail();
                    }

                    l_Register = true;
                }

                if (true == l_Register)
                {
                    l_ActCntxt.ChangeAvailability(null, l_NeweD2kMap);
                    f_RegisteredFileAvail = true;
                }
            }
            else
            {
                l_ActCntxt.ChangeAvailability(f_eD2kCompl, l_NeweD2kMap);
            }

            f_NextComplUpdateTime = a_Time.AddSeconds(c_CompletionBitmapUpdateIntervalSecs);
            f_LastComplUpdateTime = a_Time;
            f_eD2kCompl = l_NeweD2kMap;
            f_eD2kComplParts = f_eD2kCompl.p_SetBitCount;

            if (false == a_Compl.p_AllSet)
            {
                f_eD2kComplNum = IncomplFileProgress.CalcQuotient(a_Compl.p_SetBitCount, a_Compl.f_BitCount);
            }
            else
            {
                f_eD2kComplNum = 1;
                f_Known.f_LastSeenComplete = a_Time;
            }

            MainForm l_Form = f_Clnt.f_Inst.f_MainForm;
            OwnerDrawContext l_DrawCntxt = new OwnerDrawContext(l_Form.f_BarDrawer.DrawCompletenessBarImageScaled(a_Compl, null, l_Form.f_GridViewLineHeight));
            Util.VolatileWrite(out f_AvailBarImgDrawCntxt, l_DrawCntxt);

            return true;
        }

        public bool NotifyFileName(string a_FileName)
        {
            string l_NewFileName = a_FileName;

            if (0 == l_NewFileName.Length ||
                FileNames.c_MaxFileNameChars < l_NewFileName.Length)
            {
                f_PeersFileName = null;
            }
            else
            {
                f_PeersFileName = a_FileName;
            }

            return true;
        }

        public bool NotifyFileComment(FileRatings a_Rating,
                                      string a_Description)
        {
            // a_Description may be null.

            f_PeersRating = a_Rating;
            f_PeersDescription = a_Description;

            return true;
        }

        public bool RequestSources(Time a_Time)
        {
            if (a_Time < f_MinNextSourcesQueryTime)
            {
                return false;
            }

            if (false == f_Clnt.RequestSources(f_FileID))
            {
                f_Clnt.Close();

                return false;
            }

            f_MinNextSourcesQueryTime = a_Time.AddSeconds(DownloadQueue.c_SourceQueryIntervalSecs);

            return true;
        }

        public bool NotifySources(I2PAddress[] a_StreamAddresses)
        {
            f_Clnt.f_Inst.f_NetworkMngr.f_DlQueue.NotifySourceRange(f_Known, a_StreamAddresses);

            return true;
        }

        public void NotifySentContentBytes(uint a_Size)
        {
            // We can land here after Destroy() was called.

            if (false == f_Destroyed)
            {
                f_TransferVol.Sent(a_Size);
            }

            f_Known.SentContent(a_Size);
        }

        public void NotifyReceivedContentBytes(uint a_Size)
        {
            // We can land here after Destroy() was called.

            if (false == f_Destroyed)
            {
                f_TransferVol.Received(a_Size);
            }

            f_Known.RcvdContent(a_Size);
        }

        public void MeasureSpeed(Time a_Time,
                                 bool a_AlsoCalcSpeed)
        {
            f_SpeedOMeter.SaveProgress(a_Time, ref f_TransferVol);

            if (false == a_AlsoCalcSpeed)
            {
                f_DlSpeed = Speed.s_UnknownSpeed;
                f_UlSpeed = Speed.s_UnknownSpeed;
            }
            else
            {
                f_SpeedOMeter.GetSpeed(a_Time, out f_DlSpeed, out f_UlSpeed);
            }
        }

        public void CopyTransferVolumeVolatile(out TransferVolume a_Dest)
        {
            f_TransferVol.CopyToVolatile(out a_Dest);
        }

        public void GetSpeed(out Speed a_DlSpeed,
                             out Speed a_UlSpeed)
        {
            // Thread safe.

            Speed.CopyVolatile(ref f_DlSpeed, out a_DlSpeed);
            Speed.CopyVolatile(ref f_UlSpeed, out a_UlSpeed);
        }

        public void AddAsSource(Dictionary<UserContext, KnownFile.Source> a_Dic,
                                PeerContext a_ForPeerCntxt)
        {
            if (true == this.f_Clnt.p_Addr.ValueEquals(a_ForPeerCntxt.f_Clnt.p_Addr) ||
                true == a_Dic.ContainsKey(f_Clnt.p_UserCntxt))
            {
                return;
            }

            uint l_CompletedParts = 0;

            if (null != this.f_eD2kCompl)
            {
                if (null == a_ForPeerCntxt.f_eD2kCompl)
                {
                    // Assume that the asking peer has nothing.
                    l_CompletedParts = this.f_eD2kCompl.p_SetBitCount;
                }
                else if (0 != this.f_eD2kCompl.p_SetBitCount)
                {
                    l_CompletedParts = this.f_eD2kCompl.CountRelevantBits(a_ForPeerCntxt.f_eD2kCompl);
                }
            }

            if (0 == l_CompletedParts)
            {
                // Nothing useful for a_ForPeerCntxt.

                return;
            }

            KnownFile.Source l_Source = new KnownFile.Source(l_CompletedParts, f_Clnt.p_Addr, f_Clnt.p_UserCntxt);

            a_Dic.Add(l_Source.f_UserCntxt, l_Source);

            return;
        }

#region Transfer

        public bool ReinitializeForDl(out TransferDemand a_Demand)
        {
            if (false == f_Clnt.IsDownloadingFileContent())
            {
                // This is handled elsewhere, see
                // BandwidthManager.GetStreamsForTransfer().

                a_Demand = new TransferDemand();

                return false;
            }

            return f_Clnt.ReinitializeForDl(out a_Demand);
        }

        public bool ReinitializeForUl(out TransferDemand a_Demand)
        {
            if (false == f_Clnt.IsUploadingFileContent())
            {
                // This is handled elsewhere, see
                // BandwidthManager.GetStreamsForTransfer().

                a_Demand = new TransferDemand();

                return false;
            }

            return f_Clnt.ReinitializeForUl(out a_Demand);
        }

        public uint AssignDlBandwidth(uint a_ExtraBytes)
        {
            // Ignore (trust the caller).

            return a_ExtraBytes;
        }

        public uint AssignUlBandwidth(uint a_ExtraBytes)
        {
            // Ignore (trust the caller).

            return a_ExtraBytes;
        }

        public uint Download(uint a_MaxBytes)
        {
            return f_Clnt.Download(a_MaxBytes);
        }

        public uint Upload(uint a_MaxBytes)
        {
            return f_Clnt.Upload(a_MaxBytes);
        }

#endregion

        public enum DestructionInfo : byte
        {
            Keep = 0,

            RemoveAtOurRequestPeerNeedsInfo = 1,
            RemoveAtOurRequestPeerDoesNotNeedInfo = 2,
            RemoveAtPeersRequest = 3,
        }
    }

    class DlPeerContext : PeerContext
    {
        private const uint c_RecreateObtainCookieIntervalSecs = 10;

        // c_QuerySizeThreshold should be at least 64 KiB.
        public const uint c_QuerySizeThreshold = (2 * 1024 * 1024);
        // That is 128 KiB.
        public const byte c_DefaultPiecesPerRange = 2;

        private static string[] s_DlStatusStrs;

        private readonly List<ObtainCookie> f_ObtainCookies;
        // The size of the ranges we expect the content manager to
        // acquire for us just now.
        private uint f_PlannedSize;
        // The size of the ranges the content manager has already
        // acquired for us.
        private uint f_WarrantedSize;

        private FlexArray<RequestableRange> f_Ranges;
        // The number of ranges requested from the remote peer.
        private int f_RangesSent;
        // The range the remote peer is just transmitting data of:
        private RequestableRange f_ActiveRange;
        private uint f_SentToContentMngr;

        private bool f_SourceStateNotified;
        private bool f_OutOfRelevantRanges;
        private Time f_NextObtainCookieCreationTime;

        private DlStatus f_DlStatus;
        private HashQueryProgress f_HashQueryProgress;
        private uint f_HashQueryNumber;

        private bool f_ExpectFileDescriptor;
        private bool f_GotFileDescriptor;

        private uint f_Rank;
        private AverageSpeedCalculator f_AvgSpeed;

        public uint p_AverageBytesPerSec
        {
            get { return f_AvgSpeed.p_AverageBytesPerSec; }
        }

        public DlPeerContext(KnownFile a_Known,
                             FileID a_FileID,
                             DonkeyClient a_Client) :
            base(a_Known, a_FileID, a_Client)
        {
            f_ObtainCookies = new List<ObtainCookie>(2);
            FlexArray<RequestableRange>.Create(out f_Ranges);
        }

        public static void ApplyTextSet(TextSet a_TextSet)
        {
            s_DlStatusStrs = new string[5]
            {
                a_TextSet[TextKeys.UnknownValue],
                a_TextSet[TextKeys.DlStatusWaitingForFileStatus],
                a_TextSet[TextKeys.DlStatusUploadRequested],
                a_TextSet[TextKeys.DlStatusRangesQueried],
                a_TextSet[TextKeys.DlStatusContentReceived]
            };
        }

        public void Initialize()
        {
            base.InitializeBase();

            f_Clnt.SetDlPeerContext(this);
            f_Clnt.f_Inst.f_NetworkMngr.f_DlQueue.NotifyDlPeerContextCreation(this);

            if (false == RequestFileSelectionForDl(true))
            {
                f_Clnt.Close();
            }
            else
            {
                SetDlStatus(DlStatus.WaitingForInitialFileStatus);

                // We need to see if we get the file status or if the
                // peer tells us that the file is not available.
            }
        }

        private bool RequestFileSelectionForDl(bool a_MayConstruct)
        {
            // This routine does not modify the dl status.

            if (false == f_Clnt.RequestFileStatusOrComplBitmap(this, a_MayConstruct))
            {
                return false;
            }

            f_NextComplUpdateTime = f_Clnt.f_Inst.GetTime().AddSeconds(c_CompletionBitmapUpdateIntervalSecs);

            return true;
        }

        private void ClearActiveRange()
        {
            f_ActiveRange = new RequestableRange();
            f_SentToContentMngr = 0;
            f_Known.p_ActCntxt.f_DownloadsRunning--;
        }

        private void ClearDlRegisters()
        {
            foreach (ObtainCookie l_Cookie in f_ObtainCookies)
            {
                l_Cookie.SetPeerState(CookieStates.Failure);
            }

            f_ObtainCookies.Clear();
            f_WarrantedSize = f_PlannedSize = 0;

            f_Ranges.Clear();
            f_RangesSent = 0;

            if (true == f_ActiveRange.IsValid())
            {
                ClearActiveRange();
            }

            f_OutOfRelevantRanges = false;
            f_NextObtainCookieCreationTime = new Time();
        }

        protected override void DestroyChild()
        {
            if (DestructionInfo.RemoveAtOurRequestPeerNeedsInfo == f_DestructionInfo &&
                DlStatus.WaitingForInitialFileStatus <= f_DlStatus)
            {
                if (false == f_Clnt.SendDeselectDlFileMessage(f_FileID))
                {
                    f_Clnt.Close();
                }
            }

            if (true == f_ActiveRange.IsValid())
            {
                f_ActiveRange = new RequestableRange();
                f_Known.p_ActCntxt.f_DownloadsRunning--;
            }

            foreach (ObtainCookie l_Cookie in f_ObtainCookies)
            {
                l_Cookie.SetPeerState(CookieStates.Failure);
            }

            if (false == f_SourceStateNotified)
            {
                // The context is being destroyed and we don't know why
                // exactly.

                NotifySourceState(SourceStates.ConnLoss);
            }
        }

        public bool IsTransferIDRelevant(uint a_TransferID,
                                         out bool a_Excuse)
        {
            if (true == HasTransferID(a_TransferID))
            {
                a_Excuse = false;

                return true;
            }

            a_Excuse = true;

            return false;
        }

        private void SetDlStatus(DlStatus a_DlStatus)
        {
            f_DlStatus = a_DlStatus;
        }

        protected override bool ProcessChild(Time a_Time)
        {
            // Query hashes (if any).

            if (null != f_SrcCookie)
            {
                // We assume the state was just refreshed in the parent
                // routine (and thus access without volatile semantics).
                CookieStates l_ContentMngrState = f_SrcCookie.p_ContentMngrState;

                if (CookieStates.Success == l_ContentMngrState)
                {
                    if (HashQueryProgress.SentToContentMngr == f_HashQueryProgress)
                    {
                        f_HashQueryProgress = HashQueryProgress.NoQuery;
                    }
                }
                else if (CookieStates.WaitingForHashes == l_ContentMngrState)
                {
                    if (HashQueryProgress.SentToContentMngr == f_HashQueryProgress &&
                        f_SrcCookie.p_HashQueryNumber != f_HashQueryNumber)
                    {
                        // There is another job for us right away.

                        f_HashQueryProgress = HashQueryProgress.NoQuery;
                    }

                    if (HashQueryProgress.NoQuery == f_HashQueryProgress)
                    {
                        if (false == BeginHashQuery())
                        {
                            return false;
                        }
                    }
                }
            }

            if (DlStatus.DownloadGranted <= f_DlStatus &&
                false == RequestFileRanges())
            {
                return false;
            }

            f_AvgSpeed.Update(a_Time);

            return true;
        }

        public bool BeginHashQuery()
        {
            bool l_Succ;

            if (false == f_Clnt.SupportsAdvancedFileMetaData())
            {
                // We can only query the eD2k hash set as a whole.

                l_Succ = (P2PHashSets.eD2kHashSet == f_SrcCookie.f_HashQueryParams.f_HashSet &&
                          null == f_SrcCookie.f_HashQueryParams.f_HashIndices &&
                          0 == f_SrcCookie.f_HashQueryParams.f_HashOffset &&
                          f_Known.f_PartCountHashSet == f_SrcCookie.f_HashQueryParams.f_HashCount);

                if (false == l_Succ)
                {
                    Util.DebugBreak();
                }
                else
                {
                    UInt128 l_ReqFileID = UInt128.FromByteArrayBE(f_SrcCookie.f_HashQueryParams.f_ExpectedMasterHash, 0);

                    l_Succ = f_Clnt.RequestHashSet(f_FileID);
                }
            }
            else
            {
                l_Succ = f_Clnt.RequestAdvancedFileMetaData(ref f_SrcCookie.f_HashQueryParams, 0);
            }

            f_HashQueryProgress = HashQueryProgress.WaitingForPeer;
            f_HashQueryNumber = f_SrcCookie.p_HashQueryNumber;

            return (true == l_Succ ||
                    true == AbortHashQuery(0));
        }

        public bool AbortHashQuery(uint a_TicketID)
        {
            // A null delivery will be treated as failure.

            return HashDelivery(a_TicketID, null);
        }

        public bool HashDelivery(uint a_TicketID,
                                 byte[] a_HashSet)
        {
            if (HashQueryProgress.NoQuery == f_HashQueryProgress)
            {
                return false;
            }

            if (false == f_Clnt.f_Inst.f_ContentMngr.DeliverHashesAsync(f_Known, f_SrcCookie, a_HashSet, null))
            {
                return false;
            }

            f_HashQueryProgress = HashQueryProgress.SentToContentMngr;
            f_AvgSpeed.NotifyDownloadedBytes((uint)a_HashSet.Length);

            return true;
        }

        private void NotifySourceState(SourceStates a_SourceState)
        {
            f_SourceStateNotified = true;
            f_Clnt.f_Inst.f_NetworkMngr.f_DlQueue.NotifySourceState(f_Known, f_Clnt.p_Addr, a_SourceState);
        }

        public override void MarkAsRemovable(SourceStates a_SourceState,
                                             DestructionInfo a_DestructionInfo)
        {
            if (true == PrepareDestruction(a_DestructionInfo))
            {
                NotifySourceState(a_SourceState);

                f_Clnt.SpareIfDlContextIsMissing();
            }
        }

        public void DlQueueWantsContextDestroyed()
        {
            if (true == PrepareDestruction(DestructionInfo.RemoveAtOurRequestPeerNeedsInfo))
            {
                // We do not need to notify the status anymore.
                f_SourceStateNotified = true;

                f_Clnt.SpareIfDlContextIsMissing();
            }
        }

        public bool SetQueueRank(uint a_Rank)
        {
            if (false == NetworkManager.IsValidRank(a_Rank))
            {
                return false;
            }

            f_Rank = a_Rank;

            return true;
        }

        public bool HasPeerConfirmedFileRequest()
        {
            return (DlStatus.DownloadRequested <= f_DlStatus);
        }

        public string BuildDlStatusStrVolatile()
        {
            // This routine is used by the GUI.

            DlStatus l_DlStatus = f_DlStatus;
            uint l_Rank = f_Rank;
            Util.MemoryBarrier();

            string l_StatusStr = s_DlStatusStrs[(int)l_DlStatus];

            switch (l_DlStatus)
            {
                case DlStatus.DownloadGranted:
                    l_StatusStr = f_Clnt.f_Inst.f_TextSet.Format(l_StatusStr, l_Rank);
                    break;
            }

            return l_StatusStr;
        }

        public bool NotifyOnQueueAgain(bool a_NeedCompleteReset)
        {
            if ((DonkeyClient.c_NoTransferID == p_TransferID) != a_NeedCompleteReset)
            {
                return false;
            }

            if (true == f_SourceStateNotified ||
                DlStatus.DownloadGranted > f_DlStatus)
            {
                return false;
            }

            if (false == a_NeedCompleteReset)
            {
                MutateTransferID(true);

                if (false == f_Clnt.SendOnQueueAcknowledgement(p_TransferID, p_AuxiliaryTransferID))
                {
                    return false;
                }

                SetDlStatus(DlStatus.DownloadRequested);
            }
            else
            {
                if (false == RequestFileSelectionForDl(true))
                {
                    return false;
                }

                SetDlStatus(DlStatus.WaitingForInitialFileStatus);
            }

            ClearDlRegisters();

            return true;
        }

        public bool NotifyEndOfUpload()
        {
            if (HashQueryProgress.NoQuery != f_HashQueryProgress)
            {
                ClearDlRegisters();
                SetDlStatus(DlStatus.PreInit);

                return true;
            }

            MarkAsRemovable(SourceStates.PeerEndedUpload, DestructionInfo.RemoveAtOurRequestPeerNeedsInfo);

            return true;
        }

        public bool NotifyUploadRequestAccepted(bool a_AllowRepeat)
        {
            if (DlStatus.DownloadRequested != f_DlStatus)
            {
                // IMULETWEAK: iMule 2.3.2.x can send OP_ACCEPTUPLOADREQ
                // a second time without us sending OP_STARTUPLOADREQ
                // twice. Allow this.
                if (true == a_AllowRepeat &&
                    DlStatus.DownloadGranted <= f_DlStatus)
                {
                    return true;
                }

                return false;
            }

            if (true == RequestFileRanges())
            {
                SetDlStatus(DlStatus.DownloadGranted);

                return true;
            }

            return false;
        }

        private bool RequestFileRanges()
        {
            // Returns zero if the sending ranges failed.

            Time l_Time = f_Clnt.f_Inst.GetTime();

            // Always do this from zero upwards, to mind the order.
            for (int i = 0, lim = f_ObtainCookies.Count; i < lim; i++)
            {
                // Remove cookies that have become invalid and
                // process others.

                ObtainCookie l_Cookie = f_ObtainCookies[i];
                CookieStates l_PeerState = l_Cookie.p_PeerState;
                CookieStates l_ContentMngrState = l_Cookie.p_ContentMngrStateVolatile;
                bool l_Delete = false;

                if (CookieStates.RangeSet == l_ContentMngrState)
                {
                    if (CookieStates.Pending != l_PeerState)
                    {
                        continue;
                    }

                    AcceptNewObtainCookie(l_Cookie);
                }
                else if (CookieStates.Failure == l_ContentMngrState)
                {
                    if (CookieStates.Pending == l_PeerState)
                    {
                        f_PlannedSize -= c_QuerySizeThreshold;
                    }
                    else if (CookieStates.RangeSet == l_PeerState)
                    {
                        // The content manager lost interest.

                        DropRangesOfCookie(l_Cookie);
                    }
                    else throw (new ApplicationException());

                    l_Cookie.SetPeerState(CookieStates.Failure);
                    l_Delete = true;

                    SourceStates l_State;
                    if (false == HandleObtainCookieRejection(l_Cookie.p_RejectionReason, out l_State))
                    {
                        if (HashQueryProgress.NoQuery == f_HashQueryProgress)
                        {
                            MarkAsRemovable(l_State, DestructionInfo.RemoveAtOurRequestPeerNeedsInfo);

                            return true;
                        }
                        else
                        {
                            // Don't exit while we are retrieving a hash.

                            f_NextObtainCookieCreationTime = l_Time.AddSeconds(Time.c_SecsPerHalfMinute);
                        }
                    }
                    else
                    {
                        f_NextObtainCookieCreationTime = l_Time.AddSeconds(c_RecreateObtainCookieIntervalSecs);
                    }
                }
                else if (CookieStates.Success == l_ContentMngrState)
                {
                    l_Cookie.SetPeerState(CookieStates.Success);
                    l_Delete = true;
                }

                if (true == l_Delete)
                {
                    f_ObtainCookies.RemoveAt(i);
                    i--;
                    lim--;
                }
            }

            // Try to create a new obtain cookie.

            if (true == f_GotFileDescriptor &&
                true == f_NextObtainCookieCreationTime.IsEarlierThan(l_Time) &&
                (c_QuerySizeThreshold / 2) > (f_PlannedSize + f_WarrantedSize))
            {
                BitMap l_ComplMapClone;
                ObtainCookie l_NewCookie = CreateObtainCookie(c_DefaultPiecesPerRange, out l_ComplMapClone);

                if (null != l_NewCookie)
                {
                    bool l_Registered = p_ContentMngr.RegisterObtainCookieAsync(f_Known, l_NewCookie, c_QuerySizeThreshold, l_ComplMapClone, f_Known.p_ActCntxt.p_eD2kAvail);

                    if (true == l_Registered)
                    {
                        f_PlannedSize += c_QuerySizeThreshold;
                        f_ObtainCookies.Add(l_NewCookie);
                    }
                }
            }

            // Try to send ranges to the peer.

            int l_RangesRequested = (f_RangesSent + Util.BoolToByte(f_ActiveRange.IsValid()));

            if ((UlPeerContext.c_MaxRequestedRanges - DonkeyClient.c_ClassicFileRangeRequestSlots) >= l_RangesRequested &&
                f_Ranges.p_Count > f_RangesSent)
            {
                // Request additional ranges of data from the peer.

                uint l_TransferID = p_TransferID;
                int l_Limit = (DonkeyClient.c_NoTransferID == l_TransferID ? DonkeyClient.c_ClassicFileRangeRequestSlots : (UlPeerContext.c_MaxRequestedRanges - l_RangesRequested));
                int l_EffCount = Math.Min(f_Ranges.p_Count - f_RangesSent, l_Limit);

                if (false == f_Clnt.RequestFileRange(l_TransferID, ref f_Ranges, f_RangesSent, l_EffCount))
                {
                    return false;
                }

                f_RangesSent += l_EffCount;
            }

            return true;
        }

        private void AcceptNewObtainCookie(ObtainCookie a_Cookie)
        {
            f_OutOfRelevantRanges = false;
            f_PlannedSize -= c_QuerySizeThreshold;

            int l_OldCount = f_Ranges.p_Count;
            int l_NewItems = a_Cookie.f_Ranges.p_Count;
            f_Ranges.ReserveRange(l_OldCount, l_NewItems);

            ulong l_ChunkBaseOffset = a_Cookie.p_ChunkBaseOffset;
            Content.Internal.ChunkRange[] l_SrcArr = a_Cookie.f_Ranges.p_Arr;
            RequestableRange[] l_DestArr = f_Ranges.p_Arr;
            for (int k = 0; k < l_NewItems; k++)
            {
                FileRange l_FileRange;
                FileRange.CreateFromChunkRange(l_ChunkBaseOffset, ref l_SrcArr[k], out l_FileRange);

                l_DestArr[l_OldCount + k] = new RequestableRange(a_Cookie, ref l_FileRange);
                f_WarrantedSize += l_FileRange.f_Length;
            }

            a_Cookie.SetPeerState(CookieStates.RangeSet);
        }

        private void DropRangesOfCookie(ObtainCookie a_Cookie)
        {
            int l_LowIdx = f_RangesSent;
            RequestableRange[] l_Arr = f_Ranges.p_Arr;

            while (true)
            {
                // Detect and drop any clusters of ranges from that
                // cookie, (usually there will be only one cluster).
                // NOTE: We may not delete ranges that were already
                // sent to the remote peer.

                int l_RangeCount = f_Ranges.p_Count;
                int k = l_LowIdx;
                l_LowIdx = (-1);

                for (; k < l_RangeCount; k++)
                {
                    if (l_Arr[k].f_Cookie == a_Cookie)
                    {
                        if (0 > l_LowIdx)
                        {
                            l_LowIdx = k;
                        }

                        f_WarrantedSize -= l_Arr[k].f_Range.f_Length;
                    }
                    else if (0 <= l_LowIdx)
                    {
                        break;
                    }
                }

                if (0 > l_LowIdx)
                {
                    break;
                }

                f_Ranges.DeleteRange(l_LowIdx, k - l_LowIdx);
            }
        }

        private bool HandleObtainCookieRejection(ObtainCookie.RejectionReasons a_Reason,
                                                 out SourceStates a_SourceState)
        {
            // Returns true if execution should go on normally,
            // otherwise the context is to be closed with the returned
            // state.

            switch (a_Reason)
            {
                case ObtainCookie.RejectionReasons.NeedHashesFirst:
                    break;

                case ObtainCookie.RejectionReasons.NoRelevantRange:
                    if (true == f_OutOfRelevantRanges &&
                        0 == f_PlannedSize &&
                        0 == f_WarrantedSize &&
                        false == f_ActiveRange.IsValid())
                    {
                        a_SourceState = SourceStates.NoRelevantRange;

                        return false;
                    }
                    f_OutOfRelevantRanges = true;
                    break;

                case ObtainCookie.RejectionReasons.CriticalDiskSpaceLevel:
                    a_SourceState = SourceStates.CriticalDiskSpaceLevel;
                    return false;

                case ObtainCookie.RejectionReasons.BadSource:
                    a_SourceState = SourceStates.BadReputation;
                    return false;

                case ObtainCookie.RejectionReasons.IncomplFileUnloads:
                default:
                    a_SourceState = SourceStates.UnspecifiedContentMngrProblem;
                    return false;
            }

            a_SourceState = SourceStates.Undefined;

            return true;
        }

        public bool ActivateRequestedRange(ulong a_StartingOffset,
                                           out FileRange a_Range)
        {
            if (DlStatus.ContentReceived != f_DlStatus)
            {
                if (DlStatus.DownloadGranted == f_DlStatus)
                {
                    SetDlStatus(DlStatus.ContentReceived);
                }
                else
                {
                    a_Range = new FileRange();

                    return false;
                }
            }

            if (false == f_ActiveRange.IsValid())
            {
                RequestableRange[] l_Arr = f_Ranges.p_Arr;
                for (int i = 0; f_RangesSent > i; i++)
                {
                    if (l_Arr[i].f_Range.f_StartingOffset == a_StartingOffset)
                    {
                        f_ActiveRange = l_Arr[i];
                        f_SentToContentMngr = 0;
                        f_Known.p_ActCntxt.f_DownloadsRunning++;

                        f_WarrantedSize -= f_ActiveRange.f_Range.f_Length;
                        f_RangesSent--;
                        f_Ranges.DeleteAt(i);

                        a_Range = f_ActiveRange.f_Range;

                        return true;
                    }
                }
            }

            a_Range = new FileRange();

            return false;
        }

        public bool RegisterDownload(ulong a_StartingOffset,
                                     ref ByteArray a_Data,
                                     uint a_BytesRemaining,
                                     uint a_CompressionGain)
        {
            bool l_Submitted = p_ContentMngr.WriteAsync(f_Known, f_ActiveRange.f_Cookie, ref a_Data);

            if (0 == a_BytesRemaining)
            {
                ClearActiveRange();
            }
            else
            {
                f_SentToContentMngr += (uint)a_Data.f_Len;
            }

            f_AvgSpeed.NotifyDownloadedBytes((uint)a_Data.f_Len);

            if (true == l_Submitted)
            {
                Time l_Time = f_Clnt.f_Inst.GetTime();
                f_Known.f_LastReception = l_Time;
            }

            return (true == l_Submitted &&
                    true == RequestFileRanges());
        }

        public bool NotifyDisputedFileSize()
        {
            MarkAsRemovable(SourceStates.DisputedFileSize, DestructionInfo.RemoveAtOurRequestPeerNeedsInfo);

            return true;
        }

        public bool NotifyFileStatus(Time a_Time,
                                     BitMap a_Compl)
        {
            if (false == base.NotifyFileStatusBase(a_Time, a_Compl))
            {
                return false;
            }

            if (false == f_GotFileDescriptor &&
                false == f_Clnt.SupportsFileDescriptors() &&
                null != f_SrcCookie &&
                0 != f_eD2kCompl.p_SetBitCount)
            {
                // This client doesn't support file descriptors, so
                // just describe what this peer might know (we assume
                // the remote peer is in possession of the eD2k hash
                // set).

                MasterHashes l_MasterHashes = new MasterHashes(f_FileID);
                if (true == p_ContentMngr.UpdateSourceDataAsync(f_Known, f_SrcCookie, P2PHashSets.eD2kHashSet, f_Known.f_FileSize, ref l_MasterHashes))
                {
                    f_GotFileDescriptor = true;
                }
            }

            if (DlStatus.WaitingForInitialFileStatus == f_DlStatus &&
                false == f_PeerHasNoContent)
            {
                if (false == f_Clnt.RequestFileName())
                {
                    return false;
                }

                if (true == f_Clnt.SupportsFileDescriptors())
                {
                    if (false == f_Clnt.RequestFileDescriptor())
                    {
                        return false;
                    }

                    f_ExpectFileDescriptor = true;
                }

                bool l_UseTransferID = TrySetNewTransferID();

                if (false == f_Clnt.RequestUploadStart(p_TransferID, p_AuxiliaryTransferID))
                {
                    return false;
                }

                SetDlStatus(DlStatus.DownloadRequested);
            }

            return true;
        }

        public bool NotifyNoSuchFile()
        {
            // We expect this to happen before we received the file
            // status, but we allow it as a general way of stopping
            // an upload.

            MarkAsRemovable(SourceStates.FileNotAvail, DestructionInfo.RemoveAtOurRequestPeerNeedsInfo);

            return true;
        }

        public bool NotifyFileDescriptor(P2PHashSets a_ProvidableHashSets,
                                         ulong a_FileSize,
                                         ref MasterHashes a_MasterHashes)
        {
            if (true == f_ExpectFileDescriptor)
            {
                f_ExpectFileDescriptor = false;
                f_GotFileDescriptor = true;

                if (null != f_SrcCookie &&
                    true == p_IsDl)
                {
                    bool l_Succ = p_ContentMngr.UpdateSourceDataAsync(f_Known, f_SrcCookie, a_ProvidableHashSets, a_FileSize, ref a_MasterHashes);
                }

                return true;
            }

            return false;
        }

        public bool RegisterHashSet(uint a_StartingHashIndex,
                                    uint a_Count,
                                    BufferIO a_Buffer)
        {
            uint l_Pos = a_Buffer.p_Position;
            byte[] l_Data = a_Buffer.CopyRange(l_Pos, a_Buffer.p_Length - l_Pos);

            return HashDelivery(0, l_Data);
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct RequestableRange
        {
            public readonly ObtainCookie f_Cookie;
            public readonly FileRange f_Range;

            public RequestableRange(ObtainCookie a_Cookie,
                                    ref FileRange a_Range)
            {
                f_Cookie = a_Cookie;
                f_Range = a_Range;
            }

            public bool IsValid()
            {
                return (null != f_Cookie);
            }
        }

        private enum HashQueryProgress : byte
        {
            NoQuery = 0,
            WaitingForPeer = 1,
            SentToContentMngr = 2,
        }

        private enum DlStatus : byte
        {
            PreInit = 0,
            WaitingForInitialFileStatus = 1,
            DownloadRequested = 2,
            DownloadGranted = 3,
            ContentReceived = 4,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct AverageSpeedCalculator
        {
            // The default constructor may be used.

            private Time f_NextTime;
            private uint f_SpeedSum;
            private uint f_LastAvgBytesPerSec;

            public uint p_AverageBytesPerSec
            {
                get { return f_LastAvgBytesPerSec; }
            }

            public void NotifyDownloadedBytes(uint a_Count)
            {
                f_SpeedSum += a_Count;
            }

            private void Reset(Time a_Time)
            {
                f_NextTime = a_Time.AddSeconds(Time.c_SecsPerMinute);
                f_SpeedSum = 0;
            }

            public void Update(Time a_Time)
            {
                if (true == f_NextTime.IsEarlierThan(a_Time))
                {
                    if (true == f_NextTime.IsSet())
                    {
                        f_LastAvgBytesPerSec = (f_SpeedSum / Time.c_SecsPerMinute);
                    }

                    Reset(a_Time);
                }
            }
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct DownloadQueue
    {
        public const uint c_SourceQueryIntervalSecs = (20 * Time.c_SecsPerMinute);

        private const uint c_FileBlockingSecs = Time.c_SecsPerHour;
        private const uint c_RecognizeFileAsUndownloadableAfterSecs = (7 * Time.c_SecsPerMinute);

        private readonly Instance f_Inst;
        private readonly Dictionary<KnownFile, ActiveFile> f_ActiveFiles;
        private readonly Dictionary<KnownFile, Time> f_BlockedFiles;

        private GeneralAPCCookie f_QueryFilesCookie;

        private DownloadQueue(Instance a_Inst)
        {
            f_Inst = a_Inst;
            f_ActiveFiles = new Dictionary<KnownFile, ActiveFile>();
            f_BlockedFiles = new Dictionary<KnownFile, Time>();

            f_QueryFilesCookie = null;
        }

        public static void Create(Instance a_Inst,
                                  out DownloadQueue a_Dest)
        {
            a_Dest = new DownloadQueue(a_Inst);
        }

        public void Destroy()
        {
            Reset();
        }

        private void Reset()
        {
            if (0 < f_ActiveFiles.Count)
            {
                foreach (ActiveFile l_Active in f_ActiveFiles.Values)
                {
                    l_Active.Destroy();
                }

                f_ActiveFiles.Clear();
            }

            f_BlockedFiles.Clear();
        }

#region DownloadAlgo

        public void GuiResetFile(GeneralAPCCookie a_Cookie)
        {
            object[] l_Args = a_Cookie.GetArgAsObjArray();
            KnownFile[] l_Files = (KnownFile[])l_Args[0];
            ResetOperation l_ResetOp = (ResetOperation)l_Args[1];

            foreach (KnownFile l_Known in l_Files)
            {
                ActiveFile l_Active;
                if (false == f_ActiveFiles.TryGetValue(l_Known, out l_Active))
                {
                    // Unblock the file, since the user seems to be
                    // quite interested in it.

                    bool l_Unblocked = f_BlockedFiles.Remove(l_Known);
                }
                else
                {
                    switch (l_ResetOp)
                    {
                        case ResetOperation.RequerySources:
                            l_Active.RequerySources();
                            break;

                        case ResetOperation.ResetPeerStates:
                            l_Active.ResetPeerStates();
                            break;

                        default: throw (new ApplicationException());
                    }
                }
            }

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiAddSource(GeneralAPCCookie a_Cookie)
        {
            object[] l_Args = a_Cookie.GetArgAsObjArray();
            KnownFile[] l_Files = (KnownFile[])l_Args[0];
            I2PAddress l_Addr = (I2PAddress)l_Args[1];

            foreach (KnownFile l_Known in l_Files)
            {
                ActiveFile l_Active;
                if (true == f_ActiveFiles.TryGetValue(l_Known, out l_Active))
                {
                    l_Active.NotifySingleSource(l_Addr);
                }
            }

            a_Cookie.SetState(CookieStates.Success);
        }

        public void Process(Time a_Time)
        {
            if (BridgeStatus.Connected != f_Inst.f_NetworkMngr.p_BridgeStatus)
            {
                Reset();

                if (null != f_QueryFilesCookie)
                {
                    // Drop the apc.

                    f_QueryFilesCookie = null;
                }

                return;
            }

            // Query files.

            if (null == f_QueryFilesCookie)
            {
                f_QueryFilesCookie = new GeneralAPCCookie(f_Inst.f_ContentMngr.NetworkGetFilesForDownloading);
                f_Inst.f_ContentMngr.EnqueueAPC(f_QueryFilesCookie);
            }

            // Do something in the meantime and hope the apc completes
            // therewhile.

            // First remove files that are no longer suitable for
            // downloading.

            RemoveUnsuitableActiveFiles();

            // Now process files that have survived.

            ProcessFiles(a_Time);

            // Now try to update the list of downloaded files entirely.

            if (null != f_QueryFilesCookie)
            {
                CookieStates l_State = f_QueryFilesCookie.p_StateVolatile;

                if (CookieStates.Pending != l_State)
                {
                    if (CookieStates.Success == l_State)
                    {
                        KnownFile[] l_Files = (KnownFile[])f_QueryFilesCookie.GetReturnObject();

                        SelectFiles(a_Time, l_Files);
                    }

                    f_QueryFilesCookie = null;
                }
            }
        }

        private void ProcessFiles(Time a_Time)
        {
            foreach (ActiveFile l_Active in f_ActiveFiles.Values)
            {
                l_Active.Process(a_Time);
            }
        }

        private void SelectFiles(Time a_Time,
                                 KnownFile[] a_Files)
        {
            UnblockFilesIfPossible(a_Time);

            int l_MaxActiveFiles = f_Inst.f_NetworkMngr.f_NetworkPrefs.f_MaxDlFiles;
            for (int i = 0, lim = a_Files.Length; i < lim; i++)
            {
                KnownFile l_Known = a_Files[i];

                ActiveFile l_ExistingActive;
                if (true == f_ActiveFiles.TryGetValue(l_Known, out l_ExistingActive))
                {
                    // Mark this file as unremovable for a short while
                    // and go on.

                    l_ExistingActive.ProtectFile(a_Time, Time.c_SecsPerMinute);

                    continue;
                }

                FileID l_FileID;

                if (false == l_Known.IsDownloadAllowed() ||
                    false == l_Known.GetFileIDVolatile(out l_FileID))
                {
                    // This will happen very seldom.

                    continue;
                }

                int l_Avail = (l_MaxActiveFiles - f_ActiveFiles.Count);

                if (0 >= l_Avail)
                {
                    if (true == f_BlockedFiles.ContainsKey(l_Known))
                    {
                        // Currently this file is blocked, so we don't
                        // remove other files for it.

                        continue;
                    }

                    // We need to make one slot available.

                    DeactivateFiles(a_Time, -l_Avail + 1);

                    l_Avail = (l_MaxActiveFiles - f_ActiveFiles.Count);

                    if (0 >= l_Avail)
                    {
                        // There is nothing we can do anymore.

                        break;
                    }
                }

                ActiveFile l_Active = new ActiveFile(f_Inst, l_Known, l_FileID, a_Time);
                l_Active.ProtectFile(a_Time, c_RecognizeFileAsUndownloadableAfterSecs);
                f_ActiveFiles.Add(l_Known, l_Active);
            }
        }

        private void DeactivateFiles(Time a_Time,
                                     int a_Count)
        {
            // Tries to deactivate a_Count files and blocks them.

            List<ActiveFile> l_RemLs = null;

            foreach (ActiveFile l_Active in f_ActiveFiles.Values)
            {
                if (false == l_Active.IsProtected(a_Time))
                {
                    if (null == l_RemLs)
                    {
                        l_RemLs = new List<ActiveFile>(f_ActiveFiles.Count);
                    }

                    l_RemLs.Add(l_Active);
                }
            }

            if (null != l_RemLs)
            {
                l_RemLs.Sort(CompareActiveFiles);

                Time l_BlockUntilTime = a_Time.AddSeconds(c_FileBlockingSecs);

                for (int i = 0, lim = Math.Min(a_Count, l_RemLs.Count); i < lim; i++)
                {
                    l_RemLs[i].Destroy();
                    f_ActiveFiles.Remove(l_RemLs[i].f_Known);
                    BlockFile(l_RemLs[i].f_Known, l_BlockUntilTime);
                }
            }
        }

        private void RemoveUnsuitableActiveFiles()
        {
            // This routine removes files that are no longer
            // downloadable due to their user controllable
            // download settings.

            List<ActiveFile> l_RemLs = null;

            foreach (ActiveFile l_Active in f_ActiveFiles.Values)
            {
                // We don't use volatile access here on purpose
                // (performance).
                if (ComplStatus.Incomplete != l_Active.f_Known.p_Status ||
                    false == l_Active.f_Known.IsDownloadAllowed())
                {
                    if (null == l_RemLs)
                    {
                        l_RemLs = new List<ActiveFile>(f_ActiveFiles.Count);
                    }

                    l_RemLs.Add(l_Active);
                }
            }

            if (null != l_RemLs)
            {
                foreach (ActiveFile l_Active in l_RemLs)
                {
                    l_Active.Destroy();
                    bool l_Removed = f_ActiveFiles.Remove(l_Active.f_Known);
                }
            }
        }

        private static int CompareActiveFiles(ActiveFile a_A,
                                              ActiveFile a_B)
        {
            return a_A.p_BytesPerSec.CompareTo(a_B.p_BytesPerSec);
        }

        private void UnblockFilesIfPossible(Time a_Time)
        {
            List<KnownFile> l_RemLs = new List<KnownFile>();

            foreach (KeyValuePair<KnownFile, Time> l_Pair in f_BlockedFiles)
            {
                // We don't use volatile access here on purpose
                // (performance).
                if (ComplStatus.Incomplete != l_Pair.Key.p_Status ||
                    false == a_Time.IsEarlierThan(l_Pair.Value))
                {
                    l_RemLs.Add(l_Pair.Key);
                }
            }

            foreach (KnownFile l_Known in l_RemLs)
            {
                f_BlockedFiles.Remove(l_Known);
            }
        }

        private void BlockFile(KnownFile a_Known,
                               Time a_UntilTime)
        {
            f_BlockedFiles[a_Known] = a_UntilTime;
        }

#endregion

        private ActiveFile GetActiveFile(KnownFile a_Known)
        {
            // Return can be null.

            ActiveFile l_Active;
            if (true == f_ActiveFiles.TryGetValue(a_Known, out l_Active))
            {
                return l_Active;
            }

            return null;
        }

        public void NotifySingleSource(KnownFile a_Known,
                                       I2PAddress a_StreamAddr)
        {
            ActiveFile l_Active = GetActiveFile(a_Known);

            if (null != l_Active)
            {
                l_Active.NotifySingleSource(a_StreamAddr);
            }
        }

        public void NotifySourceRange(KnownFile a_Known,
                                      I2PAddress[] a_StreamAddresses)
        {
            ActiveFile l_Active = GetActiveFile(a_Known);

            if (null != l_Active)
            {
                l_Active.NotifySourceRange(a_StreamAddresses);
            }
        }

        public void NotifySourceState(KnownFile a_Known,
                                      I2PAddress a_StreamAddr,
                                      SourceStates a_State)
        {
            ActiveFile l_Active = GetActiveFile(a_Known);

            if (null != l_Active)
            {
                l_Active.NotifySourceState(a_StreamAddr, a_State);
            }
        }

        public void NotifyDlPeerContextCreation(DlPeerContext a_DlPeerCntxt)
        {
            ActiveFile l_Active = GetActiveFile(a_DlPeerCntxt.f_Known);

            if (null != l_Active)
            {
                l_Active.NotifyDlPeerContextCreation(a_DlPeerCntxt);
            }
        }

        public static uint CalcEffMaxSlotLimit(uint a_MaxSlots,
                                               Speed a_SpeedLimit)
        {
            if (0 == a_MaxSlots ||
                true == a_SpeedLimit.IsNearZero())
            {
                return 0;
            }

            if (true == a_SpeedLimit.IsMax())
            {
                return a_MaxSlots;
            }

            // We require at least 2 KiB/s for each slot.
            const uint lc_MinExpectedSpeedPerSlot = (2 * 1024);
            uint l_BytesPerSec = a_SpeedLimit.p_BytesPerSec;

            if (lc_MinExpectedSpeedPerSlot >= l_BytesPerSec)
            {
                // Allow one grace slot.

                return 1;
            }

            return Math.Min(a_MaxSlots, l_BytesPerSec / lc_MinExpectedSpeedPerSlot);
        }

        public enum ResetOperation : byte
        {
            RequerySources = 0,
            ResetPeerStates = 1,
        }

        private class ActiveFile
        {
            private const uint c_ContentMngrQueryIntervalSecs = Time.c_SecsPerHalfHour;
            private const uint c_InitialPeerQueryWaitSecs = Time.c_SecsPerMinute;
            private const ushort c_MaxSources = 512;
            private const byte c_SourcesDicInitialCapacity = 10;

            private static readonly uint[] s_WaitSecsPerSourceState;

            public readonly Instance f_Inst;
            public readonly KnownFile f_Known;
            // Always use this.f_FileID instead of KnownFile.f_FileID.
            public readonly FileID f_FileID;
            public readonly Time f_CreationTime;

            private FlexDictionary<I2PAddress, Source> f_Sources;
            private ushort f_SlotsUsed;

            private uint f_BytesPerSec;
            private Time f_ProtectedTime;

            // Kad sources.
            private Time f_NextKadSourceSearchTime;
            private uint f_SourceSearchConnGen;
            private SourceSearch f_KadSourceSearch;

            // Kad map.
            private Time f_NextKadMapQuery;

            // Content manager.
            private Time f_NextContentManagerSourceQuery;
            private SourceCookie f_SourceCookie;
            private QueryUsersCookie f_QueryUsersCookie;

            // Peer sources.
            private Time f_NextPeerSourceQuery;

            public uint p_BytesPerSec
            {
                get { return f_BytesPerSec; }
            }

            private I2PAddress p_OwnAddress
            {
                // For internal use only. May only be used while a
                // connection is up.
                get { return f_Inst.f_NetworkMngr.p_StreamAddr; }
            }

            static ActiveFile()
            {
                s_WaitSecsPerSourceState = new uint[]
                {
                    Time.c_SecsPerQuarterHour, // CantReachPeer
                    Time.c_SecsPerHalfHour, // PeerBanned
                    Time.c_SecsPerQuarterHour, // UnspecifiedConnErr
                    Time.c_SecsPerMinute, // OwnGlobalLimitReached
                    2 * Time.c_SecsPerMinute, // OwnPerUserLimitReached
                    Time.c_SecsPerQuarterHour, // OwnPerFileLimitReached
                    0, // DlContextExistsAlready
                    Time.c_SecsPerHalfMinute, // A4AFLimitation
                    Time.c_SecsPerWeek, // DisputedFileSize
                    Time.c_SecsPerHour, // FileNotAvail
                    5 * Time.c_SecsPerMinute, // PeerEndedUpload
                    Time.c_SecsPerHour, // NoRelevantRange
                    Time.c_SecsPerHalfHour, // BadReputation
                    3 * Time.c_SecsPerMinute, // CriticalDiskSpaceLevel
                    Time.c_SecsPerHour, // UnspecifiedContentMngrProblem
                    0, // AllowedBandwidthTooLow
                    7 * Time.c_SecsPerMinute, // ConnLoss
                    0, // KilledByDlQueue
                };
            }

            public ActiveFile(Instance a_Inst,
                              KnownFile a_Known,
                              FileID a_FileID,
                              Time a_Time)
            {
                f_Inst = a_Inst;
                f_Known = a_Known;
                f_FileID = a_FileID;
                f_CreationTime = a_Time;

                FlexDictionary<I2PAddress, Source>.Create(c_SourcesDicInitialCapacity, out f_Sources);

                f_NextPeerSourceQuery = a_Time.AddSeconds(c_InitialPeerQueryWaitSecs);
            }

            public void Destroy()
            {
                StopDownloads();

                if (null != f_KadSourceSearch)
                {
                    f_Inst.f_Kad.f_SearchMngr.StopSearch(f_KadSourceSearch.f_SearchID, false);
                }

                if (null != f_SourceCookie)
                {
                    EndQueryUsers();
                }
            }

            private void StopDownloads()
            {
                FlexDicEnum<I2PAddress, Source> l_Enum;
                FlexDicItem<I2PAddress, Source>[] l_Arr = f_Sources.ExposeItemArray(out l_Enum);
                for (int i; true == l_Enum.GetNextIndex(l_Arr, out i); )
                {
                    if (null != l_Arr[i].f_Val.f_DlPeerCntxt)
                    {
                        l_Arr[i].f_Val.KillDlPeerContext();
                    }
                }
            }

            public void RequerySources()
            {
                if (null == f_KadSourceSearch)
                {
                    f_NextKadSourceSearchTime = new Time();
                }

                f_NextKadMapQuery = new Time();

                if (null == f_SourceCookie)
                {
                    f_NextContentManagerSourceQuery = new Time();
                }

                f_NextPeerSourceQuery = new Time();
            }

            public void ResetPeerStates()
            {
                FlexDicEnum<I2PAddress, Source> l_Enum;
                FlexDicItem<I2PAddress, Source>[] l_Arr = f_Sources.ExposeItemArray(out l_Enum);
                for (int i; true == l_Enum.GetNextIndex(l_Arr, out i); )
                {
                    if (null == l_Arr[i].f_Val.f_DlPeerCntxt)
                    {
                        SetSourceStateAndResetWaitTime(SourceStates.NoConnAttemptYet, ref l_Arr[i]);
                    }
                }
            }

            private void MeasureSpeed(Time a_Time)
            {
                f_BytesPerSec = 0;

                FlexDicEnum<I2PAddress, Source> l_Enum;
                FlexDicItem<I2PAddress, Source>[] l_Arr = f_Sources.ExposeItemArray(out l_Enum);
                for (int i; true == l_Enum.GetNextIndex(l_Arr, out i); )
                {
                    if (null != l_Arr[i].f_Val.f_DlPeerCntxt)
                    {
                        f_BytesPerSec += l_Arr[i].f_Val.f_DlPeerCntxt.p_AverageBytesPerSec;
                    }
                }
            }

            public void Process(Time a_Time)
            {
                MeasureSpeed(a_Time);
                FindSources(a_Time);
                MantainSources(a_Time);
            }

            public void ProtectFile(Time a_Time,
                                    uint a_Secs)
            {
                f_ProtectedTime = a_Time.AddSeconds(a_Secs);
            }

            public bool IsProtected(Time a_Time)
            {
                return a_Time.IsEarlierThan(f_ProtectedTime);
            }

            private void FindSources(Time a_Time)
            {
                // Our first source is the kademlia network (published
                // sources).

                if (null == f_KadSourceSearch)
                {
                    if (a_Time >= f_NextKadSourceSearchTime &&
                        KademliaStatus.Connected == f_Inst.f_Kad.p_KadStatus &&
                        false == f_Inst.f_Kad.f_SearchMngr.StartingNewSearchIsNotRecommended())
                    {
                        // Take a small break after this in any case.
                        f_NextKadSourceSearchTime = a_Time.AddSeconds(Time.c_SecsPerHalfMinute);

                        // Ask the network.
                        f_SourceSearchConnGen = f_Inst.f_NetworkMngr.p_ConnGeneration;
                        f_KadSourceSearch = f_Inst.f_Kad.f_SearchMngr.FindSources(f_FileID, f_Known.f_FileSize, NotifyKadSource);
                    }
                }
                else if (false == f_KadSourceSearch.IsRunning())
                {
                    if (f_Inst.f_NetworkMngr.p_ConnGeneration == f_SourceSearchConnGen &&
                        true == f_KadSourceSearch.WentWellEnough())
                    {
                        // The connection was up in the meantime, so we
                        // assume another query doesn't make sense for
                        // some time.

                        f_NextKadSourceSearchTime = a_Time.AddSeconds(20 * Time.c_SecsPerMinute);
                    }

                    f_KadSourceSearch = null;
                }

                // See what our own kad map tells us.

                if (a_Time >= f_NextKadMapQuery)
                {
                    // Use the sources indexed locally.
                    List<I2PAddress> l_Addresses = f_Inst.f_Kad.f_Map.GetFileSourceArray(f_FileID, f_Known.f_FileSize);

                    if (null != l_Addresses)
                    {
                        NotifySourceRange(l_Addresses.ToArray());
                    }

                    f_NextKadMapQuery = a_Time.AddSeconds(5 * Time.c_SecsPerMinute);
                }

                // Our next source is the content manager in combination
                // with the user manager (stored sources).

                if (a_Time >= f_NextContentManagerSourceQuery)
                {
                    if (false == ContinueQueryUsers())
                    {
                        f_NextContentManagerSourceQuery = a_Time.AddSeconds(c_ContentMngrQueryIntervalSecs);
                    }
                }

                // Our final source are the connected peers who can
                // provide us with sources for our specific situation.

                if (a_Time >= f_NextPeerSourceQuery)
                {
                    KnownFile.ActivityContext l_ActCntxt = f_Known.p_ActCntxt;

                    if (null != l_ActCntxt)
                    {
                        l_ActCntxt.RequestSourcesForMyself(a_Time);
                    }

                    f_NextPeerSourceQuery = a_Time.AddSeconds(Time.c_SecsPerMinute);
                }
            }

            private bool ContinueQueryUsers()
            {
                // Returns false if queries are done. Otherwise a
                // subsequent call is required.

                if (null != f_QueryUsersCookie)
                {
                    CookieStates l_UsersState = f_QueryUsersCookie.p_StateVolatile;

                    if (CookieStates.Pending != l_UsersState)
                    {
                        // We will fail even if the request reached the
                        // content manager, but it doesn't know any users.

                        if (CookieStates.Success == l_UsersState)
                        {
                            I2PAddress l_OwnDest = p_OwnAddress;
                            UserNum[] l_Users = f_QueryUsersCookie.p_Users;

                            foreach (UserNum l_UserNum in l_Users)
                            {
                                UserContext l_UserCntxt = f_Inst.f_UserMngr.GetUserContext(l_UserNum);

                                if (null != l_UserCntxt)
                                {
                                    NotifySourceCore(l_OwnDest, l_UserCntxt.f_StreamAddr);
                                }
                            }
                        }

                        EndQueryUsers();

                        return false;
                    }

                    return true;
                }

                if (null == f_SourceCookie)
                {
                    f_SourceCookie = new SourceCookie(UserNum.s_QueryUsers);

                    if (false == f_Inst.f_ContentMngr.RegisterSourceAsync(f_Known, f_SourceCookie))
                    {
                        f_SourceCookie = null;

                        return false;
                    }
                }

                CookieStates l_State = f_SourceCookie.p_ContentMngrStateVolatile;

                if (CookieStates.Pending != l_State)
                {
                    if (CookieStates.Success == l_State)
                    {
                        f_QueryUsersCookie = new QueryUsersCookie();

                        if (false == f_Inst.f_ContentMngr.QueryUsersAsync(f_Known, f_QueryUsersCookie))
                        {
                            EndQueryUsers();

                            return false;
                        }
                    }
                    else
                    {
                        f_SourceCookie = null;

                        return false;
                    }
                }

                return true;
            }

            private void EndQueryUsers()
            {
                bool l_OK = f_Inst.f_ContentMngr.UnregisterSourceAsync(f_Known, f_SourceCookie);

                f_SourceCookie = null;
                f_QueryUsersCookie = null;
            }

            private void NotifyKadSource(FileEntry a_FileEntry)
            {
                NotifySingleSource(a_FileEntry.f_StreamAddr);
            }

            public void NotifySingleSource(I2PAddress a_StreamAddr)
            {
                NotifySourceCore(p_OwnAddress, a_StreamAddr);
            }

            public void NotifySourceRange(I2PAddress[] a_StreamAddresses)
            {
                NotifySourceRange(a_StreamAddresses, 0, a_StreamAddresses.Length);
            }

            public void NotifySourceRange(I2PAddress[] a_StreamAddresses,
                                          int a_Index,
                                          int a_Count)
            {
                I2PAddress l_OwnDest = p_OwnAddress;

                for (int i = a_Index, lim = (a_Index + a_Count); i < lim; i++)
                {
                    // Although it's not allowed, still check if there
                    // are null elements and filter them.
                    if (null == a_StreamAddresses[i])
                    {
                        continue;
                    }

                    NotifySourceCore(l_OwnDest, a_StreamAddresses[i]);
                }
            }

            private void NotifySourceCore(I2PAddress a_OwnDest,
                                          I2PAddress a_InputAddr)
            {
                if (true == f_Sources.ContainsKey(a_InputAddr) ||
                    true == a_OwnDest.ValueEquals(a_InputAddr))
                {
                    return;
                }

                FlexDicItem<I2PAddress, Source>[] l_Arr;

                if (c_MaxSources <= f_Sources.p_Count)
                {
                    // Try to drop an entry.

                    FlexDicEnum<I2PAddress, Source> l_Enum;
                    l_Arr = f_Sources.ExposeItemArray(out l_Enum);
                    for (int i; true == l_Enum.GetNextIndex(l_Arr, out i); )
                    {
                        SourceStates l_State = l_Arr[i].f_Val.f_State;

                        if (SourceStates.CantReachPeer == l_State ||
                            SourceStates.PeerBanned == l_State ||
                            SourceStates.BadReputation == l_State)
                        {
                            // We need to stop enumerating here
                            // (we are changing the dictionary).

                            bool l_Removed = f_Sources.Remove(l_Arr[i].p_Key);

                            break;
                        }
                    }

                    if (c_MaxSources <= f_Sources.p_Count)
                    {
                        // Still no slot available.

                        return;
                    }
                }

                int l_Idx;
                bool l_Added = f_Sources.PrepareAdd(a_InputAddr, out l_Idx, out l_Arr);

                l_Arr[l_Idx].f_Val = new Source(a_InputAddr);
            }

            private void MantainSources(Time a_Time)
            {
                if (true == f_Sources.IsEmpty())
                {
                    return;
                }

                int l_EffMaxSlots = (int)CalcEffMaxSlotLimit(f_Inst.f_NetworkMngr.f_NetworkPrefs.f_MaxDownloadConnsPerFile, Speed.Min(f_Inst.f_NetworkMngr.f_NetworkPrefs.f_DlLimit, f_Known.f_DlLimit));
                int l_AvailSlots = (l_EffMaxSlots - f_SlotsUsed);

                if (0 >= l_AvailSlots)
                {
                    return;
                }

                FlexDicEnum<I2PAddress, Source> l_Enum;
                FlexDicItem<I2PAddress, Source>[] l_Arr = f_Sources.ExposeItemArray(out l_Enum);

                if (0 < l_AvailSlots)
                {
                    // We should create new connections.

                    Time l_Time = f_Inst.GetTime();
                    bool l_IsDiskSpaceLevelCritical = f_Inst.f_ContentMngr.p_CriticalDiskSpaceLevelReachedVolatile;

                    for (int i; true == l_Enum.GetNextIndex(l_Arr, out i); )
                    {
                        SourceStates l_State = l_Arr[i].f_Val.f_State;

                        if (SourceStates.ConnAttemptInProgress == l_State ||
                            SourceStates.DlPeerContextCreated == l_State)
                        {
                            continue;
                        }

                        if (false == l_Arr[i].f_Val.f_WaitTime.IsEarlierThan(l_Time))
                        {
                            continue;
                        }

                        if (true == l_IsDiskSpaceLevelCritical)
                        {
                            SetSourceStateAndWaitTime(l_Time, SourceStates.CriticalDiskSpaceLevel, ref l_Arr[i]);

                            continue;
                        }

                        I2PAddress l_Addr = l_Arr[i].p_Key;
                        SourceStates l_CantConnSourceState;

                        if (false == f_Inst.f_NetworkMngr.CanConnectTo(l_Addr, out l_CantConnSourceState))
                        {
                            if (SourceStates.Undefined != l_CantConnSourceState)
                            {
                                SetSourceStateAndWaitTime(l_Time, l_CantConnSourceState, ref l_Arr[i]);
                            }

                            continue;
                        }

                        NetworkManager.ConnResultContext l_ConnResCntxt = new NetworkManager.ConnResultContext(f_Known.TryCreateDlPeerContext, f_Known, f_FileID, l_Addr);
                        f_Inst.f_NetworkMngr.RequestConnection(l_Addr, l_ConnResCntxt);

                        f_SlotsUsed++;

                        SetSourceStateAndResetWaitTime(SourceStates.ConnAttemptInProgress, ref l_Arr[i]);

                        if (0 >= --l_AvailSlots)
                        {
                            break;
                        }
                    }
                }
                else
                {
                    // We need to close connections.

                    for (int i; true == l_Enum.GetNextIndex(l_Arr, out i); )
                    {
                        if (null == l_Arr[i].f_Val.f_DlPeerCntxt)
                        {
                            continue;
                        }

                        f_SlotsUsed--;

                        l_Arr[i].f_Val.KillDlPeerContext();

                        if (0 <= ++l_AvailSlots)
                        {
                            break;
                        }
                    }
                }
            }

            private void SetSourceStateAndResetWaitTime(SourceStates a_State,
                                                        ref FlexDicItem<I2PAddress, Source> a_Item)
            {
                a_Item.f_Val.SetStateAndWaitTime(a_State, new Time());
            }

            private void SetSourceStateAndWaitTime(Time a_Time,
                                                   SourceStates a_State,
                                                   ref FlexDicItem<I2PAddress, Source> a_Item)
            {
                uint l_WaitSecs = GetWaitSecs(a_State);
                a_Item.f_Val.SetStateAndWaitTime(a_State, a_Time.AddSeconds(l_WaitSecs));
            }

            public void NotifySourceState(I2PAddress a_StreamAddr,
                                          SourceStates a_NewState)
            {
                int l_Idx;
                if (false == f_Sources.TryGet(a_StreamAddr, out l_Idx))
                {
                    // We've lost interest in that source.

                    return;
                }

                FlexDicItem<I2PAddress, Source>[] l_Arr = f_Sources.ExposeItemArray();
                SourceStates l_OldState = l_Arr[l_Idx].f_Val.f_State;

                if (SourceStates.ConnAttemptInProgress == l_OldState ||
                    SourceStates.DlPeerContextCreated == l_OldState)
                {
                    uint l_WaitSecs;
                    if (true == WasConnKilled(a_NewState, out l_WaitSecs))
                    {
                        // BEGIN OF TWEAK
                        if (SourceStates.ConnLoss == a_NewState &&
                            null != l_Arr[l_Idx].f_Val.f_DlPeerCntxt &&
                            true == l_Arr[l_Idx].f_Val.f_DlPeerCntxt.f_Clnt.IsiMule())
                        {
                            l_WaitSecs = Math.Min(l_WaitSecs, 5 * Time.c_SecsPerMinute);
                        }
                        // END OF TWEAK

                        f_SlotsUsed--;

                        l_Arr[l_Idx].f_Val.f_DlPeerCntxt = null;
                        l_Arr[l_Idx].f_Val.SetStateAndWaitTime(a_NewState, f_Inst.GetTime().AddSeconds(l_WaitSecs));
                    }
                    else
                    {
                        Util.DebugBreak();
                    }
                }
            }

            public void NotifyDlPeerContextCreation(DlPeerContext a_DlPeerCntxt)
            {
                int l_Idx;
                if (false == f_Sources.TryGet(a_DlPeerCntxt.f_Clnt.p_Addr, out l_Idx))
                {
                    // We've lost interest in that source.

                    a_DlPeerCntxt.DlQueueWantsContextDestroyed();

                    return;
                }

                FlexDicItem<I2PAddress, Source>[] l_Arr = f_Sources.ExposeItemArray();

                if (SourceStates.ConnAttemptInProgress == l_Arr[l_Idx].f_Val.f_State)
                {
                    l_Arr[l_Idx].f_Val.SetState(SourceStates.DlPeerContextCreated);
                    l_Arr[l_Idx].f_Val.f_DlPeerCntxt = a_DlPeerCntxt;
                }
                else
                {
                    // This might happen if a file was active, then
                    // inactive and then active again in a brief
                    // period of time (very unlikely).

                    a_DlPeerCntxt.DlQueueWantsContextDestroyed();
                }
            }

            private static bool WasConnKilled(SourceStates a_SourceState,
                                              out uint a_WaitSecs)
            {
                if (SourceStates._CONN_KILLED_START <= a_SourceState &&
                    SourceStates._CONN_KILLED_END >= a_SourceState)
                {
                    a_WaitSecs = GetWaitSecs(a_SourceState);

                    return true;
                }

                a_WaitSecs = 0;

                return false;
            }

            private static uint GetWaitSecs(SourceStates a_SourceState)
            {
                int l_Idx = (a_SourceState - SourceStates._CONN_KILLED_START);

                return s_WaitSecsPerSourceState[l_Idx];
            }

#if !MONO
            [StructLayout(LayoutKind.Auto)]
#endif
            private struct Source
            {
                public readonly I2PAddress f_Addr;
                public SourceStates f_State;
                public Time f_WaitTime;
                public DlPeerContext f_DlPeerCntxt;

                public Source(I2PAddress a_Addr)
                {
                    f_Addr = a_Addr;
                    f_State = SourceStates.NoConnAttemptYet;
                    f_WaitTime = new Time();
                    f_DlPeerCntxt = null;
                }

                public void SetState(SourceStates a_State)
                {
                    f_State = a_State;
                }

                public void SetStateAndWaitTime(SourceStates a_State,
                                                Time a_WaitTime)
                {
                    f_State = a_State;
                    f_WaitTime = a_WaitTime;
                }

                public void KillDlPeerContext()
                {
                    f_State = SourceStates.KilledByDlQueue;
                    f_DlPeerCntxt.DlQueueWantsContextDestroyed();
                    f_DlPeerCntxt = null;
                }
            }
        }
    }

    enum SourceStates : byte
    {
        // Changes here need to be reflected in
        // ActiveFile.s_WaitSecsPerSourceState.

        Undefined = 0,

        NoConnAttemptYet,
        ConnAttemptInProgress,
        DlPeerContextCreated,

        // Errors during connection buildup.
        CantReachPeer,
        PeerBanned,
        UnspecifiedConnErr,
        OwnGlobalLimitReached,
        OwnPerUserLimitReached,
        OwnPerFileLimitReached,
        DlContextExistsAlready,
        A4AFLimitation,

        // Errors while trying to download (each means connection
        // killed).
        DisputedFileSize,
        FileNotAvail,
        PeerEndedUpload,
        NoRelevantRange,
        BadReputation,
        CriticalDiskSpaceLevel,
        UnspecifiedContentMngrProblem,
        AllowedBandwidthTooLow,
        ConnLoss,
        KilledByDlQueue,

        _CONN_KILLED_START = CantReachPeer,
        _CONN_KILLED_END = KilledByDlQueue,
    }

    class UlPeerContext : PeerContext
    {
        // This value shouldn't be lowered below seven (otherwise
        // client compatibility may drop).
        public const byte c_MaxRequestedRanges = 7;

        private byte f_ClustersOnQueue;
        private List<FileRange> f_SettledRanges = new List<FileRange>();
        private Queue<ReadCookie> f_ReadCookies = new Queue<ReadCookie>();

        private UlStatus f_UlStatus;
        // f_Rank can be NetworkManager.c_InvalidRank.
        private ushort f_Rank;

        private Time f_StoppingTime;

        private BitField f_Flags;

        public UploadQueue.UlPeerContextFieldsOfUlQueue f_UploadQueueFields;

        public UlPeerContext(KnownFile a_Known,
                             FileID a_FileID,
                             DonkeyClient a_Client) :
            base(a_Known, a_FileID, a_Client)
        {
        }

        public void Initialize()
        {
            Packet l_Packet = f_Clnt.FindAnyFileContentPacketOnOutQueue(f_Known);

            if (null != l_Packet)
            {
                UpdateClustersOnQueue(1);
                l_Packet.e_SentCompletely += NotifyFileContentPieceSent;
            }

            base.InitializeBase();
        }

        protected override void DestroyChild()
        {
            if (DestructionInfo.RemoveAtOurRequestPeerNeedsInfo == f_DestructionInfo)
            {
                if (false == f_Clnt.SendDeselectUlFileMessage())
                {
                    f_Clnt.Close();
                }
            }

            bool l_Cleared = ClearPacketQueueAndUlRegisters(true);

            RemoveFromUlQueueIfNecessary();
        }

        public void MarkAsRemovable(DestructionInfo a_DestructionInfo)
        {
            if (true == PrepareDestruction(a_DestructionInfo))
            {
                f_Clnt.SpareIfUlContextIsMissing();
            }
        }

        public override void MarkAsRemovable(SourceStates a_SourceState,
                                             DestructionInfo a_DestructionInfo)
        {
            this.MarkAsRemovable(a_DestructionInfo);
        }

        public bool NotifyFileStatus(Time a_Time,
                                     BitMap a_eD2kCompl)
        {
            return base.NotifyFileStatusBase(a_Time, a_eD2kCompl);
        }

        public bool NotifyNoSuchFile()
        {
            f_NeedToReleaseRemoteUlResources = false;

            return true;
        }

        private void SetUlStatus(UlStatus a_UlStatus)
        {
            f_UlStatus = a_UlStatus;
        }

        public bool IsTransferIDRelevant(uint a_TransferID,
                                         out bool a_Excuse)
        {
            if (true == HasTransferID(a_TransferID))
            {
                a_Excuse = false;

                return true;
            }

            a_Excuse = f_Flags.CheckFlag(FlagsToUInt(UlPeerCntxtFlags.UnknownTransferIDsAllowed));

            return false;
        }

        public bool MaySwitchToOtherFile()
        {
            return (UlStatus.UploadNotRequested == f_UlStatus);
        }

        public void UlQueueUpdateQueueRank(uint a_Rank)
        {
            // May only be used by the UploadQueue struct. Control
            // must not re-enter the upload queue.

            f_Rank = (ushort)a_Rank;
        }

        public void UlQueueAdjustMayUploadState(bool a_MayUpload)
        {
            // May only be used by the UploadQueue struct. Control
            // must not re-enter the upload queue.

            if (false == f_Flags.AdjustFlagWithReturn(FlagsToUInt(UlPeerCntxtFlags.InputMayUpload), a_MayUpload))
            {
                return;
            }

            f_Flags.SetFlag(FlagsToUInt(UlPeerCntxtFlags.InputMayUploadStateUpdated));
            f_Rank = (false == a_MayUpload ? NetworkManager.c_InvalidRank : NetworkManager.c_BestRank);
        }

        public void UlQueueContextWasForciblyRemovedFromUlQueue()
        {
            // May only be used by the UploadQueue struct. Control
            // must not re-enter the upload queue.

            f_Flags.ClearCombination(FlagsToUInt(UlPeerCntxtFlags.RegisteredWithUlQueue | UlPeerCntxtFlags.InputMayUpload | UlPeerCntxtFlags.InputMayUploadStateUpdated));
            f_Flags.SetFlag(FlagsToUInt(UlPeerCntxtFlags.InputRemovedFromQueueByForce));
            f_Rank = NetworkManager.c_InvalidRank;
        }

        protected override bool ProcessChild(Time a_Time)
        {
            if (UlStatus.UploadNotRequested == f_UlStatus)
            {
                return true;
            }

            if (false == HandleUlQueueInputChanges(a_Time))
            {
                return false;
            }

            if (UlStatus.PeerWaitsForAccept == f_UlStatus)
            {
                if (false == HandleWaiting(a_Time))
                {
                    return false;
                }

                // NOTE: f_UlStatus may have changed.
            }

            if (UlStatus.AcceptedUlRequest == f_UlStatus)
            {
                if (false == HandleUploads(a_Time))
                {
                    return false;
                }
            }

            return true;
        }

        public bool NotifyStartUploadRequestNoTransferID()
        {
            return NotifyStartUploadRequest(DonkeyClient.c_NoTransferID, DonkeyClient.c_NoTransferID);
        }

        public bool NotifyStartUploadRequest(uint a_PrimTransferID,
                                             uint a_AuxTransferID)
        {
            if (DonkeyClient.c_NoTransferID == a_PrimTransferID)
            {
                if (DonkeyClient.c_NoTransferID != a_AuxTransferID)
                {
                    return false;
                }
            }
            else if (DonkeyClient.c_NoTransferID == a_AuxTransferID ||
                     a_PrimTransferID == a_AuxTransferID)
            {
                return false;
            }

            if (DonkeyClient.c_NoTransferID != p_TransferID)
            {
            }

            if (UlStatus.UploadNotRequested == f_UlStatus)
            {
                if (DonkeyClient.c_NoTransferID == a_PrimTransferID)
                {
                    ClearTransferID();
                }
                else
                {
                    SetTransferID(a_PrimTransferID, a_AuxTransferID);
                }

                SetUlStatus(UlStatus.PeerWaitsForAccept);
            }

            bool l_Succ = NotifyQueueRankRequest();

            return l_Succ;
        }

        private bool HandleWaiting(Time a_Time)
        {
            if (0 == f_ClustersOnQueue)
            {
                bool l_Registered = f_Flags.CheckFlag(FlagsToUInt(UlPeerCntxtFlags.RegisteredWithUlQueue));

                if (false == l_Registered)
                {
                    if (false == f_Clnt.f_Inst.f_NetworkMngr.f_UlQueue.TryAdd(this))
                    {
                        // We're not on queue.

                        if (false == HandleForcefulRemovalFromQueue())
                        {
                            return false;
                        }
                    }
                    else
                    {
                        // We're on queue.

                        f_Flags.SetFlag(FlagsToUInt(UlPeerCntxtFlags.RegisteredWithUlQueue));
                        l_Registered = true;
                    }
                }

                if (true == l_Registered &&
                    true == f_Flags.CheckFlag(FlagsToUInt(UlPeerCntxtFlags.InputMayUpload)) &&
                    false == f_Flags.CheckFlag(FlagsToUInt(UlPeerCntxtFlags.InputMayUploadStateUpdated)))
                {
                    if (false == HandleMayUploadStateUpdate(a_Time, true))
                    {
                        return false;
                    }
                }
            }

            return true;
        }

        public bool NotifyQueueRankRequest()
        {
            if (UlStatus.UploadNotRequested == f_UlStatus)
            {
                return false;
            }

            if (UlStatus.PeerWaitsForAccept == f_UlStatus)
            {
                f_Flags.SetFlag(FlagsToUInt(UlPeerCntxtFlags.PeerWaitsForRank));
            }

            return true;
        }

        public bool NotifyEndOfDownload()
        {
            if (UlStatus.UploadNotRequested == f_UlStatus ||
                false == ReturnToInitialState(false))
            {
                return false;
            }

            return true;
        }

        private bool ReturnToInitialState(bool a_AllowUnknownTransferIDs)
        {
            ClearTransferID();

            if (UlStatus.PeerWaitsForAccept != f_UlStatus)
            {
                bool l_Cleared = ClearPacketQueueAndUlRegisters(false);

                f_Flags.ClearFlag(FlagsToUInt(UlPeerCntxtFlags.PeerWaitsForRank));
            }

            RemoveFromUlQueueIfNecessary();

            f_Flags.AdjustFlag(FlagsToUInt(UlPeerCntxtFlags.UnknownTransferIDsAllowed), a_AllowUnknownTransferIDs);
            SetUlStatus(UlStatus.UploadNotRequested);

            return true;
        }

        private bool HandleUlQueueInputChanges(Time a_Time)
        {
            if (true == f_Flags.AdjustFlagWithReturn(FlagsToUInt(UlPeerCntxtFlags.InputRemovedFromQueueByForce), false))
            {
                if (false == HandleForcefulRemovalFromQueue())
                {
                    return false;
                }

                // NOTE: f_UlStatus may have changed.
            }

            if (true == f_Flags.AdjustFlagWithReturn(FlagsToUInt(UlPeerCntxtFlags.InputMayUploadStateUpdated), false))
            {
                bool l_MayUpload = f_Flags.CheckFlag(FlagsToUInt(UlPeerCntxtFlags.InputMayUpload));

                if (false == HandleMayUploadStateUpdate(a_Time, l_MayUpload))
                {
                    return false;
                }
            }

            if (UlStatus.PeerWaitsForAccept == f_UlStatus &&
                NetworkManager.c_InvalidRank != f_Rank &&
                false == SendRankIfNecessary(f_Rank))
            {
                return false;
            }

            return true;
        }

        private bool HandleForcefulRemovalFromQueue()
        {
            if (DonkeyClient.c_NoTransferID == p_TransferID)
            {
                MarkAsRemovable(DestructionInfo.RemoveAtOurRequestPeerNeedsInfo);
            }
            else if (false == f_Clnt.SendEndUploadMessage(p_TransferID) ||
                     false == ReturnToInitialState(true))
            {
                return false;
            }

            return true;
        }

        private bool HandleMayUploadStateUpdate(Time a_Time,
                                                bool a_MayUpload)
        {
            if (false == a_MayUpload)
            {
                if (UlStatus.AcceptedUlRequest == f_UlStatus &&
                    false == f_StoppingTime.IsSet())
                {
                    if (0 == f_ClustersOnQueue)
                    {
                        ClearUlRegisters();

                        f_StoppingTime = a_Time;
                    }
                    else
                    {
                        const uint lc_TimeBeforeAggressivelyReducingQueueInSecs = (6 * Time.c_SecsPerMinute);
                        f_StoppingTime = a_Time.AddSeconds(lc_TimeBeforeAggressivelyReducingQueueInSecs);
                        f_Flags.SetFlag(FlagsToUInt(UlPeerCntxtFlags.StopTransferRevocationIsNotAbortable));
                    }
                }
            }
            else
            {
                if (UlStatus.PeerWaitsForAccept == f_UlStatus)
                {
                    if (false == SendRankIfNecessary(NetworkManager.c_BestRank) ||
                        false == f_Clnt.SendAcceptUploadNotification(p_TransferID))
                    {
                        return false;
                    }

                    f_Flags.ClearFlag(FlagsToUInt(UlPeerCntxtFlags.UnknownTransferIDsAllowed));
                    SetUlStatus(UlStatus.AcceptedUlRequest);
                    f_StoppingTime = new Time();
                    f_Flags.ClearCombination(FlagsToUInt(UlPeerCntxtFlags.StopTransferRevocationIsNotAbortable | UlPeerCntxtFlags.StopTransferWaitingForAck));
                }
                else
                {
                    if (true == f_StoppingTime.IsSet() &&
                        false == f_Flags.CheckFlag(FlagsToUInt(UlPeerCntxtFlags.StopTransferRevocationIsNotAbortable)))
                    {
                        f_StoppingTime = new Time();
                    }
                }
            }

            return true;
        }

        private bool SendRankIfNecessary(uint a_Rank)
        {
            // Returns zero if a critical error occurred.

            if (true == f_Flags.AdjustFlagWithReturn(FlagsToUInt(UlPeerCntxtFlags.PeerWaitsForRank), false) &&
                false == f_Clnt.SendQueueRanking(p_TransferID, a_Rank))
            {
                return false;
            }

            return true;
        }

        private bool HandleUploads(Time a_Time)
        {
            if (true == f_StoppingTime.IsSet())
            {
                return HandleRevocation(a_Time);
            }

            while (0 < f_ReadCookies.Count &&
                   1 >= f_ClustersOnQueue)
            {
                ReadCookie l_Cookie = f_ReadCookies.Peek();
                CookieStates l_State = l_Cookie.p_StateVolatile;

                if (CookieStates.Pending == l_State)
                {
                    break;
                }

                f_ReadCookies.Dequeue();

                if (CookieStates.Success != l_State)
                {
                    return false;
                }

                uint l_ComprGain = 0;

                if (false == f_Clnt.SendFileContent(l_Cookie.f_Offset, ref l_Cookie.f_Bytes, NotifyFileContentPieceSent, out l_ComprGain))
                {
                    return false;
                }

                UpdateClustersOnQueue(f_ClustersOnQueue + 1u);
            }

            return true;
        }

        private bool ClearPacketQueueAndUlRegisters(bool a_ZeroCountByForce)
        {
            // Returns a nonzero value if the packet queue is clear
            // of file content packets. Otherwise one piece remained
            // on the queue.

            ClearUlRegisters();

            if (0 == f_ClustersOnQueue)
            {
                return true;
            }

            uint l_RemainingPieces = f_Clnt.RemoveFileContentPacketClustersFromOutQueue(f_Known, true);

            bool l_SetCountToOne = (false == a_ZeroCountByForce &&
                                    0 != l_RemainingPieces);
            UpdateClustersOnQueue(Util.BoolToByte(l_SetCountToOne));

            return (0 == l_RemainingPieces);
        }

        private void ClearUlRegisters()
        {
            f_SettledRanges.Clear();
            f_ReadCookies.Clear();
        }

        private void RemoveFromUlQueueIfNecessary()
        {
            if (false == f_Flags.AdjustFlagWithReturn(FlagsToUInt(UlPeerCntxtFlags.RegisteredWithUlQueue), false))
            {
                return;
            }

            bool l_Removed = f_Clnt.f_Inst.f_NetworkMngr.f_UlQueue.TryRemove(this);

            f_Flags.ClearCombination(FlagsToUInt(UlPeerCntxtFlags.InputMayUpload | UlPeerCntxtFlags.InputMayUploadStateUpdated | UlPeerCntxtFlags.InputRemovedFromQueueByForce));
            f_Rank = NetworkManager.c_InvalidRank;
        }

        private void UpdateClustersOnQueue(uint a_NewValue)
        {
            if (a_NewValue == f_ClustersOnQueue)
            {
                return;
            }

            bool l_OldTransferState = (0 != f_ClustersOnQueue);
            bool l_NewTransferState = (0 != a_NewValue);

            f_ClustersOnQueue = (byte)a_NewValue;

            if (l_OldTransferState != l_NewTransferState)
            {
                if (true == f_Flags.CheckFlag(FlagsToUInt(UlPeerCntxtFlags.RegisteredWithUlQueue)))
                {
                    bool l_Notified = f_Clnt.f_Inst.f_NetworkMngr.f_UlQueue.TryNotifyIsUploadingStateChanged(this, l_NewTransferState);
                }

                f_Known.p_ActCntxt.ModifyUploadingUlPeerContexts(l_NewTransferState);
            }
        }

        public bool NotifyRefuseUploadAcknowledgement(uint a_AuxTransferID)
        {
            uint l_TransferID = p_TransferID;
            bool l_Valid = (UlStatus.AcceptedUlRequest == f_UlStatus &&
                            true == f_StoppingTime.IsSet() &&
                            true == f_Flags.CheckFlag(FlagsToUInt(UlPeerCntxtFlags.StopTransferWaitingForAck)) &&
                            DonkeyClient.c_NoTransferID != l_TransferID &&
                            l_TransferID != a_AuxTransferID);

            if (false == l_Valid)
            {
                return false;
            }

            SetTransferID(l_TransferID, a_AuxTransferID);

            f_Flags.ClearFlag(FlagsToUInt(UlPeerCntxtFlags.UnknownTransferIDsAllowed));
            SetUlStatus(UlStatus.PeerWaitsForAccept);

            return true;
        }

        private bool HandleRevocation(Time a_Time)
        {
            if (0 == f_ClustersOnQueue)
            {
                if (true == f_Flags.CheckFlag(FlagsToUInt(UlPeerCntxtFlags.StopTransferWaitingForAck)))
                {
                    if (a_Time < f_StoppingTime)
                    {
                        return true;
                    }

                    goto REMOVE_UL_CONTEXT;
                }
            }
            else if (a_Time < f_StoppingTime)
            {
                return true;
            }

            if (DonkeyClient.c_NoTransferID != p_TransferID &&
                0 == f_ClustersOnQueue)
            {
                bool l_Cleared = ClearPacketQueueAndUlRegisters(false);
                f_Flags.SetFlag(FlagsToUInt(UlPeerCntxtFlags.StopTransferRevocationIsNotAbortable));

                if (false == f_Clnt.SendOnQueueNotification(p_TransferID))
                {
                    return false;
                }

                MutateTransferID(false);

                const uint lc_MaxWaitingTimeForAckInSecs = (4 * Time.c_SecsPerMinute);
                f_StoppingTime = a_Time.AddSeconds(lc_MaxWaitingTimeForAckInSecs);
                f_Flags.SetCombination(FlagsToUInt(UlPeerCntxtFlags.UnknownTransferIDsAllowed | UlPeerCntxtFlags.StopTransferWaitingForAck));

                return true;
            }

        REMOVE_UL_CONTEXT:

            if (DonkeyClient.c_NoTransferID == p_TransferID)
            {
                if (false == f_Clnt.SendOnQueueNotification(DonkeyClient.c_NoTransferID))
                {
                    return false;
                }

                MarkAsRemovable(DestructionInfo.RemoveAtOurRequestPeerDoesNotNeedInfo);
            }
            else
            {
                MarkAsRemovable(DestructionInfo.RemoveAtOurRequestPeerNeedsInfo);
            }

            return true;
        }

        private void NotifyFileContentPieceSent(Packet a_Packet)
        {
            if (true == p_Destroyed)
            {
                return;
            }

            if (false == a_Packet.IsLastPieceOfCluster())
            {
                return;
            }

            UpdateClustersOnQueue(f_ClustersOnQueue - 1u);

            if (UlStatus.AcceptedUlRequest == f_UlStatus)
            {
                Time l_Time = f_Clnt.f_Inst.GetTime();

                if (false == HandleUploads(l_Time))
                {
                    f_Clnt.Close();
                }
            }
        }

        public bool AddRequestedRanges(ulong[] a_Arr,
                                       int a_Slots)
        {
            if (UlStatus.AcceptedUlRequest != f_UlStatus)
            {
                // TWEAK: We're handling a range request that was sent
                // before the peer realized that a former ul peer context
                // got destroyed.
                if (true)
                {
                    MarkAsRemovable(DestructionInfo.RemoveAtOurRequestPeerDoesNotNeedInfo);

                    return true;
                }
            }

            if (true == f_StoppingTime.IsSet())
            {
                if (true == f_Flags.CheckFlag(FlagsToUInt(UlPeerCntxtFlags.StopTransferRevocationIsNotAbortable)))
                {
                    // Ignore this.

                    return true;
                }
            }

            // Invalidate duplicates.
            for (int i = 0; i < a_Slots; i++)
            {
                for (int k = (i + 1); k < a_Slots; k++)
                {
                    if (a_Arr[i] == a_Arr[k] &&
                        a_Arr[a_Slots + i] == a_Arr[a_Slots + k])
                    {
                        a_Arr[a_Slots + k] = 0;
                    }
                }
            }

            // Invalidate the ranges we've already settled and remove
            // obsolete items from the list of settled ranges.
            for (int k = (f_SettledRanges.Count - 1); 0 <= k; k--)
            {
                FileRange l_Range = f_SettledRanges[k];
                bool l_Keep = false;

                for (int i = 0; i < a_Slots; i++)
                {
                    if (l_Range.f_StartingOffset == a_Arr[i] &&
                        l_Range.p_FinalOffset == a_Arr[a_Slots + i])
                    {
                        a_Arr[a_Slots + i] = 0;
                        l_Keep = true;

                        break;
                    }
                }

                if (false == l_Keep)
                {
                    f_SettledRanges.RemoveAt(k);
                }
            }

            for (int i = 0; i < a_Slots; i++)
            {
                if (0 != a_Arr[a_Slots + i] &&
                    false == AddSingleRange(a_Arr[i], a_Arr[a_Slots + i]))
                {
                    return false;
                }
            }

            if (c_MaxRequestedRanges < (f_ReadCookies.Count + f_ClustersOnQueue))
            {
                if (false == f_Clnt.IsThisClient() ||
                    0x10412 < f_Clnt.GetClientVersion() ||
                    (c_MaxRequestedRanges + 1) < (f_ReadCookies.Count + f_ClustersOnQueue))
                {
                    return false;
                }
            }

            Time l_Time = f_Clnt.f_Inst.GetTime();

            return HandleUploads(l_Time);
        }

        private bool AddSingleRange(ulong a_StartingOffset,
                                    ulong a_FinalOffset)
        {
            // No checking for duplicates is done.

            FileRange l_Range;

            if (true == f_Known.CreateFileRange(a_StartingOffset, a_FinalOffset, out l_Range) ||
                DonkeyClient.c_MaxRangeLength >= l_Range.f_Length)
            {
                ReadCookie l_NextUploadCookie = base.ReadAsync(f_Known, ref l_Range);

                if (null != l_NextUploadCookie)
                {
                    f_SettledRanges.Add(l_Range);
                    f_ReadCookies.Enqueue(l_NextUploadCookie);

                    return true;
                }
            }

            return false;
        }

        private static uint FlagsToUInt(UlPeerCntxtFlags a_Flags)
        {
            return (uint)a_Flags;
        }

        private enum UlStatus : byte
        {
            UploadNotRequested = 0,
            PeerWaitsForAccept = 1,
            AcceptedUlRequest = 2,
        }

        private enum UlPeerCntxtFlags : byte
        {
            RegisteredWithUlQueue = 0x01,

            InputMayUpload = 0x2,
            InputMayUploadStateUpdated = 0x4,
            InputRemovedFromQueueByForce = 0x8,

            PeerWaitsForRank = 0x10,
            StopTransferRevocationIsNotAbortable = 0x20,
            StopTransferWaitingForAck = 0x40,
            UnknownTransferIDsAllowed = 0x80,
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct UploadQueue
    {
        private const uint c_ProcessingIntervalMs = 2503;

        // f_LimitMaxDemandItems can be zero, it will always be
        // at least as great as f_LimitEffMaxUploadSlots. f_Demand
        // can have more items than this limit.
        private ushort f_LimitMaxDemandItems;
        // f_EffMaxUploadSlots can be zero.
        private byte f_LimitEffMaxUploadSlots;
        // f_MaxGoodWillSlots and f_MaxSlotsOccupiedBySamePeer
        // start to become effective only when there are too few
        // slots to satisfy all peers that demand a slot.
        private byte f_LimitMaxGoodWillSlots;
        private byte f_LimitMaxSlotsOccupiedBySamePeer;
        private byte f_LimitDefaultMaxSlotsPerFile;

        private FlexDictionary<UlPeerContext, DummyStruct> f_Demand;
        private Time f_NextProcessingTime;

        private uint f_UlQueueFieldsGeneration;

        private FlexArray<QueueSlot> f_Slots;

        private UploadQueue(bool a_Dummy)
        {
            f_LimitMaxDemandItems = 0;
            f_LimitDefaultMaxSlotsPerFile = f_LimitMaxSlotsOccupiedBySamePeer = f_LimitMaxGoodWillSlots = f_LimitEffMaxUploadSlots = 0;

            const byte lc_InitialQueueCapacity = 16;
            FlexDictionary<UlPeerContext, DummyStruct>.Create(lc_InitialQueueCapacity, out f_Demand);
            f_NextProcessingTime = new Time();

            f_UlQueueFieldsGeneration = 0;

            FlexArray<QueueSlot>.Create(out f_Slots);
        }

        public static void Create(out UploadQueue a_Dest)
        {
            a_Dest = new UploadQueue(false);
        }

        public void NotifyPreferencesUpdated(ref NetworkPrefs a_NetworkPrefs)
        {
            const byte lc_Divisor = 5;
            uint l_MaxSlots = DownloadQueue.CalcEffMaxSlotLimit(a_NetworkPrefs.f_MaxGlobalUlSlots, a_NetworkPrefs.f_UlLimit);

            const ushort lc_QueueLength = 1024;
            f_LimitMaxDemandItems = (ushort)Math.Min(ushort.MaxValue, lc_QueueLength + l_MaxSlots);

            if (0 == l_MaxSlots)
            {
                f_LimitMaxSlotsOccupiedBySamePeer = f_LimitMaxGoodWillSlots = f_LimitEffMaxUploadSlots = 0;
            }
            else
            {
                f_LimitEffMaxUploadSlots = (byte)l_MaxSlots;
                f_LimitMaxGoodWillSlots = (byte)(l_MaxSlots / lc_Divisor);
                f_LimitMaxSlotsOccupiedBySamePeer = (byte)((l_MaxSlots / lc_Divisor) + (0 == (l_MaxSlots % lc_Divisor) ? 0u : 1u));
            }

            f_LimitDefaultMaxSlotsPerFile = byte.MaxValue;

            DemandQueueProcessing();
        }

        public bool TryAdd(UlPeerContext a_UlPeerCntxt)
        {
            // This routine returns zero if the context could not
            // be added (because data size limits are reached).

            KnownFile.ActivityContext l_ActCntxt = a_UlPeerCntxt.f_Known.p_ActCntxt;
            UserContext l_UserCntxt = a_UlPeerCntxt.f_Clnt.p_UserCntxt;

            if (0 == f_LimitMaxDemandItems ||
                true == l_ActCntxt.f_UploadQueueFields.IsRegisteredCountMaxed() ||
                true == l_UserCntxt.f_UploadQueueFields.IsRegisteredCountMaxed())
            {
                return false;
            }

            l_ActCntxt.f_UploadQueueFields.IncRegistered();
            l_UserCntxt.f_UploadQueueFields.IncRegistered();

            int l_Junk0;
            FlexDicItem<UlPeerContext, DummyStruct>[] l_Junk1;

            if (false == f_Demand.PrepareAdd(a_UlPeerCntxt, out l_Junk0, out l_Junk1))
            {
                throw (new UnexpectedStateException());
            }

            DemandQueueProcessing();

            return true;
        }

        public bool TryRemove(UlPeerContext a_UlPeerCntxt)
        {
            bool l_Removed = TryRemoveCore(a_UlPeerCntxt);

            return l_Removed;
        }

        private bool TryRemoveCore(UlPeerContext a_UlPeerCntxt)
        {
            // May be called even if Add() was never called
            // with this context. Returns a nonzero value
            // if the context was removed.

            bool l_OccupiedSlot = a_UlPeerCntxt.f_UploadQueueFields.p_OccupiesSlot;

            if (true == l_OccupiedSlot)
            {
                ClearSlotByUlPeerContext(a_UlPeerCntxt);
            }

            bool l_Removed = f_Demand.Remove(a_UlPeerCntxt);

            if (false == l_Removed)
            {
                // No change.

                return false;
            }

            KnownFile.ActivityContext l_ActCntxt = a_UlPeerCntxt.f_Known.p_ActCntxt;

            if (true == l_ActCntxt.f_UploadQueueFields.DecRegistered())
            {
                l_ActCntxt.f_UploadQueueFieldsGeneration = 0;
            }

            bool l_WasLast = a_UlPeerCntxt.f_Clnt.p_UserCntxt.f_UploadQueueFields.DecRegistered();

            DemandQueueProcessing();

            return true;
        }

        private void DemandQueueProcessing()
        {
            f_NextProcessingTime = new Time();
        }

        public void Process(Time a_Time)
        {
            // May only get called by the NetworkManager class.

            ProcessCore(a_Time);
        }

        private void ProcessCore(Time a_Time)
        {
            if (a_Time < f_NextProcessingTime)
            {
                return;
            }

            f_NextProcessingTime = a_Time.AddMilliseconds(c_ProcessingIntervalMs);

            if (true == f_Demand.IsEmpty())
            {
                return;
            }

            DemandContext[] l_DemandList = TryCreateSortedDemandList(a_Time);

            if (null == l_DemandList)
            {
                return;
            }

            ResetPlanning(a_Time, l_DemandList);

            uint a_PrioLen;
            uint a_GoodWillLen;
            UlPeerContext[] l_SuggAllotment = SuggestSlotAllotment(a_Time, l_DemandList, out a_PrioLen, out a_GoodWillLen);

            UpdateSlots(a_Time, l_SuggAllotment, a_PrioLen, a_GoodWillLen);

            UpdateRanks(l_DemandList);
        }

        private DemandContext[] TryCreateSortedDemandList(Time a_Time)
        {
            // This routine returns null if the demand list is
            // empty. It removes surplus items from the queue
            // (where possible), if the maximum length is exceeded.

            uint l_Len = (uint)f_Demand.p_Count;

            DemandContext[] l_DemandList = new DemandContext[l_Len];

            {
                FlexDicItem<UlPeerContext, DummyStruct>[] l_Arr;
                FlexDicEnum<UlPeerContext, DummyStruct> l_Enum;
                l_Arr = f_Demand.ExposeItemArray(out l_Enum);
                uint l_DemandListIdx = 0;

                for (int i; true == l_Enum.GetNextIndex(l_Arr, out i); )
                {
                    UlPeerContext l_UlPeerCntxt = l_Arr[i].p_Key;
                    l_DemandList[l_DemandListIdx++] = new DemandContext(a_Time, l_UlPeerCntxt);
                }
            }

            uint l_AimedMaxDemand = f_LimitMaxDemandItems;

            if (0 == l_AimedMaxDemand)
            {
                BeginDeacceptingAllContexts();

                for (uint i = 0; i != l_Len; i++)
                {
                    UlPeerContext l_UlPeerCntxt = l_DemandList[i].f_UlPeerCntxt;

                    if (true == l_UlPeerCntxt.f_UploadQueueFields.p_OccupiesSlot)
                    {
                        continue;
                    }

                    bool l_Removed = TryRemoveCore(l_UlPeerCntxt);

                    l_UlPeerCntxt.UlQueueContextWasForciblyRemovedFromUlQueue();
                }

                return null;
            }

            if (2 <= l_Len)
            {
                FlexArray<DemandContext> l_FlexArr;
                FlexArray<DemandContext>.Create(l_DemandList, out l_FlexArr);
                l_FlexArr.Sort(DemandContext.CompareDemandContexts);
                l_DemandList = l_FlexArr.ToArray(true);
            }

            // Remove surplus items form the queue (if any and if they
            // don't occupy a slot).

            if (l_Len > l_AimedMaxDemand)
            {
                uint l_NewLen = l_AimedMaxDemand;

                for (uint i = l_AimedMaxDemand; i != l_Len; i++)
                {
                    UlPeerContext l_UlPeerCntxt = l_DemandList[i].f_UlPeerCntxt;

                    if (false == l_UlPeerCntxt.f_UploadQueueFields.p_OccupiesSlot)
                    {
                        bool l_Removed = TryRemoveCore(l_UlPeerCntxt);

                        l_UlPeerCntxt.UlQueueContextWasForciblyRemovedFromUlQueue();
                    }
                    else
                    {
                        // Keep l_DemandList[i].

                        if (i != l_NewLen)
                        {
                            l_DemandList[l_NewLen] = l_DemandList[i];
                        }

                        l_NewLen++;
                    }
                }

                if (l_Len != l_NewLen)
                {
                    Array.Resize(ref l_DemandList, (int)l_NewLen);
                }
            }

            return l_DemandList;
        }

        private void ResetPlanning(Time a_Time,
                                   DemandContext[] a_DemandList)
        {
            if (0 == ++f_UlQueueFieldsGeneration)
            {
                f_UlQueueFieldsGeneration = 1;
            }

            uint l_Gen = f_UlQueueFieldsGeneration;
            uint l_DefaultMaxSlotsPerFile = f_LimitDefaultMaxSlotsPerFile;

            for (uint i = Util.GetArrayLength(a_DemandList); 0 != i--; )
            {
                UlPeerContext l_UlPeerCntxt = a_DemandList[i].f_UlPeerCntxt;
                l_UlPeerCntxt.f_UploadQueueFields.ResetAlloted();

                // Handle file releated data.

                KnownFile l_Known = l_UlPeerCntxt.f_Known;
                KnownFile.ActivityContext l_ActCntxt = l_Known.p_ActCntxt;

                if (l_Gen != l_ActCntxt.f_UploadQueueFieldsGeneration)
                {
                    // We haven't reset this file yet.

                    uint l_FileSpecificLimit = l_Known.p_MaxUlSlots;
                    uint l_MaxUlSlotsForFile = DownloadQueue.CalcEffMaxSlotLimit(KnownFile.c_DefaultSlotCount == l_FileSpecificLimit ? l_DefaultMaxSlotsPerFile : l_FileSpecificLimit, l_Known.f_UlLimit);
                    l_ActCntxt.f_UploadQueueFieldsGeneration = l_Gen;
                    l_ActCntxt.f_UploadQueueFields.ResetPlanningAndSetMaxUlSlots(l_MaxUlSlotsForFile);
                }

                // Handle peer releated data.

                UserContext l_UserCntxt = l_UlPeerCntxt.f_Clnt.p_UserCntxt;

                l_UserCntxt.f_UploadQueueFields.ResetPlanning();
            }
        }

        private UlPeerContext[] SuggestSlotAllotment(Time a_Time,
                                                     DemandContext[] a_DemandList,
                                                     out uint a_PrioLen,
                                                     out uint a_GoodWillLen)
        {
            // The returned array can be of zero length.

            uint l_DemandListLen = Util.GetArrayLength(a_DemandList);

            // Begin deaccepting good will slots that are at the end of
            // their life time.

            if (0 < f_Slots.p_Count)
            {
                QueueSlot[] l_ExistingSlotArr = f_Slots.p_Arr;

                for (uint i = (uint)f_Slots.p_Count; 0 != i--; )
                {
                    if (false == l_ExistingSlotArr[i].p_IsGoodWillSlot ||
                        false == l_ExistingSlotArr[i].p_MayUpload ||
                        false == l_ExistingSlotArr[i].HasReachedEndOfLife(a_Time))
                    {
                        continue;
                    }

                    BeginDeacceptingContextWithPotentialRemove(l_ExistingSlotArr, i);
                }
            }

            // Find the contexts for the prio list.

            uint l_MaxSlotsBySamePeer = f_LimitMaxSlotsOccupiedBySamePeer;
            UlPeerContext[] l_SuggAllotment = new UlPeerContext[f_LimitEffMaxUploadSlots];
            uint l_SuggAllotIdx = 0;

            for (uint l_CurAllowedSlotsPerPeer = l_MaxSlotsBySamePeer, l_RemPrioSlots = ((uint)f_LimitEffMaxUploadSlots - f_LimitMaxGoodWillSlots); 0 != l_RemPrioSlots; )
            {
                uint l_MinPeersRequired = uint.MaxValue;

                for (uint i = 0; i != l_DemandListLen; i++)
                {
                    if (true == a_DemandList[i].p_Ignore)
                    {
                        continue;
                    }

                    UlPeerContext l_UlPeerCntxt = a_DemandList[i].f_UlPeerCntxt;
                    KnownFile l_Known = l_UlPeerCntxt.f_Known;
                    KnownFile.ActivityContext l_ActCntxt = l_Known.p_ActCntxt;
                    uint l_FileMaxSlots = l_ActCntxt.f_UploadQueueFields.p_MaxUlSlots;

                    if (l_ActCntxt.f_UploadQueueFields.p_ContextsPlanned >= l_FileMaxSlots)
                    {
                        goto IGNORE_THIS_CONTEXT_FROM_NOW_ON;
                    }

                    UserContext l_UserCntxt = l_UlPeerCntxt.f_Clnt.p_UserCntxt;
                    uint l_PeerSlotsPlanned = l_UserCntxt.f_UploadQueueFields.p_ContextsPlanned;

                    if (l_PeerSlotsPlanned >= l_CurAllowedSlotsPerPeer)
                    {
                        if (true == l_UserCntxt.f_UploadQueueFields.IsPlannedCountMaxed())
                        {
                            goto IGNORE_THIS_CONTEXT_FROM_NOW_ON;
                        }

                        l_MinPeersRequired = Math.Min(l_MinPeersRequired, l_PeerSlotsPlanned + 1);

                        continue;
                    }

                    l_ActCntxt.f_UploadQueueFields.IncPlanned();
                    l_UserCntxt.f_UploadQueueFields.IncPlanned();

                    if (true == l_UlPeerCntxt.f_UploadQueueFields.p_OccupiesSlot ||
                        l_ActCntxt.f_UploadQueueFields.p_ContextsRegistered < l_FileMaxSlots)
                    {
                        l_UlPeerCntxt.f_UploadQueueFields.SetAlloted();
                        l_SuggAllotment[l_SuggAllotIdx++] = l_UlPeerCntxt;
                    }

                    if (false == l_UserCntxt.f_UploadQueueFields.AreOccupiedAndPlannedCountBelow(l_MaxSlotsBySamePeer))
                    {
                        l_UserCntxt.f_UploadQueueFields.SetShouldCloseGoodWillSlots();
                    }

                    if (0 == --l_RemPrioSlots)
                    {
                        a_DemandList[i].SetIgnore();

                        break;
                    }

                IGNORE_THIS_CONTEXT_FROM_NOW_ON:

                    a_DemandList[i].SetIgnore();
                }

                if (uint.MaxValue == l_MinPeersRequired)
                {
                    // Nothing more to be done.

                    break;
                }

                l_CurAllowedSlotsPerPeer = l_MinPeersRequired;
            }

            // Begin deaccepting contexts that currently occupy good will
            // slots if they are no longer suitable. Also count the
            // number of good will slots (and alike).

            uint l_ExistingGoodWillSlots = 0;

            if (0 < f_Slots.p_Count)
            {
                QueueSlot[] l_ExistingSlotArr = f_Slots.p_Arr;

                for (uint i = (uint)f_Slots.p_Count; 0 != i--; )
                {
                    if (false == l_ExistingSlotArr[i].p_IsGoodWillSlot)
                    {
                        if (true == l_ExistingSlotArr[i].WasRecentlyConverted(a_Time))
                        {
                            l_ExistingGoodWillSlots++;
                        }

                        continue;
                    }

                    l_ExistingGoodWillSlots++;

                    if (false == l_ExistingSlotArr[i].p_MayUpload)
                    {
                        continue;
                    }

                    UlPeerContext l_UlPeerCntxt = l_ExistingSlotArr[i].f_UlPeerCntxt;

                    if (true == l_UlPeerCntxt.f_UploadQueueFields.p_Alloted ||
                        false == l_UlPeerCntxt.f_Clnt.p_UserCntxt.f_UploadQueueFields.p_ShouldCloseGoodWillSlots)
                    {
                        continue;
                    }

                    BeginDeacceptingContextWithPotentialRemove(l_ExistingSlotArr, i);
                }
            }

            // Determine the number of good will slots we may newly
            // create.

            uint l_NewPrioSlots = l_SuggAllotIdx;
            uint l_MaxGoodWillContextsNeeded = (l_ExistingGoodWillSlots >= f_LimitMaxGoodWillSlots ? 0u : (f_LimitMaxGoodWillSlots - l_ExistingGoodWillSlots));

            if (0 != l_MaxGoodWillContextsNeeded)
            {
                uint l_SlotCount = (uint)f_Slots.p_Count;

                if (l_SlotCount >= f_LimitEffMaxUploadSlots)
                {
                    l_MaxGoodWillContextsNeeded = 0;
                }
                else
                {
                    l_MaxGoodWillContextsNeeded = Math.Min(l_MaxGoodWillContextsNeeded, f_LimitEffMaxUploadSlots - l_SlotCount);
                }
            }

            // Find new candidates for good will slots.

            for (uint k = 0; k != l_MaxGoodWillContextsNeeded; k++)
            {
                uint l_CandidateDemandIdx = Util.c_NoIndex;
                Time l_MinLastAcceptedForUlTime = new Time();

                for (uint g = 0, i = (uint)Util.GetRandomIndexFast((int)l_DemandListLen); g != l_DemandListLen; g++, i++)
                {
                    if (l_DemandListLen == i)
                    {
                        i = 0;
                    }

                    if (true == a_DemandList[i].p_Ignore)
                    {
                        continue;
                    }

                    UlPeerContext l_UlPeerCntxt = a_DemandList[i].f_UlPeerCntxt;
                    UserContext l_UserCntxt = l_UlPeerCntxt.f_Clnt.p_UserCntxt;

                    if (true == l_UlPeerCntxt.f_UploadQueueFields.p_OccupiesSlot ||
                        false == l_UlPeerCntxt.f_Known.p_ActCntxt.f_UploadQueueFields.AreOccupiedAndPlannedCountBelowMaxUlSlots() ||
                        false == l_UserCntxt.f_UploadQueueFields.AreOccupiedAndPlannedCountBelow(l_MaxSlotsBySamePeer))
                    {
                        a_DemandList[i].SetIgnore();

                        continue;
                    }

                    Time l_LastAcceptedForUlTime = l_UserCntxt.p_LastAcceptedForUlTime;

                    if (Util.c_NoIndex == l_CandidateDemandIdx ||
                        l_LastAcceptedForUlTime < l_MinLastAcceptedForUlTime)
                    {
                        l_CandidateDemandIdx = i;
                        l_MinLastAcceptedForUlTime = l_LastAcceptedForUlTime;

                        if (false == l_LastAcceptedForUlTime.IsSet())
                        {
                            // It won't get any better.

                            break;
                        }
                    }
                }

                if (Util.c_NoIndex == l_CandidateDemandIdx)
                {
                    break;
                }

                a_DemandList[l_CandidateDemandIdx].SetIgnore();
                UlPeerContext l_CandidateCntxt = a_DemandList[l_CandidateDemandIdx].f_UlPeerCntxt;

                l_CandidateCntxt.f_Known.p_ActCntxt.f_UploadQueueFields.IncPlanned();
                l_CandidateCntxt.f_Clnt.p_UserCntxt.f_UploadQueueFields.IncPlanned();

                l_SuggAllotment[l_SuggAllotIdx++] = l_CandidateCntxt;
            }

            a_PrioLen = l_NewPrioSlots;
            a_GoodWillLen = (l_SuggAllotIdx - l_NewPrioSlots);

            return l_SuggAllotment;
        }

        private void UpdateSlots(Time a_Time,
                                 UlPeerContext[] a_Cntxts,
                                 uint a_PrioLen,
                                 uint a_GoodWillLen)
        {
            if ((uint)f_Slots.p_Arr.Length > f_LimitEffMaxUploadSlots)
            {
                f_Slots.TrimExcess();
            }

            uint l_SuggestedSlots = (a_PrioLen + a_GoodWillLen);

            if (0 == l_SuggestedSlots)
            {
                // We can simple de-accept everything.

                BeginDeacceptingAllContexts();

                return;
            }

            if (0 == f_Slots.p_Count)
            {
                // We can simply take over everything from
                // a_Cntxts.

                f_Slots.ReserveRange(0, (int)l_SuggestedSlots);

                QueueSlot[] l_Arr = f_Slots.p_Arr;

                for (uint i = 0; i != l_SuggestedSlots; i++)
                {
                    bool l_GoodWill = (i >= a_PrioLen);
                    OccupySlot(a_Cntxts[i], a_Time, l_GoodWill, out l_Arr[i]);
                }

                return;
            }

            {
                int l_MinCap = (int)Math.Min(f_LimitEffMaxUploadSlots, (uint)f_Slots.p_Count + l_SuggestedSlots);

                if (f_Slots.p_Arr.Length < l_MinCap)
                {
                    f_Slots.SetCapacity(l_MinCap);
                }
            }

            QueueSlot[] l_SlotsArr = f_Slots.p_Arr;
            uint l_ExistingGoodWillSlots = 0;

            for (uint i = (uint)f_Slots.p_Count; 0 != i--; )
            {
                UlPeerContext l_CurCntxt = l_SlotsArr[i].f_UlPeerCntxt;
                bool l_Match = false;

                for (uint k = 0; k != a_PrioLen; k++)
                {
                    if (l_CurCntxt == a_Cntxts[k])
                    {
                        l_Match = true;

                        break;
                    }
                }

                if (false == l_Match)
                {
                    if (false == l_SlotsArr[i].p_IsGoodWillSlot)
                    {
                        if (true == l_SlotsArr[i].p_MayUpload)
                        {
                            BeginDeacceptingContextWithPotentialRemove(l_SlotsArr, i);
                        }
                    }
                    else
                    {
                        l_ExistingGoodWillSlots++;
                    }

                    continue;
                }

                if (true == l_SlotsArr[i].p_IsGoodWillSlot)
                {
                    l_SlotsArr[i].ClearGoodWillSlotState(a_Time);
                }

                if (false == l_SlotsArr[i].p_MayUpload)
                {
                    ReacceptContext(ref l_SlotsArr[i]);
                }
            }

            uint l_EmptyPrioSlots = 0;

            if (0 != a_PrioLen)
            {
                uint l_SlotCount = (uint)f_Slots.p_Count;

                if (l_SlotCount < f_LimitEffMaxUploadSlots)
                {
                    uint l_SlotsPossible = ((uint)f_LimitEffMaxUploadSlots - f_LimitMaxGoodWillSlots);
                    uint l_SlotsTaken = (l_SlotCount - l_ExistingGoodWillSlots);

                    if (l_SlotsPossible > l_SlotsTaken)
                    {
                        l_EmptyPrioSlots = (l_SlotsPossible - l_SlotsTaken);
                    }
                }
            }

            if (0 != l_EmptyPrioSlots ||
                0 != a_GoodWillLen)
            {
                for (uint i = 0; i != l_SuggestedSlots; i++)
                {
                    bool l_GoodWill = (i >= a_PrioLen);

                    if (false == l_GoodWill)
                    {
                        if (0 == l_EmptyPrioSlots)
                        {
                            // Monkey with the loop variable to
                            // skip all priority slots.

                            i = (a_PrioLen - 1);

                            continue;
                        }

                        if (true == a_Cntxts[i].f_UploadQueueFields.p_OccupiesSlot)
                        {
                            continue;
                        }

                        l_EmptyPrioSlots--;
                    }

                    // Add the context to an unoccupied slot.

                    uint l_NewSlotIdx = (uint)f_Slots.ReserveAtEnd();

                    l_SlotsArr = f_Slots.p_Arr;

                    OccupySlot(a_Cntxts[i], a_Time, l_GoodWill, out l_SlotsArr[l_NewSlotIdx]);
                }
            }
        }

        private void BeginDeacceptingAllContexts()
        {
            QueueSlot[] l_SlotsArr = f_Slots.p_Arr;

            for (uint i = (uint)f_Slots.p_Count; 0 != i--; )
            {
                if (false == l_SlotsArr[i].p_MayUpload)
                {
                    continue;
                }

                BeginDeacceptingContextWithPotentialRemove(l_SlotsArr, i);
            }
        }

        private void UpdateRanks(DemandContext[] a_DemandList)
        {
            const byte lc_RankStepping = 5;
            // 3.
            const byte lc_FirstGroupRank = (lc_RankStepping - (lc_RankStepping / 2));
            // 10.
            const byte lc_RankIncrementIfPeerLimitIsReached = (2 * lc_RankStepping);

            uint l_Rank = lc_FirstGroupRank;
            uint l_RemGroupSize = lc_RankStepping;

            for (uint i = 0, lim = Util.GetArrayLength(a_DemandList); i != lim; i++)
            {
                UlPeerContext l_UlPeerCntxt = a_DemandList[i].f_UlPeerCntxt;

                if (true == l_UlPeerCntxt.f_UploadQueueFields.p_OccupiesSlot)
                {
                    // This context already has a slot.

                    continue;
                }

                if (0 == l_RemGroupSize--)
                {
                    l_Rank += lc_RankStepping;
                    l_RemGroupSize = lc_RankStepping;
                }

                bool l_PeerLimitReached = (l_UlPeerCntxt.f_Clnt.p_UserCntxt.f_UploadQueueFields.p_ContextsOccupyingSlots >= f_LimitMaxSlotsOccupiedBySamePeer);
                uint l_CurRank = l_Rank;

                if (true == l_PeerLimitReached)
                {
                    l_CurRank += lc_RankIncrementIfPeerLimitIsReached;
                }

                l_UlPeerCntxt.UlQueueUpdateQueueRank(Math.Min(NetworkManager.c_MaxRank, l_CurRank));
            }
        }

        private void ClearSlotByUlPeerContext(UlPeerContext a_UlPeerCntxt)
        {
            uint l_SlotIndex = FindSlotIndex(a_UlPeerCntxt);
            ClearSlotByIndex(a_UlPeerCntxt, l_SlotIndex);
        }

        private void ClearSlotByIndex(UlPeerContext a_UlPeerCntxt,
                                      uint a_SlotIdx)
        {
            // a_UlPeerCntxt must match a_SlotIdx.

            f_Slots.DeleteAt((int)a_SlotIdx);

            OccupiesSlotStateChanged(a_UlPeerCntxt, false);
            UserContext l_UserCntxt = a_UlPeerCntxt.f_Clnt.p_UserCntxt;

            if (0 == l_UserCntxt.f_UploadQueueFields.p_ContextsOccupyingSlots)
            {
                l_UserCntxt.SetLastAcceptedForUlTime(Software.sp_MainForm.f_Inst.GetTime());
            }
        }

        private void OccupySlot(UlPeerContext a_UlPeerCntxt,
                                Time a_TimeOfAccept,
                                bool a_IsGoodWillSlot,
                                out QueueSlot a_Slot)
        {
            // This routine also allows uploading.

            OccupiesSlotStateChanged(a_UlPeerCntxt, true);

            a_Slot = new QueueSlot(a_UlPeerCntxt, a_TimeOfAccept, a_IsGoodWillSlot);
            PropagateMayUploadState(ref a_Slot);
        }

        private void OccupiesSlotStateChanged(UlPeerContext a_UlPeerCntxt,
                                              bool a_OccupiesSlot)
        {
            a_UlPeerCntxt.f_UploadQueueFields.AdjustOccupiesSlot(a_OccupiesSlot);
            a_UlPeerCntxt.f_Known.p_ActCntxt.f_UploadQueueFields.ModifyOccupiedByOne(a_OccupiesSlot);
            a_UlPeerCntxt.f_Clnt.p_UserCntxt.f_UploadQueueFields.ModifyOccupiedByOne(a_OccupiesSlot);
        }

        private void BeginDeacceptingContextWithPotentialRemove(QueueSlot[] a_SlotArr,
                                                                uint a_SlotIdx)
        {
            // a_SlotArr[a_SlotIdx].p_MayUpload must be nonzero on
            // entry. It is possible that a_SlotIdx gets removed from
            // f_Slots during the call.

            a_SlotArr[a_SlotIdx].AdjustMayUploadState(false);

            PropagateMayUploadState(ref a_SlotArr[a_SlotIdx]);

            if (true == a_SlotArr[a_SlotIdx].p_IsUploading)
            {
                return;
            }

            ClearSlotByIndex(a_SlotArr[a_SlotIdx].f_UlPeerCntxt, a_SlotIdx);
        }

        private void ReacceptContext(ref QueueSlot a_Slot)
        {
            // a_Slot.p_MayUpload must be zero on entry.

            a_Slot.AdjustMayUploadState(true);

            PropagateMayUploadState(ref a_Slot);
        }

        private void PropagateMayUploadState(ref QueueSlot a_Slot)
        {
            a_Slot.f_UlPeerCntxt.UlQueueAdjustMayUploadState(a_Slot.p_MayUpload);
        }

        public bool TryNotifyIsUploadingStateChanged(UlPeerContext a_UlPeerCntxt,
                                                     bool a_IsUploading)
        {
            if (false == a_UlPeerCntxt.f_UploadQueueFields.p_OccupiesSlot)
            {
                return false;
            }

            uint l_SlotIdx = FindSlotIndex(a_UlPeerCntxt);
            QueueSlot[] l_SlotArr = f_Slots.p_Arr;

            if (false == a_IsUploading &&
                false == l_SlotArr[l_SlotIdx].p_MayUpload)
            {
                // This slot is meant to be closed.

                ClearSlotByIndex(a_UlPeerCntxt, l_SlotIdx);
            }
            else if (l_SlotArr[l_SlotIdx].p_IsUploading != a_IsUploading)
            {
                l_SlotArr[l_SlotIdx].AdjustIsUploadingState(a_IsUploading);
            }
            else
            {
                // No change.

                return true;
            }

            DemandQueueProcessing();

            return true;
        }

        private uint FindSlotIndex(UlPeerContext a_UlPeerCntxt)
        {
            // a_UlPeerCntxt must own an upload slot on entry.

            QueueSlot[] l_SlotArr = f_Slots.p_Arr;

            for (uint i = (uint)f_Slots.p_Count; 0 != i--; )
            {
                if (l_SlotArr[i].f_UlPeerCntxt == a_UlPeerCntxt)
                {
                    return i;
                }
            }

            throw (new UnexpectedStateException());
        }

        public void GetMicroStats(out uint a_UploadingSlots,
                                  out uint a_TotalSlots,
                                  out uint a_QueueLength)
        {
            // On return a_UploadingSlots can be greater than
            // a_TotalSlots (if the slot count got adjusted recently).

            uint l_SlotCount = (uint)f_Slots.p_Count;
            a_UploadingSlots = 0;

            if (0 != l_SlotCount)
            {
                QueueSlot[] l_SlotArr = f_Slots.p_Arr;

                for (uint i = 0; i != l_SlotCount; i++)
                {
                    if (true == l_SlotArr[i].p_IsUploading)
                    {
                        a_UploadingSlots++;
                    }
                }
            }

            a_TotalSlots = f_LimitEffMaxUploadSlots;
            a_QueueLength = ((uint)f_Demand.p_Count - l_SlotCount);
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct DemandContext
        {
            public readonly UlPeerContext f_UlPeerCntxt;
            public readonly byte f_FilePriority;
            public readonly int f_CompareWeight;

            public bool f_Ignore;

            public bool p_Ignore
            {
                get { return f_Ignore; }
            }

            public DemandContext(Time a_Time,
                                 UlPeerContext a_UlPeerCntxt)
            {
                f_UlPeerCntxt = a_UlPeerCntxt;
                f_FilePriority = (byte)a_UlPeerCntxt.f_Known.f_UlPriority.f_Value;

                UserContext l_UserCntxt = a_UlPeerCntxt.f_Clnt.p_UserCntxt;
                l_UserCntxt.UpdateUlQueueCompareWeight(a_Time, false);

                f_CompareWeight = l_UserCntxt.p_UlQueueCompareWeight;
                f_Ignore = false;
            }

            public void SetIgnore()
            {
                f_Ignore = true;
            }

            public static int CompareDemandContexts(ref DemandContext a_A,
                                                    ref DemandContext a_B)
            {
                int l_Res = a_B.f_FilePriority.CompareTo(a_A.f_FilePriority);

                if (0 == l_Res)
                {
                    l_Res = a_B.f_CompareWeight.CompareTo(a_A.f_CompareWeight);
                }

                return l_Res;
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct QueueSlot
        {
            // Don't use the struct default constructor.

            private const uint c_GoodWillMinLifeInSecs = (5 * Time.c_SecsPerMinute);
            private const uint c_GoodWillMaxLifeInSecs = (8 * Time.c_SecsPerMinute);
            private const uint c_ConversionFinishTimeInSecs = (3 * Time.c_SecsPerMinute);

            public readonly UlPeerContext f_UlPeerCntxt;

            private Time f_GoodWillMaxLifeTime;
            private Time f_ConversionFinishTime;
            private bool f_IsGoodWillSlot;

            private bool f_MayUpload;
            private bool f_IsUploading;

            public bool p_IsGoodWillSlot
            {
                get { return f_IsGoodWillSlot; }
            }

            public bool p_MayUpload
            {
                get { return f_MayUpload; }
            }

            public bool p_IsUploading
            {
                get { return f_IsUploading; }
            }

            public QueueSlot(UlPeerContext a_UlPeerCntxt,
                             Time a_TimeOfAccept,
                             bool a_IsGoodWillSlot)
            {
                f_UlPeerCntxt = a_UlPeerCntxt;

                if (false == a_IsGoodWillSlot)
                {
                    f_GoodWillMaxLifeTime = new Time();
                }
                else
                {
                    uint l_Secs = (c_GoodWillMinLifeInSecs + (uint)Util.GetRandomIndexFast((int)(c_GoodWillMaxLifeInSecs - c_GoodWillMinLifeInSecs + 1)));
                    f_GoodWillMaxLifeTime = a_TimeOfAccept.AddSeconds(l_Secs);
                }

                f_ConversionFinishTime = new Time();
                f_IsGoodWillSlot = a_IsGoodWillSlot;
                f_MayUpload = true;
                f_IsUploading = false;
            }

            public bool HasReachedEndOfLife(Time a_Time)
            {
                return (a_Time >= f_GoodWillMaxLifeTime);
            }

            public bool WasRecentlyConverted(Time a_Time)
            {
                return (a_Time < f_ConversionFinishTime);
            }

            public void ClearGoodWillSlotState(Time a_Time)
            {
                f_GoodWillMaxLifeTime = new Time();
                f_ConversionFinishTime = a_Time.AddSeconds(c_ConversionFinishTimeInSecs);
                f_IsGoodWillSlot = false;
            }

            public void AdjustMayUploadState(bool a_MayUpload)
            {
                f_MayUpload = a_MayUpload;
            }

            public void AdjustIsUploadingState(bool a_IsUploading)
            {
                f_IsUploading = a_IsUploading;
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct UlPeerContextFieldsOfUlQueue
        {
            // The default constructor may be used.

            // This structure is meant to be inlined into the ul
            // peer context objects. It is maintained by
            // the UploadQueue structure alone.

            private bool f_OccupiesSlot;
            private bool f_Alloted;

            public bool p_OccupiesSlot
            {
                get { return f_OccupiesSlot; }
            }

            public bool p_Alloted
            {
                get { return f_Alloted; }
            }

            public void AdjustOccupiesSlot(bool a_OccupiesSlot)
            {
                f_OccupiesSlot = a_OccupiesSlot;
            }

            public void ResetAlloted()
            {
                if (true == f_Alloted)
                {
                    f_Alloted = false;
                }
            }

            public void SetAlloted()
            {
                f_Alloted = true;
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct UploadQueueFields
        {
            // The default constructor may be used.

            // This structure is meant to be inlined into known
            // file and user context objects. It is maintained by
            // the UploadQueue structure alone.

            private ushort f_ContextsRegistered;
            private byte f_ContextsOccupyingSlots;

            private byte f_ContextsPlanned;
            private bool f_ShouldCloseGoodWillSlots;
            private byte f_MaxUlSlots;

            public uint p_ContextsRegistered
            {
                get { return f_ContextsRegistered; }
            }

            public uint p_ContextsOccupyingSlots
            {
                get { return f_ContextsOccupyingSlots; }
            }

            public uint p_ContextsPlanned
            {
                get { return f_ContextsPlanned; }
            }

            public bool p_ShouldCloseGoodWillSlots
            {
                get { return f_ShouldCloseGoodWillSlots; }
            }

            public uint p_MaxUlSlots
            {
                get { return f_MaxUlSlots; }
            }

            public bool IsRegisteredCountMaxed()
            {
                return (ushort.MaxValue == f_ContextsRegistered);
            }

            public bool DecRegistered()
            {
                // Returns zero if the count got zero due to the
                // operation.

                return (0 == --f_ContextsRegistered);
            }

            public void IncRegistered()
            {
                f_ContextsRegistered++;
            }

            public void DecOccupied()
            {
                f_ContextsOccupyingSlots--;
            }

            public void IncOccupied()
            {
                f_ContextsOccupyingSlots++;
            }

            public void ModifyOccupiedByOne(bool a_Inc)
            {
                if (false == a_Inc)
                {
                    DecOccupied();
                }
                else
                {
                    IncOccupied();
                }
            }

            public bool IsPlannedCountMaxed()
            {
                return (byte.MaxValue == f_ContextsPlanned);
            }

            public void ResetPlanning()
            {
                if (0 != f_ContextsPlanned)
                {
                    f_ContextsPlanned = 0;
                }

                if (true == f_ShouldCloseGoodWillSlots)
                {
                    f_ShouldCloseGoodWillSlots = false;
                }
            }

            public void ResetPlanningAndSetMaxUlSlots(uint a_MaxUlSlots)
            {
                f_ContextsPlanned = 0;
                f_ShouldCloseGoodWillSlots = false;
                f_MaxUlSlots = (byte)a_MaxUlSlots;
            }

            public void IncPlanned()
            {
                f_ContextsPlanned++;
            }

            public bool AreOccupiedAndPlannedCountBelowMaxUlSlots()
            {
                return AreOccupiedAndPlannedCountBelow(f_MaxUlSlots);
            }

            public bool AreOccupiedAndPlannedCountBelow(uint a_Count)
            {
                return (a_Count > f_ContextsOccupyingSlots &&
                        a_Count > f_ContextsPlanned);
            }

            public void SetShouldCloseGoodWillSlots()
            {
                f_ShouldCloseGoodWillSlots = true;
            }
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct ArrivalTimeEstimator
    {
        // Don't use the struct default constructor.

        private const uint c_DirtyTime = uint.MaxValue;
        public const uint c_UnknownTime = (uint.MaxValue - 1);
        private const uint c_InfiniteTime = (uint.MaxValue - 2);
        private const uint c_MaxTime = (uint.MaxValue - 3);

        private const byte c_SecsPerSlot = (byte)(2 * Time.c_SecsPerMinute);
        private const ushort c_MaxUnitsPerSlot = checked((ushort)((uint)c_SecsPerSlot * 10));
        private const ushort c_MaxSlots = 65;
        private const byte c_DesiredWeightSegments = 8;

        private Timeline<ulong> f_Speeds;
        private uint f_CachedETAInSecs;

        private ArrivalTimeEstimator(bool a_Dummy)
        {
            Timeline<ulong>.Create(c_MaxUnitsPerSlot, c_MaxSlots, out f_Speeds);
            f_CachedETAInSecs = c_UnknownTime;
        }

        public static void Create(out ArrivalTimeEstimator a_Dest)
        {
            a_Dest = new ArrivalTimeEstimator(false);
        }

        public void Clear()
        {
            f_Speeds.Clear();
            f_CachedETAInSecs = c_UnknownTime;
        }

        public void Update(Time a_Time,
                           ulong a_BytesDone)
        {
            uint l_SlotIdx;

            if (Timeline<ulong>.SlotResult.Outdated == f_Speeds.TryCreateSlot(a_Time, out l_SlotIdx))
            {
                return;
            }

            ulong[] l_Arr = f_Speeds.ExposeArray();
            ulong l_OldVal = l_Arr[l_SlotIdx];

            if (l_OldVal == a_BytesDone)
            {
                // Nothing to do.

                return;
            }

            l_Arr[l_SlotIdx] = a_BytesDone;
            f_CachedETAInSecs = c_DirtyTime;
        }

        public uint CalcETAInSeconds(Time a_Time,
                                     ulong a_FileSize)
        {
            // a_Time should be the time of the last update. The return
            // value can be c_UnknownTime if an estimation is not
            // possible.

            uint l_ETAInSecs = f_CachedETAInSecs;

            if (c_DirtyTime == l_ETAInSecs)
            {
                ulong l_BytesDone;

                if (false == f_Speeds.TryGetMostRecent(out l_BytesDone))
                {
                    l_ETAInSecs = c_UnknownTime;
                }
                else
                {
                    l_ETAInSecs = CalcETAInSecondsCore(a_FileSize - l_BytesDone);
                }

                f_CachedETAInSecs = l_ETAInSecs;
            }

            if (c_MaxTime >= l_ETAInSecs)
            {
                // Apply the perspective of a_Time to l_ETAInSecs.

                Time l_Time = f_Speeds.GetNextTime().SubSeconds(c_SecsPerSlot);

                if (l_Time < a_Time)
                {
                    long l_DiffSecs = a_Time.SecondsSince(l_Time);

                    if (l_DiffSecs >= l_ETAInSecs)
                    {
                        l_ETAInSecs = 1;
                    }
                    else
                    {
                        l_ETAInSecs -= (uint)l_DiffSecs;
                    }
                }
            }

            return l_ETAInSecs;
        }

        private uint CalcETAInSecondsCore(ulong a_BytesLeft)
        {
            if (0 == a_BytesLeft)
            {
                return 0;
            }

            uint l_Count = f_Speeds.p_Count;

            if (9 > l_Count)
            {
                return c_UnknownTime;
            }

            l_Count--;
            uint l_ETAInSecs = CalcSpeedAndReturnETA(a_BytesLeft, l_Count);

            if (2 < l_Count &&
                c_MaxTime >= l_ETAInSecs)
            {
                uint l_SecsInvolved = ((l_Count - 1) * c_SecsPerSlot);

                if (l_ETAInSecs < l_SecsInvolved)
                {
                    // We may get a better result by taking less time into
                    // account, as the arrival time is quite low already.

                    uint l_SlotLimit = (l_ETAInSecs / c_SecsPerSlot);

                    if (0 != (l_ETAInSecs % c_SecsPerSlot))
                    {
                        l_SlotLimit++;
                    }

                    l_SlotLimit++;

                    if (l_SlotLimit < l_Count)
                    {
                        // Recalculate.

                        l_ETAInSecs = CalcSpeedAndReturnETA(a_BytesLeft, l_SlotLimit);
                    }
                }
            }

            return l_ETAInSecs;
        }

        private uint CalcSpeedAndReturnETA(ulong a_BytesLeft,
                                           uint a_Slots)
        {
            // a_BytesLeft must not be zero.

            const byte lc_MinSlots = 2;

            if (lc_MinSlots > a_Slots)
            {
                return c_UnknownTime;
            }

            long l_AvgBytesPerSec = CalcBytesPerSec(a_Slots);

            if (0 >= l_AvgBytesPerSec)
            {
                return c_InfiniteTime;
            }

            ulong l_ETAInSecs = (a_BytesLeft / (ulong)l_AvgBytesPerSec);

            if (0 != (a_BytesLeft % (ulong)l_AvgBytesPerSec))
            {
                l_ETAInSecs++;
            }

            return (uint)Math.Min(l_ETAInSecs, c_MaxTime);
        }

        private long CalcBytesPerSec(uint a_Slots)
        {
            // May only get called by CalcSpeedAndReturnETA(). a_Slots
            // may not be lower than 2. Returns the average bytes per
            // second of the recent slots, not counting the most recent.

            long l_Sum = 0;
            uint l_SumOfWeights = 0;

            {
                bool l_MiniSegments = (1 >= c_DesiredWeightSegments ||
                                       (2 * c_DesiredWeightSegments) > a_Slots);
                uint l_SegmentSize = (false == l_MiniSegments ? (a_Slots / c_DesiredWeightSegments) : 1);
                uint l_RemSlots = (a_Slots - 1);

                uint l_Idx;
                ulong[] l_Arr = f_Speeds.ExposeArray(out l_Idx);
                l_Idx = f_Speeds.IncreaseIndex(l_Idx, 1);

                ulong l_BytesDone = l_Arr[l_Idx];

                for (uint l_CurWeight = (2 * c_DesiredWeightSegments);; l_CurWeight--)
                {
                    uint l_CurSegmentSize = Math.Min(l_SegmentSize, l_RemSlots);
                    uint l_NextIdx = f_Speeds.IncreaseIndex(l_Idx, l_CurSegmentSize);
                    ulong l_NextBytesDone = l_Arr[l_NextIdx];
                    long l_BytesDiff = ((long)l_BytesDone - (long)l_NextBytesDone);
                    uint l_Secs = (l_CurSegmentSize * c_SecsPerSlot);

                    l_BytesDiff = (0 == l_BytesDiff ? 0 : (l_BytesDiff / l_Secs));
                    l_Sum += (l_CurWeight * l_BytesDiff);
                    l_SumOfWeights += l_CurWeight;

                    l_RemSlots -= l_CurSegmentSize;

                    if (0 == l_RemSlots)
                    {
                        break;
                    }

                    l_Idx = l_NextIdx;
                    l_BytesDone = l_NextBytesDone;
                }
            }

            long l_AvgBytesPerSec = (0 == l_Sum ? 0 : (l_Sum / l_SumOfWeights));

            return l_AvgBytesPerSec;
        }

        public static bool IsUnknownTime(uint a_ETASecs)
        {
            return (c_UnknownTime == a_ETASecs);
        }
    }

    class Packet
    {
        public const uint c_MaxAllowedBodySize = (50 * 1024 * 1024);

        public delegate void PacketBytesSentHandler(uint a_NetworkPayloadSize);
        public delegate void PacketCompletelySentHandler(Packet a_Packet);

        public readonly bool f_Received;
        public readonly I2PAddress f_RemoteAddr;

        public readonly Protocols f_Prot;
        private readonly byte f_Opcode;
        // f_RelatedUlFile can be null, but won't be for certain stream
        // uploads.
        public readonly KnownFile f_RelatedUlFile;

        private readonly BufferIO f_ReceivedData;
        private readonly byte[] f_SendReadyData;
        // f_TransferSize indicates the in-transfer size of the packet,
        // including everything that belongs to the protocols
        // (see enumeration Protocols) we support, but not including
        // overhead due to SAM or I2P.
        public readonly uint f_TransferSize;

        // When sending a compressed file range as a series of packets,
        // f_CumComprOffset represents the incremental (compressed data)
        // body size of all former packets of that series. It does not
        // contain overhead.
        private uint f_CumComprOffset;
        // When sending a packet, f_CanBeReorderedToBack will tell us if
        // other packets may travel earlier even if they were enqueued
        // later. Should be true for certain large packets.
        private bool f_CanBeReorderedToBack;

        private uint f_ClusterNumber;
        private byte f_PieceNumber;
        private byte f_PieceCountOfCluster;
        private bool f_WasClusterAborted;

        // The following events may be set after the packet was
        // already enqueued.
        public event PacketBytesSentHandler e_SentBytes;
        public event PacketCompletelySentHandler e_SentCompletely;

        public BufferIO p_Buffer
        {
            get
            {
                return f_ReceivedData;
            }
        }

        public byte p_RawOpcode
        {
            get { return f_Opcode; }
        }

        public KadOpcodes p_KadOpcode
        {
            get
            {
                return (KadOpcodes)f_Opcode;
            }
        }

        public Opcodes p_DonkeyOpcode
        {
            get
            {
                return (Opcodes)f_Opcode;
            }
        }

        public OpcodesEx p_DonkeyOpcodesEx
        {
            get
            {
                return (OpcodesEx)f_Opcode;
            }
        }

        public uint p_CumComprOffset
        {
            get { return f_CumComprOffset; }
        }

        public bool p_CanBeReorderedBehind
        {
            get { return f_CanBeReorderedToBack; }
        }

        public uint p_ClusterNumber
        {
            get { return f_ClusterNumber; }
        }

        public uint p_PieceNumber
        {
            get { return f_PieceNumber; }
        }

        private Packet(I2PAddress a_Origin,
                       Protocols a_Prot,
                       byte a_Opcode,
                       BufferIO a_ReceivedData,
                       uint a_TransferSize)
        {
            // Constructor for received packets.

            f_Received = true;
            f_RemoteAddr = a_Origin;
            f_Prot = a_Prot;
            f_Opcode = a_Opcode;
            f_ReceivedData = a_ReceivedData;
            f_TransferSize = a_TransferSize;
        }

        private Packet(I2PAddress a_Destination,
                       Protocols a_Prot,
                       byte a_Opcode,
                       KnownFile a_RelatedUlFile,
                       byte[] a_SendReadyData)
        {
            // Constructor for sendable packets.

            f_Received = false;
            f_RemoteAddr = a_Destination;
            f_Prot = a_Prot;
            f_Opcode = a_Opcode;
            f_RelatedUlFile = a_RelatedUlFile;
            f_SendReadyData = a_SendReadyData;
            f_TransferSize = (uint)f_SendReadyData.Length;
        }

        public static Packet BuildOwnPacket(I2PAddress a_Dest,
                                            KadOpcodes a_Opcode,
                                            bool a_AllowCompression,
                                            ref ByteArray a_Data)
        {
            return BuildOwnPacketCore(a_Dest, Protocols.OP_KADEMLIAHEADER, (byte)a_Opcode, null, a_AllowCompression, false, ref a_Data);
        }

        public static Packet BuildOwnPacket(I2PAddress a_Dest,
                                            Opcodes a_Opcode,
                                            KnownFile a_RelatedFile,
                                            bool a_AllowCompression,
                                            ref ByteArray a_Data)
        {
            return BuildOwnPacketCore(a_Dest, DonkeyClient.SuggestProtocol(a_Opcode), (byte)a_Opcode, a_RelatedFile, a_AllowCompression, false, ref a_Data);
        }

        public static Packet BuildOwnPacket(I2PAddress a_Dest,
                                            OpcodesEx a_OpcodeEx,
                                            bool a_AllowCompression,
                                            bool a_LegacyTranslation,
                                            ref ByteArray a_Data)
        {
            return BuildOwnPacketCore(a_Dest, Protocols.PROT_NB_PLAIN, (byte)a_OpcodeEx, null, a_AllowCompression, a_LegacyTranslation, ref a_Data);
        }

        private static Packet BuildOwnPacketCore(I2PAddress a_Dest,
                                                 Protocols a_Prot,
                                                 byte a_Opcode,
                                                 KnownFile a_RelatedFile,
                                                 bool a_AllowCompression,
                                                 bool a_LegacyTranslation,
                                                 ref ByteArray a_Data)
        {
            // a_RelatedFile and a_Data.f_Arr may be null.

            Protocols l_SendProt;
            byte l_SendOpcode;

            if (false == a_LegacyTranslation)
            {
                l_SendProt = a_Prot;
                l_SendOpcode = a_Opcode;
            }
            else
            {
                l_SendProt = Protocols.OP_EDONKEYPROT;
                l_SendOpcode = (byte)(a_Opcode - (uint)OpcodesEx.OP_NOOP + (uint)Opcodes.OP_LEGACY_START);
            }

            byte[] l_SendReadyData = MakeSendReady(l_SendProt, l_SendOpcode, a_AllowCompression, ref a_Data);

            return new Packet(a_Dest, a_Prot, a_Opcode, a_RelatedFile, l_SendReadyData);
        }

        public static Packet RecreateOwnPacket(Packet a_Template,
                                               I2PAddress a_NewDestination)
        {
            // Recreates with changed destination.

            return new Packet(a_NewDestination, a_Template.f_Prot, a_Template.f_Opcode, a_Template.f_RelatedUlFile, a_Template.f_SendReadyData);
        }

        public void SetClusterNumber(uint a_ClusterNumber,
                                     uint a_PieceNumber,
                                     uint a_PieceCountOfCluster)
        {
            f_ClusterNumber = a_ClusterNumber;
            f_PieceNumber = (byte)a_PieceNumber;
            f_PieceCountOfCluster = (byte)a_PieceCountOfCluster;
        }

        public bool IsLastPieceOfCluster()
        {
            return ((f_PieceNumber + 1) == f_PieceCountOfCluster ||
                    true == f_WasClusterAborted);
        }

        public void SetClusterAborted()
        {
            f_WasClusterAborted = true;
        }

        public void SetCumulativeCompressionOffset(uint a_CumComprOffset)
        {
            f_CumComprOffset = a_CumComprOffset;
        }

        public void SetReorderableToBack()
        {
            f_CanBeReorderedToBack = true;
        }

        public uint GetReceivedBodySize()
        {
            return (uint)f_ReceivedData.p_Length;
        }

        public bool IsNoop()
        {
            return (Protocols.PROT_NB_PLAIN == f_Prot &&
                    OpcodesEx.OP_NOOP == (OpcodesEx)f_Opcode);
        }

        public bool IsContent()
        {
            return DonkeyClientBase.IsContent(f_Prot, (Opcodes)f_Opcode);
        }

        public bool IsUlContent(KnownFile a_File)
        {
            return (false == f_Received &&
                    f_RelatedUlFile == a_File &&
                    true == IsContent());
        }

        public void BuildByteArrayForSending(uint a_Idx,
                                             out ByteArray a_Data)
        {
            a_Data = new ByteArray(f_SendReadyData, (int)a_Idx);
        }

        public void BuildByteArrayForSending(uint a_Idx,
                                             uint a_Len,
                                             out ByteArray a_Data)
        {
            a_Data = new ByteArray(f_SendReadyData, (int)a_Idx, (int)a_Len);
        }

        public void WriteToFileForSending(BufferIO a_DestFile,
                                          uint a_SrcIdx,
                                          uint a_Len)
        {
            ByteArray l_Bytes = new ByteArray(f_SendReadyData, (int)a_SrcIdx, (int)a_Len);
            a_DestFile.WriteBytes(ref l_Bytes);
        }

        public void NotifySendingPacket(uint a_Bytes,
                                        bool a_Completed)
        {
            // a_Bytes is the number of bytes just sent, a_Completed
            // tells if the final byte was sent. This routine will be
            // called exactly once with a_Completed set for every
            // packet that was completely sent.

            // There are no restrictions what the callee can do
            // in the handler routine (new packets may be enqueued
            // into the sending queue).

            if (0 != a_Bytes &&
                null != e_SentBytes)
            {
                e_SentBytes(a_Bytes);
            }

            if (true == a_Completed &&
                null != e_SentCompletely)
            {
                e_SentCompletely(this);
            }
        }

        public static Packet InterpretIncoming(I2PAddress a_Origin,
                                               Protocols a_Prot,
                                               byte a_Opcode,
                                               ref ByteArray a_Body,
                                               uint a_MaxDecompressedSize,
                                               uint a_TransferSize)
        {
            // The newly created instance will take ownership of the
            // passed byte array.

            Protocols l_EffProt;
            MemoryIO l_Mem;

            bool l_IsPacked = (Protocols.OP_PACKEDPROT == a_Prot ||
                               Protocols.OP_KADEMLIAPACKEDPROT == a_Prot ||
                               Protocols.PROT_NB_COMPRESSED == a_Prot);

            if (false == l_IsPacked)
            {
                l_EffProt = a_Prot;
                l_Mem = MemoryIO.CreateReadable(ref a_Body);
            }
            else
            {
                ByteArray l_PlainData;

                if (false == zlib.Uncompress(ref a_Body, a_MaxDecompressedSize, out l_PlainData))
                {
                    return null;
                }

                l_Mem = MemoryIO.CreateReadable(ref l_PlainData);

                switch (a_Prot)
                {
                    case Protocols.OP_PACKEDPROT:
                        l_EffProt = Protocols.OP_EMULEPROT;
                        break;

                    case Protocols.OP_KADEMLIAPACKEDPROT:
                        l_EffProt = Protocols.OP_KADEMLIAHEADER;
                        break;

                    case Protocols.PROT_NB_COMPRESSED:
                        l_EffProt = Protocols.PROT_NB_PLAIN;
                        break;

                    default: throw (new UnexpectedStateException());
                }
            }

            byte l_EffOpcode;

            if (Protocols.OP_EDONKEYPROT == l_EffProt &&
                Opcodes.OP_LEGACY_START <= (Opcodes)a_Opcode &&
                Opcodes.OP_LEGACY_END >= (Opcodes)a_Opcode)
            {
                l_EffProt = Protocols.PROT_NB_PLAIN;
                l_EffOpcode = (byte)(a_Opcode - (uint)Opcodes.OP_LEGACY_START + (uint)OpcodesEx.OP_NOOP);
            }
            else
            {
                l_EffOpcode = a_Opcode;
            }

            return new Packet(a_Origin,
                              l_EffProt,
                              l_EffOpcode,
                              l_Mem,
                              a_TransferSize);
        }

        private static byte[] MakeSendReady(Protocols a_Prot,
                                            byte a_Opcode,
                                            bool a_AllowCompression,
                                            ref ByteArray a_Body)
        {
            // a_Body.f_Arr can be null.

            if (true == a_AllowCompression &&
                300 <= a_Body.f_Len &&
                ((Protocols.OP_KADEMLIAHEADER == a_Prot &&
                // I2P addresses are not really compressable, so skip
                // all kad opcodes that are mainly about addresses.
                  (byte)KadOpcodes.KADEMLIA_SEARCH_REQ <= a_Opcode) ||
                 Protocols.PROT_NB_PLAIN == a_Prot))
            {
                ByteArray l_CompData;
                if (true == zlib.Compress(ref a_Body, true, out l_CompData))
                {
                    Protocols l_EffProt;

                    switch (a_Prot)
                    {
                        case Protocols.OP_EMULEPROT:
                            l_EffProt = Protocols.OP_PACKEDPROT;
                            break;

                        case Protocols.OP_KADEMLIAHEADER:
                            l_EffProt = Protocols.OP_KADEMLIAPACKEDPROT;
                            break;

                        case Protocols.PROT_NB_PLAIN:
                            l_EffProt = Protocols.PROT_NB_COMPRESSED;
                            break;

                        default: throw (new UnexpectedStateException());
                    }

                    return MakeSendReadyCore(l_EffProt, a_Opcode, ref l_CompData);
                }
            }

            return MakeSendReadyCore(a_Prot, a_Opcode, ref a_Body);
        }

        private static byte[] MakeSendReadyCore(Protocols a_EffProt,
                                                byte a_Opcode,
                                                ref ByteArray a_Data)
        {
            bool l_IncludeLenInfo = (Protocols.OP_KADEMLIAHEADER != a_EffProt &&
                                     Protocols.OP_KADEMLIAPACKEDPROT != a_EffProt);
            int l_HeaderLen = (sizeof(byte) + (false == l_IncludeLenInfo ? 0 : sizeof(uint)));
            int l_BodyLen = (sizeof(byte) + a_Data.f_Len);
            byte[] l_SendBytes = new byte[l_HeaderLen + l_BodyLen];

            // Copy protocol (byte).
            l_SendBytes[0] = (byte)a_EffProt;

            if (true == l_IncludeLenInfo)
            {
                // Copy size of body (uint).
                Util.WriteUInt32LE((uint)l_BodyLen, l_SendBytes, sizeof(byte));
            }

            // Copy opcode (byte).
            l_SendBytes[l_HeaderLen] = a_Opcode;

            if (0 < a_Data.f_Len)
            {
                // Copy body.
                Buffer.BlockCopy(a_Data.f_Arr, a_Data.f_Idx, l_SendBytes, l_HeaderLen + sizeof(byte), a_Data.f_Len);
            }

            return l_SendBytes;
        }
    }
}

namespace NB.Content
{
    enum DownloadControl : byte
    {
        DlDisabled = 0,
        DlEnabled = 1,
    }

    partial class KnownFile
    {
        public DownloadControl f_DlControl;

        public Speed f_DlLimit;
        public Speed f_UlLimit;

        public Priority f_DlPriority;
        public Priority f_UlPriority;

        private TransferVolume f_TransferVol;
        private uint f_SecsElapsed;

        public Time f_AddedOn;
        public Time f_LastReception;
        public Time f_LastSeenComplete;

        // f_ActCntxt can be null.
        private ActivityContext f_ActCntxt;

#region eD2kSpecificFields

        // The following four values can be safely casted to a
        // (signed) int.
        public readonly uint f_PartCount;
        public readonly uint f_PartCountFileStatus;
        public readonly uint f_PartCountHashSet;
        public readonly uint f_eD2kBitArrayByteSize;

        // f_FileID.f_Val can be null and is owned by the content
        // manager.
        public FileID f_FileID;

        public FileRatings f_OwnRating;
        // f_OwnComment can be null.
        public string f_OwnComment;

#endregion

        public ActivityContext p_ActCntxt
        {
            // May only be used by the network thread and only while
            // the file is active.
            get { return f_ActCntxt; }
        }

        public bool p_eD2kCompleted
        {
            // May only be used by the network thread and only while
            // the file is active.
            get { return (f_PartCount == f_ActCntxt.p_eD2kPartsDone); }
        }

        private void WriteNetworkParamsToBuffer(BufferIO a_Buffer)
        {
            // This routine is used by the content manager.

            if (ComplStatus.Incomplete == f_Status)
            {
                a_Buffer.WriteByte((byte)f_DlControl);
            }

            a_Buffer.WriteUInt32LE(f_DlLimit.p_BytesPerSec);
            a_Buffer.WriteUInt32LE(f_UlLimit.p_BytesPerSec);

            a_Buffer.WriteByte(f_DlPriority.f_Value);
            a_Buffer.WriteByte(f_UlPriority.f_Value);
            f_TransferVol.ToBuffer(a_Buffer);

            // Ignore f_ActiveSince here.
            a_Buffer.WriteUInt32LE(f_SecsElapsed);
            a_Buffer.WriteTime(f_AddedOn);
            a_Buffer.WriteTime(f_LastReception);
            a_Buffer.WriteTime(f_LastSeenComplete);

            a_Buffer.WriteByte((byte)f_OwnRating);
            a_Buffer.WriteUTF16LEStringEncLen(f_OwnComment);
        }

        private bool ReadNetworkParamsFromBuffer(BufferIO a_Buffer)
        {
            // This routine is used by the content manager.

            if (ComplStatus.Incomplete == f_Status)
            {
                f_DlControl = (DownloadControl)a_Buffer.ReadByte();

                if (DownloadControl.DlEnabled < f_DlControl)
                {
                    f_DlControl = DownloadControl.DlEnabled;
                }
            }

            f_DlLimit = Speed.CreateSafe(a_Buffer.ReadUInt32LE(), Speed.s_MaxSpeed);
            f_UlLimit = Speed.CreateSafe(a_Buffer.ReadUInt32LE(), Speed.s_MaxSpeed);

            f_DlPriority = Priority.CreateSafe(a_Buffer.ReadByte());
            f_UlPriority = Priority.CreateSafe(a_Buffer.ReadByte());

            TransferVolume.FromBuffer(a_Buffer, out f_TransferVol);

            f_SecsElapsed = a_Buffer.ReadUInt32LE();
            f_AddedOn = a_Buffer.ReadFormerDateTime();
            f_LastReception = a_Buffer.ReadFormerDateTime();
            f_LastSeenComplete = a_Buffer.ReadFormerDateTime();

            f_OwnRating = (FileRatings)a_Buffer.ReadByte();

            if (FileRatings.Excellent < f_OwnRating)
            {
                f_OwnRating = FileRatings.NotRated;
            }

            f_OwnComment = a_Buffer.ReadUTF16LEStringEncLen(true, Kademlia.c_MaxCommentChars);

            return true;
        }

        private void InitSpeedRelatedParams(bool a_Started)
        {
            f_DlControl = (false == a_Started ? DownloadControl.DlDisabled : DownloadControl.DlEnabled);
            f_DlPriority = Priority.s_Default;
            f_UlPriority = Priority.s_Default;
            f_DlLimit = Speed.s_MaxSpeed;
            f_UlLimit = Speed.s_MaxSpeed;
        }

        public bool IsDownloadAllowed()
        {
            // Tests only the fields that are under the user's control.

            return (DownloadControl.DlEnabled == f_DlControl &&
                    false == f_DlLimit.IsNearZero());
        }

        public bool IsUlWorthy()
        {
            return (false == f_UlLimit.IsNearZero() &&
                    0 != p_MaxUlSlots);
        }

        public bool GetFileIDVolatile(out FileID a_FileID)
        {
            // Returns true if there is a valid file id.

            a_FileID = f_FileID;
            Util.MemoryBarrier();

            return (null != a_FileID.f_Val);
        }

        public void CopyTransferVolumeVolatile(out TransferVolume a_Dest)
        {
            // May be used by any thread.

            f_TransferVol.CopyToVolatile(out a_Dest);
        }

        public void SentContent(uint a_Size)
        {
            f_TransferVol.Sent(a_Size);
        }

        public void RcvdContent(uint a_Size)
        {
            f_TransferVol.Received(a_Size);
        }

        private void Activate(Instance a_Inst)
        {
            Util.VolatileWrite(out f_ActCntxt, new ActivityContext(this, a_Inst.GetTime()));

            a_Inst.f_NetworkMngr.RegisterFile(this);
        }

        private void Deactivate(Instance a_Inst)
        {
            long l_Elapsed = f_ActCntxt.GetSecsActive(a_Inst.GetTime());
            f_SecsElapsed += (uint)l_Elapsed;

            f_ActCntxt = null;

            a_Inst.f_NetworkMngr.UnregisterFile(this);
        }

        public bool IsActive()
        {
            // May be used by any thread.

            return (null != f_ActCntxt);
        }

        public bool GetActivityContextVolatile(out ActivityContext a_ActCntxt)
        {
            // May be used by any thread. Return can be null.

            if (null == f_ActCntxt)
            {
                // Try to prevent the expensive memory barrier in this
                // case.

                a_ActCntxt = null;

                return false;
            }

            a_ActCntxt = Util.VolatileRead(ref f_ActCntxt);

            return (null != a_ActCntxt);
        }

        public void TryCreateDlPeerContext(Instance a_Inst,
                                           NetworkManager.ConnResultContext a_ConnResCntxt)
        {
            ConnectResult l_ConnRes = a_ConnResCntxt.p_Result;

            if (ConnectResult.Success != l_ConnRes)
            {
                SourceStates l_SourceState;

                switch (l_ConnRes)
                {
                    case ConnectResult.NoResponse:
                        l_SourceState = SourceStates.CantReachPeer;
                        break;

                    case ConnectResult.DestinationBanned:
                    case ConnectResult.UserBanned:
                        l_SourceState = SourceStates.PeerBanned;
                        break;

                    default:
                        l_SourceState = SourceStates.UnspecifiedConnErr;
                        break;
                }

                a_ConnResCntxt.NotifySourceState(a_Inst, l_SourceState);

                return;
            }

            DonkeyClient l_Clnt = a_ConnResCntxt.p_Client;

            if (true == l_Clnt.p_UserCntxt.PreventsA4AFCreatingAnotherDlContext(false))
            {
                a_ConnResCntxt.NotifySourceState(a_Inst, SourceStates.A4AFLimitation);
                l_Clnt.Close();

                return;
            }

            bool l_First = (null == f_ActCntxt);

            if (true == l_First)
            {
                Activate(l_Clnt.f_Inst);
            }

            DlPeerContext l_DlPeerCntxt = new DlPeerContext(this, a_ConnResCntxt.f_FileID, l_Clnt);

            SourceStates l_FailReason;
            if (false == f_ActCntxt.TryAddDlPeerContext(l_DlPeerCntxt, out l_FailReason))
            {
                l_DlPeerCntxt.Destroy();

                if (true == l_First)
                {
                    Deactivate(l_Clnt.f_Inst);
                }

                a_ConnResCntxt.NotifySourceState(a_Inst, l_FailReason);
                l_Clnt.Close();

                return;
            }

            l_DlPeerCntxt.Initialize();
        }

        public UlPeerContext TryCreateUlPeerContext(FileID a_FileID,
                                                    DonkeyClient a_Client)
        {
            // Returns null if there is already a context for this
            // destination or the limit is reached or this file isn't
            // meant to be uploaded.

            if (false == IsUlWorthy() ||
                ContentManager.AccessLevel.AccessGranted != a_Client.f_Inst.f_ContentMngr.CheckUlAccess(this, ShareKinds.eD2k_on_I2P, a_Client.p_UserCntxt))
            {
                return null;
            }

            bool l_First = (null == f_ActCntxt);

            if (true == l_First)
            {
                Activate(a_Client.f_Inst);

                if (0 == f_ActCntxt.p_eD2kPartsDone)
                {
                    // This file isn't uploadable currently.

                    Deactivate(a_Client.f_Inst);

                    return null;
                }
            }

            UlPeerContext l_UlPeerCntxt = new UlPeerContext(this, a_FileID, a_Client);

            if (false == f_ActCntxt.TryAddUlPeerContext(l_UlPeerCntxt))
            {
                l_UlPeerCntxt.Destroy();

                if (true == l_First)
                {
                    Deactivate(a_Client.f_Inst);
                }

                return null;
            }

            l_UlPeerCntxt.Initialize();

            return l_UlPeerCntxt;
        }

        public void DestroyDlPeerContext(DlPeerContext a_DlPeerCntxt)
        {
            a_DlPeerCntxt.Destroy();
            a_DlPeerCntxt.f_Clnt.DeleteDlPeerContext();

            f_ActCntxt.RemoveDlPeerContext(a_DlPeerCntxt);

            if (true == f_ActCntxt.IsEmpty())
            {
                Deactivate(a_DlPeerCntxt.f_Clnt.f_Inst);
            }
        }

        public void DestroyUlPeerContext(UlPeerContext a_UlPeerCntxt)
        {
            a_UlPeerCntxt.Destroy();
            a_UlPeerCntxt.f_Clnt.DeleteUlPeerContext();

            f_ActCntxt.RemoveUlPeerContext(a_UlPeerCntxt);

            if (true == f_ActCntxt.IsEmpty())
            {
                Deactivate(a_UlPeerCntxt.f_Clnt.f_Inst);
            }
        }

        public uint GetBaseSecsActive()
        {
            // The return value does not include the active session.

            return f_SecsElapsed;
        }

        public bool IsDownloadedVolatile()
        {
            // May be used by any thread.

            ActivityContext l_ActCntxt;

            return (true == GetActivityContextVolatile(out l_ActCntxt) &&
                    true == l_ActCntxt.HasAnyDlPeerContext());
        }

        public bool IsUploadedVolatile()
        {
            // May be used by any thread.

            ActivityContext l_ActCntxt;

            return (true == GetActivityContextVolatile(out l_ActCntxt) &&
                    true == l_ActCntxt.HasAnyUlPeerContext());
        }

        public bool CreateFileRange(ulong a_StartingOffset,
                                    ulong a_FinalOffset,
                                    out FileRange a_FileRange)
        {
            // Invalid ranges will be rejected.

            if (f_FileSize >= a_FinalOffset)
            {
                return FileRange.CreateFromUnsafeBounds(a_StartingOffset, a_FinalOffset, out a_FileRange);
            }

            a_FileRange = new FileRange();

            return false;
        }

#region Transfer

        public bool ReinitializeForDl(out TransferDemand a_Demand)
        {
            return f_ActCntxt.ReinitializeForDl(out a_Demand);
        }

        public bool ReinitializeForUl(out TransferDemand a_Demand)
        {
            return f_ActCntxt.ReinitializeForUl(out a_Demand);
        }

        public uint AssignDlBandwidth(uint a_ExtraBytes)
        {
            return f_ActCntxt.AssignDlBandwidth(a_ExtraBytes);
        }

        public uint AssignUlBandwidth(uint a_ExtraBytes)
        {
            return f_ActCntxt.AssignUlBandwidth(a_ExtraBytes);
        }

        public uint Download(uint a_MaxBytes)
        {
            return f_ActCntxt.Download(a_MaxBytes);
        }

        public uint Upload(uint a_MaxBytes)
        {
            return f_ActCntxt.Upload(a_MaxBytes);
        }

#endregion

        public class ActivityContext
        {
            private readonly KnownFile f_Known;

            private Time f_ActiveSince;
            private SpeedOMeter f_SpeedOMeter;
            private ArrivalTimeEstimator f_ArrivalTimeEstimator;
            // f_ETAInSecs can be an unknown value, in which case
            // ArrivalTimeEstimator.IsUnknownTime() will return a
            // nonzero value. f_ETAInSecs allows thread-safe retrieval.
            private uint f_ETAInSecs;
            private Speed f_DlSpeed;
            private Speed f_UlSpeed;
            private Time f_NextMeasureSpeedTime;

            // Both can be null.
            private Dictionary<I2PAddress, DlPeerContext> f_DlPeers;
            private Dictionary<I2PAddress, UlPeerContext> f_UlPeers;
            private byte f_UploadingUlPeerCntxts;
            // f_GuiNeedsPeerInformation is updated by the GUI and
            // indicates demand for the below arrays.
            private bool f_GuiNeedsPeerInformation;
            // They following arrays can be safely used by the GUI
            // (but can be null).
            private DlPeerContext[] f_DlPeersForGui;
            private UlPeerContext[] f_UlPeersForGui;
            // f_PeersForGuiDirty is used by the network thread to
            // indicate of the arrays.
            private bool f_PeersForGuiDirty;

            private BandwidthPrioritizer f_DlPrioritizer;
            private BandwidthPrioritizer f_UlPrioritizer;

            public uint f_DownloadsRunning;

#region eD2kSpecificFields

            private bool f_MapRevSet;
            private ushort f_MapRev;
            // The following fields may only be changed by the network
            // thread, other threads may access them, but have to mind
            // their volatile nature.
            private readonly BitMap f_NativeOwnCompl;
            // f_eD2kOwnCompl has a granularity of 9,500 KiB.
            private readonly BitMap f_eD2kOwnCompl;
            private uint f_PartsDone;
            private byte f_CompleteNodes;
            private byte f_PartialNodes;
            private readonly NetworkAvailability f_eD2kAvail;
            private float f_eD2kAvailRatio;
            private bool f_Fruitfulness;

#endregion

            public uint f_UploadQueueFieldsGeneration;
            public UploadQueue.UploadQueueFields f_UploadQueueFields;

            public uint p_eD2kPartsDone
            {
                get { return f_PartsDone; }
            }

            public uint p_CompleteNodes
            {
                get { return f_CompleteNodes; }
            }

            public uint p_PartialNodes
            {
                get { return f_PartialNodes; }
            }

            public NetworkAvailability p_eD2kAvail
            {
                get { return f_eD2kAvail; }
            }

            public ActivityContext(KnownFile a_Known,
                                   Time a_Time)
            {
                f_Known = a_Known;

                f_ActiveSince = a_Time;
                SpeedOMeter.Create(out f_SpeedOMeter);
                ArrivalTimeEstimator.Create(out f_ArrivalTimeEstimator);
                ClearCachedETA();

                f_NativeOwnCompl = BitMap.CreateFromFileSize(BitMap.c_NativeGranularity, f_Known.f_FileSize);
                f_eD2kOwnCompl = new BitMap(BitMap.c_eD2kGranularity, f_Known.f_PartCount);
                f_eD2kAvail = new NetworkAvailability(f_Known.f_PartCount);

                UpdateOwnCompleteness(a_Time);
            }

            public bool IsEmpty()
            {
                return (null == f_DlPeers &&
                        null == f_UlPeers);
            }

            public bool IsDownloading()
            {
                // Meant to be used by the GUI.

                return (0 != f_DownloadsRunning);
            }

            public uint GetSecsActive(Time a_Time)
            {
                // Meant to be used by the GUI.

                return (uint)a_Time.SecondsSince(f_ActiveSince);
            }

            public bool TryAddDlPeerContext(DlPeerContext a_DlPeerCntxt,
                                            out SourceStates a_FailReason)
            {
                if (null == f_DlPeers)
                {
                    f_DlPeers = new Dictionary<I2PAddress, DlPeerContext>(1);
                }
                else
                {
                    if (a_DlPeerCntxt.f_Clnt.f_Inst.f_NetworkMngr.f_NetworkPrefs.f_MaxDownloadConnsPerFile <= f_DlPeers.Count)
                    {
                        a_FailReason = SourceStates.OwnPerFileLimitReached;

                        return false;
                    }

                    if (true == f_DlPeers.ContainsKey(a_DlPeerCntxt.f_Clnt.p_Addr))
                    {
                        a_FailReason = SourceStates.DlContextExistsAlready;

                        return false;
                    }
                }

                f_DlPeers.Add(a_DlPeerCntxt.f_Clnt.p_Addr, a_DlPeerCntxt);
                f_PeersForGuiDirty = true;

                a_FailReason = SourceStates.Undefined;

                return true;
            }

            public bool TryAddUlPeerContext(UlPeerContext a_UlPeerCntxt)
            {
                if (null == f_UlPeers)
                {
                    f_UlPeers = new Dictionary<I2PAddress, UlPeerContext>(1);
                }
                else if (a_UlPeerCntxt.f_Clnt.f_Inst.f_NetworkMngr.f_NetworkPrefs.f_MaxUploadConnsPerFile <= f_UlPeers.Count ||
                         true == f_UlPeers.ContainsKey(a_UlPeerCntxt.f_Clnt.p_Addr))
                {
                    return false;
                }

                f_UlPeers.Add(a_UlPeerCntxt.f_Clnt.p_Addr, a_UlPeerCntxt);
                f_PeersForGuiDirty = true;

                return true;
            }

            public void RemoveDlPeerContext(DlPeerContext a_DlPeerCntxt)
            {
                bool l_OK = f_DlPeers.Remove(a_DlPeerCntxt.f_Clnt.p_Addr);

                if (0 == f_DlPeers.Count)
                {
                    ClearCachedETA();
                    f_DlPeers = null;
                }

                f_PeersForGuiDirty = true;
            }

            public void RemoveUlPeerContext(UlPeerContext a_UlPeerCntxt)
            {
                bool l_OK = f_UlPeers.Remove(a_UlPeerCntxt.f_Clnt.p_Addr);

                if (0 == f_UlPeers.Count)
                {
                    f_UlPeers = null;
                }

                f_PeersForGuiDirty = true;
            }

            public bool HasAnyDlPeerContext()
            {
                // May be used by any thread.

                return (null != f_DlPeers);
            }

            public bool HasAnyUlPeerContext()
            {
                // May be used by any thread.

                return (0 != f_UploadingUlPeerCntxts);
            }

            public bool HasDlPeerContext(I2PAddress a_Addr)
            {
                return (null != f_DlPeers &&
                        true == f_DlPeers.ContainsKey(a_Addr));
            }

            public void ModifyUploadingUlPeerContexts(bool a_Inc)
            {
                if (false == a_Inc)
                {
                    f_UploadingUlPeerCntxts--;
                }
                else
                {
                    f_UploadingUlPeerCntxts++;
                }
            }

            public UlPeerContext TryGetUlPeerContext(I2PAddress a_Addr)
            {
                // Return can be null.

                UlPeerContext l_UlPeerCntxt;
                if (null != f_UlPeers &&
                    true == f_UlPeers.TryGetValue(a_Addr, out l_UlPeerCntxt))
                {
                    return l_UlPeerCntxt;
                }

                return null;
            }

            public void SetGuiPeerContextsSubscription(bool a_State)
            {
                // This routine is meant to be used by the GUI tread.

                if (true == a_State)
                {
                    f_PeersForGuiDirty = true;
                }

                Util.VolatileWrite(out f_GuiNeedsPeerInformation, a_State);
            }

            public bool LoadGuiAccessablePeerContexts(out DlPeerContext[] a_DlPeers,
                                                      out UlPeerContext[] a_UlPeers)
            {
                // This routine is meant to be used by the GUI tread.
                // Returns true if at least one array is ready.

                a_DlPeers = f_DlPeersForGui;
                a_UlPeers = f_UlPeersForGui;
                Util.MemoryBarrier();

                return (null != a_DlPeers ||
                        null != a_UlPeers);
            }

            public void GetFilesPanelAttribsVolatile(Time a_Time,
                                                     out uint a_SecsToGo,
                                                     out Speed a_DlSpeed,
                                                     out Speed a_UlSpeed,
                                                     out uint a_CompleteNodes,
                                                     out uint a_PartialNodes,
                                                     out float a_AvailRatio)
            {
                // May be used by any thread.

                a_SecsToGo = f_ETAInSecs;
                a_DlSpeed = f_DlSpeed;
                a_UlSpeed = f_UlSpeed;
                a_CompleteNodes = f_CompleteNodes;
                a_PartialNodes = f_PartialNodes;
                a_AvailRatio = f_eD2kAvailRatio;
                Util.MemoryBarrier();
            }

            public void GetDetailsPanelAttribsVolatile(out uint a_CompleteNodes,
                                                       out uint a_PartialNodes,
                                                       out NetworkAvailability a_Avail,
                                                       out float a_AvailRatio)
            {
                // May be used by any thread.

                a_CompleteNodes = f_CompleteNodes;
                a_PartialNodes = f_PartialNodes;
                a_Avail = f_eD2kAvail;
                a_AvailRatio = f_eD2kAvailRatio;
                Util.MemoryBarrier();
            }

            public void ProcessPeerContexts(Time a_Time)
            {
                UpdateOwnCompleteness(a_Time);

                if (null != f_DlPeers)
                {
                    foreach (DlPeerContext l_DlPeerCntxt in f_DlPeers.Values)
                    {
                        if (false == l_DlPeerCntxt.f_Clnt.Process(a_Time) ||
                            false == l_DlPeerCntxt.Process(a_Time))
                        {
                            l_DlPeerCntxt.MarkAsRemovable(SourceStates.UnspecifiedConnErr, PeerContext.DestructionInfo.RemoveAtOurRequestPeerDoesNotNeedInfo);
                            l_DlPeerCntxt.f_Clnt.Close();
                        }
                    }
                }

                if (null != f_UlPeers)
                {
                    foreach (UlPeerContext l_UlPeerCntxt in f_UlPeers.Values)
                    {
                        if (false == l_UlPeerCntxt.f_Clnt.Process(a_Time) ||
                            false == l_UlPeerCntxt.Process(a_Time))
                        {
                            l_UlPeerCntxt.MarkAsRemovable(PeerContext.DestructionInfo.RemoveAtOurRequestPeerDoesNotNeedInfo);
                            l_UlPeerCntxt.f_Clnt.Close();
                        }
                    }
                }

                bool l_GuiNeedsInfo = Util.VolatileRead(ref f_GuiNeedsPeerInformation);

                if (true == f_NextMeasureSpeedTime.IsEarlierThan(a_Time))
                {
                    MeasureSpeed(a_Time, l_GuiNeedsInfo);
                    f_NextMeasureSpeedTime = a_Time.AddMilliseconds(SpeedOMeter.c_SuggestedMeasureSpeedInterval);
                }

                ManageGuiAccessToPeers(l_GuiNeedsInfo);
            }

            public void ManageGuiAccessToPeers(bool a_GuiDemands)
            {
                // May only get called by the network thread. Should
                // be invoked periodically.

                if (false == f_PeersForGuiDirty)
                {
                    // Leave data available as long as it is not dirty,
                    // even if a_GuiDemands is false.

                    return;
                }

                if (false == a_GuiDemands)
                {
                    f_DlPeersForGui = null;
                    f_UlPeersForGui = null;
                }
                else
                {
                    DlPeerContext[] l_DlPeers = Util.DictionaryToArray(f_DlPeers);
                    UlPeerContext[] l_UlPeers = Util.DictionaryToArray(f_UlPeers);
                    Util.MemoryBarrier();
                    f_DlPeersForGui = l_DlPeers;
                    f_UlPeersForGui = l_UlPeers;
                }

                f_PeersForGuiDirty = false;
            }

            private void UpdateETA(Time a_Time,
                                   ulong a_BytesDone)
            {
                f_ArrivalTimeEstimator.Update(a_Time, a_BytesDone);

                if (true == f_Fruitfulness)
                {
                    f_ETAInSecs = f_ArrivalTimeEstimator.CalcETAInSeconds(a_Time, f_Known.f_FileSize);
                }
            }

            private void UpdateFruitfulness()
            {
                bool l_NewFruitfulness = f_eD2kAvail.CheckFruitfulness(f_eD2kOwnCompl);

                if (l_NewFruitfulness == f_Fruitfulness)
                {
                    return;
                }

                f_Fruitfulness = l_NewFruitfulness;

                if (false == l_NewFruitfulness)
                {
                    ClearCachedETA();
                }
            }

            private void ClearCachedETA()
            {
                f_ETAInSecs = ArrivalTimeEstimator.c_UnknownTime;
            }

            public void MeasureSpeed(Time a_Time,
                                     bool a_AlsoCalcPeerSpeed)
            {
                bool l_AlsoCalcPeerSpeed = Util.VolatileRead(ref f_GuiNeedsPeerInformation);

                if (true == a_AlsoCalcPeerSpeed)
                {
                    if (null != f_DlPeers)
                    {
                        foreach (DlPeerContext l_DlPeerCntxt in f_DlPeers.Values)
                        {
                            l_DlPeerCntxt.MeasureSpeed(a_Time, true);
                        }
                    }

                    if (null != f_UlPeers)
                    {
                        foreach (UlPeerContext l_UlPeerCntxt in f_UlPeers.Values)
                        {
                            l_UlPeerCntxt.MeasureSpeed(a_Time, true);
                        }
                    }
                }

                f_SpeedOMeter.SaveProgress(a_Time, ref f_Known.f_TransferVol);
                f_SpeedOMeter.GetSpeed(a_Time, out f_DlSpeed, out f_UlSpeed);
            }

            private void UpdateOwnCompleteness(Time a_Time)
            {
                // Updates f_eD2kOwnCompl and f_PartsDone.

                ComplStatus l_Status = f_Known.p_StatusVolatile;

                if (ComplStatus.Incomplete == l_Status)
                {
                    IncomplFileProgress l_IncomplProgress = f_Known.p_IncomplProgressVolatile;

                    if (null == l_IncomplProgress)
                    {
                        if (0 != f_PartsDone)
                        {
                            f_NativeOwnCompl.SetAll(false);
                            f_eD2kOwnCompl.SetAll(false);
                            f_PartsDone = 0;
                            UpdateFruitfulness();
                        }
                    }
                    else
                    {
                        if (false == f_MapRevSet ||
                            f_MapRev != l_IncomplProgress.f_MapRevision)
                        {
                            f_MapRevSet = true;
                            f_MapRev = l_IncomplProgress.f_MapRevision;

                            // We will convert the bitmap from our native
                            // 8 MiB format into the eD2k 9,500 KiB format.

                            l_IncomplProgress.f_DoneMap.CopyVolatile(f_NativeOwnCompl);
                            f_eD2kOwnCompl.TakeOver(f_NativeOwnCompl, BitMap.TakeOverMode.RoundDown);
                            f_PartsDone = f_eD2kOwnCompl.p_SetBitCount;
                            UpdateFruitfulness();
                        }

                        if (null != f_DlPeers)
                        {
                            UpdateETA(a_Time, l_IncomplProgress.f_CompletedBytesInclUnverified);
                        }
                    }
                }
                else if (ComplStatus.Complete == l_Status)
                {
                    if (f_Known.f_PartCount != f_PartsDone)
                    {
                        f_NativeOwnCompl.SetAll(true);
                        f_eD2kOwnCompl.SetAll(true);
                        f_PartsDone = f_Known.f_PartCount;
                        UpdateFruitfulness();
                    }
                }
                else if (ComplStatus.Missing == l_Status)
                {
                    if (0 != f_PartsDone)
                    {
                        f_NativeOwnCompl.SetAll(false);
                        f_eD2kOwnCompl.SetAll(false);
                        f_PartsDone = 0;
                        UpdateFruitfulness();
                    }
                }
            }

            public void ChangeAvailability(BitMap a_Old,
                                           BitMap a_New)
            {
                if (null != a_Old)
                {
                    if (true == a_Old.p_AllSet)
                    {
                        f_CompleteNodes--;
                    }
                    else if (false == a_Old.p_AllClear)
                    {
                        f_PartialNodes--;
                    }
                }

                BitMap.UpdateAvailabilityArray(f_eD2kAvail, a_Old, a_New);
                f_eD2kAvailRatio = ((float)f_eD2kAvail.p_NonzeroItems / f_Known.f_PartCount);
                UpdateFruitfulness();

                if (null != a_New)
                {
                    if (true == a_New.p_AllSet)
                    {
                        f_CompleteNodes++;
                    }
                    else if (false == a_New.p_AllClear)
                    {
                        f_PartialNodes++;
                    }
                }
            }

            public Source[] FindSuitableSourcesForPeer(int a_MaxCount,
                                                       bool a_RestrictToClassicAddresses,
                                                       PeerContext a_ForPeerCntxt)
            {
                if (null == f_DlPeers &&
                    null == f_UlPeers)
                {
                    return null;
                }

                Dictionary<UserContext, Source> l_Dic = new Dictionary<UserContext, Source>(a_MaxCount);

                if (null != f_DlPeers)
                {
                    foreach (DlPeerContext l_DlPeerCntxt in f_DlPeers.Values)
                    {
                        if (true == a_RestrictToClassicAddresses &&
                            false == l_DlPeerCntxt.f_Clnt.p_Addr.p_IsClassicAddress)
                        {
                            continue;
                        }

                        l_DlPeerCntxt.AddAsSource(l_Dic, a_ForPeerCntxt);
                    }
                }

                if (null != f_UlPeers)
                {
                    foreach (UlPeerContext l_UlPeerCntxt in f_UlPeers.Values)
                    {
                        if (true == a_RestrictToClassicAddresses &&
                            false == l_UlPeerCntxt.f_Clnt.p_Addr.p_IsClassicAddress)
                        {
                            continue;
                        }

                        l_UlPeerCntxt.AddAsSource(l_Dic, a_ForPeerCntxt);
                    }
                }

                int l_Count = l_Dic.Count;

                if (a_MaxCount >= l_Count)
                {
                    if (0 == l_Count)
                    {
                        return null;
                    }

                    Source[] l_Sources = new Source[l_Count];
                    l_Dic.Values.CopyTo(l_Sources, 0);

                    return l_Sources;
                }

                // We've found too many sources.

                List<Source> l_Ls = new List<Source>(l_Dic.Values);
                l_Ls.Sort();
                l_Ls.RemoveRange(a_MaxCount, l_Count - a_MaxCount);

                return l_Ls.ToArray();
            }

            public bool RequestSourcesForMyself(Time a_Time)
            {
                // Returns true if at least one peer was asked.

                bool l_Asked = false;

                if (null != f_DlPeers)
                {
                    foreach (DlPeerContext l_DlPeerCntxt in f_DlPeers.Values)
                    {
                        if (true == l_DlPeerCntxt.HasPeerConfirmedFileRequest())
                        {
                            if (true == l_DlPeerCntxt.RequestSources(a_Time))
                            {
                                l_Asked = true;
                            }
                        }
                    }
                }

                return l_Asked;
            }

            public BitMap ExposeOwnComplBitmap(bool a_Native)
            {
                return (false == a_Native ? f_eD2kOwnCompl : f_NativeOwnCompl);
            }

#region Transfer

            public bool ReinitializeForDl(out TransferDemand a_Demand)
            {
                if (null == f_DlPeers)
                {
                    a_Demand = new TransferDemand();

                    return false;
                }

                uint l_NeededBw;
                BandwidthPrioritizer.Create(false, f_DlPeers.Values, f_DlPeers.Count, out l_NeededBw, out f_DlPrioritizer);
                a_Demand = new TransferDemand(f_Known.f_DlPriority, l_NeededBw, f_Known.f_DlLimit);

                return (0 != a_Demand.f_NeededBw);
            }

            public bool ReinitializeForUl(out TransferDemand a_Demand)
            {
                if (null == f_UlPeers)
                {
                    a_Demand = new TransferDemand();

                    return false;
                }

                uint l_NeededBw;
                BandwidthPrioritizer.Create(true, f_UlPeers.Values, f_UlPeers.Count, out l_NeededBw, out f_UlPrioritizer);
                a_Demand = new TransferDemand(f_Known.f_UlPriority, l_NeededBw, f_Known.f_UlLimit);

                return (0 != a_Demand.f_NeededBw);
            }

            public uint AssignDlBandwidth(uint a_ExtraBytes)
            {
                return f_DlPrioritizer.DistributeBandwidth(a_ExtraBytes);
            }

            public uint AssignUlBandwidth(uint a_ExtraBytes)
            {
                return f_UlPrioritizer.DistributeBandwidth(a_ExtraBytes);
            }

            public uint Download(uint a_MaxBytes)
            {
                return f_DlPrioritizer.Transfer();
            }

            public uint Upload(uint a_MaxBytes)
            {
                return f_UlPrioritizer.Transfer();
            }

#endregion
        }

        public class Source : IComparable<Source>
        {
            public readonly uint f_CompletedParts;
            public readonly I2PAddress f_StreamAddr;
            public readonly UserContext f_UserCntxt;

            public Source(uint a_CompletedParts,
                          I2PAddress a_StreamAddr,
                          UserContext a_UserCntxt)
            {
                f_CompletedParts = a_CompletedParts;
                f_StreamAddr = a_StreamAddr;
                f_UserCntxt = a_UserCntxt;
            }

            public int CompareTo(Source a_Other)
            {
                return this.f_CompletedParts.CompareTo(a_Other.f_CompletedParts);
            }
        }
    }
}

// End of file.
